;-*- Mode: Common-lisp; Package: langutils; Readtable: ytools; -*-
(in-package :langutils)
;;; $Id: defectexp.lisp,v 1.9 2006/10/30 21:43:13 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %ytools/ debug %langutils/ synutils)

(end-header :continue-slurping)

(needed-by-macros
   (export '(Defective-exp make-Defective-exp Defective-exp-has-target
	     Defective-exp-target is-Defective-exp
	     Defective-exp-observation Defective-exp-context
	     Defective-exp-condition make-Defective-exp-condition
	     Defective-exp-condition-ill ife-show simple-ill-formed-exp
	     Flagged-subexpression make-Flagged-subexpression
	     new-Flagged-subexpression flag-count*
	     is-Flagged-subexpression
	     Flagged-subexpression-exp Flagged-subexpression-ill-exps
	     flagexp
	     note-defective-exp defective-exp-notify
	     noting-syntactic-context
	     collecting-defective-exps 
	     break-on-flag* break-on-defective-exp* collect-bad-fields
	     check-number-of-args)))

(defun flagexp (mess exp)
   (new-Flagged-subexpression
      exp (list (simple-ill-formed-exp mess exp))))

(defvar target-placeholder*
        (make-Printable
	   (\\ (srm) (out (:to srm) "<<target>>"))))

(defvar invis-target*
        (make-Printable (\\ (_) (values))))

(def-class Defective-exp (:options :key)
   (target false)
   ;; The object singled out for attention, if any.
   ;; If not false, means the object (e.g., a Typed-exp) that this Defective-exp
   ;; is attached to as a bug.
   observation
   ;; -- observation is a function that takes the actual
   ;; target available when printing is appropriate (e.g., the source field of
   ;; a Typed-exp), plus a stream, 
   ;; and prints information about the defective expression.
   (context false)
   ;; -- false or a Syn-context
   (continuable true)
   (signaler (\\ (the-exp)
		(signal-problem :noplace
		   "Defective expression: " the-exp
		   (:continue "I'll record it somewhere"))))
;;; function that signals a problem.   
   (:handler (print (ie srm)
		(funcall (Defective-exp-observation ie)
			 (or (Defective-exp-target ie)
			     target-placeholder*)
			 srm))))
	  
(defun Defective-exp-has-target (de) (not (not (Defective-exp-target de))))

;;; For debugging
(defun ife-show (ife)
   (let ((cxt (Defective-exp-context ife)))
      (out (:to *query-io*)
	   "Defective expression"
	   (:q ((Defective-exp-has-target ife)
	       " Target: " (Defective-exp-target ife)))
	   :% "Problem: " (:e (funcall (Defective-exp-observation ife)
				     target-placeholder*
				     *query-io*))
	   :% (:q (cxt
		   (:q ((cadr cxt)
			"Immediate context: " (cadr cxt) :%))
		   (:q ((car cxt)
			"Stack: " (expstack-show (car cxt) 5) :%))))
	   :%)))

(defun simple-ill-formed-exp (mess exp)
   (make-Defective-exp
      ;;;; :has-target true
      :target exp
      :observation (\\ (target srm)
		      (out (:to srm) "(" (:a mess) 1 target ")"))
      :continuable true
      :signaler (\\ (this-exp)
		   (signal-problem :place simple-ill-formed-exp
		      "Ill-formed expression: " this-exp
		      (:continue "I'll record it and continue")))))

(define-condition Defective-exp-condition (error)
   ((ill :initarg :defective-exp :reader Defective-exp-condition-exp)
    (reporter :initarg :reporter :reader Defective-exp-condition-reporter))
   (:report (lambda (ill-ec srm)
	       (out (:to srm) ">>>: " (Defective-exp-condition-exp ill-ec)
		    :%)
	       (funcall (Defective-exp-condition-reporter ill-ec) srm))))

(def-class Flagged-subexpression
   exp
   ill-exps ;; List of Defective-exp, usually subexpressions of exp.

   (:handler
      (print (fs srm)
	 (let ((ill-exps (Flagged-subexpression-ill-exps fs)))
	    (multi-let (((has-main main others)
			 (repeat :for ((ie :in ill-exps))
			  :result (values false false ill-exps)
			  :until (and (Defective-exp-target ie)
				      (eq (Defective-exp-target ie)
					  (Flagged-subexpression-exp
						fs)))
			  :result (values true
					 ie
					 (remove1q ie ill-exps)))))
	       (letrec ((print-main (possibly-pretty-stream)
			   (cond (has-main
				  (funcall (Defective-exp-observation
					      main)
					   (Flagged-subexpression-exp fs)
					   possibly-pretty-stream))
				 (t
				  (out (:to possibly-pretty-stream)
				    (:q ((null ill-exps)
					"?? "))
				    (Flagged-subexpression-exp fs)))))
			(print-others (possibly-pretty-stream)
			   (repeat :for ((ie :in others))
			       (out (:to possibly-pretty-stream)
				    1
				    (:pp-nl :linear)
				    ie
;;; What the hell was this?
;;;;				    (:e (funcall (Defective-exp-observation
;;;;						   ie)
;;;;						invis-target*
;;;;						possibly-pretty-stream))
				    ))))
		  (cond (*print-pretty*
			 (pprint-logical-block (srm nil
						 :prefix "<<"
						 :suffix ">>")

			    (print-main srm)
			    (pprint-indent :block 0 srm)
			    (print-others srm)))
			(t
			 (out (:to srm) "<<")
			 (print-main srm)
			 (print-others srm)
			 (out (:to srm) ">>")))))))))

(defvar flag-count* 0
   "Count of number of expressions flagged")

(defvar break-on-flag* false)

(defun new-Flagged-subexpression (exp ill-exps)
   (cond ((null ill-exps) exp)
         
	 (t
          (cond ((is-Flagged-subexpression exp)
                 (!= ill-exps (append (Flagged-subexpression-ill-exps exp)
                                      *-*))
                 (!= exp (Flagged-subexpression-exp *-*)))
                (t
                 (setq flag-count* (+ flag-count* 1))))
	  (let ((fe (make-Flagged-subexpression exp ill-exps)))
	     (cond (break-on-flag*
		    (dbg-save :run-loud fe)
		    (signal-problem new-Flagged-subexpression 
			"Flagging subexpression " fe
			(:proceed
			 "Flagged subexpression will be inserted in output"))))
	     fe))))

;;;;(defun new-Flagged-subexpression (exp ill-exps)
;;;;   (make-Flagged-subexpression exp ill-exps))

(defvar defective-exp-collector* false)
;;; -- If not false, a function of two arguments, a Defective-exp and a Boolean.
;;; Returns true if producer of Defective-exp can assume it's been handled 
;;; ("collected") and do nothing; false otherwise.
;;; Second arg is true if this is not a "re-notification", i.e., the Defective-exp
;;; is fresh.  

;;; Convenience macro.  Check if there are exactly n args.  If not,
;;; complain, and if told to proceed, use makeups to make up the difference
;;; Return a list of exactly n args.
(defmacro check-number-of-args (args^ n^ fctr makeups)
   (let ((argv (gensym))
	 (shortfallv (gensym)))
      `(let ((,argv ,args^))
	  (let ((,shortfallv (- ,n^ (len ,argv))))
	     (cond ((= ,shortfallv 0)
		    ,argv)
		   (t
		    (note-defective-exp
		       ((_) "Wrong number of arguments " `(,',fctr ,@,argv))
		       :place ,fctr
		       (:novalue "I will try to guess what was meant"))
		    (cond ((< ,shortfallv 0)
			   (drop ,shortfallv ,argv))
			  ,@(<# (\\ (k)
				   `((= ,shortfallv ,k)
				     (append ,argv
					     (list ,@(take (- k) makeups)))))
				(series 1 (len makeups))))))))))

;;; NOTE-DEFECTIVE-EXP [(:obs ob) | ((targvar) -out-args-)]
;;;   [:TARGET targ] [:CONTEXT c]
;;;    -signal-stuff-)
;;; Builds a Defective-exp 
;;; with an 'observation' explicitly given or defined by the out-args 
;;; and a 'signaler' defined by the signal-stuff.
;;; If and whenever the observation is displayed, the -out-args- are
;;; printed, with targvar bound to some printable
;;; representation of the target.
;;; 'defective-exp-notify' is then called to tell the nearest Defective-exp
;;; collector that the Defective-exp has been produced.
(defmacro note-defective-exp (out-stuff &rest stuff)
   (let ((has-targ false)
	 (targ false)
	 (context false)
	 (sigstuff stuff)
	 (srmvar (gensym))
	 observation)
      (repeat
       :while (and (not (null sigstuff))
		   (memq (car sigstuff) '(:target :context)))
	 (cond ((null (cdr sigstuff))
		(signal-problem note-defective-exp
		   "Defective call to note-defective-exp (ironically): "
		   stuff
		   (:novalue "I'll ignore the bogus stuff"))
		(!= sigstuff '()))
	       ((eq (car sigstuff) ':target)
		(cond (has-targ
		       (signal-problem note-defective-exp
			  ":target specified twice in " stuff
			  (:novalue "I'll ignore all but the first")))
		      (t
		       (!= has-targ true)
		       (!= targ (cadr sigstuff))))
		(!= sigstuff (cddr *-*)))
	       (t
		(cond (context
		       (signal-problem note-defective-exp
			  ":context specified twice in " stuff
			  (:novalue "I'll ignore all but the first")))
		      (t
		       (!= context (cadr sigstuff))))
		(!= sigstuff (cddr *-*)))))
      (!= observation
	  (match-cond out-stuff
	     ?( (:obs ?fcnexp)
	       fcnexp)
	     ?( ((?(:+ ?targvar is-Symbol))
		 ?@description)
	       `(\\ (,targvar ,srmvar)
		  (out (:to ,srmvar) ,@description)))
	      (t 
	       (cond (has-targ
		      (signal-problem note-defective-exp
			 "Missing target variable in " out-stuff
			 (:novalue "I will treat it as (_)"))))
	       `(\\ (_ ,srmvar)
		  (out (:to ,srmvar) ,@out-stuff)))))
      (multi-let (((_ place-string _ condspec proceedspec)
		   (yt::signal-problem-analyze sigstuff))
		  (ievar (gensym)))
	 (cond ((not (is-String (car condspec)))
		(signal-problem note-defective-exp
		   "Can't specify condition in 'note-defective-exp': "
		   condspec)))
	 `(defective-exp-notify
	     (make-Defective-exp
	        ;;;; :has-target ',has-targ
		:target ,targ
		:observation ,observation
		:context ,(cond (context context) (t 'false))
		:continuable ',(not (or (not proceedspec)
					(eq proceedspec ':fatal)))
		:signaler (\\ (,ievar)
			     (signal-problem
			        :place ,place-string
				,@(include-if proceedspec)
				(:class Defective-exp-condition
				   :defective-exp ,ievar
				   :reporter (\\ (,srmvar)
						(format ,srmvar
						   ,@condspec))
				   ;;;; :continuable ,continuable
				   ))))
	     true))))

(defvar syntactic-context* false)

;;; Nothing is actually signaled unless there is no 'defective-collector*',
;;; or it refuses to collect this defective-exp
;;; 'fresh' means no collector has been notified of this Defective-exp
;;; before.
(defun defective-exp-notify (defective fresh)
   (maybe-set-context defective syntactic-context*)
;;;;   (!= last-defective-exp* defective)
;;;;   (dbg-save defective)
;;;;   (breakpoint defective-exp-notify
;;;;      "Notifying about: " defective
;;;;      :% " defective-exp-collector* = " defective-exp-collector*)
   (cond ((or (not defective-exp-collector*)
	      (not (funcall defective-exp-collector* defective fresh)))
	  (funcall (Defective-exp-signaler defective) defective))))

;;; If 'must' = false, then if there's no collector do nothing.
(defun defective-exps-renotify (explist must)
;;;;   (dbg-save explist)
;;;;   (breakpoint defective-exps-renotify
;;;;      "Renotifying: " explist)
   (cond (defective-exp-collector*
	  (repeat :for ((de :in explist))
	     (defective-exp-notify de false)))
	 (must
	  (repeat :for ((de :in explist))
	   :result (signal-problem defective-exps-renotify
		     "Screwed up handling of defective exps: " explist)
	   :until (not (Defective-exp-continuable de))
	   :result (funcall (Defective-exp-signaler de) de)))))

(needed-by-macros
(defun signal-stuff-is-fatal (stuff)
   (or (memq ':fatal stuff)
       (not (exists (s :in stuff)
	       (matchq ?(:\| :continue :novalue
			    (?(:\| :continue :novalue :prompt-for) ?@_))
		       s)))))
)

(defun try-restart (name &rest args)
   (let ((r (find-restart name)))
      (cond (r (apply #'invoke-restart r args))
	    (t false))))

(defmacro noting-syntactic-context (c &body b)
;;;;   (multiple-value-let (syn-stack current-syn)
;;;;		       (context-specifier-parse c) ...)
      `(bind ((syntactic-context* ,c))
;;;;	       (tuple ,(val-if-not-missing syn-stack)
;;;;		      ,(val-if-not-missing current-syn))
	  ,@b))

(defun maybe-set-context (ie c)
   (cond ((not (Defective-exp-context ie))
	  (!= (Defective-exp-context ie) c))))


;;; (collecting-defective-exps [context-specifier] 
;;;                    (cvar --other-vars--)
;;;                    e
;;;    [(:if-aborted [:restart-report s] -q-)]
;;;     -r-)

;;; binds cvar to all the Defective-exps found (i.e., those produced by
;;; 'note-defective-exp') while evaluating e,
;;; and bind other-vars to the values returned by e,
;;; then evaluates r.  
;;; The value(s) of the whole thing are the value(s) of (the last expression
;;; in) r.
;;; IMPORTANT: Don't forget to use the value of 'cvar' (typically to 
;;; flag the expression returned by 'r').  If you forget, the compiler
;;; won't warn you unless you also forget to use it in the :if-aborted
;;; clause
;;; The optional context-specifier is used to set the context of any
;;; collected Defective-exp that doesn't already have one set.
;;; If an :if-aborted clause is present anywhere in r, 
;;; then when a fatal 'note-defective-exp' occurs, control
;;; transfers to q, with cvar bound to the defective-exp's collected
;;; so far, and the other-vars not bound.
;;; If there is no :if-aborted clause, then fatal 'note-defective-exp's 
;;; cause a Defective-exp-condition to be signaled.
;;; If 'break-on-defective-exp*' is true, then the condition is signaled
;;; whenever a 'note-defective-exp' is executed, fatal or not.  
;;; In this case, if there
;;; is an :if-aborted clause, and the debugger is invoked, a restart 
;;; of type 'terminate-defective-exp-collection' will appear.
;;; The :report for this restart can be changed by including the
;;; :restart-report segment in the :if-aborted clause.
;;; Alternative syntax (after the :if-aborted clause has been extracted):
;;; (collecting-defective-exps [context-specifier]
;;;                 (cvar --other-vars--)
;;;                 -e- 
;;;    (:result r))
;;; Never used, but might come in handy.  This is equivalent to
;;; (collecting-defective-exps [context-specifier]
;;;                    (cvar --other-vars--)
;;;                    (progn ,@e)
;;;    r)

(defmacro collecting-defective-exps (&rest s)
   (let ((context false)
	 (stuff s)
	 cvar other-vars parse res
	 (abort-clause false)
	 restart-reporter)
      (cond ((is-context-specifier (car stuff))
	     (!= context (car stuff))
	     (!= stuff (cdr *-*))))
      (cond ((consp (car stuff))
	     (!= cvar (caar stuff))
	     (!= other-vars (cdar stuff))
	     (let ((abl (repeat :for ((x :in stuff :tail tx))
			 :result false
			 :until (car-eq x ':if-aborted)
			 :result tx)))
		(cond (abl
		       (!= abort-clause (cdar abl))
		       (!= stuff `(,@(ldiff *-* abl) ,@(cdr abl)))
		       (let ((tl (memq ':restart-report abort-clause)))
			 (cond (tl
				(!= restart-reporter (cadr tl))
				(!= abort-clause
				    `(,@(ldiff *-* tl) ,@(cddr tl))))
			       (t
				(!= restart-reporter
				    "Terminate defective expression collection"
				)))))))
	     (!= parse (cdr stuff))
	     (!= res (lastelt parse))
	     (cond ((car-eq res ':result)
		    (!= parse (drop -1 *-*))
		    (!= res (cdr *-*)))
		   (t
		    (!= res (cdr parse))
		    (!= parse (list (car parse)))))
	     (build-collector-form
	        context cvar other-vars parse res
		abort-clause restart-reporter))
	    (t
	     (signal-problem collecting-defective-exps
		"Missing or defective local variables in "
		`(collecting-defective-exps ,@s))))))

(defun build-collector-form (context cvar other-vars parse res
			     abort-clause restart-reporter)
   (let ((collect-block-name (gensym))
	 (fail-tag-name (cond (abort-clause false) (t (gensym))))
	 (normal-completion-var (gensym))
	 (contextvar (cond (context (gensym)) (t false))))
      (letrec ((maybe-wrap-context (form)
		  (cond (context
			 `(let ((,contextvar ,(cadr context)))
			     ,form))
			(t form)))
	       (wrap-restart-or-tagbody (form)
		  (cond (abort-clause
			 `(restart-case ,form
			     (terminate-defective-exp-collection
				 ()
				,@(cond (restart-reporter
					   `(:report ,restart-reporter))
					(t '()))
				(!= ,cvar (dreverse *-*))
				,@abort-clause)))
			(t
			 `(tagbody
			      ,form
			   ,fail-tag-name
			      ;; Can't handle fatal defective exp --
			      ;; look for someone who can.
			      (defective-exps-renotify (dreverse ,cvar) true)
			      (!= ,normal-completion-var true)
			    ))))
	       (coll-finish-form ()
		  (cond (abort-clause
			 `(defective-exp-collected ie fresh true))
			(t
			 `(or (defective-exp-collected ie fresh false)
			      (go ,fail-tag-name))))))
	      (maybe-wrap-context
		 `(let ((,cvar '())
			(,normal-completion-var false))
		     ;; Random tinkering introduced this variable, which
		     ;; should be flushed.
		     (declare (ignore ,normal-completion-var))
		     (unwind-protect 
			(block ,collect-block-name
			   ,(wrap-restart-or-tagbody
			       `(multi-let
				     ((,other-vars
				       (bind ((defective-exp-collector*
						 (\\ (ie fresh)
						    ,@(include-if context
							 `(maybe-set-context
							     ie ,contextvar))
						    (!= ,cvar (cons ie *-*))
						    ,(coll-finish-form))))
					    ,@parse)))
				   (!= ,cvar (dreverse *-*))
				   (let ((vals (values->list (progn ,@res))))
				      (!= ,normal-completion-var true)
				      (return-from ,collect-block-name
					 (list->values vals))))))
;;;;		       (cond ((and defective-exp-collector*
;;;;				   (not ,normal-completion-var))
;;;;			      (defective-exps-renotify (dreverse ,cvar)
;;;;				                       false)))
		       ))))))

(defvar break-on-defective-exp* false)

;;; Returns true if no need to branch to fail-tag.
;;; 'restart-in-place' is true if we have an :if-aborted clause
;;; handy.
;;; So a continuable defective-exp will be collected and then
;;; execution will resume at the point where the d.e. was noted
;;; without the user being asked to intervene.
(defun defective-exp-collected (defective fresh restart-in-place)
   (cond ((and break-on-defective-exp*
	       (or fresh (not (Defective-exp-continuable defective))))
	  (dbg-save :run-loud defective restart-in-place)
	  (funcall (Defective-exp-signaler defective) defective)))
   (cond ((Defective-exp-continuable defective)
	  true)
	 (restart-in-place
	  (invoke-restart 'terminate-defective-exp-collection))
	 (t false)))

(needed-by-macros
(defun is-context-specifier (e)
   (matchq (:context ?_) e)
;;;;   (and (consp e)
;;;;	(forall (x :in e)
;;;;	    (matchq (?(:\| :exp :stack) ?_)
;;;;		    x)))
)
)

(needed-by-macros
(defun context-specifier-parse (c)
   (let ((syn-stack '*missing)
	 (current-syn '*missing))
      (cond ((atom c)
	     (signal-problem context-specifier-parse :fatal
		"Illegal context spec: " c)))
      (repeat :for ((x :in c))
	 (cond ((or (atom x)
		    (not (memq (car x) '(:stack stack :exp exp))))
		(signal-problem context-specifier-parse :fatal
		   "Illegal in context spec: " x))
	       ((memq (car x) '(:stack stack))
		(cond ((not (eq syn-stack '*missing))
		       (signal-problem context-specifier-parse :fatal
			  "Stack specified twice: " c))
		      (t
		       (!= syn-stack (cadr x)))))
	       (t
		(cond ((not (eq current-syn '*missing))
		       (signal-problem context-specifier-parse :fatal
			  "Current expression specified twice: " c))
		      (t
		       (!= current-syn (cadr x)))))))
      (values syn-stack current-syn)))

(defun val-if-not-missing (v)
   (cond ((eq v '*missing) 'false)
	 (t v)))
)

(defun collect-bad-fields (good fields)
   (labels ((coll (fields)
               (cond ((null fields) '())
                     ((member (caar fields) good :test #'eq)
                      (coll (cdr fields)))
                     (t
                      (cons (car fields)
                            (coll (cdr fields)))))))
      (let ((b (coll fields)))
	 (cond ((null b) '())
	       (t
		`(,(flagexp "Meaningless in this context"
			    b)))))))

