;-*- Mode: Common-lisp; Package: ytools; Readtable: ytools; -*-
(in-package :ytools)
;;;$Id: symboid-check.lisp,v 1.2 2006/11/08 01:17:03 dvm Exp $

(depends-on %module/ ytools)

(depends-on (:at :run-time) %langutils/ symboid %misc/ fasl-save)

(defpackage wheresym-test)

(defparameter wheresym-test-package* (find-package :wheresym-test))

(test "Symboids"
   ;;; Clear the test package
   (do-symbols (sym wheresym-test-package*)
       (unintern sym wheresym-test-package*))
   (bind ((intern-printed-wheresyms* true)
          (wheresym-package* wheresym-test-package*))
      (let ((s1 (make-Symboid "foo"))
            bogus)
         (check 
            (string= (!= bogus (out-to-string s1)) "!:foo")
            (:e (dbg-save bogus))
            "Printing failed")
         (let ((wh (read-from-string (out-to-string s1))))
            (check (eq wh s1)
               "Couldn't read " s1 " back in"))
         (let ((s2 (make-Symboid "foo")))
            (dbg-save (bogus-s2 s2))
            (check
                (string= (Wheresym-name s2) "foo")
                "Duplicate names failed even before printing: "
                (Wheresym-name s2))
            (let ((print-s2 (out-to-string s2)))
               (dbg-save print-s2)
               (check
                  (not (string= (Wheresym-name s2) "foo"))
                  "Printing failed to change name of symbol"))))))

(test "Uninterned symboids"
   (bind ((intern-printed-wheresyms* false)
          (wheresym-package* wheresym-test-package*))
     (let ((s1 (make-Symboid "foo baz")))
        (check
           (string= (out-to-string s1) "!:|foo baz|")
           (:e (dbg-save (bad-s1 s1)))
           "Can't print symboids with odd characters properly")
        (bind ((intern-printed-wheresyms* true))
           (check 
              (string= (out-to-string s1) "!:|foo baz|")
              "With interning back on, printing falters")))))

(test "Multiple symboids with same name"
   (bind ((intern-printed-wheresyms* true)
          (wheresym-package* wheresym-test-package*)
          (create-wheresyms-when-read* true))
      (let ((c1 (build-symboid collide))
            (c2 (build-symboid collide)))
         (check (not (eq c1 c2))
            "Symboids prematurely interned")
         (let ((c3 (read-from-string "!:collide")))
            (check (not (or (eq c3 c1)
                            (eq c3 c2)))
                "Somehow uninterned symbols got interned")
            (let ((err-output
                     (with-output-to-string (ssrm)
                        (bind ((*error-output* ssrm))
                           (check (and (string= (out-to-string (list c1 c2))
                                                "(!:collide-1 !:collide-2)")
                                       (string= (Wheresym-name c1)
                                                "collide-1")
                                       (string= (Wheresym-name c2)
                                                "collide-2"))
                               (:e (dbg-save c1 c2 c3))
                               "Colliding symboid names not changed")))))
               (check (string= err-output
                               !"Changing symbol associated with wheresym to ~
                                 wheresym-test::collide-1~%~
                                 Changing symbol associated with wheresym to ~
                                 wheresym-test::collide-2~%")
                      (:e (dbg-save err-output))
                      "Didn't predict err-output exactly:" :%
                      err-output))))))

(defparameter object-extn-pathname*
    (make-Pathname
       :host false
       :device false
       :directory false
       :name false
       :type lisp-object-extn*))


(test "Symboid save and restore"
   (bind ((intern-printed-wheresyms* true)
          (wheresym-package* wheresym-test-package*)
          (create-wheresyms-when-read* true))
      (let* ((target-name (Symbol-name ':rrr))
             (s1 (make-Wheresym target-name)))
         (check (eq (where->here s1)
                    'wheresym-test::rrr)
            "Symbol not properly associated with wheresym " s1)
         (!= (wheresym-get s1 'easy) 'now)
         (!= (wheresym-get s1 'hard)
             (list s1 s1))
         (save-in-compiled-file (list s1 s1) "%langutils/symboid-debris.lisp")
         ;; Now simulate restoring in environment with
         ;; bogus wheresym of same name
         (!= (wheresym-get s1 distinguisher-prop*) false)
         (load (merge-pathnames object-extn-pathname*
                                (->pathname "%langutils/symboid-debris")))
         (let ((restored (first recovered-item*)))
            (check (eq (second recovered-item*) restored)
               "Failed to restore two occurrences of wheresym to 'eq' values: "
               :% 3 restored 3 (second recovered-item*))
            (check (not (eq restored s1))
               "Failed to distinguish restored wheresym from " s1)
            (check (and (string= (Wheresym-name restored)
                                 target-name)
                        (not (string= (Wheresym-name s1)
                                      target-name)))
               "Failed to standardize existing wheresym apart")
            (check (equal (<# first (Wheresym-plist restored))
                          (list 'easy 'hard distinguisher-prop*))
               (:e (dbg-save restored))
               "Failed to restore wheresym properties in expected order")
            (let ((hard (wheresym-get restored 'hard)))
               (check (and (eq (wheresym-get restored 'easy)
                               'now)
                           (is-list-of hard #'is-Wheresym)
                           (= (len hard) 2)
                           (eq (first hard) restored)
                           (eq (second hard) restored))
                  (:e (dbg-save restored))
                  "Failed to restored wheresym properties"))))))
