;-*- Mode: Common-lisp; Package: langutils; Readtable: ytools; -*-
(in-package :langutils)
;;; $Id: synutils.lisp,v 1.13 2006/10/30 21:43:13 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %ytools/ debug)

(end-header :continue-slurping)

;;; Tools for tracking syntax errors in complex input expressions

(needed-by-macros
   (export '(Expstack make-Expstack Expstack-expressions Expstack-relations
	     Context
	     expstack-push expstack-elt-push expstack-advance expstack-next
	     make-Syn-context cons-Syn-context nconc-Syn-context
             context-push-and-cons
	     Syn-context-props Syn-context-dialect-handler
	     syn-context-lookup
	     exp-stack-pair exp-syn-context
	     context-expstack-advance context-expstack-push
	     context-expstack-next context-expstack-show context-expstack-aug
	     context-find-expstack context-push-and-cons
	     expstack-show make-Noted-piece
	     make-Exp-with-rel Exp-with-rel-exp Exp-with-rel-rel
	     is-Exp-with-rel
	     collect-field append-field)))

(def-class Expstack
   expressions  ; List of expressions, innermost first
   relations    ; How each expression relates to its parent.
                ; Typically a number giving its position in parent.
                ; false means no relation.  (Last one is always false.)
   (:handler
       (print (estack srm)
	  (let ((expressions (Expstack-expressions estack)))
	     (out (:to srm)
		  "#<Expstack "(:q ((null expressions) "empty")
				  (t
				   (car expressions)
				   "@" (car (Expstack-relations estack))
				   ", " (len (cdr expressions)) " more"))
		  ">")))))

(defvar show-dialect-handler* false)

(def-class Syn-context  ;; syntactic context
    dialect-handler  ;; dialect-specific data 
    props    ;; (Lst (Tup Symbol Obj))
             ;; The symbol :stack is used to keep track of the Expstack
	     ;; Other symbols keep track of things like polarity.
   (:handler
       (print (sc srm)
	  (let ((hf (Syn-context-dialect-handler sc)))
	     (out (:to srm)
		  (:pp-block
		     "#<Syn-context "
		     (:pp-ind :block 2)
		     (:pp-nl :linear)
		     (first-alist-occurrences (Syn-context-props sc))
		     (:q (show-dialect-handler*
			 (:pp-nl :linear)
			 " --" hf))
		     ">")))))
    )

(defun nconc-Syn-context (pairs syncon)
   (make-Syn-context (Syn-context-dialect-handler syncon)
		     (nconc pairs (Syn-context-props syncon)))) 

(defun cons-Syn-context (pair syncon)
   (make-Syn-context (Syn-context-dialect-handler syncon)
		     (cons pair (Syn-context-props syncon)))) 

(defun context-push-and-cons (exp k prop val context)
   (cons-Syn-context (tuple prop val)
		     (context-expstack-push exp k context)))

(defun exp-syn-context (e dialect-handler props)
   (make-Syn-context
      dialect-handler
      (cons (exp-stack-pair e) props)))

(defun exp-stack-pair (e)
  (tuple ':stack
	 (make-Expstack (list e) (list false))))

(defun syn-context-lookup (con a)
   (assq a (Syn-context-props con)))

(defun context-expstack-advance (con)
   (multi-let (((expstack new-props)
		(context-find-expstack con true)))
      (context-expstack-aug
         con (expstack-advance expstack)
	 new-props)))

(defun context-expstack-push (e k con)
   (multi-let (((expstack props)
		(context-find-expstack con false)))
      (context-expstack-aug
         con (expstack-push e k expstack) props)))

(defun context-expstack-elt-push (k con)
   (multi-let (((expstack props)
		(context-find-expstack con false)))
      (context-expstack-aug
         con (expstack-elt-push k expstack)
	 props)))

(defun context-expstack-next (e k con)
   (multi-let (((expstack new-props)
		(context-find-expstack con true)))
      (context-expstack-aug
         con (expstack-next e k expstack)
	 new-props)))

(defun context-expstack-show (con)
   (expstack-show (context-find-expstack con false)
		  5))

;;; Returns < expstack new-props > 
;;; new-props = old props if remove = false, else old props
;;; with :stack pair removed.
(defun context-find-expstack (con remove)
   (let ((props (Syn-context-props con)))
      (let ((p (assq ':stack props)))
	 (cond (p
		(values (cadr p)
			(cond (remove (remove1q p props))
			      (t props))))
	       (t
		(signal-problem context-find-expstack :fatal
		   "Can't find :stack in context"))))))

(defun context-push-and-cons (exp k prop val context)
   (cons-Syn-context (tuple prop val)
		     (context-expstack-push exp k context)))

(defun context-expstack-aug (con expstack props)
   (make-Syn-context (Syn-context-dialect-handler con)
		     (cons (tuple ':stack expstack) props)))

;;; A list of "expstack relations" (r1 r2 ... rK) is interpreted 
;;; as the r1 of the r2 of... the rK of the expression.
;;; Written as (:compose r1 ... rK) when a rel is expected.

(defun expstack-push (e r expstack)
   (make-Expstack
      (cons e (Expstack-expressions expstack))
      (cons r (Expstack-relations expstack))))

(defun expstack-elt-push (r expstack)
   (let ((expl (Expstack-expressions expstack)))
      (cond ((null expl)
	     (signal-problem expstack-elt-push :fatal
		"Can't push an empty Expstack: " r 1 expstack))
	    (t
	     (let ((l (length (car expl))))
		(cond ((>= r l)
		       (dbg-save :run-loud (e (cadr expl)) r l)
		       (signal-problem expstack-elt-push :fatal
			  "Ran off end of enclosing expression: "
			  expstack))
		      (t
		       (expstack-push 
			  (elt (car expl) r)
			  r expstack))))))))

(defun expstack-advance (expstack)
   (let ((expl (Expstack-expressions expstack))
	 (rl (Expstack-relations expstack)))
      (cond ((or (null expl)
		 (not (is-Number (car rl)))
		 (null (cdr expl)))
	     (signal-problem expstack-advance :fatal
		"Can't advance Expstack: " expstack))
	    (t
	     (let ((l (length (cadr expl)))
		   (k (+ (car rl) 1)))
		(cond ((>= k l)
		       (dbg-save :run-loud (e (cadr expl)) k l)
		       (signal-problem expstack-advance :fatal
			  "Ran off end of enclosing expression: "
			  expstack))
		      (t
		       (rebuild-expstack
			  (elt (cadr expl) k)
			  k expl rl))))))))

(defun expstack-next (e r expstack)
   (let ((expl (Expstack-expressions expstack))
	 (rl (Expstack-relations expstack)))
      (cond ((null expl)
	     (signal-problem expstack-next :fatal
		"Can't pop expstack: " expstack))
	    (t
	     (rebuild-expstack e r expl rl)))))

(defun rebuild-expstack (e r expl rl)
   (make-Expstack
      (cons e (cdr expl))
      (cons r (cdr rl))))

(def-class Noted-piece
   exps
  (:handler
     (print (np srm)
	(out (:to srm) "*[*")
	(repeat :for ((e :in (Noted-piece-exps np)))
	   (out (:to srm) 1 e))
	(out (:to srm) " *]*"))))

(defun noted-piece-pprint (srm np)
   (out (:to srm)
      (:pp-block
         (:pre "*[*")
	 (:e (repeat :for ((e :in (Noted-piece-exps np)))
	       (:o (:to srm)
		   1 e (:pp-nl :fill))))
	 (:suf " *]*"))))

(set-pprint-dispatch 'Noted-piece #'noted-piece-pprint)

(defun expstack-show (expstack depth)
   (let ((expl (Expstack-expressions expstack))
	 (rl (Expstack-relations expstack)))
      (repeat :for ((i = 1 :to depth)
		   (rel ':start))
       :until (null expl)
         (expstack-frame-show (car expl) rel)
	 (!= expl (cdr expl))
	 (!= rel (car rl))
	 (!= rl (cdr rl))
       :result (cond ((not (null expl))
		     (out (:to *query-io*)
			  "..." :%))))))

(defun expstack-frame-show (e rel)
    (multi-let (((okay flagged-e)
		 (funcall (expstack-frame-rel-finder rel true)
			  e)))
       (out (:to *query-io*)
	    (:q (okay flagged-e)
		(t (out "?[" (:q (rel rel "> "))  e "]?")))
	    :%)))

;;; A "rel finder" is a function that takes an expression 'e'
;;; and returns two values: 
;;; < okay, piece>,
;;; where 'okay' is a Boolean.
;;; and 'piece' is a representation of a piece of 'e' 
;;; If 'okay' is false, the piece couldn't be found and
;;; 'piece' is meaningless.

;;; There are two kinds of rel finders: grabbers and noters.
;;; A grabber just returns a subexpression.
;;; A noter actually returns a superexpression with the thing
;;; found marked as a Noted-piece.

(defun expstack-frame-rel-finder (rel want-noter)
   (cond ((memq rel '(start same :start))
	  (cond (want-noter
		 (\\ (e) (values true (make-Noted-piece (list e)))))
		(t
		 (\\ (e) (values true e)))))
	 ((not rel)
	  (\\ (_) (values false nil)))
	 ((is-Integer rel)
	  (expstack-pos-finder rel want-noter))
	 ((is-Keyword rel)
	  (expstack-keyword-finder rel want-noter))
	 ((is-Pair rel)
	  (selq (car rel)
	     ((:tail)
	      (expstack-seg-finder (cadr rel) false want-noter))
	     ((:seg)
	      (expstack-seg-finder (cadr rel) (caddr rel) want-noter))
	     ((:compose)
	      (expstack-finders-compose (cdr rel) want-noter))
	     (t
	      (signal-problem expstack-frame-rel-finder
		 "Can't decipher: " rel))))
	 (t
	  (cond ((not want-noter)
		 (signal-problem expstack-frame-rel-finder
		    "Noter function " rel " where grabber required"
		    (:continue "I'll use it anyway"))))
	  (cond ((is-Symbol rel)
		 (symbol->fun rel))
		(t rel)))))

(defun expstack-pos-finder (rel want-noter)
   (\\ (e)
      (cond ((and (>= rel 0)
		  (is-list-of e (\\ (_) true))
		  (> (length e) rel))
	     (repeat :for ((el e (cdr el))
			  x
			  (i = 0 :by 1)
			  (rev '()))
		(!= x (car el))
	      :until (= i rel)
	        (!= rev (cons x *-*))
	      :result (values true
			     (cond (want-noter
				    `(,@(reverse rev)
				      ,(make-Noted-piece (list x))
				      ,@(cdr el)))
				   (t x)))))
	    (t
	     (values false nil)))))

(defun expstack-seg-finder (rel end want-noter)
   (\\ (e)
      (cond ((and (is-list-of e (\\ (_) true))
		  (>= rel 0)
		  (>= (length e) rel)
		  (or (not end)
		      (and (>= end 0)
			   (>= (length e) end)
			   (>= end rel))))
	     (values true
		     (let ((seg (nthcdr rel e))
			   (tail (cond (end (nthcdr end e))
				       (t '()))))
			(cond (want-noter
			       `(,@(ldiff e seg)
				 ,(make-Noted-piece (ldiff seg tail))
				 ,@tail))
			      (t
			       (ldiff seg tail))))))
	    (t
	     (values false nil)))))

(defun expstack-keyword-finder (rel want-noter)
   (\\ (e)
      (repeat :for ((el e :then (cddr el))
		    reached-end)
	 (!= reached-end
	     (or (null el)
		 (null (cdr el)))) 
       :until (or reached-end
		 (eq (car el) rel))
       :result (values (not reached-end)
		      (and (not reached-end)
			   (cond (want-noter
				  `(,@(ldiff e el)
				    ,(make-Noted-piece 
					`(,(car el) ,(cadr el)))
				    ,@(cddr el)))
				 (t
				  (cadr el))))))))

;;; Functional-programming overkill
(defun maybe-wrap-as-noted (want-noter x wrapper)
   (cond (want-noter
	  (funcall x wrapper))
	 (t x)))

(defun expstack-finders-compose (rels want-noter)
   (cond ((null rels)
	  (expstack-frame-rel-finder ':start want-noter))
	 (t
	  (let ((fa (expstack-frame-rel-finder (car rels) want-noter))
		(fd (expstack-finders-compose (cdr rels) false)))
	     (\\ (e)
		(multi-let (((d-okay d-piece)
			     (funcall fd e)))
		   (cond (d-okay
			  (funcall fa d-piece))
			 (t
			  (values false nil)))))))))

;;; An expression together with its position ("relation") to its
;;; superexpression.
(def-class Exp-with-rel
   (:options (:medium :list))
   rel
   exp)

(defun function-name (fcn)
  (nth-value 2 (function-lambda-expression fcn)))

;;; Remove "hidden" pairs to reduce clutter
(defun first-alist-occurrences (alist)
   (letrec ((prune (al already)
	       (cond ((null al) '())
		     ((memq (caar al) already)
		      (prune (cdr al) already))
		     (t
		      (cons (car al)
			    (prune (cdr al)
				   (cons (caar al) already)))))))
      (prune alist '())))

;;;;(defun expstack-mystery-rel (x)
;;;;   (\\ (e)
;;;;      (make-Noted-piece `("?[" (q (x x "> "))  e "]?"))))

; Like collect-field, but append results together
(defun append-field (name fields)
   (<< append (collect-field name fields)))

; Collect all e such that (name . e) occurs in fields.
(defun collect-field (name fields)
   (cond ((null fields) '())
         ((eq (caar fields) name)
          (cons (cdar fields) (collect-field name (cdr fields))))
         (t
          (collect-field name (cdr fields)))))
