;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: altformer.lisp,v 1.6 2005/12/27 01:54:43 dvm Exp $

(depends-on %module/ ytools)

(depends-on :at-run-time %nity/ desig alttype)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(types-common-supertype make-alt-type)))

(def-type-former Alt (:system :universal)
		 (alts typename vartypes)
		 (ignore typename)
   (let ((alt-types (<# (\\ (a) (designated-type a false vartypes))
			   alts)))
      (cond ((null alt-types) void-type*)
	    ((null (cdr alt-types))
	     (car alt-types))
	    (t
	     (build-alt-type alt-types)))))

(defun build-alt-type (alt-types)
      (cond ((= (len alt-types) 1)
	     (car alt-types))
	    (t
	     (make-nontrivial-alt-type
	        alt-types
		(</ common-supertype void-type* alt-types)
		no-env*))))

(defun make-alt-type (alt-types vartypes)
   (!= alt-types
       (types-max-elements alt-types vartypes))
;;;;   (dbg-save alt-types)
;;;;   (breakpoint make-alt-type
;;;;      "alt-types = " alt-types)
   (cond ((null alt-types) void-type*)
	 ((null (cdr alt-types))
	  (new-Type-closure (car alt-types) vartypes))
	 (t
	  (make-nontrivial-alt-type
	     alt-types
	     (types-common-supertype false alt-types
				     (empty-undo-stack) vartypes)
	     vartypes))))

(defun types-min-elements (types vartypes)
   (types-extreme-elements types false vartypes))

(defun types-max-elements (types vartypes)
   (types-extreme-elements types true vartypes))

(defvar types-extreme-elements-dbg* false)

(defun types-extreme-elements (types want-max vartypes)
   (let ((components (<! (\\ (ty) (alt-flatten ty vartypes))
			 types))
	 (winners-so-far '()))
      (repeat :for ((c :in components))
	 (dbg-out types-extreme-elements-dbg*
	     "Considering component c = " c (:i> 3) :%)
;;;;	 (cond ((exists (u :in winners-so-far)
;;;;		   (types-equal c u vartypes global-env*))
;;;;		(dbg-out types-extreme-elements-dbg*
;;;;		   "Duplicate, discarded" :%))
;;;;	       (t
	 (multiple-value-let (ct cenv)
			     (follow-var-ref c vartypes)
	    (let ((dominated false))
	      (dbg-out types-extreme-elements-dbg*
		    "Looking for dominators of c" :%)
	      (repeat :for ((u :in winners-so-far)
			   ut uenv)
		 (dbg-out types-extreme-elements-dbg*
		     "u = " u :%)
		 (!= < ut uenv > (follow-var-ref u empty-env*))
		 (cond ((cond (want-max
			       (type-acceptable
				   false ct ut '() cenv uenv '()))
			      (t
			       (type-acceptable
				   false ut ct '() uenv cenv '())))
			(dbg-out types-extreme-elements-dbg*
			    "c is " (:q (want-max "subtype")
				       (t "supertype"))
			    " of u, hence discarded" :%)
			(!= dominated true)))
		 :until dominated)
	      (cond ((not dominated)
		     (dbg-out types-extreme-elements-dbg*
			 "c survives; weeding out types dominated by c" :%)
		     (repeat :for ((u :in winners-so-far)
				  (survivors '())
				  ut uenv)
			(dbg-out types-extreme-elements-dbg*
			    "u = " u :%)
			(!= < ut uenv > (follow-var-ref u empty-env*))
			(cond ((cond (want-max
				      (type-acceptable
					  false ut ct '()
					  uenv cenv '()))
				     (t
				      (type-acceptable
					  false ct ut
					  '() cenv uenv '())))
			       (dbg-out types-extreme-elements-dbg*
				   "u is " (:q (want-max "subtype")
					      (t "supertype"))
				   " of c, hence discarded" :%))
			      (t
			       (dbg-out types-extreme-elements-dbg*
				  "u survives")
			       (!= survivors (cons u *-*))))
		      :result (!= winners-so-far
				 (reverse (cons c survivors))))))
	      (dbg-out types-extreme-elements-dbg*
		 "winners-so-far = " winners-so-far :% (:i< 3)))))
      winners-so-far))

(defun alt-flatten (ty vartypes)
   (multiple-value-let (ty vartypes)
		       (follow-var-ref ty vartypes)
      (multiple-value-let (etl found)
			  (type-find-feature ty 'alt-types vartypes)
	 (cond (found
		(<! (\\ (ty) (alt-flatten ty vartypes))
		    etl))
	       ((is-tvar-type ty)
		(let ((b (tvar-type-bdg ty)))
		   (cond ((null (Typebdg-supers b))
			  '())
			 (t
			  (<! (\\ (ub) (alt-flatten ub vartypes))
			      (Typebdg-supers b))))))
	       (t
		(list (new-Type-closure ty vartypes)))))))

(defvar take-min-supertype* true)

;;; Heavy-duty version.
;;; Deals with Type-closures etc.
;;; If constrain-tvars is true, constrains so as to minimize the bound.
(defun types-common-supertype (constrain-tvars types undo-stack vartypes)
   (repeat :for ((ty :in types)
		 (boundsofar void-type*)
		 (boundenv '())
		 (ustack undo-stack))
    :result (values boundsofar ustack)
      (multi-let (((sty styenv)
		   (follow-var-ref ty vartypes)))
	 (multi-let (((bound-could-be-high u)
		      (type-acceptable constrain-tvars
					  sty boundsofar ustack
					  styenv boundenv '())))
	    (!= ustack u)
	    (cond ((not bound-could-be-high)
		   (track-extra-vals :extra-vars ((inner-ustack ustack))
		      (repeat
		       :until (extra-vals
				 (type-acceptable constrain-tvars
				    boundsofar sty inner-ustack
				    boundenv styenv '())
				 :+ inner-ustack)
		       :result (progn (!= boundsofar sty)
				      (!= boundenv styenv))
			 (!= < sty styenv >
			     (follow-var-ref (type-find-supertype sty styenv)
					     styenv))
		       :while sty
		       else :result (signal-problem types-common-supertype
				      "Can't find common supertype of"
				      :% 3 ty
				      :% 3 "and " boundsofar))
		    :values ((!= ustack inner-ustack)))))))))
