;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: argspair.lisp,v 1.5 2006/03/09 16:11:02 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ typevar)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(params-args-pair)))

(defun Arglistspec-all-nonrest-argspecs (alspec)
   (cond ((null (Arglistspec-rest alspec))
	  (Arglistspec-argspecs alspec))
	 (t
	  (append (Arglistspec-required alspec)
		  (Arglistspec-optional alspec)
		  (Arglistspec-key alspec)))))

;;; This file does not obey the Lisp rules exactly, but an idealized version.
;;; E.g.: unquoted keywords are assumed to be argument labels.  So
;;; (list :a :b :c) is an error.  You should write (list ':a ':b ':c).

;;; This should be called by decl stuff as well.
;;; Returns < pairs, rest-pair, pathology >
;;; where pairs is a list of (Argspec argval) pairs for every
;;; element of argvals;
;;; rest-pair = if non-redundant &rest then (restarg val)
;;;                else (false other-key-args)
;;; ("val" is not really an expression, just a list of expressions for
;;;  future reference)
;;; A &rest is redundant if there are keyargs or &allow-other-keys is
;;; present; in this case every args is accounted for as a key args.
;;; other-key-args are "other" keys if &allow-other-keys is present
;;; (as an alist).
;;; Note: You need all of this information because other-key-args
;;; tells you some args that may need to be processed because they
;;; don't occur in pairs, and rest-pair tells you the exact order
;;; in which they need to be processed.
;;; pathology is false if args match vals; else too-many-args or too-few-args
(defun params-args-pair (param-specs args)
   (let ((reqparms (Arglistspec-required param-specs))
	 (optparms (Arglistspec-optional param-specs))
	 (keyparms (Arglistspec-key param-specs)) 
	 (restparms (Arglistspec-rest param-specs))
	 (allow-other-key (Arglistspec-allow-other-keys param-specs))
	 (pathology false))
      (let ((really-rest
	       (and (not (null restparms))
		    (null keyparms))))
	 (let ((keyvals (cond (really-rest '())
			      (t (find-first-key args)))))
            (let ((nonkeyvals (cond ((null keyparms) args)
				    (t (ldiff args keyvals)))))
	       (let ((arglen (len nonkeyvals))
		     (reqnum (len reqparms))
		     (optnum (len optparms))
		     (rest '())
		     (keypairs '())
		     (other-keypairs '()))
		  (letrec ()
		     (cond ((< arglen reqnum)
;;;;			    (dbg-save arglen reqnum param-specs args)
;;;;			    (breakpoint params-args-pair
;;;;			       "About to flag 'too-few-args'")
			    (!= pathology 'too-few-args)
			    (!= reqparms (take arglen *-*)))
			   ((> arglen (+ reqnum optnum))
			    ;; More args than there are arg names
			    (cond ((or (not (null restparms))
				       (not (null keyparms)))
				   (!= rest (drop (+ reqnum optnum)
						  args)))
				  (t
				   (!= pathology 'too-many-args)))
			    (!= reqparms (append *-* optparms))
			    (!= optparms '())
			    (!= reqnum (+ reqnum optnum)))
			   ((> arglen reqnum)
			    (!= reqparms (append *-* (take (- arglen reqnum)
							  optparms)))
			    (!= optparms (drop (- arglen reqnum) optparms))
			    (!= reqnum arglen)))
		     (cond ((not really-rest)
			    (look-for-keyvals)))
		     ;; Return < l, re, r, k>
		     ;; where l is list of (Argspec, val) pairs
		     ;; in order.
		     ;; re is true iff r is essential.
		     ;; r is false or (restargspec, restval)
		     ;; k is list of (keysym, val) pairs for unknown key parms
		     ;; unknown keys if allow-other-key is set.
		     (values
			(append (<# tuple reqparms nonkeyvals)
				(reverse keypairs))
			(cond (really-rest
			       (tuple (car restparms) rest))
			      (t
			       (tuple false other-keypairs)))
			pathology)
		       
	      :where

	        (:def look-for-keyvals ()
		   ;; At this point (len reqparms)
		   ;; = (len nonkeyvals) = reqnum
		   ;; optparms are optionals that did *not* appear
		   ;; If optparms are non-() then keyvals should be ()
		   (no-keys-without-all-optionals)
		   (repeat :for ((kl keyvals (cddr kl)))
		    :until (or (null kl)
			      (null (cdr kl)))
		    :result (progn
			      (cond ((not (null kl))
				     (odd-key-msg kl keypairs))))
		      (cond ((is-Keyword (car kl))
			     (repeat :for ((a :in keyparms))
			      :result (cond (allow-other-key
					    (!= other-keypairs
						(cons (tuple
							 (car kl)
							 (cadr kl))
						      *-*)))
					   (t
					    (illegal-key-msg
					       (car kl) keypairs)))
			      :until (eq (car kl) (Argspec-mode a))
			      :result (progn
					(!= keypairs
					    (cons (tuple a (cadr kl))
						  *-*))
					(!= keyparms (remove1q a *-*)))))
			    (t
			     (illegal-key-msg (car kl) keypairs)))))

	        (:def no-keys-without-all-optionals ()
		    (cond ((and (not (null optparms))
				(not (null keyvals)))
			   (error-break argcheck 
			      "Optional parms " optparms " missing before"
			      :% " keyword parms " keyvals
			      (:novalue "I will assume their default values"))
			   (!= nonkeyvals
			       (append *-* (<# Argspec-default optparms)))
			   (!= optparms '()))))			   

	        (:def illegal-key-msg (k keypairs)
		   (error-break argcheck
		      "Illegal keyword label: " k
		      :% " Already saw: " keypairs
		      :% " Still looking for: " keyparms 
		      (:novalue "I will ignore it and the argument following it")))

		(:def odd-key-msg (stuff keypairs)
		   (error-break argcheck
		      "Odd stuff at end of keyword list: " stuff
		      :% " After finding: " keypairs
		      :% " Still looking for: " keyparms
		      (:novalue "I will ignore it")))

;		(:def wna-msg (which)
;		   (error-break argcheck
;		      which " arguments"
;		      :% " Got: " args
;		      :% " Require: " reqparms
;		      (:novalue "I will ignore the excess formal or actual parameters")))
	     )))))))

(defun find-first-key (keyvals)
   (repeat
    :until (or (null keyvals)
	      (is-Keyword (car keyvals)))
      (!= keyvals (cdr *-*))
    :result keyvals))

(defun Rest-pair-just-keys (rp) (not (car rp)))
(defun Rest-pair-spec (rp) (car rp))
(defun Rest-pair-args (rp) (cadr rp))
(defun Arg-pair-spec (ap) (car ap))
(defun Arg-pair-arg (ap) (cadr ap))


