;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: auxdefs.lisp,v 1.3 2004/02/22 00:46:57 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(end-header :continue-slurping)

;;; If they are allowed, this variable is bound to a list of them
(defvar aux-defs* ':not-allowed)

;;; If true, don't collect aux-defs; eval them.
(defvar eval-aux-defs* false)

;;; kind is either :fun or :var.
;;; If kind=:fun, def is (name args -body-)
;;; If kind=:var, def is (name val)
(def-class Aux-def
    kind
    def)

(defun note-aux-def (kind def)
  (cond (eval-aux-defs* 
	 (repeat :for ((e :in (aux-defs-global (list (make-Aux-def kind def)))))
	    (eval e))))
  (cond ((eq aux-defs* ':not-allowed)
	  (error-break note-aux-def
	     "Auxiliary definition in illegal context: "
	     "[" kind "] = " def
	     (:novalue "I will ignore it")))
	 (t
	  (!= aux-defs* (cons (make-Aux-def kind def) *-*)))))

;;(defvar aux-var-num* 0)

(defun aux-defs-global (aux-defs)
   (let ((funs '())
	 (vars '()))
      (repeat :for ((a :in aux-defs))
	 (cond ((not (is-Aux-def a))
		(error-break aux-defs-global
		   "Bad aux-def: " a
		   (:novalue "It will be ignored")))
	       ((memq (Aux-def-kind a) '(:fun fun))
		(!= funs (cons `(defun ,@(Aux-def-def a))
			       *-*)))
	       (t
		(!= vars (cons `(defvar ,@(Aux-def-def a))
			       *-*)))))
      `(,@vars ,@funs)))

;;; Useful macro: (lambda-or-aux-def test symform args -body-)
;;; Evaluates either to name of (aux-def'ed) function or a lambda-expression,
;;; depending on test.
(defmacro lambda-or-aux-def (test^ fun-name^ def)
   (let ((ev (gensym))
	 (nv (gensym)))
      `(let ((,ev ,def))
	  (cond (,test^
		 (let ((,nv ,fun-name^))
		    (note-aux-def ':fun
		       (cons ,nv ,ev))
		    ',nv))
		(t
		 (cons ':inline ,ev))))))

