;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: conformer.lisp,v 1.6 2004/07/08 14:16:07 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ desig contype dottypes subtype
			 %langutils/ synutils)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(make-const-type 
	     const-type const-of-type true-type* false-type*)))

(def-type-former Con (:system :universal)
		 (range typename vartypes)
		 (ignore typename vartypes)
   (make-const-type (consts-canonicalize range)))

;;; To avoid ambiguities, the symbols empty-list, false, and true are
;;; not interpreted as symbols.  To force them to be so interpreted,
;;; quote them.
;;; All constants *besides* these are quoted in the 'consts' property
;;; of the constructed type.
(defun consts-canonicalize (consts)
   (let ((l (len consts)))
      (!= consts
	  (<# (\\ (c k)
		 (letrec ((ask (choices)
			    (repeat :for (r)
			       (err-out
				  "Type one of " choices
				  " to replace ("
				  (:e (repeat :for
					     ((x :in consts)
					      (i = 0 :by 1))
					(:o (:q ((= i k)
					       ">>> "
					       x " <<<")
					      (t x))
					    (:q ((< i (- l 1))
						 1)))))
				  "): ")
			       (!= r (read *query-io*))
			     :until (member  r choices :test #'equal)
			     :result r)))
		    (cond ((eq c t)
			   (ask '(true 'true)))
			  ((eq c nil)
			   (ask '(false empty-list 'nil)))
			  ((car-eq c 'quote)
			   c)
			  ((memq c '(empty-list true false))
			   c)
			  (t
			   `(quote ,c)))))
	      consts
	      (series 0 (- l 1))))
      (!= consts (nodup  *-* :test #'equal))
;      (err-out "Revised consts = " consts)
      consts))

;;; Every const is represented as ('quote' c) except
;;; empty-list, true, and false.
(defun make-const-type (consts)
   (cond ((null consts) void-type*)
	 ((and (memq 'true consts)
	       (memq 'false consts)
	       (forall (x in consts)
		  (memq x '(true false))))
	  bool-type*)
	 ((forall (x in consts)
	     (eq x 'empty-list))
	  empty-tup-type*)
	 (t
	  (make-nontrivial-const-type
	      consts 
	      (</ (\\ (ty a)
		     (common-supertype
			ty
			(cond ((or (eq a 'true)
				   (eq a 'false))
			       bool-type*)
			      ((eq a 'empty-list) empty-tup-type*)
			      ((or (is-Symbol a)
				   (is-Canonized-symbol a))
			       (signal-problem make-const-type
				  "Illegal in const list: " a))
			      ((car-eq a 'quote)
			       (const-type (cadr a)))
			      (t
			       (const-type a)))))
		  void-type* consts)))))


(defvar true-type* (nisptype::make-const-type '(true)))
(defvar false-type* (nisptype::make-const-type '(false)))

(def-op const-type (c) univ-type*)

(def-meth const-type ((c cons))
   (let ((cdrtype
	    (cond ((null (cdr c))
		   empty-tup-type*)
		  (t
		   (const-type (cdr c))))))
      (make-dot-type
	 'Tup
	 (const-type (car c))
	 cdrtype global-env*)))

(def-meth const-type ((c null))
   (let ((intended
	    (signal-problem const-type
	       c " can't be typed when it occurs as a constant; "
	       :% "Use 'false' or !(..) instead"
	       (:prompt-for
		  ":false or :empty-list"))))
      (cond ((eq intended ':false)
	     false-type*)
	    (t
	     (make-lstype univ-type*)))))

