;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: deftyp.lisp,v 1.12 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

;;;;(depends-on :at-run-time extra/ dbghck)
(depends-on (:at :slurp-time) %nity/ foundation
	    (:at :run-time) %ytools/ debug
			 %nity/ types subtype desig tloader parmutils
				auxdefs parmsparse parmsgen
			 %langutils/ synutils namespace)

;;; Slurp to get macro definitions
(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(forward-define def-type def-type-fun def-type-constructor
	     Clump)))

;; What's tricky here:
;; Conceptually there are three phases that a Nisp file must undergo:
;; slurping, compiling, and loading.
;; As far as DEFTYPE is concerned, what must happen at each phase is
;; as follows--
;;   Slurping: Set up type structure and declarations.
;;   Compiling:  Type structure and declarations must already be set up.
;;      Compile auxiliary functions.
;;   Loading: Set up type structure and load functions

;; But the three phases can be scrambled up--
;;   Typing at terminal: slurping, compiling, and
;;      loading happen all at once
;;   Loading source file: slurping happens, then compiling and loading
;;   Compiling & loading file: slurping happens, then compiling; later loading

;; Tentatively we arrange for all this as follows.  SLURP handles DEFTYPEs
;; by finding the type structure they specify and setting it up.
;; During compiling, type-structure setups are generated for later use
;; by loading.
;; During loading, both type-structure setups and function definitions are
;; executed, *unless* the type-structure setups think they're coming from a source
;; file, when they do nothing, on the assumption that SLURP already did it.

(in-type-system universal)

;;; Simplified version of Cluster
;;; We don't have decl yet, so we can have no declarations 
(def-type-former Clump (args typename vartypes)
   (:system universal)
   (match-cond args
      (:? (?which-sort ?base ?@patches)
;;;;	(out "which-sort = " which-sort
;;;;	     " base = " base " patches = " patches :%)
	(cond ((not (memq which-sort '(:subtype :modified)))
	       (error-break clump-type-former :fatal
		  "Second field of Clump must be :subtype or :modified: "
		  `(clump ,which-sort ,base ,@patches))))
	(let ((just-base (and (eq which-sort ':modified)
			      (null patches))))
	   (let ((base (designated-type-with-name
		          base true
			  (cond (just-base typename)
				(t false))
			  vartypes)))
	      (cond (just-base
		     base)
		    (t
		     (let ((tclass (patches-class patches which-sort vartypes)))
			(let ((newtype
				 (new-Type tclass
					   `(Clump ,@args)
					   (Type-freevars base)
					   base
					   '()
					   (clump-features
					      typename patches which-sort))))
			   (repeat :for ((p :in patches))
			      ;(out "patch: " p :%)
			      (cond ((atom p)
				     (error-break clump-type-former
					"Illegal in Clump: " p
					(:novalue "I will ignore it")))
				    ((eq (car p) 'is)
				     (!= (Type-features newtype)
					 (cons (is-extract (cdr p)
							   typename
							   vartypes)
					       *-*)))
				    ((eq (car p) ':feature)
					(!= (Type-features newtype)
					    (cons (tuple (cadr p)
							 (feature-decipher
							    (cddr p) vartypes))
						  *-*)))
				    ((not (eq (car p) ':class))
				     ;; It's a slot
				     (!= (Type-slotfns newtype)
					 (cons (make-Slot (car p)
							  (cadr p)
							  false
							  (cond ((cddr p)
								 (designated-type
								    (caddr p)
								    false
								    vartypes))
								(t
								 univ-type*))
							  no-args-spec*
							  false
							  false)
					       *-*)))))
;;;;			   (cond ((not (null (Type-slotfns newtype)))
;;;;				  (dbg-save newtype)
;;;;				  (breakpoint clump-type-former
;;;;				     "got slots")))
			   newtype)))))))
      (t
       (error-break clump-type-former :fatal
	  "Ill-formed Clump: " `(Clump ,@args)))))
	
(defun clump-features (typename patches which-sort)
		      (ignore patches)
   (nconc (cond (typename
		 (list (tuple 'home
			      (ensure-top-level-type-sys))))
		(t '()))
	  (cond ((eq which-sort ':modified)
		 (list (tuple 'mod-equal-tester
			      (simple-loadable-function
				    compare-patches))
		       (tuple 'mod-got-me-want-other
			      (simple-loadable-function
			         patches-subtyper))
		       (tuple 'mod-got-other-want-me
			      (simple-loadable-function
			         patches-subtyper))))
		(t '()))))

(defun is-extract (is-clause typename vartypes)
   (let ((i (cond ((eq (car is-clause) 'boolean)
		   (cadr is-clause))
		  (t (car is-clause)))))
      (letrec ()
	 (match-cond i
	    ((atom i)
	     (is-test-feat i))
	    (:? (?(:\| lambda \\) ?@body)
	      (is-test-feat `(lambda ,@body)))
	    (:? (function ?fn)
	      (is-test-feat fn))
	    (:? (?f x)
	      (is-test-feat f))
	    (t
	     (multiple-value-let (b params has-an-x)
				 (is-body-resolve i vartypes)
		(cond ((not has-an-x)
		       (error-break is-extract :fatal
			  "Is test with no occurrence of X: "
			  i)))
		(cond ((null params)
		       (is-test-feat `(lambda (x) ,b)))
		      (t
		       (let ((ib-name (typename-is-builder-name typename vartypes)))
			  (is-builder-define ib-name b params)
			  (tuple 'is-builder ib-name)))))))
       :where
         (is-test-feat (fcn)
	    (tuple 'is-tester fcn)))))
			      
(defun typename-is-builder-name (typename env)
   (cond (typename
	  (let-fun ()
	     (cond ((is-symbolish typename)
		    (is-builder-sym typename))
		   (t
		    (is-builder-sym (car typename))))
	   :where
	      (is-builder-sym (name)
		 (let ((ns (env-find-namespace env)))
		    (build-symbol (:< (local-name name ns true)) -is-builder)))))
	 (t
	  (gensym))))

(defun is-builder-define (ib-name form params)
   ;; We've smuggled the total number into every param, so we just pluck it out
  ;; of the first
  (let ((numparms (caddr (car params))))
     (let ((arg-is-testers
	      (<# (\\ (i) (build-symbol |IS | (:< i)))
		  (series 0 numparms))))
	(note-aux-def 'fun
	   `(,ib-name ,arg-is-testers
	       (fill-is-holes '(lambda (x) ,form)
			      (vector ,@arg-is-testers)))))))

;;; This belongs elsewhere (must be available at run time)

(defun fill-is-holes (form is-tester-vec)
		(letrec ((fill-the-holes (form)
			    (cond ((atom form) form)
				  ((eq (car form) 'is-hole)
				   (aref is-tester-vec (cadr form)))
				  (t
				   (<# fill-the-holes form)))))
		   (fill-the-holes form)))

;;; Create the body of a function that, applied to a vector of argtypes, returns
;;; an is-tester.
;;; Also return the subscripts of argtypes that are actually seen
;;; and true if x occurs in exp.				       
(defun is-body-resolve (exp vartypes)
   (letrec ()
      (cond ((or (atom exp)
		 (eq (car exp) 'quote))
	     (values exp '() (eq exp 'x)))
	    ((eq (car exp) 'is)
	     (cond ((is-symbolish (cadr exp))
		    (let ((p (var-lookup (cadr exp) vartypes)))
		       (cond ((not p)
			      (error-break is-body-resolve :fatal
				 "Undefined type in `is' form: " exp)))
		       (cond ((eq (Vartype-type p) type-type*)
			      (cond ((is-param-type (Vartype-initial p))
				     ;;; We have to postpone
				     (let ((c (vartype-feat p
							    'tyfun-context)))
					(cond (c
					       (multiple-value-let
							  (a pl h)
							  (is-body-resolve
							      (caddr exp)
							      vartypes)
						  (cond ((eq a '*untestable)
							 (values '*untestable
								 '() false))
							(t
							 (values `((is-hole
								    ,(cadr c))
								   ,a)
								 (cons c pl)
								 h)))))
					      (t
					       (error-break is-body-resolve
							    :fatal
						  "Unrecognizable parameter"
						  " in `is' form: "
						  exp)))))
				    (t
				     (easy-is (Vartype-initial p)))))
			     (t
			      (error-break is-body-resolve :fatal
				 "Nontype in `is' form: " exp)))))
		   (t
		    (easy-is (designated-type (cadr exp) true vartypes)))))
	    (t
	     (do ((al exp (cdr al))
		  (res '())
		  (params '())
		  (has-an-x false))
		 ((or (null al) (eq res '*untestable))
		  (values (cond ((eq res '*untestable) '*untestable)
				(t (reverse res)))
			  params has-an-x))
	       (multiple-value-let (r p h)
				   (is-body-resolve (car al) vartypes)
		  (cond ((eq r '*untestable)
			 (!= res '*untestable))
			(t
			 (!= res `(,r ,@*-*))
			 (!= params (append p *-*))
			 (cond (h (!= has-an-x true)))))))))
    :where
      (easy-is (ty)
	 (let ((isf (get-is-tester ty false vartypes '())))
	    (cond ((eq isf '*untestable)
		   (values '*untestable '() false))
		  (t
		   (multiple-value-let (a pl has-an-x)
				       (is-body-resolve
					  (caddr exp) vartypes)
		      (cond ((eq a '*untestable)
			     (values '*untestable '() false))
			    (t
			     (values 
				(build-funform isf (list a))
				pl has-an-x))))))))))

(defun patches-class (patches which-sort vartypes)
   (repeat :for ((p :in patches))
    :result (cond ((eq which-sort ':modified)
		  modified-class*)
		 (t
		  subtype-class*))
    :until (car-eq p ':class)
    :result (letrec ((better-be-class (c)
		       (cond ((is-Type-class c)
			      c)
			     (t (error-break patches-class :fatal
				   "Clump class specifier " (cadr p)
				   " refers to " c :% " which is not a class")))))
	      (let ((c (var-lookup (param-name (cadr p))
				   vartypes)))
	         (cond (c (better-be-class c))
		       (t
			(error-break patches-class :fatal
			   "Undefined class " (cadr p) " in Clump")))))))

(defun feature-decipher (l vartypes)
;;;;   (dbg-save l vartypes)
;;;;   (breakpoint feature-decipher
;;;;      "vartypes = " vartypes)
   (cond ((eq (car l) ':value)
	  (eval (cadr l)))
	 ((eq (car l) ':type)
	  (designated-type (cadr l) false vartypes))
	 ((or (is-Number (car l))
	      (is-String (car l)))
	  (car l))
	 (t
	  (error-break feature-decipher :fatal
	     "Indecipherable feature spec: " l))))

(defvar deftype-slurping* false)

(datafun :slurp-types def-type-fun
   (defun :^ (dtform _)
      (match-cond dtform
	  (:? (def-type-fun ?name ?@stuff)   ;?args ?type-exp
	    (bind ((aux-defs* ':not-allowed)
		   (deftype-slurping* true)
		   (eval-aux-defs* true))
	       (multiple-value-let (name params sys type-exp)
				   (def-type-fun-parse name stuff)
		  (cond (+namespace-mode+
			 (!= name (symbol-env-place
				     *-* (place-type-system-env sys)))))
		  (typedecl
		     name
		     tyfun-type*
		     (cond ((eq type-exp ':forward)
			    (new-Tyfun name params 'forward nil))
			   (t
			    (bind ((top-level-type-system* sys))
			       (let ((defn (typefun-body-designation
					      name type-exp params
					      (place-type-system-env sys))))
				  (!= (type-feature defn 'is-builder-name)
				      '*untestable)
				  (new-Tyfun name params defn
					     (empty-vartypes sys))))))
		     sys))))
	  (t
	   (error-break def-type-fun :fatal
	      "Ill-formed: " dtform)))
     false))

(defmacro def-type-fun (name &rest stuff)
   (bind ((aux-defs* '()))
      (multiple-value-let (name params sys type-exp)
			  (def-type-fun-parse name stuff)
	 (cond (+namespace-mode+
		(!= name (symbol-env-place
			    *-* (place-type-system-env sys)))))
	 (tyfun-forward-define name params sys)
	 (cond ((eq type-exp ':forward)
		`(tyfun-forward-define ',name ',params
					 ,(type-system-loader sys)))
	       (t
		;; A bit kludgy, but ...
		(bind ((top-level-type-system* sys))
		   (let ((defn (typefun-body-designation
				   name type-exp params
				   (place-type-system-env sys)))
			 (ae (type-system-acc-env sys))
			 (tyf (var-val name (place-type-system-env sys))))
		      (cond ((eq (Tyfun-defn tyf) 'forward)
			     (!= (Tyfun-defn tyf) defn))
			    (t
			     (out (:to *error-output*)
				 "Tyfun " name " has unexpected definition:"
				 :% (Tyfun-defn tyf))))
		      (multiple-value-let (isb is-builder-name)
					  (sort-out-is-builder
					     tyf params defn sys)
			 (!= (type-feature defn 'is-builder-name)
			     (or is-builder-name '*untestable))
			 ;(dbg-save defn is-builder-name)
			 ;(breakpoint def-type-fun
			 ;  "Got defn and is-builder for " name)
			 (let ((defn-loader
				  (bind ((acc-env* ae))
				     `(circular-object-restore
					 ,(feature-add-loader
					     (type-loader defn '())
					     'is-builder-name
					     is-builder-name)
					 type-re-cycle))))
			    ;(dbg-save defn-loader)
			    ;(error-break def-type-fun :novalue
			    ;   "Got: " defn-loader)
			    `(progn
				(eval-when (:load-toplevel :execute)
				   (typedecl ',name
					     tyfun-type*
					     (new-Tyfun
					        ',name
						(circular-object-restore
						   ,(bind ((acc-env* ae))
						       (arglistspec-loader
							  params '()))
						   arglistspec-re-cycle)
						,defn-loader
						(empty-vartypes
						   ,(type-system-loader
						       sys)))
					     ,(type-system-loader sys)))
				,@(aux-defs-global aux-defs*)
				,@isb))))))))))

(defun def-type-fun-parse (funname stuff)
   (multiple-value-let (funname stuff sys)
		       (extract-sys funname stuff)
      (let ((env (empty-vartypes sys)))
	 (multiple-value-let (params definiens)
			     (cond ((null stuff)
				    (values '() 'Obj))
				   ((eq (car stuff) ':forward)
				    (values '() ':forward))
				   ((null (cdr stuff))
				    (values (car stuff) 'Obj))
				   (t
				    (let ((definiens (type-def-parse (cdr stuff) funname)))
				       (values (car stuff)
					       definiens))))
	    (multi-let (((params _ _ _ synerrs)
			 (lambdalist-parse params true type-type* true env)))
	       (cond ((not (null synerrs))
		      (note-defective-exp
		         ((_) "Syntax errors: " synerrs)
			 (:continue "I will ignore the syntax errors"))))
	       (cond ((not deftype-slurping*)
		      (tyfun-defaults-parse params env)))
	       (match-cond definiens
		  (:? (subtype ?d)
		    (!= definiens `(clump :subtype ,d))))
	       (values funname params sys
		       definiens))))))

(defun tyfun-defaults-parse (specs svartypes)
      (repeat :for ((spec :in (Arglistspec-argspecs specs)))
	 (cond ((not (memq (Argspec-mode spec)
			   '(:required :rest)))
		(cond ((Argspec-default spec)
		       (!= (Argspec-default spec)
			   (designated-type *-* true svartypes))))))
	       ;; Note that default is either false (missing)
	       ;; or must be a type.
	 (let ((aname (Argspec-name spec))
	       (atype (Argspec-type spec)))
	    (cond ((type-feature atype 'meta)
		   (!= svartypes
		       (cons (new-Vartype aname
					  atype
					  (make-param-type
					     aname
					     (cond ((eq (type-feature
							   atype 'meta)
							'type)
						    atype)
						   (t type-type*))))
			     *-*)))))))

(defun typefun-body-designation (funname type-exp params env)
   (let ((body-env (argspecs-declare-desig-unks
		           ;;argspecs-declare-placeholder-params
		      params env)))
      (let ((param-vtl (Env-vartypes body-env)))
	 (let ((l (- (len param-vtl) 2)))
	    (repeat :for ((vt :in param-vtl)
			 (i :in (series 0 l)))
	       (!= (vartype-feat vt 'tyfun-context)
		   `(,vt ,i ,l ,(build-symbol is " " (:< i))))))
	 (let ((base (designated-type-with-name
		        type-exp false `(,funname ,@(arglistspec-arglist
						      params))
				       body-env))
	       (typefun-params
		  (list (make-param-type 'defined-fun-param)
			(make-param-type 'defined-fun-args-param))))
            (!= (Type-features base)
		(nconc (list (tuple 'defined-fun
				    (make-type-closable
				       (car typefun-params)
				       false))
			     (tuple 'defined-fun-args
				    (make-type-closable
				       (cadr typefun-params)
				       'param-close-elements)))
		       *-*))
	    (!= (Type-freevars base)
		(append typefun-params *-*))
	    base))))

;;; Return < code to define is-builder, whether there is one >
;;; defn has free params in general, and is not a Type-closure.
(defun sort-out-is-builder (tyf arglist defn sys)
   (let ((defn-is-builder
	    (type-local-feature defn 'is-tester)))
      (cond (defn-is-builder
	     ;; is clause in def created is-builder; don't need to include it
	     ;; again
	     (values '() defn-is-builder))
	    ((not (null (Arglistspec-rest arglist)))
	     (out (:to *error-output*) "Can't build is-tester for Tyfun "
		  (Tyfun-name tyf)
		  " with :rest args" :%)
	     (values nil '*untestable))
	    (t
	     ;; We can produce the is-builder by inspecting the
	     ;; definition
	     (typefun-is-builder-definition
		tyf arglist defn (place-type-system-env sys))))))

;;; Returns < code, is-builder-name >
;;; where is-builder-name *untestable if code doesn't actually define anything.
(defun typefun-is-builder-definition (tyf params defn outer-env)
   (or (slotfn-deint
	  (type-find-feature (Tyfun-defn tyf) 'is-tester outer-env))
       (multi-let (((dummy-vartypes tester-names)
		    (is-testable-arg-vartypes params)))
	  (let ((body-env
		   (env-bindings-append
		      true dummy-vartypes outer-env))
		(tyfname (Tyfun-name tyf)))
	     (let ((is-builder-name
		      (typename-is-builder-name tyfname outer-env)))
	        (let ((testbod (typefun-body-is-tester
			       defn body-env
			       tyf is-builder-name dummy-vartypes)))
		   (cond ((or (not testbod)
			      (eq testbod '*untestable))
			  (values `((eval-when (:load-toplevel :execute)
				      (declare-typefun-untestable
					 ,tyfname ,(type-system-loader
						      (vartypes-globals
							 outer-env)))))
				  '*untestable))
			 (t
			  (!= (type-feature defn 'tester-names) tester-names)
			  (values
			     `((eval-when (:load-toplevel :execute)
				 (defun ,is-builder-name
					,(<# (\\ (tn vt)
						(or tn (Vartype-var vt)))
					     tester-names
					     dummy-vartypes)
				    #',testbod)))
			     is-builder-name)))))))))
			        
(defun is-testable-arg-vartypes (params)
   (let ((paired-vts-and-testers
	    (repeat :for ((a :in (Arglistspec-argspecs params)))
	     :collect
		  (let ((aname (Argspec-name a))
			(atype (Argspec-type a)))
		     (let ((tester-name
			      (cond ((type-feature atype 'meta)
				     (build-symbol is " " (:< aname)))
				    (t
				     false))))
			(tuple (new-Vartype
				  aname
				  atype
				  (cond (tester-name
					 (new-Type false
					    `(is-testable ,a)
					    '() univ-type* '()
					    (list (tuple
						     'is-tester
						     `(lambda (x)
							 (funcall
							    ,tester-name
							    x))))))
					(t
					 ;; This is extremely dubious,
					 ;; but at least has the
					 ;; right number of args.
					 (unknown-constrained-type
					    (make-param-type
					       (Argspec-name a))
					    a
					    (Type-kind atype)
					    false))))
			       tester-name))))))
     (values (<# car paired-vts-and-testers)
	     (<# cadr paired-vts-and-testers))))

(defvar warn-about-is-builders* false)

(defun typefun-body-is-tester (ty vartypes tyf is-builder-name param-vts)
   (multiple-value-let (ty vartypes)
		       (follow-var-ref ty vartypes)
      (let ((b (type-find-feature ty 'is-builder vartypes)))
	 (cond ((and b (not (eq b '*untestable)))
		(build-is-tester b ty false vartypes
				 (list (make-Recursive-type-frame
					  (make-typeform
					     tyf param-vts)
					  vartypes
					  is-builder-name))
				 true))
	       (t
		(dbg-out warn-about-is-builders*
		     "Warning: No is-builder for " ty :%)
		'*untestable)))))

(defmacro declare-typefun-untestable (name sys)
;   (cond ((not sys)
;	  (!= sys top-level-type-system*)))
   `(eval-when (:load-toplevel :execute)
      (note-typefun-untestable ',name ,sys)))

(datafun :slurp-types def-type
   (defun :^ (dform _)
      (match-cond dform
	  (:? (def-type ?typename ?@def)
	    (bind ((deftype-slurping* true)
		   (eval-aux-defs* true)
		   (aux-defs* ':not-allowed))
	       ;(out "Slurping def-type " typename 1 def
		;    :% " with type-sys = " top-level-type-system* :%)
	       (multiple-value-let (typename sys definiens)
				   (def-type-parse typename def)
		  (cond (+namespace-mode+
			 (!= typename (symbol-env-place
					 *-* (place-type-system-env sys)))))
		  (bind ((top-level-type-system* sys))
		     (forward-define typename sys)
		     (cond ((not (eq (car def) ':forward))
			    (let ((ty (defined-type-designation
					  typename definiens sys)))
			       (cond ((not (is-symbolish (Type-desig ty)))
				      (!= (type-feature ty 'forward) true)))
			       (global-type-define
				  typename ty sys))))))))
       (t
	(error-break def-type-to-slurp :fatal
		  "Ill-formed: " dform)))
     false))

(defmacro def-type (typename &rest def)
   (multiple-value-let (typename sys definiens)
		       (def-type-parse typename def)
      (cond (+namespace-mode+
	     (!= typename (symbol-env-place
			     *-* (place-type-system-env sys)))))
      (cond ((eq (car def) ':forward)
	     `(forward-define ',typename ,(type-system-loader sys)))
	    (t
	     (bind ((aux-defs* '())
		    (top-level-type-system* sys))
		(forward-define typename sys)
;;;		(dbg-save typename)
;;;		(breakpoint def-type
;;;		   "Ready to find designation of " definiens)
		(let ((ty (defined-type-designation
			     typename definiens sys)))
		   (cond ((not (is-symbolish (Type-desig ty)))
			  (!= (type-feature ty 'forward) true)))
;;;		   (out "Typename = " typename :%)
;;;		   (dbg-save ty)
;;;		   (breakpoint def-type
;;;		      "Ready to generate loader")
		   (let ((super-loader
			    (bind ((acc-env*
				      (type-system-acc-env sys)))
			       (cond ((and (is-symbolish (Type-desig ty))
					   (not (eq (Type-desig ty)
						    typename)))
				      `(circular-object-restore
					  ,(type-loader ty '())
					  1-arg-ignore))
				     (t
				      `(maybe-change-name
					  ',typename
					  (circular-object-restore
					     ,(type-guts-loader ty '())
					     type-guts-re-cycle)))))))
;;;		      (cond ((eq typename 'Action)
;;;			     (dbg-save ty)
;;;			     (breakpoint def-type
;;;				"Defined Action")))
		      (global-type-define typename ty sys)
		      (cond (now-loading*
			     (out (:to *error-output*)
				 "Warning-- def-type for " typename
				 " apparently unslurped" :%)))
		      `(progn (maybe-forward-define
				 ',typename ,(type-system-loader sys))
			      ,@(aux-defs-global aux-defs*)
			      (eval-when (:load-toplevel :execute)
				 ,(aux-defs-wrap aux-defs*
				     `(global-type-define ',typename
					  ,super-loader
					  ,(type-system-loader sys))))))))))))

(defun def-type-parse (typename def)
   (multiple-value-let (typename def sys)
		       (extract-sys typename def true)
      (let ((definiens (type-def-parse def typename)))
	 (values typename sys definiens))))

(defun type-def-parse (def typename)
   (cond ((null def)
	  (error-break type-def-parse
	     "Empty def in definition of " typename
	     (:novalue "I will assume you meant (Clump :subtype Obj)"))
	  '(Clump :subtype Obj))
	 (t
	  (let ((whether-sub false) definiens patches)
	     (cond ((memq (car def) '(:sub :subtype))
		    (!= whether-sub true)
		    (!= def (cdr def)))
		   ((memq (car def) '(:same =))
		    (!= def (cdr def))))
	     (!= definiens (car def))
	     (!= patches (cdr def))
	     (cond ((or whether-sub (not (null patches)))
		    (!= definiens `(Clump ,(cond (whether-sub ':subtype)
						 (t ':modified))
				    ,*-* ,@patches))))
	     definiens))))

(defun defined-type-designation (typename type-exp type-sys)
   (let ((ty (designated-type-with-name type-exp true typename
					(empty-vartypes type-sys))))
      (cond ((not (is-symbolish (Type-desig ty)))
	     (let ((cl (Type-class ty)))
		(cond ((eq ty univ-type*)
		       (dbg-save ty typename)
		       (breakpoint defined-type-designation
			  "About to set desig of " ty " to " typename)))
		(!= (Type-desig ty) typename)
;;;;		(out "defined-type-designation adding defined-fun feature"
;;;;		     :%)
		(!= (Type-features ty)
		    (nconc (cond ((or (eq cl subtype-class*)
				      (eq cl modified-class*))
				  (list (tuple 'defined-fun-args '())
					(tuple 'defined-fun ty)))
				 (t '()))
			   (list (tuple 'home (ensure-top-level-type-sys)))
			*-*)))))
      ty))

;;;(defmacro forward-define (typename)
;;;   `(eval-when (:compile-toplevel :load-toplevel)
;;;        (define-global-constant-type ',typename 'forward)))

(defun aux-defs-wrap (aux-defs form)
   (let ((funs '())
	 (vars '()))
      (repeat :for ((a :in aux-defs))
	 (cond ((not (is-Aux-def a))
		(error-break defs-wrap
		   "Bad aux-def: " a
		   (:novalue "It will be ignored")))
	       ((eq (Aux-def-kind a) 'fun)
		(!= funs (cons (Aux-def-def a) *-*)))
	       (t
		(!= vars (cons (Aux-def-def a) *-*)))))
      (cond ((not (null funs))
	     (!= form
		 `(let-fun ,funs ,form))))
      (cond ((not (null vars))
	     (!= form
		 `(let-var ,vars ,form))))
      form))
				   
(defun feature-add-loader (type-loader featname featval)
   (let ((typevar (gensym)))
      `(let ((,typevar ,type-loader))
	  (!= (type-feature ,typevar ',featname)
	      ,(thing-loader featval '()))
	  ,typevar)))

;;; A key point about 'def-type-constructor' is that it creates two kinds of
;;; terms: (C y) is a type if y is, 
;;; and (make C e) is a term of type (C y) if e is of type y.  

;;; In the Lisp context, the object produced by (make C e) is actually
;;; an object produced by (make-C1 e) for some defstruct C1, although
;;; some care must be taken if the definition of the type constructor
;;; is _not_ wrt the global env, which is impossible for
;;; def-type-constructor.

(datafun :slurp-types def-type-constructor
   (defun :^ (dtcform slurp-if-substantive)
      (match-cond dtcform
	 (:? (def-type-constructor ?name ?params ?@stuff)
	   (bind ((auxdefs* ':not-allowed))
	      (multi-let (((name high-alspec sys)
			   (def-type-constructor-parse name params stuff)))
		 (cond (+namespace-mode+
			(!= name (symbol-env-place
				    *-* (place-type-system-env sys)))))
		 (let ((cspec (make-Constructor-spec
				 false high-alspec)))
		    (let ((tyf (new-Tyfun name high-alspec cspec
					  (empty-vartypes sys))))
		       (!= (Constructor-spec-tyfun cspec) tyf)
		       (typedecl name
				 constructor-type*
				 tyf
				 sys)
;;;;			      (dbg-save name tyf sys)
;;;;			      (breakpoint def-type-constructor-to-slurp
;;;;				 "Defined " name)
)))))
	 (t
	  (signal-problem def-type-constructor-to-slurp :fatal
	     "Ill-formed: " dtcform)))
      true))

(defmacro def-type-constructor (name params &rest stuff)
   (bind ((aux-defs* '()))
      (multi-let (((name high-alspec sys)
		   (def-type-constructor-parse name params stuff))
		  (sys-var (gensym))
		  (env-var (gensym))
		  (ctyf-var (gensym))
		  (lname nil))
	 (cond (+namespace-mode+
		(!= name (symbol-env-place
			    *-* (place-type-system-env sys)))
		(!= lname (local-name name (Type-system-namespace sys) true)))
	       (t
		(!= lname name)))
	 (let ((ae (type-system-acc-env sys))
	       (cspec (make-Constructor-spec false high-alspec))
	       (tyfun-name (build-symbol (:< lname) :-tyfun*))
;;;;	       (param-var (build-symbol (:package false) param- (++ symno*)))
	       (params-without-types
		  (arglistspec-arglist high-alspec))
	       )
	    (bind ((acc-env* ae))
	       `(eval-when (:load-toplevel :execute :slurp-toplevel :compile-toplevel)
		   (let ((,sys-var ,(type-system-loader sys)))
		      (let ((,env-var (place-type-system-env ,sys-var)))
			 (let ((,ctyf-var
				(circular-object-restore
				      (new-Tyfun
					 ',name
					 ,(arglistspec-loader
						 high-alspec '())
					 ,(constructor-spec-loader
						cspec '())
					 (empty-vartypes ,sys-var))
				   constructor-tyfun-circularize)))
			    (typedecl ',name
				      constructor-type*
				      ,ctyf-var
				      ,sys-var)
			    (defvar ,tyfun-name (var-val ',name ,env-var))
			    (defun ,(build-symbol make- (:< lname) -type)
				   ,params-without-types
			       (make-constructed-type
				   ,tyfun-name         ;;;; ,param-var
				   ,(arglistspec-arg-lister high-alspec)
				   ,env-var))
			    (defun ,(build-symbol is- (:< lname) -type)
				   (ty)
			       (and (is-constructed-type ty)
				    (eq (constructed-type-tyfun ty ,env-var)
					,tyfun-name))))))))))))

(defun arglistspec-arg-lister (alspec)
   (let-fun ()
      (let ((restargs (Arglistspec-rest alspec)))
	 (cond ((null restargs)
		`(list ,@(non-rest-args-lister)))
	       (t
		(let ((nrl (non-rest-args-lister)))
		   (cond ((null nrl)
			  (Argspec-name (car restargs)))
			 (t
			  `(nconc ,nrl
				  ,(Argspec-name (car restargs)))))))))
    :where
       (non-rest-args-lister ()
	  `(,@(<# Argspec-name (Arglistspec-required alspec))
	    ,@(<# Argspec-name (Arglistspec-optional alspec))
	    ,@(<# Argspec-name (Arglistspec-key alspec))))))

(defun def-type-constructor-parse (name params stuff)
   (multi-let (((_ _ sys)
		(extract-sys false stuff)))
      (multi-let (((alspec synerrs)
		   (high-funargs-argspecs params (place-type-system-env sys))))
	 (cond ((not (null synerrs))
		(note-defective-exp
		   ((_) "Syntactic errors in definition of "
			name ": " synerrs)
		   :place def-type-constructor-parse
		   (:continue "I will ignore them"))))
	 (values name alspec sys))))



