;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: desig.lisp,v 1.9 2004/07/26 04:44:55 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat %langutils/ namespace)

(depends-on :at-run-time %nity/ subtype istester argspair parmsgen
			 %langutils/ synutils)

;;; Slurp to get macro definitions
(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(is-type-desig designated-type
	     designated-type-with-name ;;;;designated-type-with-tvars
	     is-forward-defined symbol-namespace-or-env-place symbol-env-place
;;;;	     exp-env-symbols-place exp-namespace-symbols-place
	     get-namespace env-find-namespace finish-constructed-type
	     var-place-and-lookup rest-args-prefs)))

(in-type-system universal)

#|
The designatee of an expression is its internal representation, which
is prior to any type checking, and hence prior to any constraints on
vars.  That's because the only designators that ever occur occur in
user code.  (This includes all names appearing in typefun
definitions.)  These can be transformed into internal form before any
type checking.  (At "type-designation time")

A type designator is a symbol, or an expression of the form 
   (tf -args-).
tf is a symbol, the name of either a type-former or a Tyfun, 

The args are typically types, but don't have to be.

To find the designatee: 

- of a symbol: however it was defined.
- of a form (type-former ...): call the type-former definition (a Lisp function)
- of a form (f -args-):  
  Two subcases: 
     - f is a *constructor-type*, such as Cons.  Produce a type 
       of constructed-class 



the designatee is a typeform-type 
   (see vartypes.nsp) whose
   fun is the designatee of f and whose argtypes are the designatees of args.
   fun is a Tyfun, whose defn contains param-types corresponding to its
      parameters.
   Most, but not all, operations on typeforms require that the typeform
   be *expanded*, by closing the body of the Tyfun namd by f in an
   environment that binds the param-types to values.

Note that one difference between a named type and a type function of
no arguments is that the latter is allowed to be recursive.  The
function isn't expanded until we ask about the type's internal
structure (by asking for its expansion property; see types.nsp).

type-formers take three arguments:
the designator, a typename, and an environment.
The typename is usually false, but if non-false it indicates that the 
designator occurred in a DEFTYPE, as in (DEFTYPE typename designator).
In this case, the type-former for the designator is allowed to define
auxiliary functions and to have their names appear in the type
expression, subject to these restrictions:

  -- The global variable DEFTYPE-SLURPING* must be non-()
  -- The auxiliary function is defined by adding a form defining it to the
     list DEFTYPE-FUNS*.  Actually, any evaluable expression can be
     added to this list, which will appear as part of the expansion of
     the DEFTYPE macro (before anything else).
  -- It is the responsibility of the type loader to make sure that the
     functions have unique names.  Currently only STRUCTURE and CLUSTER
     define auxiliary functions, and they are careful to obey this restriction.

Very few type-formers care about the typename argument.  If one writes
(def-type foo (Lst Integer))
the system will find the type designated by (Lst Integer).  This is a type 
with designator (Lst Integer); foo is ignored.   Of course, def-type may
reinstate it.

When finding the designatee of something like (Lst Symbol), at 
where Lst is a Tyfun with params (p) and skel
(Alt Null (Dot p (Lst p)))
at designator-expansion time p must be bound to a param-type.
Later, when someone examines the designatee, they will see a typeform-type
whose expansion is an Alt type, one of whose alt-types is a Dot type,
whose cartype feature is a param-type with name p.  This param-type is
bound to the designatee of Symbol.

E.g., in processing a function definition (\\ 1 - r (a b - t) body),
we internalize (a b - t) first.  Then we process the body, with
a and b bound to param-types a and b, with designators !/a and !/b.
When we come across an occurrence of a in a type designator in body,
it gets internalized to !/a.  Hence the type of body will include
occurrences of !/a and !/b. 
Later, when we apply the function to types, such as (Integer String),
we bind a to Integer, b to String.  The type of the resulting occurrence
of body is then parametrized with !/a and !/b standing for Integer and
String.

Designators are not in the same name space as params, but they're both
stored as vartypes.  There is no separate
designator-resolution phase.  Resolution can happen during type checking.
That means that the vartypes will include both bindings of symbols that 
haven't been internalized yet, and bindings of params.

Convention (not that well obeyed): Often there are two functions for 
building a type of some class.  One is called 'build-xxx-type' and the 
other is called 'make-xxx-type'.  The "build" version is for designator
resolution.  It assumes that some typenames will be bound to param-types 
(the types that print out as #{Type !/var}).  When a type designator
is replaced by an actual type, these names will be replaced by the
param types.  However, the resulting types are useless unless (a) they
are accessed in an environment that binds the params ; (b) they are
wrapped as Closables of some kind.  This is especially true for type
features containing param types.  These should always be Closables with
a closer that knows how to find where the types are inside them.

The term "closand" is used occasionally to mean a type that needs to be
closed (because it may contain param-types).
|#


(defconstant embrace-unfamiliar-symbols* true)


(defun td-massage (d) d)  ; for now

(defun is-type-desig (x vartypes)
   (let ((ns (and +namespace-mode+ (env-find-namespace vartypes))))
      (letrec ((declared-meta (a kinds)
;;;;		  (trace-around declared-meta
;;;;		     (:> "(declared-meta: " a 1 kinds")")
		  (let ((res
			   (and (is-symbolish a)
				(let ((real-a (symbol-namespace-or-env-place
					         a ns vartypes)))
;;;;				   (dbg-save a real-a ns vartypes)
;;;;				   (breakpoint is-type-desig
;;;;				      "real-a = " real-a)
				   (let ((e (var-lookup real-a vartypes)))
				      (and e
					   (memq (type-feature (Vartype-type e) 'meta)
						 kinds)))))))
;;;;		     (cond ((and (not res) (eq a 'opt::Prop))
;;;;			    (dbg-save vartypes)
;;;;			    (breakpoint is-type-desig
;;;;			       "Prop is not a type!")))
		     res)
;;;;		     (:< (val &rest _) "declared-meta: " val))
		  ))
	 (cond ((is-Qvar x)
		(declared-meta (Qvar-sym x) '(type typefun)))
	       ((atom x)
		(declared-meta x '(type typefun)))
	       ((atom (car x))
		(declared-meta (car x) '(typefun)))
	       (t false)))))

;;; This is for debugging only.  We *never* use type names for types any more
(defun coerce-to-type (x vartypes)
   (cond ((is-Type x) x)
	 ((is-type-desig x vartypes)
	  (designated-type x true vartypes))
	 (t
	  (error-break coerce-to-type :fatal
	     "Can't coerce to type: " x))))

;;; Returns < type, whether resolved >
(defun designated-type (d must-resolve vartypes)
   (designated-type-with-name d must-resolve false vartypes))

(defun find-new-tvars (e vartypes)
   (cond ((is-Qvar e)
	  (let ((vt (var-lookup (Qvar-sym e) vartypes)))
	     (cond (vt '())
		   (t (list (Qvar-sym e))))))
	 ((atom e) '())
	 (t
	  (<! (\\ (x) (find-new-tvars x vartypes))
	      e))))

;;; In some contexts we can "defer" resolving a designator until it is
;;; better defined.  must-resolve=false if we're in such a context.
(defun designated-type-with-name (d must-resolve typename vartypes)
   (let ((rty 
	    (cond ((is-symbolish d)
		   (resolve-type-fun d '() must-resolve typename vartypes))
		  ((is-Qvar d)
		   (cond ((eq (Qvar-sym d) '_)
			  (values any-tvar* true))
			 (t
			  (resolve-type-fun (Qvar-sym d) '()
					    must-resolve typename vartypes))))
		  ((atom d)
		   (error-break designated-type-with-name :fatal
		      "Illegal in context requiring type: " d))
		  ((is-symbolish (car d))
		   (resolve-type-fun (car d) (cdr d) must-resolve typename vartypes))
		  (t
		   (error-break designated-type-with-name :fatal
		     "Ill-formed designator: " d)))))
     (cond ((and (listp (Type-desig rty))
		 (exists (x :in (Type-desig rty))
		    (car-eq x 'Unk)))
	    (dbg-save :run-loud d vartypes rty)
	    (breakpoint designated-type-with-name
	       "Deviant designatee: " rty)))
     rty))

;;; < s yields supertype.
(defun qvar-notes-parse (notes vartypes)
   (multiple-value-let (unique-sub notes)
		       (cond ((memq '* notes)
			      (values true (remove '* notes)))
			     (t
			      (values false notes)))
      (values (match-cond notes
		 ?((< ?sup)
		   (designated-type sup false vartypes))
		 (t univ-type*))
	      unique-sub)))

;;; Returns < result, whether-resolved >
;;; If not resolved, it's "deferred, " meaning 
;;; we return a suspended typeform, the details of which
;;; won't be known until fn is defined.
(defun resolve-type-fun (fn args must-resolve typename vartypes)
;;;;   (cond ((eq fn 'obj)
;;;;	  (breakpoint resolve-type-fun
;;;;	     "Instance of 'obj' encountered"
;;;;	     :% " args = " args " typename = " typename
;;;;	     :% " vartypes = " vartypes)))
   (let ((ns (env-find-namespace vartypes)))
      (!= fn (symbol-namespace-or-env-place
		fn ns vartypes))
      (let ((p (var-lookup fn vartypes)))
	 (cond (p
		(let ((def (Vartype-initial p)))
		   (cond ((eq (Vartype-typ p) type-former-type*)
			  (cond ((and (consp def)
				      (eq (car def) 'type-former))
				 (let ((formfun (coerce-to-fun (cadr def))))
				    (let ((ty (funcall formfun
						    args typename vartypes)))
				       (values ty true))))
				(t
				 (error-break resolve-type-fun :fatal
				    "Misdefined type-former " (Vartype-var p)))))
			 ((or (eq (Vartype-type p)
				  tyfun-type*)
			      (eq (Vartype-type p)
				  constructor-type*))
			  (values (designated-typeform fn def args vartypes)
				  true))
			 ((type-feature (Vartype-type p) 'meta)
			  (cond ((is-Type def)
				 (!= def (check-for-unknown-param
					    def false fn))
				 (cond ((not (null args))
					(note-defective-exp
					   ("Type " fn " occurs with args " args)
					   (:novalue "I will ignore the args"))))
				 def)
				((and (not must-resolve) (eq def 'forward))
				 (make-param-type fn))
				(t
				 (note-defective-exp 
				    ("Bogus type in non-deferable context: "
				     fn " [=" def "]")
				    :place resolve-type-fun :fatal))))
			 (t
			  (note-defective-exp 
				 ("Illegal function in type designator " fn
				  " -- type = " (Vartype-type p))
				:place resolve-type-fun :fatal)))))
	       ((and (not must-resolve) embrace-unfamiliar-symbols*)
		(build-typeform fn args vartypes))
	       (t
		(note-defective-exp ("Undefined in type: " fn)
		   :place resolve-type-fun :fatal))))))

;;; There must be a more revealing name for this function.
(defun check-for-unknown-param (def require-desig fn)
   (cond ((is-unknown-type def)
	  (let ((p (type-feature def 'param))
		(d (or (not require-desig)
		       (type-feature def 'desig))))
	     (cond ((or (not p) (not d))
		    (error-break resolve-type-fun
		       "Type desig " fn
		       " resolves to"
		       " rogue unknown type "
		       def)))
	     p))
	 (t def)))

(defun designated-typeform (fun-name tyfun arg-desigs vartypes)
   (let ((tdefn (Tyfun-defn tyfun)))
      (cond ((is-Constructor-spec tdefn)
	     (make-constructor-typeform tdefn fun-name arg-desigs vartypes))
	    (t
	     (make-typeform tyfun
			    (tyfun-args-resolve tyfun arg-desigs vartypes
						`(,fun-name ,@arg-desigs)))))))

(defun make-constructor-typeform (cspec _ arg-desigs vartypes)
  (let ((cparms (Constructor-spec-type-params cspec))) 
      (multi-let (((vtl _ _)
		   (tyfun-alspec-args-resolve cparms arg-desigs true vartypes
					      `(,(Tyfun-name
						    (Constructor-spec-tyfun cspec))
						,@arg-desigs))))
	 (finish-constructed-type cspec vtl vartypes))))	    

(defun finish-constructed-type (cspec arg-vartypes vartypes)
   (let ((tyfun (Constructor-spec-tyfun cspec))
	 (params (Constructor-spec-type-params cspec))
	 (freevars (nodupq
		      (<$ (\\ (aty)
			     (let ((v (Vartype-val aty)))
			        (cond ((is-Type v)
				       (type-find-freevars v vartypes))
				      ((is-list-of v #'is-Type)
				       (<$ (\\ (ety)
					      (type-find-freevars ety vartypes))
					   v))
				      (t !()))))
			  arg-vartypes))))
      (cond ((not (is-list-of freevars #'is-param-type))
	     (dbg-save freevars cspec params vartypes)
	     (signal-problem finish-constructed-type
		"Bogus freevars: " freevars)))
      (repeat :for ((vt :in arg-vartypes))
         (let ((aspec (arglistspec-find-argspec-with-name
		         params (Vartype-var vt)))
	       (atype (Vartype-val vt)))
	    (multi-let (((constraints found)
			 (argspec-feat aspec 'constraints)))
	       (cond ((and found
			   (is-Type atype) ;; sanity check
			   (exists (con :in constraints)
			      (ecase (car con)
				 ((< =<)
				  (not (type-acceptable
					  false
					  atype (cadr con)
					  (empty-undo-stack)
					  vartypes vartypes

					  !())))
				 ((> >=)
				  (not (type-acceptable
					  false
					  (cadr con) atype
					  (empty-undo-stack)
					  vartypes vartypes
					  !()))))))
		      (dbg-save :run-loud atype constraints vartypes)
		      (note-defective-exp
			 ((_) "Constructor "
			      (Tyfun-name tyfun)
			  :% " arg " aspec
			  :% " assigned type " atype
			  :% " fails to satisfy constraints "
			     constraints)
			 (:continue "I will overlook this fact"))))))
       :where
          (arglistspec-find-argspec-with-name (alspec name)
	     (repeat :for ((aspec :in (Arglistspec-argspecs alspec)))
	      :result (signal-problem finish-constructed-type
			 "Fumbled argspec named " name " in " alspec)
	      :until (eq (Argspec-name aspec) name)
	      :result aspec)))
      (let ((desig
	       `(,(Tyfun-name tyfun)
		 ,@(funcall (high-alspec-args-assembler params)
			    arg-vartypes))))
	 (make-type-of-class constructed-class
	    desig freevars univ-type*
	    (cons (make-Slot
		     'conser false false false
		     (closure-or-closable
			(arglistspec-make-more-specific params arg-vartypes)
			vartypes
			'arglistspec-close
			freevars)
		     false `(conser ,@(cdr desig)))
		  (constructed-type-slots arg-vartypes params vartypes))
	    (list (tuple 'constructor cspec)
		  (tuple 'arg-vartypes
			 (closure-or-closable
			    arg-vartypes vartypes 'vartypelist-close freevars)))))))

   ;;;;	       (tuple 'argtypes
   ;;;;		      (closure-or-closable 
   ;;;; 			 argtypes vartypes 'typelist-close freevars))


(defun constructed-type-slots (arg-vartypes alspec vartypes)
   (repeat :for ((aspec :in (Arglistspec-argspecs alspec)))
    :within
      (let ((parm (Argspec-name aspec)))
	 (let ((vt (vartypes-lookup parm arg-vartypes)))
	    (cond (vt
		   (let ((aty (Vartype-val vt)))
		      (cond ((is-Type aty)
			     (:continue
			      :collect (ctype-slot
					     parm aty (Argspec-position aspec)))))))
		  ((eq (Argspec-type aspec) type-type*)
		   (:continue
		    :collect (ctype-slot
				parm type-type* (Argspec-position aspec)))))))

    :where
       (ctype-slot (name ty pos)
	  (make-Slot name false false
		     (closure-or-closable ty vartypes false
					  (type-find-freevars
					     ty vartypes))
		     no-args-spec*
		     (make-Loadable-function-with-extras
		      :name 'constructed-type-pos-invert
		      :fcn false
		      :towit (list arg-vartypes pos))
		   `(,name ,pos)))))

(defun constructed-type-pos-invert (arg-vartypes pos type-params)
   (\\ (type args)
      (values true
	      (list (new-Vartype
		       (car args)
		       (finish-constructed-type
			  (type-feature type 'constructor)
			  (repeat :for ((vt :in arg-vartypes))
			   :collect
			     (let ((vtname (Vartype-var vt)))
				(new-Vartype 
				   vtname
				   (Vartype-type vt)
				   (repeat :for ((asp :in (Arglistspec-argspecs
							      type-params)))
				    :result (signal-problem constructed-type-pos-invert
					       "Can't find argspec with name "
					       vtname)
				    :until (eq (Argspec-name asp)
					       vtname)
				    :result (cond ((eql (Argspec-position asp)
							pos)
						   type)
						  (t (Vartype-val vt)))))))
			  global-env*)
		       false)))))

;;; Args are mostly type designators, but possibly other things as well.
;;; ur-env is env when typeform was constructed,
;;; which is necessary because the args haven't been internalized yet.
;;; This is the only time we need look directly at the arglistspec for 
;;; tyf; the results are summarized in Closables for tyfun-apply.
(defun tyfun-args-resolve (tyf arg-desigs ur-env tyform)
   (multi-let (((vtl freevars _)
		(tyfun-alspec-args-resolve
		   (Tyfun-params tyf) arg-desigs true ur-env tyform)))
      ;;(!= (Type-freevars tyform) freevars)
      (values vtl freevars)))

;;; Used in typecheck.
;;; Returns < vartypes, arg prefs, freevars >
(defun tyfun-alspec-args-resolve (alspec arg-desigs require-desigs env form)
   (multi-let (((pairs rest-pair patho)
	        (params-args-pair alspec arg-desigs))
	       (ns (and +namespace-mode+ (env-find-namespace env))))
      (cond ((not (null patho))
	     (note-defective-exp
		((_) "Syntax errors in resolution of args of " form ":"
		     :% arg-desigs
		     :% patho)
		:place tyfun-alspec-args-resolve
		(:proceed "I will try to ignore them"))))
      (repeat :for ((as :in (Arglistspec-all-nonrest-argspecs
			     alspec))
		    (freevars !())
		    :collector vtl prefs)
       :within
	 (let ((ap (assq as pairs)))
	    (cond (ap
		   (multi-let (((arg fvl)
				(tyfun-arg-check (cadr ap)
						 (car ap)
						 require-desigs
						 form ns env)))
		      (!= freevars (freevars-union fvl *-*))
		      (:continue
		       :collect (:into vtl arg)
		       :collect (:into prefs (argspec-arg-position (car ap))))))
		  ((eq (Argspec-mode as) ':required)
		   (note-defective-exp
		      ((tg) "Missing required arg " (Argspec-name as)
			    " in " tg)
		      :target form
		      (:proceed "I'll try to do without it")))))
       :result
	 (cond ((car rest-pair)
		(let ((rest-as (car rest-pair)))
		   (cond ((not (null (argspec-feat rest-as 'template-vars)))
			  (signal-problem tyfun-alspec-args-resolve
			     "Rest argspec with template vars in Tyfun: "
			     rest-as)))
		   (let ((rest-type (designated-type `(Val ,@(cadr rest-pair))
						     true
						     env)))
		      (values (nconc vtl
				    (list (new-Vartype
					      (Argspec-name rest-as)
					      type-type*
					      rest-type)))
			      (freevars-union freevars (Type-freevars rest-type))
			      (nconc prefs
				     (rest-args-prefs (cadr rest-pair) rest-as))))
		   ))
	       (t
		(values vtl freevars prefs))))))

(defun rest-args-prefs (args rest-argspec)
  (let ((first-rest-pos
	   (car (Argspec-position rest-argspec))))
     (series
	first-rest-pos
	(+ first-rest-pos (len args) -1))))

(defun argspec-arg-position (spec)
   (Argspec-position spec))
;   (cond ((eq (Argspec-mode spec) ':rest)
;	  '&rest)
;	 (t (Argspec-position spec))))

;(defun argspec-pref (as)
;   (cond ((eq (Argspec-mode as) ':key)
;	  `(,(Argspec-position as)))
;	 (t '())))

;;; Returns < vartype, freevars >
;;; where vartype gives name, type, and value of arg, and freevars
;;; gives its free params ('()' if not a type).
(defun tyfun-arg-check (arg-desig aspec require-desigs form ns env)
   (let ((atype (Argspec-type aspec))
	 (aname (Argspec-name aspec))
	 (is-a-qvar (is-Qvar arg-desig))
	 )
      (letrec ((check-const-arg (c)
		  (cond ((car-eq c 'quote)
			 (!= c (cadr c))))
		  (cond ((not (const-of-type c atype env))
			 (error-break tyfun-arg-check
			    "Bad as arg " aname
			    " in " form ":"
			    :% " Got: " c
			    :% " Wanted type: " atype
			    (:novalue
			     "I will proceed, but something will crash"))))
		  c))
	 (cond (is-a-qvar
		(!= arg-desig (Qvar-sym arg-desig))))
	 (cond ((eq arg-desig '_)
		(cond (is-a-qvar any-tvar*)
		      (t (check-const-arg '_))))
	       ((is-symbolish arg-desig)
		(!= arg-desig (symbol-namespace-or-env-place *-* ns env))
		(let ((vt (var-lookup arg-desig env)))
		   (cond (vt
			  (let ((vtype (Vartype-type vt))
				(vval (Vartype-val vt)))
			     (cond ((try-accept
				       vtype atype '() env '() '())
				    (cond ((type-feature vtype 'meta)
					   (cond (vval
						  (!= vval
						      (check-for-unknown-param
							 *-* require-desigs
							 arg-desig)))
;;;;						 (is-a-qvar
;;;;						  (!= vval
;;;;						      (make-tvar-type
;;;;						         arg-desig
;;;;							 vtype)))
						 (t
						  (!= vval
						      (make-param-type
						         arg-desig))))
					   (values
					      (new-Vartype aname vtype vval
;;;;						 (make-type-closable
;;;;						    vval false)
					      )
					      (type-find-freevars
							vval env)))
					  (t
					   (!= vval (check-const-arg vval))
					   (values
					      (new-Vartype
					         aname
						 (thing-kind
						    vval env univ-type*)
						 vval)
					      '()))))
				   (t
				    (error-break tyfun-arg-check
				       "Wrong type in arg " aname
				       " in " form ": "
				       :% " Got: " vtype
				       :% " Wanted: " atype
				       :novalue)))))
			 (t
			  (error-break tyfun-arg-check :fatal
			     "Unbound symbol " arg-desig " as arg "
			     (Argspec-name aspec) " in " form)))))
	       ((is-type-desig arg-desig env)
		(let ((aty (designated-type arg-desig true env)))
		   (values
		      (new-Vartype
		         aname
			 (Type-kind aty)
			 (make-type-closable aty false))
		      (Type-freevars aty))))
	       (t
		(let ((arg-val (check-const-arg arg-desig)))
		   (values
		      (new-Vartype
			 aname
			 (thing-kind arg-val env univ-type*)
			 arg-val)
			 ;(make-Closable arg-val 'no-close)
		      '())))))))

(def-op thing-kind (v env default)
   (multi-let (((v1 env1)
	        (follow-var-ref v env)))
      (cond ((eq v v1) default)
	    (t (thing-kind v1 env1 default)))))

(def-meth thing-kind ((ty Type) env _)
   (or (type-find-feature ty 'kind env)
       type-type*))

(defmacro def-type-former (sym &rest defn)
   (multiple-value-let (sym defn sys)
		       (extract-sys sym defn true)
      (let ((fun-name (cond ((eq sys universal-type-sys*)
			     (build-symbol (< sym) -type-former))
			    (t
			     (build-symbol (< sym) "/" (< (Type-system-name sys))
					   -type-former)))))
         (cond (+namespace-mode+
		(!= sym (symbol-env-place *-* (place-type-system-env sys)))))
	 `(progn (defun ,fun-name ,@defn)
		 (typedecl 
		    ',sym
		    type-former-type*
		    (tuple 'type-former #',fun-name)
		    ,(type-system-loader sys))))))

(defmacro syn-type-former (syn orig &rest sys)
   (multi-let (((orig _ sys)
		(extract-sys orig sys true)))
   `(typedecl
       ',syn
       type-former-type*
       (tuple 'type-former
	      (find-type-former ',orig ,(type-system-loader sys)))
       ,(type-system-loader sys))))

(defun form-type (former d typename vartypes)
   (cond ((eq (car former) 'type-former)
	  (funcall (cadr former) d typename vartypes))
	 (t
	  (eval (funcall (cadr former) d typename vartypes)))))

(defun arglistspec-make-more-specific (alspec vtl)
   (let ((new-argspecs
	    (<# argspec-copy (Arglistspec-argspecs alspec))))
      (repeat :for ((na :in new-argspecs))
	 (let ((vt (vartypes-lookup (Argspec-name na)
				    vtl)))
	    (cond ((and vt (is-Type (Vartype-val vt)))
		   (!= (Argspec-type na)
		       (Vartype-val vt))))))
      (new-Arglistspec new-argspecs)))

;;; We probably need local versions of these as well:

(defun is-forward-defined (typename sys)
  (let ((def (global-declaration typename sys false)))
     (and def
	  (eq (Vartype-typ def) type-type*)
	  (type-feature (Vartype-initial def)
			'forward))))
			
(defun not-allowed-in-def-type (typename former)
   (cond (typename
	  (error-break not-allowed-in-def-type 
	     "Attempt to define named type " typename 
	     " with type former " former
	     (:novalue
	      "I will allow it, but it's not necessarily meaningful")))))

#| These are too crude.  A better idea is to interleave symbol translation
   with whatever other processing we are doing.  (The main problem with these
   procedures is that they unnecessarily translate bound-variable names.)

(defun exp-env-symbols-place (e env)
   (declare (ignorable env))
   (cond (+namespace-mode+
	  (let ((ns (env-find-namespace env)))
	     (cond (ns
		    (exp-namespace-symbols-place e ns env))
		   (t
		    (let-fun ((elim-!syms (sub-e)
				 (cond ((atom sub-e) sub-e)
				       ((eq (car sub-e) ':!sym)
					(cadr sub-e))
				       (t
					(<# elim-!syms sub-e)))))
		       (elim-!syms e))))))
;;;;		    (signal-problem exp-env-symbols-place
;;;;		       "No namespace in " env
;;;;		       :% " so no way to canonicalize symbols in " e
;;;;		       (:continue "I'll assume all the symbols are global"))
	 (t e)))

(defun exp-namespace-symbols-place (exp ns env)
   (declare (ignorable ns env))
   (cond (+namespace-mode+
	  (let-fun ((:def rebuild (sub-e)
;;;;		       (trace-around rebuild
;;;;			  (:> "(rebuild: " sub-e ")")
		       (cond ((yt::is-Qvaroid sub-e)
			      (yt::make-Qvaroid
			         (yt::Qvaroid-atsign sub-e)
				 (yt::Qvaroid-comma sub-e)
				 (rebuild (yt::Qvaroid-sym sub-e))
				 (let ((nl (yt::Qvaroid-notes sub-e)))
				    (cond ((null nl) !())
					  (t (rebuild nl))))))
			     ((is-symbolish sub-e)
			      (symbol-namespace-or-env-place sub-e ns env))
			     ((atom sub-e) sub-e)
			     ((eq (car sub-e) ':!sym)
			      (cadr sub-e))
			     (t
			      (<# rebuild sub-e)))
;;;;			  (:< (val &rest _) "rebuild: " val))
		       ))
	     (rebuild exp)))
	 (t exp)))
|#

(defun var-place-and-lookup (var env)
   (var-lookup (cond ((-- +namespace-mode+)
		      (symbol-env-place var env))
		     (t var))
	       env))

(defun symbol-env-place (sym env)
   (let ((ns (env-find-namespace env)))
      (symbol-namespace-or-env-place sym ns env)))

;;; Returns < canonized-symbol, found-as-local >.  Second val is
;;; true iff found in a productive namespace wrt 'ns'. 
(defun symbol-namespace-or-env-place (sym ns env)
   (declare (ignorable ns env))
   (let-fun ()
      (cond ((-- +namespace-mode+)
	     (etypecase sym
		(symbol
		 (cond ((or (not ns) (is-Keyword sym) (var-lookup sym env))
			(values sym true))
		       (t
			(multi-let (((sym found-loc)
				     (sym-ns-resolve sym ns true)))
			   (values (check-exists sym)
				   found-loc)))))
		(Prefixed-symbol
		 (cond (ns
			(multi-let (((sym found-loc)
				     (prefixed-symbol-canonicalize sym !() ns)))
			   (values (check-exists sym)
				   found-loc)))
		       (t
			(signal-problem symbol-namespace-or-env-place
			   "Can't canonicalize " sym
			   " in namespace, because there isn't any namespace"))))
		(Canonized-symbol sym)))
	    (t
	     (values sym true)))
    :where
       (:def check-exists (can-sym)
	   (or can-sym
	       (progn
		  (note-defective-exp
		     ((_) " Attempt to resolve " sym " in " ns " failed")
		     :target sym
		     "Can't resolve " sym " in " ns
		     (:proceed "I will use local sym"))
		  sym)))))

(defun env-find-namespace (env)
   (repeat :for ((x :in (Env-vartypes env)))
    :result false
    :within
       (let ((ns (get-namespace x)))
	  (:continue
	   :until ns
	   :result ns))))
