;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: desigsubst.lisp,v 1.4 2004/02/22 00:46:58 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ vartypes)

(needed-by-macros
   (defvar desig-subst-tab* (make-eq-hash-table))

(datafun attach-datafun desig-subst
   (defun (_ sym fname)
      (!= (table-entry desig-subst-tab* sym)
          (symbol->fun fname))))
)

;;; Make desig look more intelligible
;;; bvars are params from the desig itself that should be
;;; treated as "bound" locally, not in the env
(defun desig-subst (d bvars env)
   (cond ((and (is-Qvar d) (null (Qvar-notes d)))
	  (let ((sym (Qvar-sym d)))
	     (let ((tvar-type (symboid-tvar sym)))
	        (let ((tvarval (Typebdg-val (tvar-type-bdg tvar-type))))
		   (cond (tvarval
			  (desig-subst
			     (Type-desig (Type-closure-type tvarval))
			     '()
			     (Type-closure-env tvarval)))
			 (t d))))))
	 ((atom d) d)
	 (t
	  (let ((ds-handler (table-entry desig-subst-tab* (car d))))
	     (cond (ds-handler
		    (funcall ds-handler d bvars env))
		   (t
		    (<# (\\ (e) (desig-subst e bvars env))
			d)))))))
			     
(datafun desig-subst param
   (defun (d bvars env)
      (let ((vname (cadr d)))
	 (cond ((memq vname bvars)
		d)
	       (t
		(let ((vt (var-lookup vname env)))
		   (cond (vt
			  (let ((val (Vartype-val vt)))
			     (cond ((eq val '*unbound)
				    d)
				   ((is-Type val)
				    (desig-subst (Type-desig val)
						 '() env))
				   ((is-Type-closure val)
				    (desig-subst (Type-closure-type val)
						 '()
						 (Type-closure-env val)))
				   (t val))))
			 (t '<*unbound*>))))))))

(datafun desig-subst Unk
   (defun (d _ _)
      (let ((p (unk-desig-param d)))
	 (cond (p `(param ,(cadr p)))
	       (t d)))))

(defun typeclo-pprint (srm tc)
   (let ((ty (Type-closure-type tc))
	 (env (Type-closure-env tc)))
      (let ((desig (Type-desig ty)))
	 (let ((subst-desig (desig-subst desig '() env)))
	    (out (:to srm)
	      (:pp-block
		    "[/"    ;"#{Typeclo "
		    (:q ((is-Type ty)
			desig
			(:q ((not (equal subst-desig desig))
			    (:pp-nl :linear srm)
			    " =" subst-desig)))
		       (t ty))
		    "/"
		    (:pp-nl :linear srm)
		    (env-condense env)
		    (:pp-nl :linear srm)
		    (:e (freevars-print (Type-closure-freevars tc) srm))
		    "/]"
		    ))))))

(set-pprint-dispatch 'Type-closure #'typeclo-pprint 1)
