;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: dotformers-valtup.lisp,v 1.1 2006/01/16 06:22:46 dvm Exp $

(depends-on :at-run-time %nity/ vartypes dottypes)

(depends-on :at-compile-time %nity/ desig)

(depends-on %langutils/ defectexp)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(TT T R Img <R< Tup alspec-args-sort)))

(in-type-system :universal)

;; This might as well go here.
(def-type-former T (params typename vartypes)
		   (ignore typename vartypes)
   (check-number-of-args params 0 T ())
   type-type*)

(def-type-former TT (params typename vartypes)
		    (ignore typename vartypes)
   (check-number-of-args params 0 TT ())
   type-tup-type*)

(def-type-former R (params typename vartypes)
		   (ignore typename vartypes)
   (check-number-of-args params 0 R ())
   row-type*)

(def-type-former Dot (params _ vartypes)
   (let ((params (check-number-of-args params 2 Dot
				      ('Obj 'Obj))))
      (build-dot-type 'Tup
		     (designated-type (car params)
				      false vartypes)
		     (designated-type (cadr params)
				      false vartypes))))

(def-type-former Lin (params _ vartypes)
   (let ((params (check-number-of-args params 2 Dot
				      ('Obj 'Obj))))
;;;;      (signal-problem Lin-type-former
;;;;	 "Attempt to find designatum of " `(Lin ,@params))
      (build-dot-type 'Row
		     (designated-type (car params)
				      false vartypes)
		     (designated-type (cadr params)
				      false vartypes))))

(defun build-dot-type (med cartype cdrtype)
   (cond ((and (memq med '(Row Arg Val))
	       (eq cdrtype empty-row-type*))
	  cartype)
	 (t 
	  (let ((car-desig (Type-desig cartype))
		(cdr-desig (Type-desig cdrtype)))
	     (multi-let (((class desig)
			  (dot-type-class+desig med car-desig cdr-desig)))
		(new-Type class
			  desig
			  (freevars-union (Type-freevars cartype)
					  (Type-freevars cdrtype))
			  univ-type*
			  '()
			  (list (tuple 'cartype (make-type-closable
						   cartype false))
				(tuple 'cdrtype (make-type-closable
						   cdrtype false))
				(tuple 'medium med)
				(tuple 'kind (dot-medium-kind med)))))))))


(def-type-former Ktp (params _ vartypes)
   (multi-let (((keyspecs synerrs)
		(keyword-params-parse params univ-type* vartypes)))
      (cond ((not (null synerrs))
	     (note-defective-exp ("Syntactic errors in Ktp type: "
			       synerrs)
		:novalue)))
      (build-key-type 'Tup keyspecs)))
		     
(def-type-former Key (params _ vartypes)
   (multi-let (((keyspecs synerrs)
		(keyword-params-parse params univ-type* vartypes)))
      (cond ((not (null synerrs))
	     (note-defective-exp ("Syntactic errors in Key type: "
			       synerrs)
		:novalue)))
      (build-key-type 'Row keyspecs)))

;;; Handles only keyword parameters.
(defun keyword-params-parse (params default-type vartypes)
   (let ((res-aspecs '())
	 (to-be-typed '())
	 (syntax-errors '()))
      (repeat :for ((p :in params :tail pl))
	 (cond ((eq p '-)
		(cond ((null (cdr pl))
		       (!= syntax-errors
			   (cons '("No type after - in arglist")
				 *-*)))
		      ((is-type-desig (cadr pl) vartypes)
		       (let ((ty (designated-type (cadr pl) false vartypes)))
			  (fill-in-param-types to-be-typed ty vartypes)
			  (!= to-be-typed '()))
		       (!= pl (cdr pl)))
		      (t
		       (!= syntax-errors
			   (cons `("Meaningless type designator " ,(cadr pl))
				 *-*))
		       (!= pl (cdr pl)))))
	       ((is-Keyword p)
		(cond ((exists (r :in res-aspecs)
			  (eq (Argspec-position r)
			      p))
		       (!= syntax-errors
			   (cons '("Keyword occurs twice in keyword list: "
				   p)
				 *-*)))
		      (t
		       (let ((r (new-Argspec '_ p '*notype)))
			  (!= res-aspecs (cons r *-*))
			  (!= to-be-typed (cons r *-*))))))
	       (t
		(!= syntax-errors
		    (cons `("Illegal in keyword list: " ,p)
			  *-*))))
       :result (progn (fill-in-param-types to-be-typed default-type vartypes)
		      (values res-aspecs syntax-errors)))))

(defvar +missing-default+ (printable-as-string "Unspecified default"))

;;; Called by bound-variable-list processors after seeing a set of
;;; Argspecs ('to-be-typed') and then encountering their type ('param-type').
;;; If 'param-type' is not a type, it is the symbol ':tvar', which means
;;; that each Argspec gets a new Typevar.
(defun fill-in-param-types (to-be-typed param-type vartypes)
   (multi-let (((init-exp has-init)
		(cond ((or (eq param-type ':tvar)
			   (not (Env-is-closing vartypes)))
		       (values nil false))
		      (t
		       (type-find-feature
			   param-type 'initexp vartypes)))))
     (repeat :for ((prevspec :in to-be-typed)
		   (new-tvars '()))
      :result new-tvars
	(cond ((eq (Argspec-type prevspec) '*notype)
	       (cond ((eq param-type ':tvar)
		      (let ((new-tvar (create-tvar-for-bvar
				         (Argspec-name prevspec))))
			 (!= new-tvars (cons new-tvar *-*))
			 (!= (type-feature new-tvar 'kind)
			     type-type*)
			 (!= (Argspec-type prevspec) new-tvar)))
		     (t
		      (!= (Argspec-type prevspec) param-type)
		      (cond ((and has-init
				  (eq (Argspec-default prevspec)
				      +missing-default+))
			     (!= (Argspec-default prevspec)
				 init-exp))))))))))

(defun build-key-type (med keyspecs)
      (let ((keyspecs (key-args-sort keyspecs)))
	 (let ((freevars 
		   (</ (\\ (fvl ks)
			  (freevars-union
			     fvl (Type-freevars (Argspec-type ks))))
		       '() keyspecs)))
	    (new-Type (cond ((eq med 'Tup) ktp-class*)
			    (t key-class*))
		      `(,(cond ((eq med 'Tup) 'Ktp)
			       (t 'Key))
			,@(keyspecs-typed-arglist keyspecs))
		      freevars
		      univ-type*
		      '()
		      (list (tuple 'keyspecs
				   (new-Closable keyspecs freevars
						  'argspecs-close))
			    (tuple 'medium med)
			    (tuple 'kind (dot-medium-kind med)))))))
		     
;;; Returns a list of alspecs, in this order: required, then optional,
;;; then if keys then keys in alphabetical order, otherwise rest arg if any.
;;; Used in opt/action.
(defun alspec-args-sort (alspec)
   (append (Arglistspec-required alspec)
	   (Arglistspec-optional alspec)
	   (cond ((null (Arglistspec-key alspec))
;;;;		  (cond ((null (Arglistspec-rest alspec))
;;;;			 !())
;;;;			(t ...))
                  (Arglistspec-rest alspec))
		 (t
		  (key-args-sort (Arglistspec-key alspec))))))

(defun key-args-sort (keyspecs)
   (sort (list-copy keyspecs)
	 #'string<
	 :key (\\ (as) (symbol->string (Argspec-position as)))))

(defun keyspecs-typed-arglist (keyspecs)
   (letrec ((expand (kl)
	       (cond ((null kl)
		      (values '() univ-type*))
		     (t
		      (let ((ae (Argspec-position (car kl)))
			    (at (Type-desig (Argspec-type (car kl)))))
			 (multi-let (((de dt)
				      (expand (cdr kl))))
			    (cond ((equal at dt)
				   (values `(,ae ,@de)
					   dt))
				  (t
				   (values `(,ae - ,at ,@de)
					   at)))))))))
      (multi-let (((ke _)
		   (expand keyspecs)))
	 ke)))

;; (Img (parms) template (args))
(def-type-former Img (params typename vartypes)
                     (ignore typename)
   (let ((params (check-number-of-args params 3 Img
				      ('() (car params) (list (car params))))))
      (letrec ((param-or-designatum (d kind)
		  (cond ((or (is-Symbol d) (is-Canonized-symbol d))
			 (make-param-type d kind))
			(t
			 (kind-trans
			    (designated-type d true vartypes)
			    kind no-env* no-env*)))))
	 (let ((iter (<# (\\ (p) (param-or-designatum p type-type*))
			 (car params)))
	       (template
		  (kind-trans (designated-type
				   (cadr params) true
				   (bind-iterparams-to-unks
				      (car params) vartypes))
			      type-type* no-env* no-env*))
	       (args (<# (\\ (p) (param-or-designatum p type-tup-type*))
			 (caddr params))))
	    (cond ((not (is-list-of iter (\\ (x) (and (is-Type x)
						      (is-param-type x)))))
		   (note-defective-exp
		      ("In type " `(Img ,params)
		       " first arg to Img must be list of param types: " iter)
		      :place Img-type-former))
		  ((not (= (len iter) (len args)))
		   (note-defective-exp
		      ("Number of params must match number of args in Img type:"
			 :% 3 iter
			 :% 3 "vs. " args)
		      :place Img-type-former)))
	    (build-image-type iter template args vartypes)))))

(defun seg-image-type (parm-vtl ty env)
   (let ((parmtypes
	    (<# (\\ (vt) (type-feature (Vartype-val vt) 'param))
		parm-vtl)))
      (build-image-type parmtypes
		       ty
		       (<# (\\ (p) (make-param-type
				      (param-type-varname p)
				      type-tup-type*))
			   parmtypes)
		       env)))

(defun make-image-type (iterparams ty args hide-unks vartypes)
   (new-Type-closure
      (build-image-type
          iterparams ty args vartypes
	  (<# (\\ (a)
		 (let ((d (type-find-designator a vartypes)))
		    (cond ((and hide-unks
				(car-eq d 'Unk))
			   (let ((p (type-find-feature
				       a 'param vartypes)))
			      (or (Type-desig p) d)))
			  (t d))))
	      args)
	  (</ (\\ (fv arg)
		 (freevars-union (type-find-freevars arg vartypes)
				 fv))
	      (freevars-complement
	          (type-find-freevars ty
				      (bind-iterparams-to-unks
				            iterparams vartypes))
		  iterparams)
	      args))
      vartypes))

(defun bind-iterparams-to-unks (iterparams vartypes)
   (env-bindings-nconc false
		       (<# unk-vartype
			   iterparams)
		       vartypes))

(defun unk-vartype (param)
   (let ((v (param-name param)))
      (new-Vartype v
		   type-type*
		   (unknown-type-for-param
		       v type-type* univ-type* true))))

;; args is a list of tup-types.  Initially it's the same as params, but
;; we expect those to be bound externally as type-tups.
;; The optional args can be omitted *only* during designator-resolution
;; time.  make-image-type supplies them because at expansion time    
;; the defaults are completely wrong.
(defun build-image-type (iterparams ty args env
			 &optional (argdesigs (<# Type-desig args))
				   (freevars (</ (\\ (fv pt)
						    (freevars-union
						       (Type-freevars pt) fv))
						 (Type-freevars ty)
						 iterparams)))
   (!= args (<# (\\ (a) (kind-trans a type-tup-type* env env))
		*-*))
;   (repeat :for ((a :in args))
;      (cond ((not (eq (Type-kind a) type-tup-type*))
;	     (error-break build-image-type :fatal
;		"Non-(TT) as arg to Img: " a
;		:% " [in " argdesigs "]"))))
   (cond ((null iterparams) void-type*)
	 ((and (is-param-type ty)
	       (= (len iterparams) 1)
	       (eq (param-type-varname ty) 
		   (param-type-varname (car iterparams))))
	  (car args))
	 (t
	  (cond ((is-Type-closure ty)
		 (dbg-save :run-loud iterparams ty args env)
		 (breakpoint build-image-type
		    "Template is Type-closure: " ty :% " args = " args)))
	  (make-type-of-class image-class
	     `(Img ,(<# param-type-varname iterparams)
		   ,(Type-desig ty)
		   ,argdesigs)
	     freevars
	     univ-type* '() 
	     (list (tuple 'template-params iterparams)
		   (tuple 'template-type
			  (new-Closable ty (Type-freevars ty)
					(simple-loadable-function
					   unclose)))
		   (tuple 'template-args
			  (new-Closable args freevars 'close-elements))
		   (tuple 'kind type-tup-type*))))))

(defun unclose (ty _) ty)

;; The type of '()' (and nothing else)
;(defvar null-type*
;      (new-Type boring-class* 'Null '() univ-type* '()
;                 (list (tuple 'home universal-type-sys*))))
;
;(define-global-constant-type Null null-type* :universal)

;;(defvar lst-type-no* 0)

;;(def-type-fun (Lst elty) (Alt Null (Dot elty (Lst elty))))

(defun build-to-row-type (orig-type)
   (make-type-of-class to-row-class
      `(<R< ,(Type-desig orig-type))
      (Type-freevars orig-type)
      univ-type*
      '()
      (list (tuple 'source-type
		   (make-type-closable orig-type false))
	    (tuple 'elt-types
		   (new-Arglistspec
		      '()
		      :rest (list (new-seg-Argspec '_ orig-type))))
	    (tuple 'kind row-type*))))

(defun make-to-row-type (ty env)
   (multi-let (((ty env)
		(follow-var-ref ty env)))
      (cond ((eq ty empty-tup-type*)
	     empty-row-type*)
	    (t
	     (make-type-of-class to-row-class
		`(<R< ,(Type-desig ty))
		(type-find-freevars ty env)
		univ-type*
		'()
		(list (tuple 'source-type
			     (closure-or-closable
			        ty env
				false (Type-freevars ty)))
		      (tuple 'kind row-type*)))))))
