;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: feature.lisp,v 1.5 2004/03/13 04:20:44 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ vartypes)

; Routines for accessing slots and features of types.
;;; ... except slot stuff has been moved to slot.nsp .

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(type-find-designator type-find-supertype
	     type-find-feature type-must-find-feature type-find-feature-spread
	     type-feature-remove)))

(defun type-find-designator (oty env)
   (multiple-value-let (ty env)
		       (follow-var-ref oty env)
      (cond ((is-Type ty)
	     (let ((d (Type-desig ty)))
	        (cond ((eq d '*as-feature)
		       (type-find-feature ty 'designator env))
		      (t d)
;;;;		      (t
;;;;		       (letrec ()
;;;;			  (params->desigs d)
;;;;			:where
;;;;			  (params->desigs (d)
;;;;			     (cond ((atom d) d)
;;;;				   ((eq (car d) 'param)
;;;;				    (let ((vt (var-lookup (cadr d) env)))
;;;;				       (cond (vt
;;;;					      (let ((vv (Vartype-val vt)))
;;;;						(cond ((eq (Vartype-val vt)
;;;;							   '*unbound)
;;;;						       d)
;;;;						      ((and
;;;;							 (not
;;;;							  (Env-is-closing env))
;;;;							 (is-unknown-type vv))
;;;;						       d)
;;;;						      (t
;;;;						       (type-find-designator
;;;;							  vv env)))))
;;;;					     (t d))))
;;;;				   (t
;;;;				    (<# params->desigs d))))))
)))
	    (t
	     (error-break type-find-designator :fatal
		"Non-type passed to type-find-designator: " ty
		t " originally " oty)))))

;;; Find supertype, looking through vars.
;;; If you find *as-feature as a supertype, it means to start all over again
;;; treating supertype as a feature (which is probably what it should 
;;; normally be anyway).
(defun type-find-supertype (x type-env)
   (multi-let (((ty type-env)
		(follow-var-ref x type-env)))
	 (let ((sup (Type-super ty)))
	    (cond ((eq sup '*as-feature)
		   (multi-let (((sup found)
				(type-find-feature ty 'supertype type-env)))
		      (cond (found sup)
			    (t
			     (signal-problem type-find-supertype
				"Supertype feature missing on "
				ty
				t " in environment " type-env)))))
		  (sup
		   (type-clos/able sup type-env))
		  (t
		   false)))))

(defvar brk-count* 0)
(defvar stop-at-brk* 50)

(defun must-be-kind (ty kind)
   (let ((kt (Type-kind ty)))
      (cond ((eq kt kind)
	     ty)
	    (t
	     (error-break must-be-kind :fatal
		"Type of wrong kind: " ty
		t " Kind = " kt  " Wanted: " kind)))))

;;; Do type-find-feature and then make wrapper go away, 
;;; returning the wrappee and env.
(defun type-find-feature-spread (ty feat vartypes)
   (multiple-value-let (val found)
		       (type-find-feature ty feat vartypes)
      (cond (found
	     (multiple-value-let (ty vartypes)
				 (follow-var-ref val vartypes)
		(values ty vartypes true)))
	    (t
	     (values false false false)))))

(defun type-must-find-feature (ty feat vartypes)
   (multiple-value-let (val found)
		       (type-find-feature ty feat vartypes)
      (cond (found val)
	    (t
	     (error-break type-must-find-feature
	        "Missing feature " feat " on " ty
		(:novalue "I will proceed with val false,"
			  " but more errors are likely"))
	     false))))

;;; type-find-feature returns < value, present >, where present is a boolean
;;; specifying whether the feature was found or not.  IF present = false,
;;; then value = false, but not vice versa.
;;; A FEATURE-FILLER feature is a function that generates new feature
;;; pairs.
;;; The assumption is that this never assumes anything, and that any variable
;;; in the result is with respect to the same vartypes
;;; as were passed in.  To circumvent that restriction, make sure 
;;; feature-filler returns a Type-closure.
(defun type-find-feature (ty feat orig-vartypes)
   (multiple-value-let (ty vartypes)
		       (follow-var-ref ty orig-vartypes)
      (letrec ()
	 (multiple-value-let (v found) (find-it (Type-features ty))  ; orig
	    (cond (found
		   (values v true))
		  (t
		   (multiple-value-let (v found)
				       (find-it (Type-cached-features ty))
		      (cond (found
			     (values v true))
			    (t
			     (let ((cf (Type-class ty)))
				(cond (cf 
				       (find-it (Type-class-featurelist cf)))
						;orig
				      (t (values false false))))))))))
    :where
      (find-it (fl)
;;;;	 (trace-around find-it
;;;;	    (:> "(find-it: " feat 1 fl")")
	 (letrec ()
            ;(out (to ttyout*) "Trying " fl t)
	    (let ((fp (assq feat fl)))
	       ;(out "fp = " fp t)
	       (cond (fp 
		      (match-cond (cadr fp)
			 ?((*feature-fcn ?fcn)
			   (multiple-value-let (val okay)
					       (funcall fcn ty feat vartypes)
			      (cond (okay
				     (succeed val fp))
				    (t
				     (check-missing-val val)
				     (try-feature-filler)))))
			 (t
			  ;(out "no local feature fcn" t)
			  (values (close-or-unclose (cadr fp))
				  true))))
		     (t
		      (try-feature-filler))))
	  :where

	    (try-feature-filler ()
	       (let ((fp (assq 'feature-filler fl)))
		  (cond (fp
			 (multiple-value-let (val found)
					     (funcall (cadr fp)
						      ty feat vartypes)
			    (cond (found
				   (succeed val false))
				  (t
				   (check-missing-val val)
				   (values false false)))))
			(t (values false false)))))

	    ;; We cache the result, but on the type we found after following
	    ;; var-refs, not necessarily the original type.
	    ;; And we don't cache if that type has freevars, because result
	    ;; will be different the next time we look.
	    ;; cache-loc is either false, in which case a
	    ;; new feature-pair is created, or a pair containing a *feature-fcn,
	    ;; in which case the value overwrites the *feature-fcn.
	    (succeed (val cache-loc) 
	       (!= val (close-or-unclose val))
	       (cond ((null (Type-freevars ty))
		      (cond ((not cache-loc)
			     (!= cache-loc (lrecord feat nil))
			     (!= (Type-cached-features ty) (cons cache-loc *-*))))
		      (!= (cadr cache-loc) val)))
	       (values val true))
	    
	    ;;; If we had to change envs to get to val, wrap it in inner env.
	    (close-or-unclose (val)
	       (close-if-closable val vartypes))

	    (check-missing-val (val)
	       (cond (val
		      (error-break type-find-feature
		         "Feature filler returned non-false absent value: " val
			 (:novalue "I will discard it"))))))
;;;;	    (:< (val &rest _) "find-it: " val))
	 
      ))))

;;; Note that neither of the two functions above check the cached-feature list.
;;; Rationale: Whether anything is found would depend on the exact time we asked.

;;; Like type-feature, but it removes what it finds
(defun type-feature-remove (ty feat)
   (let ((fp (assq feat (Type-features ty))))
      (cond (fp
	     (!= (Type-features ty)
		 (dremove1 fp *-*))
	     (values (cadr fp) true))
	    (t
	     (values false false)))))

(defun Type-home (ty)
   (let ((h (type-local-feature ty 'home)))
      (cond ((and (not h)
		  (is-CSymbol (Type-desig ty)))
	     (error-break Type-home
		"Type with no home: " ty
		(:novalue "I'll assume it's in the universal type system"))
	     universal-type-sys*)
	    (t h))))


;;; Needed by deftyp
(defun note-typefun-untestable (name type-sys)
       (let ((b (global-declaration name type-sys false)))
	  (cond ((and b (eq (Vartype-type b) tyfun-type*))
		 (let ((tyf (Vartype-val b)))
		    (!= (type-feature (Tyfun-result-type tyf)
				      'is-builder-name)
			'*untestable))))))

