;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: foundation.lisp,v 1.7 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on %nity/ types)

(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
  (export '(univ-type* void-type* Obj Void tyfun-type* constructor-type*
	    Constructor is-Constructor-spec
	    Constructor-spec Constructor-spec-type-params
	    arglistspec-bind-to-tvars
	    Boolean boolean-class bool-type*
	    Slot make-Slot Slot-name Slot-acc Slot-set Slot-atypes
	    Slot-access-fcn Slot-set-fcn
)))

(needed-by-macros

 (renew-Type-system 'universal 'universal-type-sys* !() false)

)

(!= (Type-features type-type*)
    (list (tuple 'meta 'type)
	  (tuple 'home universal-type-sys*)
	  (tuple 'var 'type-type*)))

(in-type-system universal)

;;; The absolute minimal set of types we require are Obj (supertype of everything)
;;; Void (subtype of everything), and (typeform fcn ...) which is used for
;;; recursive types.  (Plus "syntactic" types tvar-type and param-type, the types
;;; denoted by polymorphic variables and type parameter variables.)
;;; Plus "meta" types type-type* and typefun-type*, which describe types and
;;; typefuns themselves.

;; The class of types with designators like (T), i.e., consisting of a
;; type-former with no args.
(def-type-class simple-typeform-class
    (:system universal))

(def-type-class boring-class
    (:system universal)
    (:var boring-class*))
  
(!= (Type-class type-type*)
    (find-type-class 'simple-typeform-class universal-type-sys* true))

(defvar univ-type*
             (new-Type boring-class* 'Obj '() false '()
                 (list (tuple 'accept-any true)
		       (tuple 'home universal-type-sys*))))
;;; Any expression will work as an obj.

;;; The type of rows of objects, which, in Lisp, exist only as 
;;; multiple return vals and argument tuples.
(defvar row-type* (make-type-of-class simple-typeform-class
		     '(R) '() univ-type* '()
		     (list (tuple 'meta 'type)
			   (tuple 'home universal-type-sys*)
			   (tuple 'var 'row-type*)
			   )))

;;; A row of size 1 is just an ordinary type.
(!= (Type-super type-type*) row-type*)

;;; type-type* is now kosher.

(define-global-constant-type Obj univ-type* :universal)

;;(!= (Type-super type-type*) univ-type*)

;;(define-global-constant-type Type type-type* :universal)

(defvar void-type*
      (new-Type boring-class* 'Void '() univ-type* '()
                 (list (tuple 'home universal-type-sys*))))

(define-global-constant-type Void void-type* :universal)

;;(define-global-constant-type Type type-type* :universal)

;;; The type of things like Lst, as in (Lst Integer)
(defvar tyfun-type* (new-Type boring-class* 'Tyfun '() univ-type* '()
			      (list (tuple 'meta 'typefun)
				    (tuple 'home universal-type-sys*))))

(def-type-class boolean-class
    (:system :universal)
    (:feature consts '(true false)))

(defun is-boole (x)
   (or (eq x true)
       (eq x false)))

;;; See file slot.lisp for documentation.
(def-class Slot 
   name acc set type atypes invert descriptor)

;;; We define boolean here instead of using the 
;;; seemingly more elegant (Con true false)
;;; for a variety of reasons, including the oddness of Lisp's boolean type and
;;; the difficulty with the printed representation of false.
(defvar bool-type*
	(let ((bt (make-type-of-class boolean-class 'Boolean '() univ-type* '()
				      (list (tuple 'is-tester #'is-boole)
					    (tuple 'consts '(true false))))))
	   (let ((eq-slot (make-Slot '= '(lambda (x y) (cond (x y) (t (not y))))
				     false bt
				     (args->spec (list bt))
				     false
				     '(= boole-eq))))
	      (!= (Type-slotfns bt) (list eq-slot))
	      bt)))
		     
(define-global-constant-type Boolean bool-type* :universal)

(def-type-class typeform-class (:system :universal)
  (:var typeform-class*))

;;; This following are needed at run time by code generated by deftyp.nsp

;;; The type of types defined as a base type + modifications
(def-type-class modified-class (:system :universal) (:var modified-class*))

;;; The type of types with a given base type, but no specification
;;; of what sets its elements apart.
(def-type-class subtype-class (:system :universal) (:var subtype-class*))

(defvar type-former-type* (new-Type boring-class* 'Type-former () univ-type*
				    '()
				    (list (tuple 'meta 'typefun))))

(defun find-type-former (name sys)
   (let ((vt (global-declaration name sys true)))
      (cond ((eq (Vartype-type vt) type-former-type*)
	     (let ((v (Vartype-val vt)))
	        (match-let (type-former ?tf) v
		   tf)))
	    (t
	     (signal-problem find-type-former
		"Not the name of a type-former: " name)))))

(def-type-class const-class (:system :universal))

(def-class Constructor-spec
   ;; The type-params are redundant, since they're the same
   ;; as those for 'tyfun'
   tyfun
   ;;;; make-params -- a blind alley
   type-params)

;;;;   tyfun params)

(defvar constructor-type*
    (new-Type boring-class* 'Constructor () univ-type*
	      '()
	      (list (tuple 'meta 'typefun))))

(def-type-class constructed-class (:system :universal)
   (:var constructed-class*))

;;; Returns false if can't find, else pointer to tail of
;;; list beginning with thing after keyword.
;;; Used by code compiled with tupformers.
(defun find-field (keyword l)
   (repeat :for ((xl l (cond ((null (cdr xl)) '())
			     (t (cddr xl)))))
    :until (null xl)
    :result false
    :until (eq (car xl) keyword)
    :result (cdr xl)))

(def-op inside-print (x srm)
   (out (:to srm) ">>>...<<<"))

(def-op ll-add (l x)
  (cond ((listp l)
	 (cons x l))
	(t
	 (signal-problem ll-add
	    "Don't know how to add to " l))))

(def-op ll-car (l)
  (cond ((consp l) (car l))
	(t
	 (signal-problem ll-car
	    "Don't know how to take ll-car of " l))))

;;; Should also have ll-cons, ll-cdr, etc. 

(def-class List-with-length
    (:handler (print (ll srm)
		 (out (:to srm) "([Length: "
		      (List-with-length-size ll)
		      "]"
		      (:e (inside-print ll srm))
		      ")"))
	      (inside-print (ll srm)
		 (let ((cl (List-with-length-contents ll)))
		    (cond ((not (null cl))
			   (repeat :for ((x :in cl))
			      (out (:to srm) 1 x))))))
	      (ll-car (ll)
		(car (List-with-length-contents ll))))
    size
    contents)

(def-class Prioritized-handler
   priority fcn
   (:handler
      (print-object (ah srm)
	 (out (:to srm) "#{Prioritized-handler "
	      (Prioritized-handler-priority ah) "}"))))
