;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: funacceptors.lisp,v 1.2 2004/02/22 00:47:00 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ funtypes tupacceptors parmutils)

(depends-on :at-run-time %nity/ funformers vartypes subtype)

(class-feature-datafun equal-tester fun-class
   (defun (fty1 fty2 env1 env2)
      (let ((l1 (type-find-feature fty1 'level env1))
	    (l2 (type-find-feature fty2 'level env2))
	    (a1 (type-find-feature fty1 'argtype env1))
	    (a2 (type-find-feature fty2 'argtype env2))
	    (s1 (type-find-feature fty1 'side-effects env1))
	    (s2 (type-find-feature fty2 'side-effects env2)))
	 (and l1 l2 
	      (= l1 l2)
	      (eq (not s1) (not s2))
	      (types-equal a1 a2 env1 env2)
	      (let ((pl1 (funtype-extract-arglist fty1 env1))
		    (pl2 (funtype-extract-arglist fty2 env2)))
		 (let ((got-arglists (and pl1 pl2)))
		    (and (or (not got-arglists)
			     (arglistspecs-congruent pl1 pl2))
			 (letrec ()
			    (cond ((> l1 0)
				   (cond (got-arglists
					  (multiple-value-call
					     #'result-types-equal 
					     (paramlists-declare-identically
						pl1 pl2 env1 env2)))
					 (t
					  (signal-problem fun-class-equal-tester
					     "One or both high-level function type"
					     " lacks an arglist:"
					     :% 3 fty1
					     :% 3 fty2))))
				  ((or got-arglists
				       (and (is-tvar-type a1)
					    (is-tvar-type a2)
					    (same-tvar a1 a2)))
				   (result-types-equal env1 env2))
				  (t
				   false))
			  :where
			    (result-types-equal (env1 env2)
			       (let ((r1 (type-find-feature
					     fty1 'resulttype env1))
				     (r2 (type-find-feature
					     fty2 'resulttype env2)))
				  (and r1 r2
				       (types-equal r1 r2 env1 env2))))))))))))

;;; In general, if two types agree on everything except their
;;; conser, then they are equal.  The type says how they behave, not
;;; how they were built.  However, there is one exception: The 
;;; types of functions.  For two functions to be the same, it
;;; isn't enough that their argtypes be the same; they must also
;;; have the same "constructor," which is (in a sense) the compiler, 
;;; or (in a sense) the function itself.
;;; That is, the only way to construct an Arg for function f1 is
;;; to evaluate (f1 ....).  If f1 and f2 are of the same type, they
;;; must be callable in the same circumstances, which means that
;;; their Arglistspecs must be "congruent": their arguments don't have to 
;;; have the same names, but they have to have the same pattern of
;;; requireds, optionals, etc.

(defun arglistspecs-congruent (pl1 pl2)
   (multi-let (((min1 maxp1 max1)
		(arglistspec-min-max pl1))
	       ((min2 maxp2 max2)
		(arglistspec-min-max pl2)))
      (and (= min1 min2)
	   (eql maxp1 maxp2)
	   (eql max1 max2))))
;;; See also 'argspecs-nums-consistent' at the end of this file,
;;; which does the corresonding calculation for a subtype test.

(defun argspecs-equivalent (a1 a2 names-matter1 names-matter2 env1 env2)
   (and (or (eq (Argspec-name a1) (Argspec-name a2))
	    (and (not names-matter1) (not names-matter2)))
	(types-equal (Argspec-type a1)
		     (Argspec-type a2)
		     env1 env2)))

(defun functions-try-accept (constrain-tvars fun1 fun2 undo-stack
			     env-1 env-2 accept-stack)
   (let ((lev1 (type-find-feature fun1 'level env-1))
	 (lev2 (type-find-feature fun2 'level env-2)))
      (cond ((and lev1 lev2 (= lev1 lev2)
		  (or (type-must-find-feature
			 fun2 'side-effects env-2)
		      (not (type-must-find-feature
			      fun1 'side-effects env-1))))
	     (let ((aty2 (type-must-find-feature fun2 'argtype env-2))
		   (aty1 (type-must-find-feature fun1 'argtype env-1)))
		(multi-let (((a-okay undo-stack-2)
			     (type-acceptable
				constrain-tvars aty2 aty1
				undo-stack env-2 env-1 accept-stack)))
		   (cond (a-okay
			  (let ((als1 (type-find-feature fun1 'arglist env-1))
				(als2 (type-find-feature fun2 'arglist env-2)))
			    (cond ((not als1)
				   (!= als1 (type-arglistspec aty1 env-1))))
			    (cond ((not als2)
				   (!= als2 (type-arglistspec aty2 env-2))))
			    (cond ((> lev1 0)
				   (high-restypes-accept
				      constrain-tvars fun1 fun2 als1 als2
				      undo-stack-2 env-1 env-2 accept-stack))
				  (t
				   (low-restypes-accept
				      constrain-tvars fun1 fun2 als1 als2
				      undo-stack-2 env-1 env-2
				      accept-stack)))))
			 (t
			  (values false (undo undo-stack-2 undo-stack)))))))
	    (t
	     (values false undo-stack)))))

(defun high-restypes-accept (constrain-tvars fun1 fun2 als1 als2
			     undo-stack env-1 env-2 accept-stack)
  (cond ((and als1 als2)
	 (let ((env-2a
		  (argspecs-declare-placeholder-params
		     als2 env-2))
	       (env-1a
		  (argspecs-declare-arbitrary-types
		     als1 env-1)))
	    (!= (Env-is-closing env-2a) true)
	    ;; Note that scope-times of tvars are > scope-times of unks
	    (let ((rty1 (type-must-find-feature
			   fun1 'resulttype env-1a))
		  (rty2 (type-must-find-feature
			   fun2 'resulttype env-2a)))
;;;;	       (breakpoint functions-try-accept
;;;;		  "Ready to check result types "
;;;;		  :% rty1 :% rty2)
	       (multiple-value-let (r-accept undo-stack-3)
				   (type-acceptable
				      constrain-tvars
				      rty1 rty2 undo-stack
				      env-1a env-2a accept-stack)
;;;;		  (breakpoint functions-try-accept
;;;;		     "Acceptable? " r-accept)
		  (okay-or-undo r-accept undo-stack-3 undo-stack)))))
	(t
	 ;; Shouldn't this be an error?
	 (values false undo-stack))))

(defun low-restypes-accept (constrain-tvars fun1 fun2 als1 als2
			    undo-stack env-1 env-2 accept-stack)
  (let ((rty1 (type-must-find-feature
		 fun1 'resulttype env-1))
	(rty2 (type-must-find-feature
	         fun2 'resulttype env-2)))
     (cond ((and als1 als2)
	    (cond ((argspecs-nums-consistent als1 als2)
		   (type-acceptable constrain-tvars rty1 rty2 undo-stack
				    env-1 env-2 accept-stack))
		  (t
		   (values false undo-stack))))
	   (t
	    (multi-let (((okay undo1)
			 (type-acceptable constrain-tvars rty1 rty2
					  undo-stack env-1 env-2
 					  accept-stack)))
	       (cond (okay
		      (values true
			      (postpone-argspecs-nums-consistent-check
			         fun1 fun2 als1 als2 undo1 env-1 env-2)))
		     (t
		      (values false (undo undo1 undo-stack)))))))))

(defun postpone-argspecs-nums-consistent-check
                (fun1 fun2 als1 als2 undo-stack env-1 env-2)
   (letrec ((maybe-tvar (aty)
	       (cond ((is-tvar-type aty) (list aty))
		     (t '()))))
      (multi-let (((aty1 ae1 _)
		   (type-find-feature-spread fun1 'argtype env-1))
		  ((aty2 ae2 _)
		   (type-find-feature-spread fun2 'argtype env-2)))
	 (let ((vars (nconc (maybe-tvar aty1)
			    (maybe-tvar aty2))))
	    (cond ((null vars)
		   (signal-problem low-restypes-accept
		      "Functions have substantive argtypes but no"
		      " arglists: "
		      :% 3 fun1
		      :% 3 fun2)))
	    (let ((vc (new-Tvar-constraint
			 (\\ (_ _ undo-stack)
			    (check-tvar-arglists
			       aty1 aty2 ae1 ae2 als1 als2 undo-stack)))))
	       (repeat :for ((v :in vars))
		  (!= undo-stack
		      (tvar-add-constraint v vc *-*)))
	       undo-stack)))))

(defun check-tvar-arglists (aty1 aty2 ae1 ae2 als1 als2 undo-stack)
   (let ((als1 (or als1
		   (type-arglistspec aty1 ae1)))
	 (als2 (or als2
		   (type-arglistspec aty2 ae2))))
      (cond ((and als1 als2)
	     (values (cond ((argspecs-nums-consistent als1 als2)
			    'satisfied)
			   (t
			    'violated))
		     undo-stack))
	    (t
	     (values 'unknown undo-stack)))))

(declare-class-feature got-me-want-other fun-class
       (make-Accept-handler 5 #'functions-try-accept))

; 	    (\\ (ctv me other undo-stack my-env other-env accept-stack)
; 	       (functions-try-accept
; 		         ctv me other undo-stack my-env other-env
; 			 accept-stack))))

(declare-class-feature got-other-want-me fun-class
       (make-Accept-handler 5 #'functions-try-accept))

; 	    (\\ (ctv other me undo-stack other-env my-env accept-stack)
; 	       (functions-try-accept
; 		          ctv other me undo-stack other-env my-env
; 			  accept-stack))))


;;; When testing function arg types:
;; got = how it will be called
;; want = how it will be defined

;; Okay:
;
;; got:  rrooo
;; want: rrooo
;
;; got:  rrrooo
;; want: rroooooo
;;
;; got: rroo
;; want: rrooo...
;;
;; got: rrrooo...
;; want: rrooo...
;;
;; Not okay:
;
;; got:  rrrooooo
;; want: rrrroooo
;
;; (it might be called with 3 args -> tfa)
;
;; got:  rrrooooo
;; want: rroooo
;
;; (it might be called with 7 or 8 args -> tma)
;;
;; got:  rrooo...
;; want: rroo
;;
;; (it might be called with 18 args -> tma)

(defun argspecs-nums-consistent (got-alspec wanted-alspec)
   (multi-let (((gmin _ gmax)
		(arglistspec-min-max got-alspec))
	       ((wmin _ wmax)
		(arglistspec-min-max wanted-alspec)))
      (and (>= gmin wmin)
	   (cond ((not gmax)
		  (not wmax))
		 (t
		  (or (not wmax)
		      (>= wmax gmax)))))))

;;; Given that Arglistspecs pl1 and pl2 are equal, 
;;; bind the corresponding vars to identical unknown types.
(defun paramlists-declare-identically (alspec1 alspec2 env1 env2)
   (let ((bdgs1 (Env-vartypes env1))
	 (bdgs2 (Env-vartypes env2)))
      (letrec ((declarem (specl1 specl2)
		  (repeat :for ((spec1 :in specl1)
			       (spec2 :in specl2))
		     (cond ((type-find-feature
			        (Argspec-type spec1) 'meta env1)
			    (note-arg spec1 spec2
				      (unknown-constrained-type
					  (argspec-param spec1)
					  spec1
					  (type-find-kind (Argspec-type spec1)
							  env1)
					  false))))))

	       (note-arg (spec1 spec2 atype)
		     (push (new-Vartype (Argspec-name spec1)
					type-type*
					atype)
			   bdgs1)
		     (push (new-Vartype (Argspec-name spec2)
					type-type*
					atype)
			   bdgs2)))
					
	 (declarem (Arglistspec-key alspec1)
		   (Arglistspec-key alspec2))
	 (declarem (Arglistspec-optional alspec1)
		   (Arglistspec-optional alspec2))
	 (declarem (Arglistspec-required alspec1)
		   (Arglistspec-required alspec2))
	 (let ((rl1 (Arglistspec-rest alspec1))
	       (rl2 (Arglistspec-rest alspec2)))
	    (cond ((not (or (null rl1) (null rl2)))
		   (let ((r1 (car rl1)) (r2 (car rl2)))
		      (let ((rt1 (Argspec-type r1))
			    ;;(rt2 (Argspec-type r2)) -- same as rt1
			    (seg1 (Argspec-seg-type r1))
			    (seg2 (Argspec-seg-type r2)))
			 (cond ((and seg1 seg2)
				(note-arg r1 r2 rt1))
			       (t
				(note-arg r1 r2
					  (make-lstype
					     (unknown-type
						type-type* rt1))))))))))
	 (values (make-Env true bdgs1)
		 (make-Env true bdgs2)))))

(class-feature-datafun scope-time-checker fun-class
   (defun (ty scope-time bound-unks undo-stack env)
      (multi-let (((aokay undo-stack-1)
		   (all-freevars-scope-okay
		      (type-find-feature ty 'argtype env)
		      scope-time bound-unks undo-stack env))
		  ((lev _)
		   (type-find-feature ty 'level env)))
	 (cond (aokay
		(letrec ()
		      (cond ((> lev 0)
			     (bind-args-and-check-restype))
			    (t
			     (check-restype '() env)))
		 :where
		   (bind-args-and-check-restype ()
		      (multi-let (((arglist arglenv found)
				   (type-find-feature-spread ty 'arglist env)))
			 (cond (found
				(let ((new-params
				        (argspecs-placeholder-params
					   arglist arglenv)))
				   (check-restype
				      (<# Vartype-val new-params)
				      (env-bindings-append
				         false new-params env))))
			       (t
				(signal-problem fun-class-scope-time-checker
				   "Level-" lev " function type with no"
				   " arglist"
				   :% ty)))))

		   (check-restype (new-bound-unks arg-env)
		      (all-freevars-scope-okay
			 (type-find-feature ty 'resulttype arg-env)
			 scope-time
			 (append new-bound-unks bound-unks)
			 undo-stack-1 arg-env))))
	       (t
		(values false (undo undo-stack-1 undo-stack)))))))
