;-*- Mode: Common-lisp; Package: ytools; Readtable: ytools; -*-
(in-package :ytools)
;;; $Id: fundefs.lisp,v 1.3 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools)

SUPERSEDED by typecheck and funsyn in lisplang


(depends-on :at-run-time %nity/ funformers)


;;; Utilities for parsing Lisp-like languages with Nisp-style type declarations.


(def-class Fundef-rec
    (:options :key)
   name
   level
   special-flags
   (res-funtype false)  ;; false or resulttype of funtype if it's high
   (funtype false)
   (high-result false)
   (low-restype false)
   low-result  ;; low-result and ...
   low-args  ;; ... low-args and ...
   high-args ;; ...and high-args are initially type designators
                ;; or false if missing.
   defn rel
   (ill-formed-subexps '()))

;; Returns < fundef-recs, tifel >, where tifel is a list of Defective-exps
;;; that can't be associated with a particular fundef.
;; If 'restype' is non-false, it's the default resulttype for all the functions
;;; 'allow-res-types' is 'false' if explicit result types are not allowed
;;;    ('restype' is the *only* allowed resulttype)
;;; It's '*per-fundef' if they're allowed in each function definition
;;; It's '*wherever' if they're allowed in definitions and between them
;;; 'expect-flags' is for use in declaring predicates, where strange
;;; flags can occur in the midst of the function arguments; it's
;;; '()' in all other contexts.
;;; tvars-if-unspec means that if result types or arg types are absent
;;; they should be filled in with fresh tvars.
(defun fundefs-types (defs expect-flags tvars-if-unspec restype allow-res-types
		      allow-bodies mvartypes)
   (multi-let (((fundef-recs trailing-ill-formed-exps)
		(function-list-decompose defs expect-flags allow-res-types
					 allow-bodies mvartypes)))
      (cond (restype
	     (repeat for ((fdr in fundef-recs))
		(cond ((eq (Fundef-rec-low-result fdr)
			   '*absent)
		(!= (Fundef-rec-low-restype fdr) restype))))))
      ;; Create the types
      (repeat :for ((rec :in fundef-recs))
	 (collecting-defective-exps
	        (td-ife-list _)
		(fundef-rec-set-funtype rec tvars-if-unspec mvartypes)
	   (!= (Fundef-rec-ill-formed-subexps rec)
	       (append td-ife-list *-*))
	   (let ((atype (type-find-feature (Fundef-rec-funtype rec)
					   'nisptype::argtype
					   mvartypes)))
	      (cond ((and (type-find-feature
			     atype 'nisptype::medium mvartypes)
			  (not (type-find-feature
				  atype 'nisptype::elt-types mvartypes)))
		     (dbg-save rec mvartypes atype)
		     (signal-problem fundefs-types
			"Argtype with no elt-types: "
			rec
			:continue))))))
      (values fundef-recs trailing-ill-formed-exps)))


(defun function-list-decompose (defs expect-flags allow-res-types
				allow-bodies mvartypes)
   (let ((to-be-typed '())
	 (res-fdr-list '())
	 (new-ill-formed-exps '()))
      (letrec ()
	 (repeat :for ((fl flist)
		       (res-fdr-list !(Fundef-rec)))
	  :until (null fl)
	    (collecting-defective-exps
	            (f-ill-formed _)
		    (match-cond fl
		       ?( (- ?td ?@r)
			 (!= fl r)
			 (cond ((eq allow-res-types ':wherever)
				(multi-let (((rty rlspec _)
					     (function-row-type
						false td 0 false univ-type*
						vartypes)))
				   ;;(!= freevars (append (Type-freevars rty)
				     ;;		   *-*))
				   (fill-in-types rty rlspec)))
			       (t
				 (note-defective-exp
				   ((targ) "Can't specify a result-type"
				    " designator: " targ)
				   :target td
				   :place function-list-decompose
				   (:novalue "I will ignore it"))
			       )))
		       ?( (-)
			 (!= fl '())
			 (note-defective-exp
			     ("Hyphen at end of bvar list: " flist)
			     :place function-list-decompose
			     (:novalue "I will ignore it")))
		       ?( (?def ?@r)
			 (!= fl r)
			 (multi-let (((fname flags level
				       low-res low-arg high-arg)
				      (fundecl-decompose
				         def possible-flags
					 allow-res-types)))
			    (let ((new-fdr
				     (create-fundef-rec
					fname flags level
					'*absent low-res low-arg high-arg
					'*absent '*absent)))
			       (cond ((eq (Fundef-rec-low-result new-fdr)
					  '*absent)
				      (cond ((eq allow-res-types '*wherever)
					     (!= to-be-typed
						 (cons new-fdr *-*)))))
				     ((not allow-res-types)
				      (note-defective-exp
					 ("Explicit result type in forbidden"
					  " context: " (Fundef-rec-low-result
							  new-fdr))
					 :place function-list-decompose
					 (:novalue "I will ignore it"))
				      (!= (Fundef-rec-low-result new-fdr)
					  '*absent)))
			       (!= res-fdr-list (cons new-fdr *-*))))))
	       (!= new-ill-formed-exps
		   (nconc *-* f-ill-formed)))
	  result (values (dreverse res-fdr-list)
			 new-ill-formed-exps))

       :where

         (create-fundef-rec (name flags level high-res low-res
				  low-arg high-args
				  defn rel)
	    (let ((new-fdr (make-Fundef-rec
		              :name name
			      :special-flags flags
			      :level level
			      :high-result high-res
			      :low-result low-res
			      :low-args low-arg
			      :high-args high-args
			      :defn defn
			      :rel rel
			      :ill-formed-subexps new-ill-formed-exps)))
	       (!= res-fdr-list (cons new-fdr *-*))
	       (!= new-ill-formed-exps '())
	       new-fdr))

	 (fill-in-types (ty rlspec)
	    ;;(out (tr fill-in ("fill-in> " ty)
	    (repeat for ((prev in to-be-typed))
	       (cond ((not (Fundef-rec-low-restype prev))
		      (!= (Fundef-rec-low-restype prev)
			  ty)
		      (!= (Fundef-rec-low-result prev)
			  rlspec))))
	    (!= to-be-typed '())))))


(fundef-recs trailing-ill-formed-exps)
		(function-list-decompose defs expect-flags allow-res-types
					 mvartypes)