;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)

;;; $Id: funformers.lisp,v 1.10 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ funtypes tupformers parmutils trecycle)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(arglistspec->vartypes argspec-vartype arglist reslist
	     level side-effects resulttype argtype)))

(in-type-system universal)

(def-type-former Fun
		 (desig typename vartypes)
		 (ignore typename)
   (multi-let (((level r a s)
		(funtype-desig-decompose desig)))
      (designated-fun-type level r a s vartypes)))

(defun funtype-desig-decompose (ftype-desig)
   (let ((level 0))
      (cond ((is-Number (car ftype-desig))
	     (!= level (car ftype-desig))
	     (!= ftype-desig (cdr ftype-desig))))
      (match-cond ftype-desig
	 ?(?(:\| (?r <- ?a . ?s)
		(?a -> ?r . ?s)
		(?r ?a . ?s))
	   (values level r a s))
	 (t
	  (note-defective-exp
	     ((_) "Ill-formed function-type designator: " ftype-desig))))))

(defun designated-fun-type (level r a s vartypes)
      (let ((side-effects (or (null s)
			      (memq (car s) '(t true :side-effects)))))
	 (letrec ()
	    (let ((fty (cond ((consp a)
			      (multi-let (((low high _)
					   (args-low-high-split
					      a false vartypes)))
				 (cond ((null high)
					(build-funtype
					   r low level false vartypes))
				       (t
					(build-funtype
					       `(Fun ,level ,r ,low)
					       high
					       (+ level 1)
					       side-effects vartypes)))))
			     (t
			      (build-funtype
			         r a level side-effects vartypes)))))
	      fty))))

;;; Returns < lows, highs, split-point >
;;; 'split-point' is position of '!&' if it's there; false if highs are
;;; absent or implicit (i.e., use the '?' notation)
(defun args-low-high-split (params require-var-names vartypes)
;;;;   (trace-around args-low-high-split
;;;;      (:> "(args-low-high-split: " params ")")
   (let ((tail (memq '!& params)))
      (cond ((or (null tail) (null (cdr tail)))
	     ;(out (tr try-split ("try> ")
	     (let ((need-names (params-expect-names params require-var-names)))
		(let-fun ((find-qvars (x)
			     ;;(out (tr find-qvars ("find> " x)
			     (cond ((is-Qvar x)
				    (cond ((var-lookup (Qvar-sym x) vartypes)
					   '())
					  (t
					   (list (Qvar-sym x)))))
				   ((atom x) '())
				   (t (<! find-qvars x)))
			     ;;("<find " (car out-vals*))))
			     )
			  (remove-qvars (x)
			     (cond ((is-Qvar x) (Qvar-sym x))
				   ((atom x) x)
				   (t (<# remove-qvars x))))
			  (remove-qvars-from-type-desigs (l)
			     (cond ((null l) '())
				   ((eq (car l) '-)
				    `(- ,(remove-qvars (cadr l))
					,@(remove-qvars-from-type-desigs
					     (cddr l))))
				   (t
				    `(,(car l)
				      ,@(remove-qvars-from-type-desigs
					   (cdr l)))))))
		   (let ((imps (nodupq
				  (cond (need-names
					 (repeat :for ((y params (cdr y))
						      (res '()))
					  :until (null y)
					  :result res
					    (cond ((eq (car y) '-)
						   (!= res (nconc (find-qvars
								    (cadr y))
								  *-*))))))
					(t
					 (<! find-qvars params))))))
		      (cond ((null imps)
			     (values params false false))
			    (t
			     (values (cond (need-names
					    (remove-qvars-from-type-desigs 
					       params))
					   (t (remove-qvars params)))
				     imps false))))))
		;("<try " out-vals*)))
		)
	    (t
	     (let ((low (ldiff params tail))
		   (high (cdr tail)))
		(values
		   (cond ((eq (lastelt low) '-)
			  ;; let (... - !& v ...) abbreviate
			  ;;     (... - v !& v ...)
			  `(,@low ,(car high)))
			 (t low))
		   high
		   (len low))))))
;;;;      (:< (lo hi ll) "args-low-high-split: " lo " + " hi " [" ll "]"
;;;;;;;;	  (:e (cond ((equal hi '(opt::c12))
;;;;;;;;		     (out "Saving..." :%)
;;;;;;;;		     (dbg-save lo hi params require-var-names vartypes)
;;;;;;;;		     (breakpoint args-low-high-split
;;;;;;;;			"Unexpected high: " hi))))
;;;;	  ))
   )

(defun build-funtype (rdesig adesig lev side-effects vartypes)
   (multi-let (((atype alspec _)
		(function-row-type
		   true adesig lev false
		   univ-type* vartypes )))
      (multi-let (((rtype rlspec _)
		   (result-type
		       rdesig lev alspec univ-type*
		       vartypes)))
	 (let ((fty (make-funtype lev rtype atype
				  rlspec alspec
				  side-effects
				  vartypes)))
;;;;	    (dbg-save fty vartypes atype rtype lev)
;;;;	    (breakpoint build-funtype
;;;;	       "fty = " fty)
	    fty))))

(defun result-type (resdesig lev alspec default-type vartypes)
   (function-row-type false resdesig 0 false default-type
		      (cond ((> lev 0)
			     (argspecs-declare-desig-unks
				alspec vartypes))
			    (t vartypes))))

;;; Returns < type, arglistspec, new-tvars >.
(defun function-row-type (is-arg desig level require-var-names default-type
			  vartypes)
   (letrec ()
      (cond ((matchq (?(:\| Row Arg Val) ?@_)
		     desig)
	     (params-row-type (cdr desig)))
	    ((is-type-desig desig vartypes)
	     (res-type-desig->type desig vartypes))
	    ((listp desig)
	     (params-row-type desig))
	    (t
	     (note-defective-exp
	        ("Ill-formed " (:q (is-arg "argument") (t "result"))
		 " designator " desig)
		:place function-row-type)))
    :where
      (params-row-type (parms)
	 (cond (is-arg
		(designated-argtype parms level require-var-names
				    default-type vartypes))
	       (t
		(tup-val-former 'Row parms false default-type vartypes))))))

;;;;(defvar some-voids*
;;;;        (compile-time-designated-type
;;;;	   (Arg &rest Void)
;;;;	   universal))

(defun res-type-desig->type (desig vartypes)
   (let ((ty (designated-type desig false vartypes)))
      (values (kind-trans ty row-type* vartypes vartypes)
	      (new-Arglistspec
		 (list (new-Argspec '_ ':required ty)))
	      '())))

(defun range-general-funtype (level range)
   (let ((gen-arg (anon-tvar 'arg true row-type*)))
      (dbg-out binding-dbg*
	 "Creating ?" (tvar-type-varname gen-arg)
	 " as general argument for range " range :%)
      (let ((rgfty
	       (make-funtype level range gen-arg
		    (args->spec (list range))
		    false
		    false empty-env*)))
	 rgfty)))

;;; Tricky thing about level-1 funtypes: The result type *must* have
;;; freevars no matter when it is built.  This means that 'type-find-feature'
;;; can't be called to get the restype unless those vars are bound.

;;; rlspec and alspec are false if unknown (i.e., restype and argtype
;;; are tvars).
(defun make-funtype (level restype argtype rlspec alspec side-effects env)
   (letrec ()
      (cond ((and (not alspec) (> level 0))
	     (signal-problem make-funtype
		"Can't have high-order function with unknown arglist"
		:% "  argtype = " argtype
		:% "  restype = " restype)))
      (let ((rlist (arglistspec-printable rlspec restype env))
	    (alist (arglistspec-printable alspec argtype env))
	    (rvars (type-find-unbound-vars restype env))
	    (avars (type-find-freevars argtype env)))
	(let ((fundesig
	       	 `(Fun ,@(include-if (> level 0) level)
		       ,rlist <- ,alist
		       ,@(include-if side-effects side-effects))))
	   (make-type-of-class fun-class
	      fundesig
	      (freevars-union
		 (cond ((> level 0)
			(<? (\\ (fv)
			      (not (exists (as in (Arglistspec-argspecs
						     alspec))
				      (eq (Argspec-name as)
					  (param-type-varname fv)))))
			    rvars))
		       (t rvars))
		 avars)
	      univ-type*
	      '()
	      (nconc
		 (cond (rlspec
			(list (tuple 'reslist
				     (new-Closable
					rlspec (Typoid-freevars restype)
					'arglistspec-close))))
		       (t '()))
		 (cond (alspec
			(list (tuple 'arglist 
				     (closure-or-closable
					alspec env
					'arglistspec-close
					(Type-freevars argtype)))))
		       (t '()))
		 (list (tuple 'level level)
		       (tuple 'side-effects (eq side-effects ':side-effects))
		       (tuple 'resulttype
			      (rtype-closable restype alspec (> level 0)))
		       (tuple 'argtype
			      (type-clos/able argtype env)))))))))

(defun has-occurrence-of-any-unk (e)
   (letrec ((search-for-unk (x)
	       (match-cond x
		  ((atom x) false)
		  ?( (Unk ?_ ?@_)
		    true)
		  (t
		   (exists (y in x) (search-for-unk y))))))
      (search-for-unk e)))

(needed-by-macros
(defun resulttype-close (close level0 rtype alspec vartypes)
   (cond ((and close (Env-is-closing vartypes))
	  (new-Type-closure rtype vartypes))
	 (level0
	  (type-clos/able rtype vartypes))
	 (t
	  (rtype-closable rtype alspec true))))

(defun rtype-closable (rtype alspec high-level)
   (let ((type-params 
	    (cond ((not high-level) '())
		  (alspec 
		   (repeat :for ((as :in (Arglistspec-argspecs alspec))
				(res '()))
		      (let ((aty (Argspec-type as)))
			 (cond ((and (null (type-find-unbound-vars
					      aty empty-env*))
				     (type-find-feature
				        aty 'meta empty-env*))
				(!= res (cons (Argspec-name as) *-*)))))
		    :result res))
		  (t
		   (signal-problem argtype-type-params
		      "High-level funtype with no arglist")))))
      ;;(out "Params to check: " type-params :%)
;;;;     (cond ((and (not (null type-params))
;;;;		 (is-Type-closure rtype))
;;;;	    (dbg-save :run-loud rtype alspec high-level)
;;;;	    (breakpoint rtype-closable
;;;;	       "type-params = " type-params 5 "rtype = " rtype)))
      (make-type-closable
	 rtype
	 (cond ((null type-params)
		false)
	       (t
		(make-Loadable-function-with-extras
		   :name 'check-type-params-bound
		   :fcn false
		   :towit (list type-params rtype)))))))

(defun check-type-params-bound (type-params rtype)
   (\\ (rt env)
;;;;(trace-around check-params-fcn
;;;;   (:> "(check-params-fcn: " rt 1 env")")
      (cond ((not (bound-last type-params env))
	     (error-break resulttype-close
		"Attempt to close result type "
		rtype
		:% "  in env that fails to bind "
		type-params ": "
		:% (env-condense env)
		(:novalue
		 "I will proceed; expect trouble"))))
      (new-Type-closure rt env)
;;;;   (:< (val &rest _) "check-params-fcn: " val))
))

(defun argtype-close (close ty env)
   (type-clos/able ty (cond (close env)
			    (t no-env*))))
;;;;   (cond (close
;;;;	  (new-Type-closure ty env))
;;;;	 (t
;;;; 	  (make-type-closable ty false))))
)


;;(defvar fun-type* (range-general-funtype 0 univ-type*))

(declaim (special fun-class*))

(defun is-funtype (ty)
   (eq (Type-class ty) fun-class*))

(defun arglistspec->vartypes (alspec)
   (<# argspec-vartype (Arglistspec-argspecs alspec)))

(defun argspec-vartype (as)
   (let ((vt (new-Vartype (Argspec-name as)
			  (Argspec-type as)
			  false)))
      (!= (Vartype-props vt)
	  (Argspec-props as))
      vt))

;; This version doesn't attempt to close anything
(defun argtype-raw-arglistspec (aty)
   (multi-let (((alspec found)
		(type-feature aty 'elt-types true)))
      (cond (found
	     alspec)
	    (t (args->spec (list aty))))))

(defun argtype-typed-arglist (atype env)
   (let ((alspec (type-find-feature atype 'elt-types env)))
      (cond (alspec
	     (arglistspec-typed-arglist alspec))
	    (t
	     `(,(error-break argtype-typed-arglist
			 "Lost name of single argument to function with argtype "
			 atype
			 (:prompt-for "the name of that argument" #'is-symbol))
	       - ,(Type-desig atype))))))


(defun fun-arg-type (ftype env)
   (type-find-feature ftype 'argtype env))

;;; Like (type-find-feature ftype 'resulttype env), but
;;; avoids common problem with unbound arg types.
(defun fun-result-type (ftype env)
   (let ((lev (type-find-feature ftype 'level env)))
      (cond ((> lev 0)
	     (!= env (env-bindings-append
		        true
			(funtype-params-declare-unbound
			        ftype env)
			*-*))))
      (type-find-feature ftype 'resulttype env)))

(defun funtype-params-declare-unbound (ftype env)
   (argspecs-declare
      (type-find-feature ftype 'arglist env)
      (\\ (_) '*unbound)
      env))

;;;;(defun argtype-type-bindings (atype vartypes)
;;;;   (let ((alspec (type-find-feature atype 'elt-types vartypes)))
;;;;      (cond (alspec
;;;;	     (argspecs-declare
;;;;		alspec
;;;;		(\\ (_) '*unbound)
;;;;		;;(create-params-for-metas vartypes)
;;;;		vartypes))
;;;;	    (t '()))))

