;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: istester.lisp,v 1.6 2004/03/13 04:20:44 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ slot feature)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(type-find-tvars get-is-tester const-subtyper is)))

(defvar is-tester-num* 0)

(defun const-of-type (c pt vartypes)
   (let ((tester (get-callable-is-tester pt vartypes)))
      (and tester
	   (funcall tester c))))

(defun const-subtyper (ctv got wanted undo-stack envg envw accept-stack)
                      (ignore ctv accept-stack)
   (multi-let (((got-consts found-got)
		(type-find-feature got 'consts envg))
	       ((wanted-consts found-wanted)
		(type-find-feature wanted 'consts envw)))
;;;;      (dbg-save got-consts wanted-consts)
;;;;      (breakpoint const-subtyper ">>")
      (cond (found-got 
	     (cond (found-wanted
		    (values (is-sublist got-consts wanted-consts :test #'equal)
			    undo-stack))
		   (t
		    (let ((wit (get-callable-is-tester wanted envw)))
		       (cond (wit
			      (values 
				 (forall (c in got-consts)
				    (funcall
				        wit
					(official-const->val c)))
				 undo-stack))
			     (t
			      (values false undo-stack)))))))
	    (t
	     (values false undo-stack)))))

(defun official-const->val (c)
   (match-cond c   
      ?( empty-list '())
      ?( true true)
      ?( false false)
      ?(((:quote quote) ?d)
	d)
      (t c)))

(defun get-callable-is-tester (ty env)
   (or (type-find-feature ty 'callable-is-tester env)
       (let ((it (get-is-tester ty true env '() false)))
	  (cond ((and it (not (eq it '*untestable)))
		 (cond ((null (Type-freevars ty))
			(let ((compiled-tester
				 (cond ((is-Symbol it)
					(let ((it-def (fdefinition it)))
					   (cond ((is-Pair it-def)
						  (compile it it-def)
						  (fdefinition it))
						 (t
						  it-def))))
				       (t
					(compile nil it)))))
			   (!= (type-feature ty 'compiled-is-tester)
			       compiled-tester)
			   compiled-tester))
		       (t
			(eval `(function ,it)))))
		(t false)))))

(def-class Recursive-type-frame
   type
   env
   recurser)
;;; 'recurser' is false unless a recursion has been detected, in
;;; which case it is the name of the recursive entity.

;; must-be-defined is true if we plan to run the (value of the) thing
;; we get back.  If false, we can generate the name of the is-tester
;; at compile time without worrying about whether it's actually runnable now.
(defun get-is-tester (ty must-be-defined vartypes is-tester-stack
		      &optional (warn-if-undefined true))
   (or (slotfn-deint
	  (type-find-feature ty 'is-tester vartypes))
       (multiple-value-let (ty vartypes)
			   (follow-var-ref ty vartypes)
	  (let ((infinite-recursion
		   (find-in-recursive-type-stack ty vartypes is-tester-stack)))
	     (cond (infinite-recursion
		    (or (Recursive-type-frame-recurser
			   infinite-recursion)
			(let ((name (build-symbol
				       is-tester-
				       (++ is-tester-num*))))
			   (!= (Recursive-type-frame-recurser
				  infinite-recursion)
			       name)
			   `,name)))
		   (t
		    (let ((b (type-find-feature ty 'is-builder vartypes)))
		       (cond ((and b (not (eq b '*untestable)))
			      (build-is-tester b ty must-be-defined vartypes
					       is-tester-stack
					       warn-if-undefined))
			     (t
			      (cond (warn-if-undefined
				     (out (:to *error-output*)
					  "Warning: No is-builder for "
					  ty :%)))
			      '*untestable)))))))))

;;; This is here because it's too small to go in a file by itself.

;;; Find all the unbound tvars that occur in type, and whose scope-time
;;; is in the interval st-int.  If st-int is false,
;;; find all unbound tvars.
(defun type-find-tvars (type st-int vartypes stack)
   (multi-let (((type vartypes)
		(follow-var-ref type vartypes)))
      (cond ((is-tvar-type type)
	     (cond ((not st-int)
		    (list type))
		   (t
		    (let ((num (type-must-find-feature
				  type 'scope-time vartypes)))
		       (cond ((and (> num (car st-int))
				   (=< num (cadr st-int)))
			      (list type))
			     (t '()))))))
	    ((find-in-recursive-type-stack type vartypes stack)
	     !())
	    (t
;;;;	     (out "type = " type :% "  finder = " (type-feature type 'tvar-finder) :%)
	     (let ((stack
		       (cons (make-Recursive-type-frame
			         type vartypes false)
			     stack)))
		(let ((h (type-feature type 'tvar-finder)))
		   (cond (h
			  (funcall h type st-int vartypes stack))
			 (t
			  (error-break type-find-tvars
			     "No tvar-finder for " type
			     (:novalue "I will assume it contains no tvars"))
			  '()))))))))

(defun find-in-recursive-type-stack (ty vartypes ist-stack)
   (repeat :for ((itfr :in ist-stack))
    :result false
    :until (types-equal (Recursive-type-frame-type itfr)
		       ty
		       (Recursive-type-frame-env itfr)
		       vartypes)

;;;;    :until (and (eq (Recursive-type-frame-type itfr)
;;;;		   ty)
;;;;	       (eq (Recursive-type-frame-env itfr)
;;;;		   vartypes))
    :result itfr))

(defun build-is-tester (builder ty must-be-defined vartypes
			is-tester-stack warn-if-undefined)
   (let ((itfr (make-Recursive-type-frame ty vartypes false)))
      (let ((it (funcall (coerce-to-fun builder)
			 ty must-be-defined vartypes
			 (cons itfr is-tester-stack)
			 warn-if-undefined)))
	 (cond ((not it)
		(!= it '*untestable)))
	 (cond ((not (eq it '*untestable))
		(let ((itname (Recursive-type-frame-recurser itfr)))
		   (cond (itname
			  (!= it
			     `(lambda (x)
				 (labels ((,itname (x)
				          (,it x)))
				   (,itname x)))))))))
	 (cond ((null (Type-freevars ty))
		(!= (type-feature ty 'is-tester)
		    it)))
	 it)))
	       
;; Similar to try-accept, except that we make no inferences, and we must
;; prove the two types equal.
;;; Important: If this returns false, the types may still be equal,
;;; but it isn't obvious.
(defun types-equal (t1 t2 e1 e2)
;;;   (cond ((and (eq (type-desig t1) 'null)
;;;	       (eq (type-desig t2) 'null)
;;;	       (not (eq t1 t2)))
;;;	  (!= t1* t1)
;;;	  (!= t2* t2)
;;;	  (error-break types-equal :novalue "Null comparison")))

       (multiple-value-let (t1 e1)
			   (follow-var-ref t1 e1)
	  (multiple-value-let (t2 e2)
			      (follow-var-ref t2 e2)
	     (let ((name1 (Type-desig t1))
		   (name2 (Type-desig t2)))
	        (cond ((and (is-CSymbol name1)
			    (is-CSymbol name2)
			    (eq name1 name2))
		       true)
		      (t
		       (let ((eh1 (type-find-feature t1 'equal-tester e1)))
			   (cond (eh1
				  (funcall eh1 t1 t2 e1 e2))
				 (t
				  (let ((eh2 (type-find-feature t2 'equal-tester e2)))
				     (cond (eh2
					    (funcall eh2 t2 t1 e2 e1))
					   (t
					    ;;(out "Calling types-look-equal" :%)
					    (types-look-equal t1 t2 e1 e2)))))))))))))
				  
(defun types-look-equal (t1 t2 e1 e2)
   (let ((v (and (eq t1 t2)
		 (forall (v in (Type-freevars t1))
		    (types-equal v v e1 e2)))))
;;;;      (cond ((and (not v) (equal (Type-desig t1) (Type-desig t2)))
;;;;	     (signal-problem types-look-equal :novalue
;;;;		(:before-break (dbg-save t1 t2))
;;;;		"Apparent crock: t1* = " t1 " t2* = " t2
;;;;		:% " ex1 = " e1  :%
;;;;		" ex2 = " e2 )))
      v))
		
(defun build-funform (f argl)
   (letrec ()
      (cond ((atom f)
	     `(,f ,@argl))
	    ((eq (car f) 'function)
	     `(,@(cond ((atom (cadr f)) `(,(cadr f)))
		       (t (lambda-hack (cadr f))))
	       ,@argl))
	    ((memq (car f) '(lambda \\))
	     `(,@(lambda-hack f) ,@argl))
	    (t
	     `(,f ,@argl)))
    :where
      (lambda-hack (lexp)
	 (match-cond lexp
	    ?((?() ?_ (funcall ?f ?@_))
	      `(funcall ,f))
	    ?((?() ?_ (?f ?@_))
	      `(,f))
	    (t `(,lexp))))))

(defun cons-funcall (f argl)
   (cond ((and (is-Pair f) (memq (car f) '(function funktion quote)))
          `(,(cadr f) ,@argl))
         ((and (is-Pair f) (eq (car f) '\\))
          `((lambda ,@(cdr f)) ,@argl))
         (t `(funcall ,f ,@argl))))
