;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: parmsparse.lisp,v 1.14 2006/01/11 14:48:52 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ dotformers desig tuptypes)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(params-parse lambdalist-parse high-funargs-argspecs)))

;; Return < arglistspec, freevars, new-tvars, syntax-errors >
;;; syntax-errors is a list of lists of S-expressions.  An entry
;;; is typically of the form (string [optional thing]).
(defun params-parse (params require-var-names default-type
		     expect-defaults allow-allow-other vartypes)
   (cond ((and (atom params)
	       (not (eq params '())))
	  (values no-args-spec* '() '() (list `("Illegal params "
						       ,params))))
	 ((params-expect-names params require-var-names)
	  (multiple-value-let (alspec freevars new-tvars allow-other synerrs)
			      (lambdalist-parse params require-var-names
						default-type expect-defaults
						vartypes)
	     (values alspec freevars new-tvars
		     (cond ((and allow-other (not allow-allow-other))
			    (cons '("&allow-other-keys in illegal place") synerrs))
			   (t synerrs)))))
	 (t
	  (typelist-parse params vartypes))))

(defun params-expect-names (params require-var-names)
   (or require-var-names
       (memq '- params) (memq '-- params) (memq ':- params)))

;;; Returns < arglistspec, freevars, new-tvars, whether-allow-other-keys,
;;;           syntax-errors >
(defun lambdalist-parse (args names-matter default-type expect-defaults
			 vartypes)
   (let ((restarg false)
	 (allow-other false)
	 (mode ':required)
	 (specs '())
	 (to-be-typed '())
	 ;(typeless-args '())
	 (freevars '())
	 (new-tvars '())  ; those that first appear here
	 (rest-out-of-place false)
	 (argnames '())
	 (duplicates '())
	 (template-vars '())
	 (syntax-errors '()))
      (repeat :for ((al args) a)
       :until (null al)
       :result (finish-args)
	 (!= a (car al))
	 ; &rest is allowed to come in exactly one place: after all required and
	 ; optional args, and before all keyword args
	 (cond ((and (not rest-out-of-place)
		     (cond ((eq a '&rest)
			    (eq mode ':key))
			   (restarg
			    (eq a '&optional))
			   (t false)))
		(!= rest-out-of-place true)
		(!= syntax-errors
		    (cons '("&rest out of place") *-*))))
;;;;         (out "Looking at " al :%)
	 (cond ((eq a '&optional)
		(!= mode ':optional)
		;(!= typeless-args (cons '&optional *-*))
		(!= al (cdr al)))
	       ((eq a '&rest)
		(cond (restarg
		       (!= syntax-errors
			   (cons '("More than one occurrence of &rest")
				 *-*))))
		(multi-let (((okay tvl argname r)
			     (post-rest-parse (cdr al))))
		   (!= template-vars tvl)
		   (cond (okay
			  (!= restarg
			      (argspec-dup-check argname ':rest
						 '*notype false false))
			  (push restarg specs)
			  (push restarg to-be-typed))
			 (t
			  (!= syntax-errors
			      (cons `(,(format nil "Nonsense after &rest ~s"
						   al))
				      *-*))))
		   (!= al r)))
	       ((eq a '&key)
		;(!= typeless-args (cons '&key *-*))
		(!= mode ':key)
		(!= al (cdr al)))
	       ((eq a '&allow-other-keys)
		;(!= typeless-args (cons '&allow-other-keys *-*))
		(!= allow-other true)
		(!= al (cdr al)))
	       ((memq a '(- :- --))
		(let ((seg (memq a '(:- --))))
		   (cond ((and seg (or (null to-be-typed)
				       (not (eq (Argspec-mode
						   (car to-be-typed))
						':rest))))
			  (!= syntax-errors
			      (cons `(,(out-to-string
					  "Warning: encountered `:-' in"
					  " unexpected place;"
					  " treating as `-'"))
				    *-*))
			  (!= seg false)))
		   (cond ((cdr al)
			  (cond ((is-type-desig (cadr al) vartypes)
				 (let* ((ty 
					  (arg-designated-type 
					      (cadr al) to-be-typed
					      template-vars seg vartypes))
					(fv (Type-freevars ty)))
				    (!= freevars (append fv *-*))
;;;;                                    (dbg-save to-be-typed)
				    (!= new-tvars
					(nconc (fill-in-param-types
						  to-be-typed ty vartypes)
					       *-*))
				    (!= to-be-typed !())))
				 (t
				  (!= syntax-errors
				      (cons `("Meaningless type designator: "
					      ,(cadr al))
					    *-*))))
			  (!= al (cddr al)))
			 (t
			  (!= syntax-errors
			      (cons '("No type after `-' in arglist")
				    *-*))
			  (!= al (cdr al))))))
	       (t
		(!= al (cdr al))
		(cond ((or (is-symbolish a) (is-Qvar a))
		       (cond ((is-Qvar a)
			      (!= a (Qvar-sym a))))
;;;;                       (out "Here!" :%)
		       (let* ((default 
				 (cond ((and expect-defaults
					     (memq mode '(:key :optional)))
					+missing-default+)
				       (t false)))
			      (spec (argspec-dup-check
				      a (arg-kind mode a)
				      '*notype default false)))
			  (push spec specs)
			  (push spec to-be-typed)))
		      (t
		       (cond ((eq mode ':required)
			      (on-list `("Illegal as required argument: " ,a)
				       syntax-errors))
			     ((not expect-defaults)
			      (on-list
			          `("Illegal default specifier in arglist: "
				    ,a)
				  syntax-errors)))
		       (let ((argname (car a))
			     (default (and (cdr a) (cadr a)))
			     (supplied-p-param
				(and (cdr a) (cddr a) (caddr a)))
			     (kind mode))
			 (cond ((is-Qvar argname)
				(!= argname (Qvar-sym argname))
				(!= kind (arg-kind mode a)))
			       ((consp argname)
				(cond ((and (eq mode ':key)
					    (not (null (cdr argname))))
				       (!= kind (car argname))
				       (!= argname (cadr argname)))
				      (t
				       (!= syntax-errors
					   (cons `("Ill-formed argument " ,a)
						 *-*)))))
			       (t
				(!= kind (arg-kind mode argname))))
			 (let ((spec (argspec-dup-check
					argname kind '*notype default
					supplied-p-param)))
			    (push spec specs)
			    (push spec to-be-typed)))))))

      :where

 (:def argspec-dup-check (name kind type default supp)
;;;;    (trace-around argspec-dup-check
;;;;       (:> "(argspec-dup-check: " name 1 kind 1 :% 3 type :% 1 default 1 supp")")
    (let ((s (new-Argspec name kind type default
			   (list (tuple 'supplied-param supp)))))
       (cond (names-matter
	      (cond ((eq name '_)
		     (!= syntax-errors
			 (cons `("_ where name expected")
			       *-*)))
		    ((memq name argnames)
		     (!= duplicates (adjoinq name *-*)))
		    (t
		     (!= argnames (cons name *-*))))))
       s)
;;;;       (:< (val &rest _) "argspec-dup-check: " val))
    )

 ;; al is the stuff after &rest.  Parse it and return
 ;; < okay, template-vars, variable, remainder of list >
 ;; where remainder starts with "-" or ":-".
 (:def post-rest-parse (al)
;;;;		 (trace-around post-rest-parse
;;;;		    (:> "(post-rest-parse: " al ")")
    (let-fun ((:def from-var-on (al tvl)
		 (match-cond al
		    ?((?v ?@(:& ?r ?(:\| () (?(:\| - :- -- &key) ?_ ?@_))))
		      (cond ((and (is-symbolish v)
				  (not (memq v '(- ()))))
			     (values true tvl v r))
			    ((is-Qvar v)
			     (values true tvl (Qvar-sym v) r))
			    (t
			     (values false '() nil r))))
		    (t
		     (values false '() nil '())))))
       (cond ((null al)
	      (values false '() nil '()))
	     ((and (listp (car al))
		   (not (is-Qvar (car al))))
	      ;; Template
	      (from-var-on (cdr al) (car al)))
	     (t
	      (from-var-on al '()))))
;;;;		    (:< (okay tvl argname r) "post-rest-parse: "
;;;;			okay 1 tvl 1 argname 1 r))
		 )

 (:def finish-args ()
    (cond ((not (null duplicates))
	   (!= syntax-errors
	       (cons `("Duplicate args in arglist: " ,duplicates)
		     *-*))))
    (!= new-tvars
	(nconc (fill-in-param-types to-be-typed default-type vartypes)
	       *-*))
;;;;     (repeat :for ((as :in specs))
;;;;	(cond ((and (not (eq (Argspec-mode as) ':key))
;;;;		    (not (= (Argspec-position as) 0)))
;;;;	       (error-break lambdalist-parse :novalue
;;;;		  "Argspec with position " (Argspec-position as)
;;;;		  ": " as))))
    (repeat :for ((spec :in specs))
       (cond ((eq (Argspec-default spec) +missing-default+)
	      (!= (Argspec-default spec) false)
	      (on-list `("Default unspecified for argument: "
			    ,(Argspec-name spec))
		       syntax-errors))))
    (values (argspecs-arglistspec (reverse specs) allow-other true)
	    ;;; Remove duplicates:
	    (freevars-union freevars '())
	    new-tvars
	    allow-other
	    syntax-errors)))))

(defun arg-kind (mode param)
  (cond ((eq mode ':key)
	 (intern (Symbol-name param) keyword-package*))
	(t mode)))

(defun arg-designated-type (desig to-be-typed template-vars seg vartypes)
   (cond ((null to-be-typed)
	  (signal-problem arg-designated-type
	     "Type " desig
	     " in typed lambdalist is not associated with any params"
	     (:proceed "I'll ignore it")))
	 (t
	  (let ((las (car to-be-typed)))
	     (cond ((eq (Argspec-mode las)
			':rest)
		    (let ((parm-vtl (<# unk-vartype
					template-vars)))
		       (let ((ty (designated-type
				    desig false
				    (env-bindings-append
				       (Env-is-closing vartypes)
				       parm-vtl vartypes))))
			  ;; Fudge for later externalization
			  (!= (argspec-feat las 'template-vars)
			      (<# Vartype-var parm-vtl))
			  (!= (argspec-feat las 'ext-type)
			      (Type-desig ty))
			  (cond ((or seg (not (null parm-vtl)))
				 (!= (Argspec-seg-type las) true)
				 (cond ((null parm-vtl)
					(kind-trans ty type-tup-type*
						    vartypes vartypes))
				       (t
					(seg-image-type
					    parm-vtl ty vartypes))))
				(t ty)))))
		   (t
		    (let ((ty (designated-type desig false vartypes)))
		       (kind-trans ty type-type* vartypes vartypes))))))))

(defun typelist-parse (typelist vartypes)
   (let-fun ((:def chew-list (a)
                (cond ((null a)
                       (values '() '() '()))
                      ((is-type-desig a vartypes)
                       (out (:to *error-output*)
                            "Warning --  obsolete type list; ends with "
                            a :%)
                       (dbg-save :run-loud a vartypes)
                       (breakpoint typelist-parse
                          "Obsolescence: " vartypes)
                       (rest-list (list a)))
                      ((atom a)
                       (values '() '() (list `("Illegal argtypes: " ,a))))
                      ((eq (car a) '&rest)
                       (rest-list (cdr a)))
                      (t
                       (multiple-value-let (keyword a)
                                           (cond ((or (eq (car a) '&optional)
                                                      (is-Keyword (car a)))
                                                  (values (car a) (cdr a)))
                                                 (t
                                                  (values false a)))
                          (cond ((null a)
                                 (values '() '()
                                         (list `("typelist ends with "
                                                 ,(or keyword (car a))))))
                                (t
                                 (next-type keyword a)))))))

             (:def next-type (keyword a)
                (multi-let (((as-k as-name)
                             (cond ((eq keyword '&optional)
                                    (values ':optional '_))
                                   (keyword
                                    (values keyword (intern keyword
                                                            *package*)))
                                   (t
                                    (values ':required '_))))
                            ((specs dfv syntax-errors)
                             (chew-list (cdr a))))
                   (cond ((is-type-desig (car a) vartypes)
                          (let ((ty (kind-trans
                                       (designated-type
                                          (car a) false vartypes)
                                       type-type* vartypes vartypes)))
                            (values (cons (new-Argspec as-name as-k ty)
                                          specs)
                                    (freevars-union (Type-freevars ty) dfv)
                                    syntax-errors)))
                         (t
                          (dbg-save :run-loud a vartypes)
                          (breakpoint typelist-parse
                             "Meaningless: " (car a)
                             " in " typelist)
                          (values (cons (new-Argspec as-name as-k
                                                     univ-type*)
                                        specs)
                                  dfv 
                                  (cons `("Meaningless type designator: "
                                          ,(car a))
                                        syntax-errors))))))

                (:def rest-list (l)
                   (cond ((null l)
                          (values '() '()
                                  (list '("&rest at end of type list"))))
                         (t
                          (let ((seg (car-eq (car l) '=))
                                td)
                             (cond ((and seg
                                         (not (matchq (= ?td) (car l))))
                                    (values '() '()
                                            (list `("Ill-formed segment-type specifier: " ,(car l)))))
                                   (t
                                    (let ((ty (cond (seg
                                                     (kind-trans
                                                        (designated-type
                                                           td false vartypes)
                                                        type-tup-type*
                                                        vartypes vartypes))
                                                    (t
                                                     (kind-trans
                                                        (designated-type
                                                           (car l)
                                                           false vartypes)
                                                        type-type*
                                                        vartypes vartypes)))))
                                      (values
                                       (list (cond (seg
                                                    (new-seg-Argspec '_ ty))
                                                   (t
                                                    (new-Argspec
                                                       '_ ':rest ty))))
                                       (Type-freevars ty)
                                       (cond ((null (cdr l))
                                              '())
                                             (t
                                              (list `("Extra junk after &rest"
                                                      ,(cdr l))))))))))))))
      (multiple-value-let (argspecs freevars synerrs)
			  (chew-list typelist)
	 (values (argspecs-arglistspec argspecs false false)
		 freevars '() synerrs))))

(defun high-funargs-argspecs (args vartypes)
   (let ((specs '())
	 (to-be-typed '())
	 (argnames '())
	 (duplicates '())
	 (syntax-errors '()))
     (let-fun ((:def argspec-dup-check (name kind type default supp)
		 (let ((s (new-Argspec name kind type default
					(list (tuple 'supplied-param supp)))))
		    (cond ((memq name argnames)
			   (!= duplicates (adjoinq name *-*)))
			  (t
			   (!= argnames (cons name *-*))))
		    s)))
	(repeat :for ((al args) a)
	 :until (or (null al)
		   (eq (car al) '&constraints))
	   (!= a (car al))
	   (cond ((eq a '&rest)
		  (match-cond (cdr al)
		     ?( (?(:+ ?v is-CSymbol) ?@r)
		       (let ((spec (argspec-dup-check v ':rest '*notype false false)))
			  (push spec specs)
			  (push spec to-be-typed))
		       (!= al r))
		     (t
		      (on-list `("Meaningless after &rest" ,(cdr al))
			       syntax-errors))))
		 ((memq a '(- < >))
		  (cond ((cdr al)
			 (cond ((or (is-type-desig (cadr al) vartypes)
				    (is-symbolish (cadr al)))
				(let ((ty (designated-type
					     (cadr al) false vartypes)))
				   (let ((dty (cond ((eq a '-)
						     ty)
						    (t type-type*))))
				      (repeat :for ((prevspec :in to-be-typed))
					(cond ((eq (Argspec-type prevspec) '*notype)
					       (!= (Argspec-type prevspec) dty)
					       (cond ((memq a '(< >))
						      (!= (argspec-feat prevspec
									'constraints)
							  (list (tuple a ty))))))))))
				(!= to-be-typed '()))
			       (t
				 (!= syntax-errors
				     (cons `("Meaningless type designator: "
					     ,(cadr al))
					   *-*))))
			 (!= al (cddr al)))
			(t
			 (!= syntax-errors
			     (cons '("No type after `-' in arglist")
				   *-*))
			 (!= al (cdr al)))))
		 (t
		  (!= al (cdr al))
		  (cond ((or (is-symbolish a) (is-Qvar a))
			 (cond ((is-Qvar a)
				(!= a (Qvar-sym a))))
			 (let ((spec (argspec-dup-check a ':required
							'*notype false false)))
			    (push spec specs)
			    (push spec to-be-typed)
			    ;(!= typeless-args `(,a ,@*-*))
			    ))
			(t
			 (!= syntax-errors
			     (cons `("Illegal type argument: " ,a)
				   *-*))))))
	(cond ((not (null duplicates))
	       (!= syntax-errors
		   (cons `("Duplicate args in arglist: " ,duplicates)
			 *-*)))))
     ;; Any that did not get typed must receive type Type
     (repeat :for ((prevspec :in to-be-typed))
	(cond ((eq (Argspec-type prevspec) '*notype)
	       (!= (Argspec-type prevspec) type-type*))))
     (values (argspecs-arglistspec (reverse specs)
				   false true)
	     syntax-errors))))

(defun argspecs-arglistspec (argspecs allow-other-keys names-matter)
   (new-Arglistspec argspecs :allow-other-keys allow-other-keys
		             :names-matter names-matter))   

