;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: parmutils.lisp,v 1.6 2005/12/26 20:34:24 dvm Exp $

(depends-on :at-run-time %nity/ dotformers feature vartypes tuptypes parmsgen
			 %langutils/ synutils)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(argspec-tvar-type funtype-declare-placeholder-params
	     high-argspecs-declare-placeholder-params argspecs-declare-env
	     argspecs-declare-placeholder-params argspecs-placeholder-params
	     argspecs-declare-arbitrary-types argspecs-tvar-types
	     argspecs-declare-desig-unks
	     argspecs-declare create-params-for-metas argspec-desig-unk
	     argspec-unknown-type argspec-param unk-for-param
	     unknown-constrained-type funtype-extract-arglist
	     funtype-extract-reslist type-arglistspec arglistspec-printable)))

;;; Various features used here are actually "introduced" in
;;; files tupformers (elt-types) and funformers (level, argtype, arglist)
;;; But it's convenient and harmless to use them as feature names
;;; in this file.

(defun funtype-declare-placeholder-params (funtype env)
   (high-argspecs-declare-placeholder-params
      funtype (funtype-extract-arglist funtype env) env))

;;; Sometimes the alspec is massaged before we generate placeholders
(defun high-argspecs-declare-placeholder-params (funtype arglist env)
   (let ((lev (type-find-feature funtype 'level env)))
      (cond ((> lev 0)
	     (argspecs-declare-placeholder-params arglist env))
	    (t env))))

(defun argspecs-declare-env (alspec type-trans base-env)
   (make-Env
	 false
	 (argspecs-declare alspec type-trans base-env)))

;;; (Formerly 'argspecs-declare-meta-params').
(defun argspecs-declare-placeholder-params (alspec vartypes)
   (env-bindings-nconc true
		       (argspecs-placeholder-params alspec vartypes)
		       vartypes))

(defun vartypes-declare-placeholder-params (high-bdgs vartypes)
   (nconc
      (<# (\\ (vt)
	     (new-Vartype (Vartype-var vt)
			  (Vartype-type vt)
			  (unk-for-param (Vartype-var vt)
					 (Vartype-type vt)
					 false false)))
	  high-bdgs)
      vartypes))

(defun argspecs-placeholder-params (alspec vartypes)
		       (argspecs-declare
			  alspec
			  (\\ (a)
			     (argspec-desig-unk a vartypes))
			  vartypes))

(defun argspecs-declare-arbitrary-types (alspec vartypes)
   (env-bindings-nconc true
		       (argspecs-tvar-types alspec vartypes)
		       vartypes))

(defun argspecs-tvar-types (alspec vartypes)
   (argspecs-declare
      alspec
      (\\ (a)
	 (cond ((type-find-feature
		   (Argspec-type a) 'meta vartypes)
		(argspec-tvar-type a))
	       (t false)))
      vartypes))

;;; This should be used only to declare unknowns for resolving
;;; designators.
(defun argspecs-declare-desig-unks (alspec vartypes)
   (env-bindings-nconc false
		       (argspecs-declare
			  alspec
			  (\\ (a)
			     (cond ((type-find-feature
				       (Argspec-type a) 'meta vartypes)
				    (argspec-desig-unk a vartypes))
				   (t false)))
			  vartypes)
		       vartypes))

;;;(defun atype-arspecs-declare (argtype vartypes)
;;;   (nconc (argspecs-declare
;;;	     (argtype-arglistspec argtype vartypes)
;;;	     (\\ (a) (argspec-meta-param a vartypes))
;;;	     vartypes)
;;;	  vartypes))

;;;;(defun argtype-arglistspec (aty vartypes)
;;;;   (multiple-value-let (alspec found)
;;;;		       (type-find-feature aty 'elt-types vartypes)
;;;;      (cond (found
;;;;	     alspec)
;;;;	    (t
;;;;	     (args->spec (list (new-Type-closure aty vartypes)))))))

;;; Gather three important values for an Arglistspec: number of
;;; required args, number of optional args, and whether the
;;; number of args has no upper bound.
;;; Prefix them with a spoiler: a Boolean that says whether the 
;;; presence of a seg screws up the computation.
(defun alspec-reqnum-optnum-expandable (alspec)
   (let ((r (Arglistspec-rest alspec)))
      (cond ((not (null r))
	     (!= r (car r))))
      (cond ((and r (Argspec-seg-type r))
	     ;; Punt
	     (values false nil nil nil))
	    (t
	     (values true
		     (len (Arglistspec-required alspec))
		     (len (Arglistspec-optional alspec))
		     (and r
			  (null (Arglistspec-key alspec))
			  (not (Arglistspec-allow-other-keys
				alspec))))))))

;;; Returns < min-number-of-args, max-pos-args, max-all-args >
;;; max-pos-args is maximum number of non-key args.
;;; max-all-args is maximum number of all args (note that 
;;; keywords are *not* counted as args --- they are arg
;;; position markers, so '&key a b c' has max 3 args, not 6.
;;; "max" numbers are false if there is no upper bound
;;; (The max-pos-args result is probably useless, but who knows.)
(defun arglistspec-min-max (alspec)
   (let ((min-num (len (Arglistspec-required alspec)))
	 (opt-num (len (Arglistspec-optional alspec)))
	 (key-num (len (Arglistspec-key alspec))))
      (cond ((= key-num 0)
	     (cond ((null (Arglistspec-rest alspec))
		    (values min-num
			    (+ min-num opt-num)
			    (+ min-num opt-num)))
		   (t
		    (values min-num false false))))
	    (t
	     (values min-num
		     (+ min-num opt-num)
		     (+ min-num opt-num key-num))))))

;;; This is often used to declare "meta-params," that is, params of
;;; level-1 functions that must be bound before we can look at the
;;; result types of those functions.
;;; Returns < new-vartypes, positions >
;;; where each position gives the Argspec-name of the Argspec used
;;; to generate the corresponding Vartype.
(defun argspecs-declare (arglspec type-trans mvartypes)
   (!= scope-time* (+ scope-time* 1))
   (multiple-value-let (arglspec avartypes)
		       (follow-var-ref arglspec mvartypes)
      (let ((res '()) (pos-res '()))
	 (let-fun ((:def declarem (specs)
		      (repeat :for ((spec :in specs))
			 (note-arg spec (Argspec-type spec)))
		      )

		   (:def note-arg (spec atype)
			 ;;(out (tr note-arg ("note-arg> " spec 1 atype)
			 (push (new-Vartype (Argspec-name spec)
					    (new-Type-closure atype avartypes)
					    (funcall type-trans spec))
			       res)
			 (push (argspec-arg-position spec)
			       pos-res)
			 ;;("note-arg<" (car res))))
			   )
		   (:def rest-argspec-decl-type (rspec)
		      (cond ((Argspec-seg-type rspec)
			     (Argspec-type rspec))
			    (t
			     (make-lstype (Argspec-type rspec))))))
	    (declarem (reverse (Arglistspec-key arglspec)))
	    (let ((r (Arglistspec-rest arglspec)))
	       (cond ((not (null r))
		      (!= r (car r))
		      (note-arg r
				(rest-argspec-decl-type r)))))
	    (declarem (reverse (Arglistspec-optional arglspec)))
	    (declarem (reverse (Arglistspec-required arglspec))))
	 (values res pos-res)
	 ;(reverse res)
	 )))


;;; Useful in conjunction with argspecs-declare
(defun create-params-for-metas (vartypes)
   (\\ (spec)
      ;;(out (tr create-params-for-metas ("crp> " spec)
      (let ((aty (Argspec-type spec)))
	 (cond ((type-find-feature aty 'meta vartypes)
		;; (T) or (TT)
		(let ((unk (argspec-unknown-type spec false)))
		   (dbg-out binding-dbg*
		      "Creating unknown type " (:e (unk-pprint
						    *error-output*
						    (Type-desig unk)))
		      " as binding of type var "
		     (Argspec-name spec) :%)
		  unk))
	       (t
		false)))
      ;;("crp< " (car out-vals*))))
      ))

;;; (Formerly 'argspec-meta-param')
(defun argspec-desig-unk (aspec vartypes)
   (cond ((type-find-feature (Argspec-type aspec) 'meta vartypes)
	  (argspec-unknown-type aspec true))
	 (t false)))

(defun argspec-unknown-type (spec is-desig)
   (unk-for-param (argspec-param spec)
		  (Argspec-type spec)
		  spec is-desig))

(defun argspec-param (spec)
   (make-param-type (Argspec-name spec)
		    (Argspec-type spec)))

(defun unk-for-param (param kind spec desig)
   (let ((unk (cond (spec (unknown-constrained-type
			      param spec kind desig))
		    (t (unknown-type-for-param
			   param kind univ-type* desig)))))
;;;;      (!= (type-feature unk 'param) param)
;;;;      (!= (type-feature unk 'desig) desig)
;;;;      (!= (Type-desig unk)
;;;;	  `(,@*-* (param ,(param-type-varname param) ,desig)))
      unk))

(defun unknown-constrained-type (param argspec kind desig)
;;;;				 &optional (kind (Type-kind
;;;;						    (Argspec-type argspec))))
   (constrained-type
       argspec
       (\\ (sup) (unknown-type-for-param
		    param kind sup desig))))

(defun constrained-type (argspec type-builder)
   (let ((constraints (argspec-feat argspec 'constraints))
	 (ut false))
      (repeat :for ((con :in constraints))
	 (cond ((and (eq (car con) '<) (not ut))
		(!= ut (funcall type-builder (cadr con))))
	       (t
		(out (:to *error-output*) "Warning: ignoring constraint "
		     con " from " argspec :%))))
      (cond ((not ut)
	     (!= ut (funcall type-builder univ-type*))))
      ut))

(defun argspec-tvar-type (aspec)
   (let ((ty (Argspec-type aspec)))
      (let ((tpref (cond ((eq ty type-type*)
			  "*")
			 ((eq ty type-tup-type*)
			  "**")
			 (t " "))))
	 (anon-tvar (string-concat "a" tpref "/"
				   (symbol-name (Argspec-name aspec)))
		    true ty))))

(defun funtype-extract-arglist (funtype env)
   (let ((als (type-find-feature funtype 'arglist env)))
      (or als
	  (type-arglistspec (type-find-feature funtype 'argtype env)
			    env)
	  (cond ((> (type-find-feature funtype 'level env)
		    0)
		 (signal-problem funtype-extract-arglist
		    "High-level funtype without arglist: " funtype))
		(t false)))))

(defun funtype-extract-reslist (funtype env)
   (let ((rls (type-find-feature funtype 'reslist env)))
      (or rls
	  (type-arglistspec (type-find-feature funtype 'resulttype env)
			    env))))

(defun type-arglistspec (ty env)
   (or (type-find-feature ty 'elt-types env)
       (cond ((type-generates-arglistspec
		 ty env)
	      (new-Arglistspec
		 (list (new-Argspec '_ ':required ty))))
	     (t false))))

(defun type-generates-arglistspec (ty env)
   (not (is-tvar-type (follow-var-ref ty env))))

(defun arglistspec-printable (alspec under-type env)
	 (cond (alspec
		(arglistspec-arglist-or-typelist alspec))
	       (t (type-find-designator under-type env))))
