;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: scopecheck.lisp,v 1.4 2004/03/13 04:20:45 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ feature)

;;; Check whether value for tvar does not contain unks for params created after
;;; the tvar (stored as its 'scope-time').
;;; If any unbound tvars are encountered, they have their scope-times
;;; updated to the same timestamp.
;;; bound-unks is list of param unks that we had to bind in order to
;;; explore funtypes along the way.  They are not free, so they are ignored.
;;; Returns < okay, undo-stack >
(defun all-freevars-scope-okay (ty scope-time bound-unks undo-stack env)
   (multi-let (((ty env)
		(follow-var-ref ty env)))
      (cond ((is-tvar-type ty)
	     (let ((var-time (type-feature ty 'scope-time)))
	        (cond ((=< var-time scope-time)
		       (values true undo-stack))
		      (t
		       (!=? (type-feature ty 'scope-time)
			    scope-time
			    undo-stack)
		       (let ((bdg (tvar-type-bdg ty))
			     (constraints-okay true)
			     (undo1 undo-stack))
			  (repeat :for ((sub :in (Typebdg-subs bdg)))
			     (!= < constraints-okay undo1 >
				 (all-freevars-scope-okay
				    sub scope-time bound-unks undo1 env))
			   :while constraints-okay)
			  (repeat :for ((sup :in (Typebdg-supers bdg)))
			   :while constraints-okay
			     (!= < constraints-okay undo1 >
				 (all-freevars-scope-okay
				    sup scope-time bound-unks undo1 env)))
			  (okay-or-undo constraints-okay
					undo1 undo-stack))))))
	    ((is-param-type ty)
	     ;; Must be bound during descent into funtype; not free,
	     ;; so doesn't matter
	     (values true undo-stack))
	    (t
	     (let ((h (type-find-feature ty 'scope-time-checker env)))
	        (cond (h
		       (funcall h ty scope-time bound-unks undo-stack env))
		      (t
		       (cond ((or (not (is-CSymbol (Type-desig ty)))
				  (eq (Type-desig ty) '*as-feature))
			      (signal-problem all-freevars-scope-okay
				 "No scope-time checker for " ty
				 (:novalue
				  "I will assume the scope-times are okay"))))
		       (values true undo-stack))))))))

(class-feature-datafun scope-time-checker unk-class
   (defun (ty scope-time bound-unks undo-stack _)
      (cond ((memq ty bound-unks)
	     (values true undo-stack))
            (t
	     (let ((ut (type-feature ty 'scope-time)))
		(cond (ut
		       (values
			  (cond ((= ut scope-time)
				 (signal-problem
				     unk-class-scope-time-checker
				     "Parameter "
				     (type-feature ty 'param)
				     :% " has same scope-time"
				     " as tvar whose value it"
				     " occurs in"
				     (:novalue
				      "I'll treat it as if"
				      " scope-time were greater"))
				 false)
				(t 
;;;;				 (out (:to (errout))
;;;;				    "Unknown " ty " has scope-time "
;;;;				    ut "-- okay? " (< ut scope-time)
;;;;				    " [" scope-time "]" :%)
				 (< ut scope-time)))
			  undo-stack))
		      (t
		       (signal-problem
			  all-freevars-scope-okay
			  "Unknown type " ty
			  " has no scope-time"))))))))

(class-feature-datafun scope-time-checker typeform-class
   (defun (ty scope-time bound-unks undo-stack env)
      (typelist-freevars-scope-check
         (<# Vartype-type (type-find-feature ty 'args env))
         scope-time bound-unks undo-stack env)))

(class-feature-datafun scope-time-checker constructed-class
   (defun :^ (ty scope-time bound-unks undo-stack env)
      (typelist-freevars-scope-check
         (<# Vartype-type (type-find-feature ty 'arg-vartypes env))
         scope-time bound-unks undo-stack env)))

(defun supertype-scope-time-checker (ty scope-time bound-unks undo-stack env)
   (all-freevars-scope-okay
      (type-find-supertype ty env)
      scope-time bound-unks undo-stack env))

(class-feature-datafun scope-time-checker simple-typeform-class
   (defun (_ _ _ undo-stack _)
      (values true undo-stack)))

(class-feature-datafun scope-time-checker const-class
   (defun (_ _ _ undo-stack _)
      (values true undo-stack)))

(defun type-features-freevars-scope-check
               (type features bound-unks scope-time undo0 env)
   (repeat :for ((feat :in features)
		(undo-stack undo0)
		(okay true))
      (!= < okay undo-stack >
	  (all-freevars-scope-okay
	     (type-find-feature type feat env)
	     bound-unks scope-time undo-stack env))
    :while okay
    :result (okay-or-undo okay undo-stack undo0)))

(defun typelist-freevars-scope-check (types bound-unks scope-time undo0 env)
   (repeat :for ((ty :in types)
		(undo-stack undo0)
		(okay true))
      (!= < okay undo-stack >
	  (all-freevars-scope-okay ty bound-unks scope-time undo-stack env))
    :while okay
    :result (okay-or-undo okay undo-stack undo0)))
    

	    