;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: slot.lisp,v 1.6 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ vartypes)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(maybe-type-slot-fun ;;;;type-slot-fun
	     )))

;;; The slot class is defined in foundation.lisp, but becomes usable
;;; with the functions in this file.

;;; Slots define access and set functions for types, plus related
;;; functions such as constructors, is-checkers, etc.

;;; One key difference between slots and features is that slots are
;;; inherited by subtypes.  Features are inherited only if the hooks
;;; for a given feature says to look at supertypes.

;;; A slot consists of six things: name, access fun name, set fun name,
;;; type, atypes, invert and descriptor.  
;;; Normally, type is the type of the thing in the slot
;;; and atypes is an arglistspec giving the types of all access-fun arguments 
;;; except the first (which is of the type owning this slot). 
;;; For the conser "slot," the type is irrelevant (the conser
;;; constructs the type owning this slot); and the atypes gives the
;;; types of all the arguments.
;;; The descriptor, if not false, is a symbolic description of the slot
;;; that should unambiguously differentiate it from other slots the same
;;; underlying data structure might have.  If false, then acc is assumed 
;;; to play this role.
;;; 'invert' is a function that infers the type of an object given the
;;; type of this slot.  E.g., if 'x' of type (Tup a b - Number), then 
;;; the inverter for !_b takes type 'y' and returns (Tup a - Number b - y).

(defun get-slot-descriptor (s)
   (or (Slot-descriptor s)
       (Slot-acc s)
       (Slot-set s)   ; extreme cases!
       (Slot-name s)))  

;; If first arg is true, we will constrain tvars to find those
;; slots
;;; Returns < found, truetype, new-undo-stack >
;;; (See 'type-slot-find' for further explanation.)
(defun maybe-type-slot-fun (ctv ty slot wanted undo-stack vartypes)
   (multi-let (((sl truetype vartypes undo-stack-1)
		(type-slot-find ctv ty slot undo-stack vartypes))
	       (res false))
      (cond (sl
	     (!= res (selq wanted
			  ((acc access :read)
			   (Slot-acc sl))
			  ((set :set)
			   (Slot-set sl))
			  ((type :type)
			   (close-if-closable (Slot-type sl) vartypes))
			  ((atypes :atypes)
			   (close-if-closable
			      (Slot-atypes sl) vartypes))
			  ((invert :invert)
			   (Slot-invert sl))
			  ((descriptor :descriptor)
			   (Slot-descriptor sl))
			  (t
			   (new-Closable
			      sl
			      (let ((fvl (cond ((Slot-type sl)
						(type-find-freevars
						   (close-if-closable
						       (Slot-type sl)
						       vartypes)
						   vartypes))
					       ;; conser has no type
					       (t !()))))
				 (cond ((Slot-atypes sl)
					(freevars-union
					   (arglistspec-freevars
					      (close-if-closable
					         (Slot-atypes sl)
						 vartypes))
					   fvl))
				       (t fvl)))
			      #'slot-close))))
	     (values
		(cond ((and res (memq wanted '(acc access set)))
		       (slotfn-deint res))
		      (t res))
		truetype undo-stack-1))
	    (t
	     (values false
		     nil (undo undo-stack-1 undo-stack))))))

;;; Returns < found, truetype, env, new-undo-stack >
;;; If found is true, truetype is actual type where slot was found..
;;; env binds its free variables.
;;; If found is false, other fields are not interesting (except
;;; new-undo-stack is same as original).
(defun type-slot-find (ctv ty slotname undo-stack vartypes)
   (!= < ty vartypes >
       (follow-var-ref ty vartypes))
   (cond ((is-tvar-type ty)
	  (cond (ctv
		 (multi-let (((found truetype env undo-stack-1)
			      (slot-find-in-tvar-subs ty slotname undo-stack)))
		    (cond (found
			   (values true truetype env undo-stack-1))
			  (t
			   (slot-find-in-supers ty slotname
						(undo undo-stack-1 undo-stack)
						vartypes)))))
		(t
		 (slot-find-in-supers ty slotname undo-stack vartypes))))
	 (t
	  (slot-find-in-supers ty slotname undo-stack vartypes))))

(defun slot-find-in-supers (ty slotname undo-stack env)
   (repeat :for ((supty ty)
		slotspec)
      (!= slotspec (lookup-slot slotname (Type-slotfns supty)))
    :until slotspec
    :result (values slotspec supty env undo-stack)
      (!= < supty env >
	  (cond ((is-tvar-type ty)
		 (let ((supers (Typebdg-supers (tvar-type-bdg ty))))
		    (cond ((= (len supers) 1)
			   (follow-var-ref (car supers) empty-env*))
			  (t
			   (values false nil)))))
		(t
		 (type-find-supertype supty env))))
      :while supty
      else :result (values false nil nil undo-stack)))

(defun slot-find-in-tvar-subs (tvar slotname undo-stack)
   (let ((subs (Typebdg-subs (tvar-type-bdg tvar))))
      (cond ((= (len subs) 1)
	     (multi-let (((subty subenv)
			  (follow-var-ref (car subs) empty-env*)))
		(multi-let (((found truetype env undo-stack-1)
			     (let ((slotspec
				    (lookup-slot slotname
						 (Type-slotfns subty))))
				(cond (slotspec
				       (values true subty subenv undo-stack))
				      ((is-tvar-type subty)
				       (slot-find-in-tvar-subs
						 subty slotname undo-stack))
				      (t
				       (values false nil nil undo-stack))))))
		   (cond (found
			  (multi-let (((okay undo-stack-2)
				       (set-bdg-val
					  (tvar-type-bdg tvar)
					  (force-type-closure
					     truetype env)
					  tvar undo-stack-1)))
			     (cond (okay
				    (values true truetype env undo-stack-2))
				   (t
				    (values false nil nil
					    (undo undo-stack-2 undo-stack))))))
			 (t
			  (values false nil nil
				  (undo undo-stack-1 undo-stack)))))))
	    (t
	     (values false nil nil undo-stack)))))

(defun slot-close (sl vartypes)
   (make-Slot 
      (Slot-name sl)
      (Slot-acc sl)
      (Slot-set sl)
      (new-Type-closure (Slot-type sl) vartypes)
      (let ((argtypes (Slot-atypes sl)))
	 (cond ((not argtypes)
		false)
	       (t
		(new-Closable argtypes
			      (arglistspec-freevars argtypes)
			      #'arglistspec-close))))
      (Slot-invert sl)
      (Slot-descriptor sl)))

(defun slot-atypes-add-type-arg (sl-atypes slname type env)
   (multi-let (((paramlist paramenv)
		(follow-var-ref sl-atypes env)))
      (values
	 (cond ((memq slname '(conser is))
		paramlist)
	       (t
		(new-Arglistspec
		   (cons (new-Argspec '_ ':required type)
			 (cond ((null paramlist)
				'())
			       (t
				(<# argspec-copy
				    (Arglistspec-argspecs
				       paramlist))))))))
	 paramenv)))

;;; Probably useful, but now now.
(defun type-slot-compute (ty slotname vartypes)
   (let ((superfn (type-find-feature ty 'slot-filler vartypes)))
      (cond (superfn
             (let ((new (funcall superfn ty slotname vartypes)))
                (cond (new
		       (cond ((null (Type-freevars ty))
			      (!= (Type-slotfns ty) (cons new *-*))))
		       new)
		      (t false)  )))
	    (t false)   )))
                          
(defun Slot-access-fcn (s)
   (slotfn-deint (Slot-acc s)))

(defun Slot-set-fcn (s)
   (slotfn-deint (Slot-set s)))

;;; A slot function of the form (:inline vars . body) means the same
;;; as (lambda vars . body), but is a signal to def-type not to transform it
;;; into a named function.  'slotfn-deint' strips off this form when
;;; necessary
(defun slotfn-deint (e)
   (match-cond e
      ?( (:inline ?a1 (?f ?,@a1))
	f)
      ?( (:inline ?@d)
	`(lambda ,@d))
      (t e)))

(defun type-lookup-slot (name ty env)
   (multi-let (((ty env)
		(follow-var-ref ty env)))
      (let ((sl (lookup-slot name (Type-slotfns ty))))
	 (values sl env))))

(defun lookup-slot (name slots)
   (find name slots :key #'Slot-name :test #'eq))


;;; This is *not* the only place where Typebdg vals are set.
;;; opt/fundef.nsp is another.  
(defun set-bdg-val (b val var undo0)
   (let ((undo-stack undo0))
      (cond ((and (is-Type-closure val)
		  (null (Type-closure-env val))
		  (not (null (Type-freevars (Type-closure-type val)))))
	     (error-break :novalue
		"oops: " b " !!! " val)))
      (cond ((Typebdg-val b)
	     (error-break check-changed-typebdg :fatal
		"About to assign second val to " b)))
      (!=? (Typebdg-val b) val undo-stack
	   (dbg-out binding-dbg*
	      "Bound tvar ?" var
	      " to " (Typebdg-val b) :%)
	   (dbg-out binding-dbg*
	      "Removed val for ?" var :%))
      (!=? (Typebdg-subs b) '() undo-stack)
      (!=? (Typebdg-supers b) '() undo-stack)
      (typebdg-constraints-check b 'val (list val)
				 undo-stack undo0)))


;;No longer used
#|
(defun type-slot-fun (ty slot wanted meanit vartypes)
   (let ((sl (maybe-type-slot-fun true ty slot wanted vartypes)))
      (or sl
	  (cond (meanit
		  (let ((revised
		           (error-break type-slot-fun
			       (:q (sl "No " wanted " field for")
				  (t "Missing"))
			       " (type slot) = ("
			       (Type-desig ty) 1 slot ")"
			       (:prompt-for "Intended (type slot): " nil))))
		    (repeat
		     :until (and (consp revised)
				(consp (cdr revised))
				(is-type-desig (car revised) vartypes))
		       (ttymsg "Type a list (type slot):")
		       (!= revised (stdread))   )
		    (type-slot-fun (designated-type (car revised)
						    false vartypes)
				    (cadr revised)
				    wanted meanit vartypes)))
		(t false)))))
|#
