;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: subtype.lisp,v 1.6 2004/02/22 00:47:03 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ types istester scopecheck tuptypes
			 %ytools/ debug)

;; lispworks barfs if we don't do this:
(depends-on :at-compile-time %nity/ vartypes)

; This should be somewhere else, and use designators.
(defun declare-type-acceptable (got want)
   ;(!= want (coerce-to-type want))
   (cond (got
          ;(!= got (coerce-to-type got))
          (!= (type-feature got 'acceptable-as)
               (adjoinq want *-*)))
         (t (!= (type-feature want 'accept-any)
		true))   ))

;;; Find the common supertype of two constant types.   This ignores
;;; variables, so it's not that useful.
(defun common-supertype (t1 t2)
   (cond ((eq t2 void-type*) t1)
	 ((eq t1 void-type*) t2) 
	 ((is-subtype t2 t1) t1)
         (t
          (repeat
           :until (is-subtype t1 t2)
           :result t2
             (!= t2 (Type-super t2))
           :until (or (null t2) (eq t2 '*as-feature))
           :result univ-type*))))

; Cheap, simple test that just goes up the type hierarchy.  
; Often inappropriate
(defun is-subtype (t1 t2)
   (cond ((and t1 t2)
	  (let ((orig-t1 t1))
;;;;	     (out "In 'is-subtype', t2 = " t2 :%)
	     (repeat :for ((_ = 1 :to 100))
	      :result (progn
			 (dbg-save :run-loud orig-t1 t1 t2)
			 (signal-problem is-subtype
			    "Can't find top of " orig-t1 :% " looking for " t2
			    (:prompt-for "Boolean, whether it's a subtype" false)))
;;;;		(out " t1 = " t1 :%)
	      :until (eq t1 t2)
	      when :result (progn ;;;; (out "Yes, it is a subtype" :%)
				  true)
		(!= t1 (Type-super t1))
	      :while (and t1 (not (eq t1 '*as-feature)))
	      else :result (progn ;;;; (out "No, it isn't a subtype" :%)
				  false))))
	    (t false)))

(def-class Accept-handler (:options (:include Prioritized-handler))
   (:handler
      (print (ah srm)
	(out (:to srm) "#{Accept-handler " (Accept-handler-priority ah) "}"))))

(def-class Accept-stack
      (:handler (print (ast srm)
		   (out (:to srm) "#{Accept-stack " (Accept-stack-size ast)
			"}")))
   size
   contents)  ; A list of pairs of type-closures, representing a got-want
	      ; test in progress.
;;; Accept-stack should be based on List-with-length.

(defun cons-Accept-stack (p ast)
   (cond ((null ast)
	  (make-Accept-stack 1 (list p)))
	 (t
	  (make-Accept-stack (+ (Accept-stack-size ast) 1)
			     (cons p (Accept-stack-contents ast))))))

(defun accept-stack-pop (ast)
   (cond ((null ast) '())
	 (t (make-Accept-stack
	       (- (Accept-stack-size ast) 1)
	       (cdr (Accept-stack-contents ast))))))

(defun accept-as-equal (constrain-tvars t1 t2 undo-stack
			env1 env2 accept-stack)
   (multiple-value-let (okay undo1)
		       (type-acceptable constrain-tvars t1 t2 undo-stack
					env1 env2 accept-stack)
      (cond (okay
	     (multiple-value-let (okay undo2)
				 (type-acceptable constrain-tvars t2 t1
						  undo1 env2 env1
						  accept-stack)
		(values okay
			(maybe-undo okay undo2 undo-stack))))
	    (t
	     (values false (undo undo1 undo-stack))))))

;;; The two lists are already known to be of the same length.
(defun type-lists-acceptable (constrain-tvars lg lw undo0
			      envg envw accept-stack)
   (let ((undo-stack undo0))
      (repeat :for ((gt :in lg)
		   (wt :in lw)
		   okay)
       :result (values true undo-stack)
	 (!= < okay undo-stack >
	     (type-acceptable constrain-tvars gt wt undo-stack
			      envg envw accept-stack))
       :while okay
       else :result (values false (undo undo-stack undo0)))))

;;; Returns < successful, new undo-stack >
;;; If constraint-tvars is false, then we fail without trying
;;; to bind or constrain tvars in any way.
(defun try-accept (got wanted undo-stack
		   envg envw accept-stack)
   (type-acceptable true got wanted undo-stack envg envw accept-stack))

;; Call by which role ty1 plays.  which = got or want
(defun accept-check (ctv which ty1 ty2 undo-stack env1 env2 accept-stack)
   (cond ((eq which 'wanted)
	  (type-acceptable ctv ty2 ty1 undo-stack env2 env1 accept-stack))
	 (t
	  (type-acceptable ctv ty1 ty2 undo-stack env1 env2 accept-stack))))

(defvar max-accept-stack-size* 500)

(defun type-acceptable (constrain-tvars got wanted undo-stack
			envg envw accept-stack)
   (cond ((not got)
	  (dbg-save :run-loud got wanted envg envw)
	  (signal-problem type-acceptable
	     "got = nil")))
    (let ((o-wanted wanted) (o-envw envw))
      (multiple-value-let (got envg)
			  (follow-var-ref got envg)
	 (multiple-value-let (wanted envw)
			     (follow-var-ref wanted envw)
	    (cond ((or (is-Type-closure got)
		       (is-Type-closure wanted))
		   (dbg-save :run-loud
			     got wanted envg envw o-wanted o-envw)
		   (breakpoint type-acceptable
		      "Following var ref failed to eliminate Type-closure: "
		      got 1 wanted)))
	    (let ((r (lrecord (identifying-type-closure got envg)
			      (identifying-type-closure wanted envw))))
	       (cond ((and accept-stack
			   (> (Accept-stack-size accept-stack)
			      max-accept-stack-size*))
		      (signal-problem type-acceptable
			 "Apparent infinite recursion (depth "
			 (Accept-stack-size accept-stack)
			 ") in type-checker"
			 :novalue)))
	       (cond ((accept-stack-has-loop r accept-stack)
		      ;;(out (:to *error-output*) "Accept pair on stack" :%)
		      (values true undo-stack))
		     (t
		      (multi-let (((okay undo-stack-1)
				   (really-try-accept
					 constrain-tvars
					 got wanted undo-stack envg envw
					 (cons-Accept-stack
					    r accept-stack))))
			 (cond ((and (not constrain-tvars)
				     (not (eq undo-stack-1 undo-stack)))
				(signal-problem type-acceptable
				   "Undo-stack changed when tvars"
				   " were not to be constrained: "
				   :% 3 "got = " got
				   :% 3 "wanted = " wanted
				   (:novalue
				     "Undo unexpected changes and proceed"))))
			 (values okay undo-stack-1)))))))))

(defun really-try-accept (constrain-tvars got wanted
			  undo-stack envg envw accept-stack)
   ;; Since we've used follow-var-ref, we know that even if we find tvar
   ;; bindings, they contain no value.	  
   (cond ((or (eq got any-tvar*) (eq wanted any-tvar*)
	      (eq got void-type*) (eq wanted univ-type*))
	  (values true undo-stack))
	 ((is-tvar-type wanted)
	  (cond ((eq got wanted)
		 (values true undo-stack))
		(constrain-tvars
		 (multiple-value-let (okay undo-stack-2)
				     (tvar-try-accept wanted 'wanted got
						      undo-stack
						      envw envg accept-stack)
		    (cond (okay
			   (cond ((and (is-tvar-type got)
				       (not (Typebdg-val
					       (tvar-type-bdg wanted))))
				  ;; We throw in the check that 'wanted'
				  ;; didn't get set, because if it did,
				  ;; it got set to 'got'.
				  (multiple-value-let (got envg)
						      (follow-var-ref got envg)
				     (cond ((is-tvar-type got)
					    (tvar-try-accept
					       got 'got wanted undo-stack-2
					       envg envw accept-stack))
					   (t
					    (values true undo-stack-2)))))
				 (t
				  (values true undo-stack-2))))
			  (t
			   (undo undo-stack-2 undo-stack)
			   (values false undo-stack)))))
		(t
		 (values (check-tvar-bounds wanted false got envw envg
					    accept-stack)
			 undo-stack))))
	 ((is-tvar-type got)
	  (cond (constrain-tvars
		 (tvar-try-accept got 'got wanted
				  undo-stack envg envw accept-stack))
		(t
		 (values (check-tvar-bounds got true wanted envg envw
					    accept-stack)
			 undo-stack))))
	 (t
	  (multiple-value-let (accep found)
			      (type-find-feature got 'acceptable-as envg)  
	     (cond ((or (and found
			     (exists (a :in accep)
				(types-equal a wanted '() envw)))
			(type-find-feature wanted 'accept-any envw))
		    (values true undo-stack))
		   (t
		    ; (out "got = " got :% "envg = " envg :%)
; 		    (cond ((and envg1* (not envg))
; 			   (!= envg2* envg)
; 			   (signal-problem really-try-accept :novalue
; 			      "envg1* = " envg1*
; 			      "envg2* = " envg2*)))
		    (find-acceptable-supertype constrain-tvars
					       got wanted undo-stack
					       envg envw accept-stack)))))))

(defvar accept-handler-dbg* false)

(defun find-acceptable-supertype (constrain-tvars got wanted
				  undo-stack envg envw accept-stack)
   (let ((ahw (type-find-feature wanted 'got-other-want-me envw))
	 (ahg (type-find-feature got 'got-me-want-other envg))
	 (super got)
	 (identical false))
      (repeat 
	 (!= identical
	     (types-equal super wanted envg envw))
       :until identical
	 (!= < super envg >
	     (follow-var-ref 
		(type-find-supertype super envg)
		envg))
       :while super)
      (cond (identical
	     ;(out "identical" :%)
	     (values true undo-stack))
	    ((or ahg ahw)
	     (let ((use-w (and ahw
			       (or (not ahg)
				   (> (Accept-handler-priority ahw)
				      (Accept-handler-priority ahg))))))
		(dbg-out accept-handler-dbg*
		     "Got handler ahg = " ahg " class = " (Type-class got)
		     :% " ahw = " ahw " class = " (Type-class wanted) :%
		     " use w? " use-w :%)
;;;;		(cond ((eq (Type-class wanted) unk-class*)
;;;;		       (dbg-save :comp-quiet got wanted envg envw)
;;;;		       (breakpoint find-acceptable-supertype
;;;;			  "Want unk " wanted)))
		(multi-let (((okay undo-stack-2)
			     (let ((h (cond (use-w
					     (Accept-handler-fcn ahw))
					    (t
					     (Accept-handler-fcn ahg)))))
				(funcall h
				   constrain-tvars got wanted undo-stack
				   envg envw accept-stack))))
; 		   (cond ((not (tailp undo-stack undo-stack-2))
; 			  (dbg-save undo-stack undo-stack-2)
; 			  (error-break find-acceptable-supertype
; 			     "Handler fumbled undo-stack"
; 			     :novalue)))
		   (values okay
			   (maybe-undo okay undo-stack-2 undo-stack)))))
	     (t
	      ;;;;(out (:to *query-io*) "dbg-saving" :%)
	      (dbg-save :run-loud got wanted)
	      (error-break find-acceptable-supertype :fatal
		      "No accept-handler for <got> " got
		      :% " vs. <wanted> " wanted)))))

(defun check-tvar-bounds (ty1 got-ty1 ty2 env1 env2 accept-stack)
   (let ((bdg (tvar-type-bdg ty1)))
      (exists (b :in (cond (got-ty1 (Typebdg-supers bdg))
			   (t (Typebdg-subs bdg))))
	 (cond (got-ty1
		(type-acceptable false b ty2 '() env1 env2 accept-stack))
	       (t
		(type-acceptable false ty2 b '() env2 env1 accept-stack))))))

(defvar tvar-match-hook* false
   "If non-false, a function to call when a tvar is matched")

(defun tvar-try-accept (tvar which other undo-stack envv envo accept-stack)
   (tvar-kind-reconcile tvar other which envv envo)
   (multiple-value-let (bdg undo-stack-2)
		       (record-tvar-props tvar undo-stack envv accept-stack)
      (cond (bdg
	     (cond ((Typebdg-val-only bdg)
;;;;		    (dbg-save tvar bdg other envo)
;;;;		    (breakpoint tvar-try-accept
;;;;		       "About to set " tvar)
		    (!=? (Typebdg-val bdg)
			 (force-type-closure other envo)
			 undo-stack
			 (dbg-out binding-dbg*
			    "Set = tvar ?" (tvar-type-varname tvar)
			    " to " (Typebdg-val bdg) :%)
			 (progn (dbg-out binding-dbg*
				   "Removed val for ?" tvar :%)
;;;;				(breakpoint tvar-assign-undo
;;;;				   "Removed!")
				))
		    (values true undo-stack))
		   (t
		    (tvar-try-constrain
		       tvar bdg which other undo-stack-2
		       envv envo accept-stack))))
	    (t
	     (values false (undo undo-stack-2 undo-stack))))))

(defun tvar-kind-reconcile (tvar other which envv envo)
   (let ((vk (Type-kind tvar))
	 (ok (Type-kind other)))
      (let ((v-sub-o (is-subtype vk ok))
	    (o-sub-v (is-subtype ok vk)))
	 (cond ((not (case which (got o-sub-v) (t v-sub-o)))
		(letrec ()
		   (cond ((case which (got v-sub-o) (t o-sub-v))
			  (let ((ot (kind-trans other vk envo envv)))
			     (cond (ot
				    (!= other ot))
				   (t
				    (kind-complain)))))
			 (t
			  (kind-complain)))
		 :where
		   (kind-complain ()
		      (error-break tvar-try-accept :novalue
			 "Attempt to constrain tvar " tvar
			 ": " (Type-kind tvar) 
			 :% " with " (:q ((eq which 'got) "super")
				     (t "sub"))
			 1 other ": " (Type-kind other)))))))))

;;; The following never actually does anything but return the tvar's
;;; Typebdg, because tvars are never
;;; created with the unusual properties this procedure deals with.
;;; Return < tvar-bdg, new-undo-stack >
;;; If contradiction encountered, return < false, original-undo-stack >
;;; Note that, given where this is called, we already know that tvar has no
;;; val.
(defun record-tvar-props (tvar undo-stack venv accept-stack)
   (let ((uniqsub (type-local-feature tvar 'unique-sub))
	 (sup (Type-super tvar))
	 (bdg (tvar-type-bdg tvar)))
      (cond ((or uniqsub (not (eq sup univ-type*)))
	     (err-out "Nontrivial record-tvar-props: "
		      uniqsub  " / " sup :%)))
      (letrec ()
	 (multiple-value-let (okay undo-stack-2)
			     (record-uniqueness undo-stack)
	    (cond (okay
		   (multiple-value-let (okay undo-stack-3)
				       (record-sup undo-stack-2)
		      (cond (okay
			     (values bdg undo-stack-3))
			    (t
			     (values false (undo undo-stack-3 undo-stack))))))
		  (t
		   (values false (undo undo-stack-2 undo-stack)))))
       :where
	 (record-uniqueness (undo-stack)
	    (cond ((and uniqsub (not (Typebdg-unique-sub bdg)))
		   (make-bdg-unique-sub bdg tvar undo-stack accept-stack))
		  (t
		   (values true undo-stack))))

	 (record-sup (undo-stack-2)
	    (cond ((eq sup univ-type*)
		   (values true undo-stack-2))
		  (t
		   (try-add-unique-sub-or-super
		       bdg sup 'super tvar undo-stack-2
		       venv accept-stack)))))))

(defun tvar-try-constrain (tvar bdg which other undo-stack
			   envv envo accept-stack)
   (let ((dir (cond ((eq which 'wanted) 'sub)
		    (t 'super)))
	 (siblings (cond ((eq which 'got)
			  (Typebdg-subs bdg))
			 (t
			  (Typebdg-supers bdg)))))
      (repeat :for ((sib :in siblings)
		   ok (undo-stack-2 undo-stack))
       :result (multiple-value-let (okay undo-stack-3)
				  (typevar-new-sub-or-super
				     tvar bdg other dir
				     undo-stack-2 envo accept-stack)
		 (cond (okay
			(cond (tvar-match-hook*
			       (funcall tvar-match-hook*
				  tvar which other envv envo
				  (\\ (u)
				     (push u undo-stack-3)))))
			(values true undo-stack-3))
		       (t
			(values false (undo undo-stack-3
					    undo-stack)))))
	 (!= < ok undo-stack-2 >
	     (accept-check true which
			   (Type-closure-type sib) other
			   undo-stack-2
			   (Type-closure-env sib) envo
			   accept-stack))
	:while ok
	else :result (values false
			    (undo undo-stack-2 undo-stack)))))

(defun typevar-new-sub-or-super (tvar bdg new which
				 undo-stack newenv accept-stack)
   (let ((sym (Typebdg-varname bdg)))
      (let ((occ (tvar-occurs-in sym new true newenv)))
	 (cond ((eq occ 'same)
		(values true undo-stack))
	       (t
		(multi-let (((okay undo-stack-1)
			     (all-freevars-scope-okay
				new (type-feature tvar 'scope-time)
				'() undo-stack newenv)))
		   (cond (okay
			  (let ((uniqsub (or (Typebdg-unique-sub bdg)
					     (type-local-feature
						 tvar 'unique-sub))))
			     (cond ((or uniqsub (eq which 'super))
				    (try-add-unique-sub-or-super
				       bdg new which tvar undo-stack-1
				       newenv accept-stack))
				   (t	
				    (try-add-sub
				       bdg new tvar undo-stack-1
				       newenv accept-stack)))))
			 (t
			  (values false (undo undo-stack-1 undo-stack))))))))))

;;; bdg already occurs in tvarbdgs, but we need to set its unique-sub
;;; true, which means checking that its subs are kosher.
(defun make-bdg-unique-sub (bdg tvar undo-stack accept-stack)
   (let ((subs (Typebdg-subs bdg)))
      (cond ((and binding-dbg* (> (len subs) 1))
	     (err-out
		"Changing " tvar " to have just one alternative subtype,"
		:% "  which requires comparing all its existing subtypes")))
      (multi-let (((consistent comparisons undo-stack-1)
		   (typelist-try-order subs undo-stack accept-stack)))
	 (cond (consistent
		;; Put comparisons in more useful format, making every list
		;; the same length, showing comparisons of corresponding
		;; type to all the others.
		(!= comparisons (comparisons-complete *-*))
		(let ((new-subs
			 (repeat :for ((sub :in (Typebdg-subs bdg))
				       (compl in comparisons))
			  :when (not (exists (c :in compl)
					(memq c '(=< =<?))))
			  :collect sub)))
		  (!=? (Typebdg-subs bdg) new-subs undo-stack-1)
		  (values true undo-stack-1)))
	       (t
		(values false (undo undo-stack-1 undo-stack)))))))

(defun comparisons-complete (comparisons)
   (repeat :for ((_ :in comparisons :tail tail-i)
		 (_ :in (series 0 (- (len comparisons) 1))
		    :tail tail-s))
    :collect (let ((preds (ldiff comparisons tail-i)))
		`(,@(<# (\\ (pred)
			   (let ((p (assq (nth (car tail-s)
					       pred)
					  '((=< >=) (>= '=<)
					    (=<? >=?) (>=? =<?)
					    (= =)))))
			      (cond (p (cadr p))
				    (t false))))
			preds)
		  *self
		  ,@(car tail-i)))))

;;; "Unique" here means that there's just one legitimate super (or sub)
;;; in the end.

;;; Returns < okay-flag, undo-stack >
(defun try-add-unique-sub-or-super
	  (bdg new which tvar undo-stack newenv accept-stack)
   (multiple-value-let (consistent redundant new-list undo-stack-2)
		       (compare-with-previous-unique
			  new 
			  (typebdg-subs-or-supers bdg which)
			  which tvar
			  undo-stack newenv accept-stack)
      (cond (consistent
	     (cond (redundant
		    (values true undo-stack-2))
		   (t
		    (multiple-value-let (okay undo-stack-3)
					(set-bdg-subs-or-supers
					   bdg which
					   (cons (force-type-closure
						    new
						    newenv)
						 new-list)
					   undo-stack-2)
		       (values okay
			       (maybe-undo okay undo-stack-3 undo-stack))))))
	    (t
	     (cond ((not consistent)
		    (dbg-out binding-dbg*
		       "New " (:a which) " for tvar ?" (Typebdg-varname bdg)
		       " inconsistent with some old " (:a which) :%)))
	     (values false (undo undo-stack-2 undo-stack))))))

;;; Returns < okay-flag, undo-stack >
(defun try-add-sub (bdg new tvar undo-stack newenv accept-stack)
   (multiple-value-let (redundant new-list)
		       (compare-with-previous-nonunique
			  new 
			  (Typebdg-subs bdg)
			  'sub tvar newenv accept-stack)
      (cond (redundant
	     (values true undo-stack))
	    (t
	     (multiple-value-let (okay undo-stack-2)
				 (set-bdg-subs-or-supers
				    bdg 'sub
				    (cons (force-type-closure
					     new newenv)
					  new-list)
				    undo-stack)
		(values okay
			(maybe-undo okay undo-stack-2 undo-stack)))))))

;;; Return < consistent, redundant, nonredundant-olds, new-undo-stack >
;;; where consistent means: new doesn't contradict, 
;;; redundant means it doesn't constrain,
;;; and nonredundant-olds is the list with old elements deleted 
;;; that would be made
;;; redundant by the new element.
;;; If consistent=false, the values of 'nonredundant-olds' is meaningless
(defun compare-with-previous-unique
	       (new prev which tvar undo-stack new-env accept-stack)
   (cond ((exists (old :in prev)
	     (let ((old-type (Type-closure-type old))
		   (old-env (Type-closure-env old)))
		(types-equal new old-type new-env old-env)))
	  (values true true prev undo-stack))
	 (t
	  (let ((consistent true)
		(redundant false)
		(survivors '()))
	     (cond ((and binding-dbg* (not (null prev)))
		    (err-out
		       "Testing whether " new " can be added as "
		       (:a which) " of " tvar
		       :% " given existing values " prev)))
	     (repeat :for ((old :in prev)
			  old-type old-env
			  ok-high undo-high ok-low undo-low
			  already-high already-low
			  (new-role
			     (cond ((eq which 'sub) "new SUB ")
				   (t "new SUPER ")))
			  (old-role
			     (cond ((eq which 'sub) "old SUB ")
				   (t "old SUPER "))))
		(!= old-type (Type-closure-type old))
		(!= old-env (Type-closure-env old))
		(!= < ok-low already-low undo-low
		      ok-high already-high undo-high >
		    (types-try-order new old-type new-role old-role
				     undo-stack new-env old-env accept-stack))
		(multi-let (((ok-proximal already-proximal undo-proximal
			      ok-distal already-distal undo-distal)
			     (cond ((eq which 'sub)
				    (values ok-high already-high undo-high
					    ok-low already-low undo-low))
				   (t
				    (values ok-low already-low undo-low
					    ok-high already-high undo-high))))
			    (keep-old true))
		   (cond (already-distal
			  (dbg-out binding-dbg*
			     "New redundant; not recording it" :%)
			  (!= redundant true))
			 (already-proximal
			  (dbg-out binding-dbg*
			     "New more specific than old;"
			     " replacing old with new"
			     :%)
			  (!= keep-old false))
			 (ok-distal
			  (cond (ok-proximal
				 (dbg-out binding-dbg*
				    "Consistent either way; posting"
				    " constraint" :%)
				 (require-types-ordered-constraint-post
				    new old-type undo-stack
				    new-env old-env accept-stack))
				(t
				 (dbg-out binding-dbg*
				    "Making new redundant" :%)
				 (!= redundant true)
				 (!= undo-stack (redo *-* undo-distal)))))
			 (ok-proximal
			  (dbg-out binding-dbg*
			     "Making old redundant" :%)
			  (!= keep-old false)
			  (!= undo-stack (redo *-* undo-proximal)))
			 (t
			  (dbg-out binding-dbg*
			     "Inconsistent" :%)
			  (!= consistent false)))
		   (cond (keep-old
			  (!= survivors (cons old *-*)))))
	     :while consistent)
	  (values consistent redundant survivors undo-stack)))))

;;; Returns < consistent, pairwise-comparisons, new-undo-stack >.
;;; Each element of 'pairwise-comparisons' is 
;;; =<, >=, =<?, >=?, =, or false.
;;; Comparisons are in following order: 
;;; ((1 vs. 2, ..., 1 vs. k)
;;;  (2 vs. 3, ..., 2 vs. k)
;;;  (k-1 vs. k))
;;; Unless inconsistent, when comparisons are meaningless (but may be
;;; useful for debugging).
;;; Algorithm is greedy, so early choices may cause later
;;; inconsistencies.
(defun typelist-try-order (typelist undo-stack accept-stack)
   (let ((start-undo-stack undo-stack))
      (repeat :for ((typel typelist (cdr typel))
		   (consistent true)
		   (comparisons '())
		   ty1 env1)
       :until (null typel)
       :result (values true (reverse comparisons) undo-stack)
	 (!= < ty1 env1 > (follow-var-ref (car typel) empty-env*))
	 (repeat :for ((ty2 :in (cdr typel))
		      env2 
		      ok-ty1=<ty2 ok-ty2=<ty1
		      already-ty1=<ty2 already-ty2=<ty1
		      undo1 undo2 (compl '())
		      next-comp)
	  :result (!= comparisons (cons (reverse compl) *-*))
	    (!= < ty2 env2 > (follow-var-ref ty2 empty-env*))
	    (!= < ok-ty1=<ty2 already-ty1=<ty2 undo1
		  ok-ty2=<ty1 already-ty2=<ty1 undo2 >
		(types-try-order ty1 ty2 "" "next type " undo-stack
				 env1 env2 accept-stack))
	    (cond (already-ty1=<ty2
		   (!= next-comp (cond (already-ty2=<ty1 '=) (t '=<))))
		  (already-ty2=<ty1
		   (!= next-comp '>=))
		  (ok-ty1=<ty2
		   (cond (ok-ty2=<ty1
			  (require-types-ordered-constraint-post
			     ty1 ty2 undo-stack env1 env2 accept-stack)
			  (!= next-comp false))
			 (t
			  (!= next-comp '=<?)
			  (!= undo-stack (redo *-* undo1)))))
		  (ok-ty2=<ty1
		   (!= next-comp '>=?)
		   (!= undo-stack (redo *-* undo2)))
		  (t
		   (!= consistent false)))
	  :while consistent
	    (!= compl (cons next-comp *-*))
	  else :result nil)
       :while consistent
       else :result (values false comparisons
			   (undo undo-stack start-undo-stack)))))

(defun types-try-order (ty1 ty2 role1 role2 undo-stack env1 env2 accept-stack)
   (let (ok-ty1=<ty2 ok-ty2=<ty1 already-ty1=<ty2 already-ty2=<ty1
	 undo1 undo2)
      (letrec ()
	 (dbg-out binding-dbg*
	    "Considering " (:a role1) 1 ty1
	    :% " as SUB of " (:a role2) ty2 :%)
	 (!= < ok-ty1=<ty2 undo1 >
	     (try-accept ty1 ty2 undo-stack env1 env2 accept-stack))
	 (!= already-ty1=<ty2 (and ok-ty1=<ty2 (eq undo1 undo-stack)))
	 (debug-out ok-ty1=<ty2 already-ty1=<ty2 "SUB")
;;;;	 (err-out "Undoing ... len undo1: " (len undo1) :%)
	 (!= undo-stack (undo undo1 undo-stack))
;;;;	 (err-out "... done undoing len undo-stack: " (len undo-stack) :%)
	 (dbg-out binding-dbg*
	    "Considering " (:a role1) 1 ty1
	    :% " as SUPER of " (:a role2) 1 ty2 :%)
	 (!= < ok-ty2=<ty1 undo2 >
	     (try-accept ty2 ty1 undo-stack env2 env1 accept-stack))
	 (!= already-ty2=<ty1 (and ok-ty2=<ty1 (eq undo2 undo-stack)))
	 (debug-out ok-ty2=<ty1 already-ty2=<ty1 "SUPER")
;;;;	 (cond ((and ok-ty2=<ty1 (not already-ty2=<ty1))
;;;;		(dbg-save :run-loud undo2 undo-stack)
;;;;		(breakpoint types-try-order
;;;;		   "Ok but not necessary (len undo2) = " (len undo2)
;;;;		   " (len undo-stack) = " (len undo-stack) :%)))
	 (!= undo-stack (undo undo2 undo-stack))
	 (values ok-ty1=<ty2 already-ty1=<ty2 undo1
		 ok-ty2=<ty1 already-ty2=<ty1 undo2)

       :where

	 (debug-out (ok-in-loc already-in-loc relative-loc)
	    (dbg-out binding-dbg*
	       ty1 1 (:q (already-in-loc "is")
			(t "could " (:q ((not ok-in-loc) "not "))
			   "be"))
	       1 (:a relative-loc) :%)))))


;;; Like the previous version, but does not assume that there can be only
;;; one true element in the list if all the variables' values were known.
;;; (This is called for subs of tvars because they eventually get 
;;; combined into an alt-type.  But for consistency we supply a 'which'
;;; arg that is always 'sub'.)
;;; It does nothing backtrackable, so an undo-stack is not necessary.
;;; It can detect no inconsistency, so it returns just
;;; < redundant, survivors >
(defun compare-with-previous-nonunique
          (new prev which tvar new-env accept-stack)
   (let ((redundant false)
	 (survivors '()))
      (cond ((exists (old :in prev)
		(let ((old-type (Type-closure-type old))
		      (old-env (Type-closure-env old)))
		   (types-equal new old-type new-env old-env)))
	     (values true prev))
	    (t
	     (cond ((and binding-dbg* (not (null prev)))
		    (err-out
			"Testing whether " new " needs to be added as " which
			" of " tvar
			:% " given existing values " prev)))
	     (repeat :for ((old :in prev)
			  old-type old-env
			  is-high is-low)
		(!= old-type (Type-closure-type old))
		(!= old-env (Type-closure-env old))
		(dbg-out binding-dbg*
		   "Considering new " (:a which) 1 new
		   :% " as SUPER of old " (:a which) 1 old :%)
		(!= < is-high _ >
		    (type-acceptable false old-type new '()
				           old-env new-env accept-stack))
		(dbg-out binding-dbg*
		   new 1 (:q (is-high "is") (t "is not"))
		   " SUPER" :%)
		(dbg-out binding-dbg*
		   "Considering new " (:a which) 1 new
		   :% " as SUB of old " (:a which) 1 old :%)
		(!= < is-low _ >
		    (type-acceptable false new old-type '()
				           new-env old-env accept-stack))
		(dbg-out binding-dbg*
		   new 1 (:q (is-high "is") (t "is not"))
		   " SUB" :%)
		(cond ((or (and is-high (eq which 'super))
			   (and is-low (eq which 'sub)))
		       (!= survivors (cons old *-*))))
		(cond ((and is-low is-high)
		       (dbg-out binding-dbg*
			   "New equal to old")
		       (!= redundant true))
		      (is-low
		       (dbg-out binding-dbg*
			  "Consistent only as SUB" :%)
		       (cond ((eq which 'sub)
			      (!= redundant true))))
		      (is-high
		       (dbg-out binding-dbg*
			  "Consistent only as SUPER" :%)
		       (cond ((eq which 'super)
			      (!= redundant true))))))
	     (values redundant survivors)))))

;;;;(defun add-bdg-subs-or-supers (bdg which new-elt undo-stack)
;;;;   (set-bdg-subs-or-supers bdg which
;;;;			   (cons new-elt
;;;;				 (cond ((eq which 'sub)
;;;;					(Typebdg-subs bdg))
;;;;				       (t
;;;;					(Typebdg-supers bdg))))
;;;;			   undo-stack))

;;; This is called after we're sure that the new constraint is consistent.
(defun set-bdg-subs-or-supers (bdg which new-list undo-stack)
   (let ((var (Typebdg-varname bdg))
	 (ustack undo-stack)
	 (initial-len (len undo-stack))
	 (old-list (cond ((eq which 'sub)
			  (Typebdg-subs bdg))
			 (t
			  (Typebdg-supers bdg)))))
      (let ((new-stuff (set-difference new-list old-list :test #'eq)))
	 (letrec ((add-msg ()
		     (dbg-out binding-dbg*
			   "Adding new " (:a which) " to tvar ?" var ":"
			   :% new-stuff :%))
		  (del-msg ()
			(dbg-out binding-dbg*
			   "Removing new " (:a which)
			   " from tvar ?" var  ":"
			   :% new-stuff :%)))
	    (cond ((and (is-Type-closure (car new-list))
			(not (null (Type-closure-freevars (car new-list)))))
		   (dbg-save :run-loud bdg new-list)
		   (breakpoint set-bdg-subs-or-supers
		      "Suspicious closure")))
	    (cond ((eq which 'sub)
		   (!=? (Typebdg-subs bdg) new-list ustack
			(add-msg) (del-msg)))
		  (t
		   (!=? (Typebdg-supers bdg) new-list ustack
			(add-msg) (del-msg))))
	    (multiple-value-let (okay ustack-1)
				(check-changed-typebdg
				   bdg ustack which new-stuff)
	       (cond ((=< (len ustack-1) initial-len)
		      (signal-problem set-bdg-subs-or-supers
			 "Undo-stack screwup")))
	       (cond (okay
		      (values true ustack-1))
		     (t
		      (dbg-out binding-dbg*
			 "Change to ?" var " binding fails" :%)
;;;;		      (dbg-save var)
;;;;		      (breakpoint set-bdg-subs-or-supers
;;;;			 "Abject failure")
		      (undo ustack-1 undo-stack)
		      (values false undo-stack))))))))

(defvar fixedpointnum* 0)

(defvar allow-infinite-data* false)

;;; If super and sub have collapsed to a single possible type, that becomes 
;;; the val.
(defun check-changed-typebdg (b undo0 which new-stuff)
   (let ((val false)
	 (var (Typebdg-varname b))
	 (undo-stack undo0))
      (cond ((not (Typebdg-val b))
	     (repeat :for ((super :in (cons (make-Type-closure
					       univ-type* '() global-env*)
					    (Typebdg-supers b)))
			   sup-type sup-env)
	      :result nil
		(!= sup-type (Type-closure-type super))
		(!= sup-env (Type-closure-env super))
	      :until (type-feature sup-type 'primitive)
	      :when :result (!= val super)
		(repeat :for ((sub :in (cons (make-Type-closure
						void-type* '() global-env*)
					     (Typebdg-subs b)))
			      sub-type sub-env)
		 :result nil
		   (!= sub-type (Type-closure-type sub))
		   (!= sub-env (Type-closure-env sub))
		 :until (types-equal sup-type sub-type sup-env sub-env)
		 :result (!= val super))
	      :until val
	      :result nil)))
      (cond (val
	     (let ((occ (tvar-occurs-in var
				   (Type-closure-type val)
				   true
				   (Type-closure-env val))))
		(cond (occ
		       (cond ((eq occ 'proper-part)
			      (dbg-out binding-dbg*
				 "Tvar ?" var " occurs as proper part"
				 " of putative new value "
				  :% 2 val)
;;;;			      (dbg-save b var val)
;;;;			      (breakpoint check-changed-typebdg
;;;;				 "What's going on?")
			      (values false undo-stack))
			     (t ;;(eq occ 'same)
			      (values true undo-stack))))
		      (t
		       (multi-let (((res undo-stack)
				    (set-bdg-val b val var undo-stack)))
			  (okay-or-undo res undo-stack undo0))))))
	    (t
	     (typebdg-constraints-check b which new-stuff
					undo-stack undo0)))))

;;; These constraints are normally posted after acceptability tests that
;;; appear to succeed but don't leave enough of a trace, the prime 
;;; example being where Y is acceptable as both U1 and U2, so that
;;; the test whether it's acceptable as (Alt U1 U2) doesn't constrain
;;; anything.
;;; When an acceptor posts this constraint, 
;;; *** it must pop the top frame off the accept-stack *** because that
;;; frame refers to the very types we are now testing, so the repeat
;;; test will trivially succeed.
(defun acceptability-constraint-post (constrain-tvars got want undo0
				      envg envw accept-stack)
;;;;   (breakpoint acceptability-constraint-post
;;;;      "Checking " got
;;;;      :% " vs. " want)       
   (let ((con (new-Tvar-constraint
	         (\\ (which tcl undo-then)
		     (ignore which tcl)
		    (multi-let (((okay undo-after)
				 (type-acceptable
				    constrain-tvars got want undo-then
				    envg envw accept-stack
				    )))
		       (values (cond (okay 'unknown)
				     (t 'violated))
			       (maybe-undo okay undo-after undo-then)))))))
      (types-tvars-constrain (list got want)
			     (list envg envw)
			     con undo0)))

(defun require-types-ordered-constraint-post
                       (type1 type2 undo0 env1 env2 accept-stack)
   (let ((con (new-Tvar-constraint
		 (\\ (which tc undo-then)
		     (ignore which tc)
		    (multi-let (((ok-ty1=<ty2 ok-ty2=<ty1
				  _ _ undo1 undo2)
				 (types-try-order type1 type2 "" ""
						  undo-then env1 env2
						  accept-stack)))
		       (cond (ok-ty1=<ty2
			      (cond (ok-ty2=<ty1
				     (values 'unknown undo-then))
				    (t
				     (values 'satisfied
					     (redo undo-then undo1)))))
			     (ok-ty2=<ty1
			      (values 'satisfied
				      (redo undo-then undo2)))
			     (t
			      (values 'violated undo-then))))))))
      (types-tvars-constrain (list type1 type2)
			     (list env1 env2)
			     con undo0)))

(defun types-tvars-constrain (types envs constraint undo-stack)
   (let ((tvars (</ (\\ (tvl ty env)
		       (tvars-union (type-find-tvars ty false env !())
				    tvl))
		    '() types envs)))
      (repeat :for ((v :in tvars))
	 (!= undo-stack (tvar-add-constraint v constraint undo-stack)))
      undo-stack))

(defun typebdg-subs-or-supers (bdg which)
   (cond ((eq which 'sub)
	  (Typebdg-subs bdg))
	 (t
	  (Typebdg-supers bdg))))

(defun tvar-supers (tvar)
   (cond ((eq (Type-super tvar) univ-type*)
	  '())
	 (t (list (Type-super tvar)))))

(defun type-closures-equal (c1 c2)
   (types-equal (Type-closure-type c1)
		(Type-closure-type c2)
		(Type-closure-env c1)
		(Type-closure-env c2)))

;;; Deals with the possibility that x and y are not types.
(defun types-etc-equal (x y e1 e2)
   (let ((x-is-type (is-Typoid x))
	 (y-is-type (is-Typoid y)))
      (cond ((or x-is-type y-is-type)
	     (cond ((and x-is-type y-is-type)
		    (types-equal x y e1 e2))
		   (t false)))
	    (t (equal x y)))))

(defun type-change-checker (sym oldtype newtype type-sys)
   (let ((e (empty-vartypes type-sys)))
      (cond ((not (types-equal oldtype newtype e e))
	     (bind ((*print-pretty* true))
		(out (:to *error-output*)
		   (:pp-block (:pre "Warning: ")
			     "Changing type of " sym
			     (:pp-nl :linear *error-output*)
			     " from " oldtype
			     (:pp-nl :linear *error-output*)
			     " to " newtype :%)))))))      

(!= typedecl-checker* #'type-change-checker)

(!= (type-feature univ-type* 'is-tester)
    '(lambda (x) (ignore x) true))

(!= (type-feature void-type* 'is-tester)
    '(lambda (x) (ignore x) false))

;;; As long as no new types are created during a 'type-acceptable' 
;;; computation, we are guaranteed to detect a loop eventually.
;;; So any Accept-handler that creates a type must flag it as having
;;; a "seed," that is, the type it sprang from.  The loop checker
;;; checks the seeds if they exist.  Since new seeds are never created,
;;; we retain the loop-detection property (but not necessarily
;;; correctness).
;;; Note that these must be Type-closures, even if ty has no freevars. 

(defun type-place-seed (ty env)
   (or (type-find-feature ty 'seed env)
       (force-type-closure ty env)))

(defun identifying-type-closure (ty env)
   (or (type-find-feature ty 'seed env)
       (force-type-closure ty env)))

(defun accept-stack-has-loop (r accept-stack)
   (and accept-stack
	(exists (s :in (Accept-stack-contents accept-stack))
	   (closure-pair-equal s r))))

(defun closure-pair-equal (p1 p2)
   (and (type-closures-equal (car p1) (car p2))
	(type-closures-equal (cadr p1) (cadr p2))))

(!= (type-feature univ-type* 'got-me-want-other)
    (make-Accept-handler 50 (\\ (ctv me other undo-stack my-env other-env
			         accept-stack)
			      (ignore ctv me my-env other-env accept-stack)
			      (values (eq other univ-type*)
				      undo-stack))))

(!= (type-feature void-type* 'got-me-want-other)
    (make-Accept-handler 51 (\\ (ctv me other undo-stack my-env other-env
			         accept-stack)
			      (ignore ctv me other my-env other-env
				      accept-stack)
			      (values true undo-stack))))
					
;;; If a void is expected, any value will do, because it's going to be ignored
(!= (type-feature univ-type* 'got-other-want-me)
    (make-Accept-handler 51 (\\ (ctv other me undo-stack other-env my-env 
				accept-stack)
			      (ignore ctv me other my-env
				      other-env accept-stack)
			      (values true undo-stack))))

(!= (type-feature void-type* 'got-other-want-me)
    (make-Accept-handler 50 (\\ (ctv other me undo-stack my-env other-env
			         accept-stack)
			      (ignore ctv me my-env other-env accept-stack)
			      (values (eq other void-type*) undo-stack))))

(declare-class-feature got-me-want-other boolean-class
		       (make-Accept-handler 7 #'const-subtyper))

(declare-class-feature got-other-want-me boolean-class
		       (make-Accept-handler 7 #'const-subtyper))

(defun unk-subtyper (ctv got wanted undo-stack envg envw accept-stack)
                    (ignore ctv envg envw accept-stack)
   (values
      (and (eq (Type-class got) unk-class*)
	   (eq (Type-class wanted) unk-class*)
	   (= (unknown-type-id got)
	      (unknown-type-id wanted)))
      undo-stack))

(declare-class-feature got-me-want-other unk-class
		       (make-Accept-handler 4 #'unk-subtyper))

(declare-class-feature got-other-want-me unk-class
		       (make-Accept-handler 4 #'unk-subtyper))

;; Boring types are acceptable only as their supertypes.
(defun boring-subtyper (ctv got wanted undo-stack envg envw accept-stack)
                       (ignore
			ctv got wanted envg envw accept-stack)
   (values false undo-stack))

(declare-class-feature got-me-want-other boring-class
		       (make-Accept-handler 4 #'boring-subtyper))

(declare-class-feature got-other-want-me boring-class
		       (make-Accept-handler 4 #'boring-subtyper))

;;; For type-tup-type*, type-type*, and row-type*
(declare-class-feature got-me-want-other simple-typeform-class
		       (make-Accept-handler 20 #'boring-subtyper))

(declare-class-feature got-other-want-me simple-typeform-class
		       (make-Accept-handler 20 #'boring-subtyper))

(class-feature-datafun equal-tester typeform-class
   (defun (me other envm envo)
      (let ((my-fn (type-find-feature me 'fun envm))
	    (other-fn (type-find-feature other 'fun envo)))
	 (and my-fn other-fn
	      (or (eq my-fn other-fn)
		  (and (eq envm envo)
		       (eq (Tyfun-name my-fn)
			   (Tyfun-name other-fn))))
	      (<& (\\ (my-arg other-arg)
		     (cond ((and (eq (type-feature (Vartype-type my-arg)
						   'meta)
				     'type)
				 (eq (type-feature (Vartype-type other-arg)
						   'meta)
				     'type))
			    (types-equal
			       (new-Type-closure (Vartype-val my-arg)
						 envm)
			       (new-Type-closure (Vartype-val other-arg)
						 envo)
			       envm envo))
			   (t
			    (equal (Vartype-val my-arg)
				   (Vartype-val other-arg)))))
		  (type-find-feature me 'args envm)
		  (type-find-feature other 'args envo))))))

(class-feature-datafun equal-tester tvar-type-class
    (defun (ty1 ty2 env1 env2)
           (ignore env1 env2)
       (and (is-tvar-type ty2)
	    ;; Both tvars are known to be unassigned
	    (eq (tvar-type-varname ty1)
		(tvar-type-varname ty2)))))

(defun kind-trans (ty kind envty envk)
   (let ((kt (type-find-kind ty envty)))
      (cond ((type-acceptable false kt kind '() envty envk '())
	     ty)
	    (t
	     (let ((h (type-feature kt 'kind-transfn)))
	        (or (and h (funcall h ty kind envty envk))
		    (progn
		       (dbg-save :comp-quiet ty kind)
		       (error-break kind-trans
			  "Unable to transform " ty
			  :% " to be of kind " kind
			  (:novalue "I will just take it as is"))
		       ty)))))))

;;;;(!= (type-feature row-type* 'kind-transfn)
;;;;    (\\ (_ target-kind _ envk)
;;;;       (multi-let (((tk _)
;;;;		    (follow-var-ref target-kind envk)))
;;;;	  (cond ((eq tk type-tup-type*)
;;;;		 type-tup-type*)
;;;;		(t false)))))
;;;;
;;;;(!= (type-feature row-type* 'got-me-want-other)
;;;;    (make-Accept-handler
;;;;       20
;;;;       (\\ (ctv me other undo-stack my-env other-env accept-stack)
;;;;	  (ignore ctv me my-env accept-stack)
;;;;	  (multi-let (((other _)
;;;;		       (follow-var-ref other other-env)))
;;;;	     (values (eq other type-tup-type*)
;;;;		     undo-stack)))))
					
