;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: tloader.lisp,v 1.6 2004/03/13 04:20:45 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ desig vartypes feature foundation types
			 %langutils/ namespace)

(depends-on %nity/ trecycle)

;;; Slurp to get macro definitions
(end-header :continue-slurping)

;;; This code runs at type-resolution time, creating code that can
;;; be compiled and executed at load time, using only the run-time 
;;; facilities of the types module.  
;;; It's mirrored by code in trecycle.nsp for restoring the
;;; circularities at load time.

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(type-loader +empty-tloader-stack+ acc-env*
	     type-system-acc-env)))

;;; Counter for creating symbols needed when form is built
(needed-by-macros
(defvar form-var-no* 0)
)
;;; Counter for creating symbols needed when form is evaluated.
(defvar load-time-var-no* 0)

(defmacro handle-tloader-circularity (x^ circ-stack^ &rest body)
   (let ((form-var (build-symbol form- (++ form-var-no*)
				 (:package nil)))
	 (uc-var (build-symbol uc- (++ form-var-no*)
			       (:package nil))))
      (let ((form
	       `(let ((,uc-var (make-Under-construction ,x^ false '())))
		   (let ((,circ-stack^
			    (cons ,uc-var ,circ-stack^)))
		      (let ((,form-var (progn ,@body)))
			 (note-load-tag ,uc-var ,form-var))))))

;;;			(cond ((Under-construction-tag ,uc-var)
;;;			       `(note-load-tag
;;;				   ',',(Under-construction-tag ,uc-var)
;;;				   ,',form-var))
;;;			      (t ,form-var)))))))
	 form)))

(defun note-load-tag (tag thing-builder)
   (cond ((is-Under-construction tag)
	  (!= tag (Under-construction-tag tag))))
   (cond (tag
	  `(tag-for-re-cycling ',tag ,thing-builder))
	 (t
	  thing-builder)))

;	 (cond ((memq ':check flags)
;		`(if-not-circular ,x^ ,circ-stack^ ,form))
;	       (t form)))))

(defmacro if-not-circular (x^ circ-stack^ &rest body)
   `(or (check-circularity ,x^ ,circ-stack^)
	(handle-tloader-circularity ,x^ ,circ-stack^
	   ,@body)))

(defun check-circularity (x tl-stack)
   (repeat :for ((fr :in tl-stack))
    :result false
    :until (and (is-Under-construction fr)
	       (eq (Under-construction-thing fr)
		   x))
    :result (let ((load-tag (Under-construction-tag fr)))
	      (cond ((not load-tag)
		     (!= load-tag
			 (build-symbol load- (++ load-time-var-no*)))
	     (!= (Under-construction-tag fr) load-tag)))
	      (let ((circ (make-Circularity fr load-tag)))
		 (!= (Under-construction-circularities fr)
		     (cons circ *-*))
		 circ))))

;;; 'vartypes' is a currently available env.
;;; 'load-time-exp' is an expression whose value will be 
;;; 'vartypes' at load time.
(def-class Acc-env
   load-time-exp
   vartypes)

;;; This never changes during the recursion through a type, so
;;; we make it a dynamic variable.  
;;; **It must be bound to an Acc-env when a "loader" is called.**
(defvar acc-env*)

(defconstant +empty-tloader-stack+ !())

;;; The assumption is that any feature on cached-features does not
;;; need to be reloaded because it will be recomputed if needed.

(defun type-loader (typ tl-stack)
   (let ((d (Type-desig typ)))
      (cond ((is-CSymbol d)
	     (cond ((eq typ univ-type*)
		    'univ-type*)
		   ((eq typ void-type*)
		    'void-type*)
		   (t
		    (or			;(check-circularity typ tl-stack)
		        (load-by-name d acc-env* type-type*
			       typ
			       (\\ (ty)
				  (type-feature ty 'forward)))
			(type-guts-loader typ tl-stack)))))
	    (t
	     (type-guts-loader typ tl-stack)))))

;;; For use in def-type as well as here, because it loads type
;;; contents even if it has a global name.
(defun type-guts-loader (typ tl-stack)
   (let ((special-loader (type-feature typ 'type-loader)))
      (cond (special-loader
	     (or (check-circularity typ tl-stack)
		 (funcall special-loader typ tl-stack)))
	    (t
	     (anonymous-type-loader typ tl-stack)))))

;;; Because we put the circularity check above, none of the special
;;; loaders have to do it, but they do have to push the stack.

(class-feature-datafun type-loader typeform-class
   (defun (typ tl-stack)
      (let ((susp (type-feature typ 'suspended)))
	 (cond (susp
		(let ((fun-name (type-feature typ 'fun-name))
		      (arg-desigs (type-feature typ 'arg-desigs))
		      (ur-env (type-feature typ 'ur-env)))
		   (handle-tloader-circularity typ tl-stack
		      `(build-typeform ',fun-name
				       ',arg-desigs
				       ,(env-loader ur-env tl-stack)))))
	       (t
		(let ((fun (type-feature typ 'fun))
		      (args (type-feature typ 'args)))
		   (cond ((not fun)
			  (error-break typeform-class-type-loader-fcn
			     "Attempt to create loader for"
			     " non-suspended typeform with"
			     " no fun feature: " typ)))
		   (handle-tloader-circularity typ tl-stack
		      `(make-typeform ,(tyfun-loader fun tl-stack)
				      ,(list-conser
					  (<# (\\ (vt)
						 (vartype-loader
						    vt tl-stack))
					      args))))))))))

(class-feature-datafun type-loader simple-typeform-class
   (defun (typ _)
      (let ((v (type-feature typ 'var)))
	 (or v
	     (error-break simple-typeform-class-type-loader :fatal
		"No global var for " typ)))))

(defun anonymous-type-loader (typ tl-stack)
   (if-not-circular typ tl-stack
      `(new-Type ,(type-class-loader (Type-class typ))
		 ',(Type-desig typ)
		 ,(freevars-loader (Type-freevars typ) tl-stack)
		 ,(cond ((eq (Type-super typ) '*as-feature)
			 ''*as-feature)
			(t
			 (type-loader (Type-super typ) tl-stack)))
		 ,(list-conser (<# (\\ (s) (slot-loader s tl-stack))
				   (Type-slotfns typ)))
		 ,(list-conser (<# (\\ (feat)
				      (feature-loader
					 feat tl-stack))
				   (Type-features typ))))))
   
(defun freevars-loader (vars tl-stack)
   (letrec ()
      (cond ((and vars (is-CSymbol vars))
	     `',vars)
	    (t
	     (list-conser (<# fv-loader vars))))
    :where
      (fv-loader (v)
	 (cond ((is-param-type v)
		`(make-param-type ',(param-name v)
				  ,(type-loader (Type-kind v) tl-stack)))
	       (t
		(signal-problem freevars-loader
		   "Non-param freevar: " v)

;;		`(make-tvar-type ',(tvar-type-varname v)
;;				 ,(type-loader (Type-kind v) tl-stack))

		)))))

(defun slot-loader (sl tl-stack)
   `(make-Slot ',(Slot-name sl)
	       ',(Slot-acc sl)
	       ',(Slot-set sl)
	       ,(let ((ty (Slot-type sl)))
		   (cond (ty
;;;;			  (trace-around slot-type-loader
;;;;			     ("ty: " ty)
			  (let ((tl (thing-loader ty tl-stack)))
;;;;			     (cond ((car-eq tl 'quote)
;;;;				    (dbg-save :run-loud sl tl)
;;;;				    (breakpoint slot-loader
;;;;				       "Got: " tl)))
			     tl)
;;;;			     ("loader: " (car out-vals*)))
			  )
			 (t 'false)))
	       ,(thing-loader (Slot-atypes sl) tl-stack)
	       ,(thing-loader (Slot-invert sl) tl-stack)
	       ',(Slot-descriptor sl)))

(defun arglistspec-loader (spec tl-stack)
   `(new-Arglistspec
       (<# argspec-pos-clear
	   ,(list-conser (<# (\\ (as) (argspec-loader as tl-stack))
			     (Arglistspec-argspecs spec))))
       :allow-other-keys ',(Arglistspec-allow-other-keys spec)
       :names-matter ',(Arglistspec-names-matter spec)))

(defun argspec-loader (spec tl-stack)
   `(make-Argspec ',(Argspec-name spec)
		  ',(Argspec-mode spec)
		  ',(Argspec-position spec)
		  ,(type-loader (Argspec-type spec) tl-stack)
		  ,(thing-loader (Argspec-default spec) tl-stack)
		  (tuple ,@(<# (\\ (y) `(tuple ',(car y)
					       ,(thing-loader (cadr y) tl-stack)))
			       (Argspec-props spec)))))

(defun tyfun-loader (tyf tl-stack)
   (let ((tdefn (Tyfun-defn tyf)))
      (let ((is-con (is-Constructor-spec tdefn)))
	 (or (load-by-name (Tyfun-name tyf) acc-env*
			   (cond (is-con constructor-type*)
				 (t tyfun-type*))
			   tyf
			   (\\ (tf)
;;;;			      (out (tr forward-check ("fc> " tf)
			      (eq (Tyfun-defn tf) 'forward)
;;;;			      ("fc< " (car out-vals*))))
			      ))
	     (if-not-circular tyf tl-stack
		`(make-Tyfun
		       ',(Tyfun-name tyf)
		       ,(arglistspec-loader (Tyfun-params tyf) tl-stack)
		       ,(cond (is-con (constructor-spec-loader tdefn tl-stack))
			      (t (type-loader tdefn tl-stack)))
		       ,(env-loader (Tyfun-env tyf) tl-stack)))))))

(defun constructor-spec-loader (cspec tl-stack)
   (if-not-circular cspec tl-stack
      (let ((tyf (Constructor-spec-tyfun cspec))
;;;;	    (make-alspec (Constructor-spec-make-params cspec))
	    (type-alspec (Constructor-spec-type-params cspec)))
	 `(make-Constructor-spec
	     ,(cond (tyf (tyfun-loader tyf tl-stack))
		    (t false))
;;;;	     ,(cond (make-alspec
;;;;		     (arglistspec-loader make-alspec tl-stack))
;;;;		    (t 'false))
	     ,(cond (type-alspec
		     (arglistspec-loader type-alspec tl-stack))
		    (t 'false))
	     ))))

;;;;	     ',(Constructor-spec-params cspec)

(defun feature-loader (feat tl-stack)
   `(tuple ',(car feat)
	   ,(thing-loader (cadr feat) tl-stack)))

(defun vartype-loader (vt tl-stack)
   (if-not-circular vt tl-stack
      `(make-Vartype ',(Vartype-var vt)
		     ,(type-loader (Vartype-typ vt) tl-stack)
		     ,(thing-loader (Vartype-initial vt) tl-stack)
		     ,(list-conser
			 (<# (\\ (p)
				`(tuple ',(car p)
					,(thing-loader (cadr p) tl-stack)))
			     (Vartype-props vt))))))

(defun type-closure-loader (x tl-stack)
   (if-not-circular x tl-stack
      (let ((vx (gensym)) (ex (gensym)))
	 `(let ((,vx ,(thing-loader (Type-closure-type x) tl-stack))
		(,ex ,(env-loader (Type-closure-env x)
					     tl-stack)))
	     (make-Type-closure ,vx
				(env-freevars (Typoid-freevars ,vx)
					      ,ex false)
				;;(Typoid-freevars ,vx) ??
				,ex)))))

;;; This version is a bit absurd; it's a good idea to check for
;;; duplicate envs, but there's no reason to suppose that envs are
;;; involved in "circularities" in the sense meant here.  We really
;;; need a more general hack for maintaining EQ-lity between objects
;;; across loads.
(defun env-loader (env tl-stack)
   (cond ((null (Env-vartypes env))
	  (cond ((Env-is-closing env)
		 ''empty-env*)
		(t
		 ''no-env*)))
	 (t
	  (if-not-circular env tl-stack
	     `(make-Env ',(Env-is-closing env)
		,(letrec ()
		    (vartypes-loader (Env-vartypes env)
				     tl-stack)
		  :where
		    (vartypes-loader (envl tl-stack)
		       (cond ((null envl) ''())
			     (t
			      (if-not-circular envl tl-stack
				 (let ((vt (car envl)))
				    `(cons ,(cond ((is-Vartype vt)
						   (vartype-loader vt tl-stack))
						  (t
						   (make-loader vt tl-stack)))
					   ,(vartypes-loader
					       (cdr envl) tl-stack)))))))))))))

(class-feature-datafun type-loader param-type-class
   (defun (ty tl-stack)
      (cond ((and (null (Type-slotfns ty))
		  (matchq ((param ?()))
			  (Type-features ty)))
	     `(make-param-type ',(param-name ty)
			       ,(type-loader (Type-kind ty) tl-stack)))
            (t (anonymous-type-loader ty tl-stack)))))

(class-feature-datafun type-loader tvar-type-class
   (defun (ty tl-stack)
      (cond ((forall (feat in (Type-features ty))
		 (eq (car feat) 'bdg))
	     `(make-tvar-type ',(tvar-type-varname ty)
			      ,(type-loader (Type-kind ty) tl-stack)))
            (t
	     (out (:to *error-output*) "Warning: tvar with features "
		  (<# car (Type-features ty)) :%)
	     (anonymous-type-loader ty tl-stack)))))
	     
(def-op make-loader (x tl-stack)
   (ignore tl-stack)
   (error-break make-loader :fatal
	      "Don't know how to make loader for " x))

(defun thing-loader (x tl-stack)
   (cond ((listp x)
	  (let ((elts (<# (\\ (y) (thing-loader y tl-stack))
			  x)))
	     (cond ((null elts) ''())
		   ((forall (e in elts)
		       (car-eq e 'quote))
		    `',x)
		   (t (list-conser elts)))))
	 ((or (not x) (is-CSymbol x) (is-String x) (is-Number x) (is-Char x))
	  `',x)
	 (t (make-loader x tl-stack))))

(def-meth make-loader ((s Type-system) _)
   (type-system-loader s))

(def-meth make-loader ((x Type) tl-stack)
   (type-loader x tl-stack))

(def-meth make-loader ((vt Vartype) tl-stack)
   (vartype-loader vt tl-stack))

(def-meth make-loader ((x Closable) tl-stack)
   `(make-Closable ,(thing-loader (Closable-thing x) tl-stack)
		   ,(freevars-loader (Closable-freevars x) tl-stack)
		   ,(let ((ct (Closable-trans x)))
		       (cond ((not ct)
			      'false)
			     ((or (symbolp ct)
				  (and (consp ct)
				       (symbolp (car ct))))
			      `',ct)
			     (t
			      (make-loader ct tl-stack))))))

(def-meth make-loader ((lf Loadable-function) _)
   `(make-Loadable-function ',(Loadable-function-name lf)
			    ,(cond ((or (not (Loadable-function-fcn lf))
					(eq (Loadable-function-fcn lf)
					    (symbol->fun
					       (Loadable-function-name lf))))
				    'false)
				   (t
				    (signal-problem
				       Loadable-function-make-loader
				       "Don't know how to make-loader for "
				       lf)))))

(def-meth make-loader ((lfe Loadable-function-with-extras) tl-stack)
   `(make-Loadable-function-with-extras
       :name ',(Loadable-function-name lfe)
       :fcn false
       :towit (list ,@(mapcar (\\ (x) (thing-loader x tl-stack))
			      (Loadable-function-with-extras-towit lfe)))))

(def-meth make-loader ((x Type-closure) tl-stack)
   (type-closure-loader x tl-stack))

(def-meth make-loader ((alspec Arglistspec) tl-stack)
   (arglistspec-loader alspec tl-stack))

(def-meth make-loader ((cs Constructor-spec) tl-stack)
   (constructor-spec-loader cs tl-stack))       

(def-meth make-loader ((x Tyfun) tl-stack)
   (tyfun-loader x tl-stack))

(!= (type-feature univ-type* 'type-loader)
    (\\ (ty _) (ignore ty) 'univ-type*))

(!= (type-feature void-type* 'type-loader)
    (\\ (ty _) (ignore ty) 'void-type*))

(!= (type-feature bool-type* 'type-loader)
    (\\ (ty _) (ignore ty) 'bool-type*))

(defun load-by-name (sym ae sym-type val in-progress)
   (let ((vt (var-lookup sym (Acc-env-vartypes ae))))
      (cond (vt
	     (cond ((eq (Vartype-type vt) sym-type)
		    (let ((val-in-env (Vartype-val vt)))
		       (cond ((eq val-in-env val)
			      `(tag-for-re-cycling ',sym
				  (var-val ',sym ,(Acc-env-load-time-exp
						     ae))))
			     ((funcall in-progress val-in-env)
			      false)
			     (t
			      (dbg-save :run-loud val val-in-env sym ae)
			      (signal-problem load-by-name
				 "Symbol " sym " - " (Type-desig sym-type)
				 :% 3 "has visible definition: "
				 :% 5 val-in-env
				 :% 3 "that differs from the definition I'm"
				 " producing a builder for:"
				 :% 5 val
				 :% 3 "Environment: " ae
				 :fatal)))))
		   (t
		    (signal-problem load-by-name
		       "Symbol " sym " has visible declaration as"
		       :% 5 (Vartype-type vt)
		       :% 3 "I'm trying to build an object of that name"
		       " with type:"
		       :% 5 sym-type
		       :fatal))))
	    (t
	     (signal-problem load-by-name
		"Undefined symbol " sym " is used as name of"
		:% 3 "object of type " sym-type
		:% 3 "that I'm trying to produce a builder for"
		:fatal)))))

(defun list-conser (element-forms)
   (cond ((null element-forms) ''())
	 (t
	  `(list ,@element-forms))))

(defmacro compile-time-designated-type (desig type-sys-name)
   (bind ((acc-env* (type-system-acc-env
		       (find-type-system type-sys-name true))))
      `(circular-object-restore 
	  ,(type-loader (designated-type
			   desig
			   true
			   (empty-vartypes
			      (find-type-system type-sys-name true)))
			'())
	  type-re-cycle)))
		 
(defun type-system-acc-env (sys)
   (make-Acc-env
	 `(place-type-system-env ,(type-system-loader sys))
	 (place-type-system-env sys)))

