;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: trecycle.lisp,v 1.6 2005/12/26 20:34:24 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ vartypes)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(type-re-cycle circular-object-restore)))

;;; Slurp to get macro definitions
(end-header :continue-slurping)

;;; We have to mimic the apparatus of
;;; tloader to restore circularities ("re-cycle") at load time.
;;; It is important that the mimicker of a loader examine the
;;; pieces of an object in exactly the same order that the loader's
;;; code constructs them.

;;; Some of the stuff in here is here because code generated at
;;; run time by tloader needs it.

;;; The things on the tloader stack:
(def-class Under-construction
      (:handler (print (c srm)
		   (bind ((*print-level* 5)
			  (*print-length* 5)
			  (*print-circle* true))
		      ;; We're probably dealing with a circular
		      ;; structure
		      (out (:to srm) "#{Under construction "
			   (Under-construction-thing c)
			   (:q ((not (null (Under-construction-circularities
					     c)))
			       " [" (len (Under-construction-circularities
					    c)) "] tag: "
				    (Circularity-tag c)))
			   "}"))))
   thing
   tag     ;; false until circularity detected, then unique sym.
   circularities)
   
;;; owner is an Under-construction (at form-build time); false at load
;;; time.  var is false at form-build time; at load time,  a var that
;;; is bound to an object that contains this circularity as a subpart.  
(def-class Circularity (:options (:kind :object))
   (:handler (print (circ srm)
		(out (:to srm)
		     "#{Circularity "
		     (Circularity-tag circ) 
		     " : " (Circularity-owner circ)
		     "}"))
	     (make-load-form (c &optional env)
		             (declare (ignore env))
		`(make-Circularity
		    false ',(Circularity-tag c))))
   owner
   tag)

(defvar non-circularities* !())

(def-meth Circularity-tag ((x t))
   (let ((tag (alref non-circularities* x)))
      (cond ((not tag)
             (!= tag (gensym))
             (!= (alref non-circularities* x)
                 tag)))
      tag))

(defvar obj-frag-table* '*not-properly-bound)

(defun tag-for-re-cycling (tag thing)
   (add-to-obj-frag-table tag thing)
   thing)

(defun add-to-obj-frag-table (tag thing)
   (make-sure-frag-table-okay)
   (let ((p (assq tag obj-frag-table*)))
      (cond (p
	     (cond ((not (eq (cadr p) thing))
		    (signal-problem add-to-obj-frag-table
		       "Tag " p " being defined for second time"
		       (:novalue
			"I'll ignore all but the first definition")))))
	    (t
	     (!= p (tuple tag thing))
	     (!= obj-frag-table*
		 (cons p *-*))))))

(defun find-circ-in-obj-frag-table (circ)
   (let ((p (find-in-obj-frag-table (Circularity-tag circ))))
      (or p
	  (signal-problem find-circ-in-obj-frag-table
	     "Can't find circularity tag for " circ))))

(defun find-in-obj-frag-table (tag)
   (make-sure-frag-table-okay)
   (let ((p (assq tag obj-frag-table*)))
      (and p (cadr p))))

(defun make-sure-frag-table-okay ()
   (cond ((eq obj-frag-table* '*not-properly-bound)
	  (signal-problem make-sure-frag-table-okay
	     "obj-frag-table* not bound before attempting to rebuild"
	     :% " circular structure"))))

(needed-by-macros

(defvar load-var-no* 0)

(defun lambda-for-\\ (fcn)
   (cond ((matchq (\\ ?@_) fcn)
	  `(lambda ,@(cdr fcn)))
	 (t fcn)))

(defun acc-normal-setter (acc)
   (match-cond acc
      ?( (?(:\| \\  lambda) (?v) ?form)
	`(lambda (,v | newval |) (!= ,form | newval |)))
      (t
       `(lambda (| whole | | newval |)
	    (!= (,acc | whole |) | newval |)))))

)

(defmacro circular-object-restore (object-builder^ object-re-cycler)
   (let ((new-obj-var (build-symbol load- (++ load-var-no*))))
      `(bind ((obj-frag-table* '()))
	  (let ((,new-obj-var ,object-builder^))
	     (,object-re-cycler ,new-obj-var)
	     ,new-obj-var))))

;;whole^ and tag-tab^ must be variables.
;; Returns true if piece was a circularity, and hence was replaced.
;; Returns false if piece was not a circularity, and hence
;; must be re-cycled.
(defmacro see-circularity (whole^ acc &optional set)
   (!= acc (lambda-for-\\ *-*))
   (cond ((not set)
	  (!= set (acc-normal-setter acc))))
   (let ((load-var (build-symbol load- (++ load-var-no*))))
      `(let ((,load-var (,acc ,whole^)))
	  (cond ((is-Circularity ,load-var)
		 (,set ,whole^ (find-circ-in-obj-frag-table ,load-var))
		 true)
		(t false)))))

;;; Streamlined version
(defmacro see-circ-or-recurse (form re-cycler)
   (match-cond form
      ?( (?acc ?whole^)
	(!= acc (lambda-for-\\ *-*))
	`(or (see-circularity ,whole^ ,acc)
	     (,re-cycler (,acc ,whole^))))
      (t
       (signal-problem see-circ-or-recurse
	  "Ill-formed: " `(see-circ-or-recurse ,form ,re-cycler)))))

(defmacro list-elts-circ-or-recurse (l^ re-cycler)
   (let ((tail-var (gensym)))
      `(repeat :for ((,tail-var ,l^ (cdr ,tail-var)))
        :until (null ,tail-var)
	  (see-circ-or-recurse (car ,tail-var)
			       ,re-cycler))))

;;; Up to this point, types are not involved.  The whole mechanism
;;; could be used to restore an arbitrary object.

(defmacro feature-circ-or-recurse (type^ feat re-cycler)
   `(see-circ-or-recurse ((\\ (ty) (type-feature ty ',feat))
			  ,type^)
			 ,re-cycler))

(defun type-re-cycle (typ)
   (let ((d (Type-desig typ)))
      (cond ((is-CSymbol d)
	     (cond ((and (not (eq typ univ-type*))
			 (not (eq typ void-type*))
			 (not (loaded-by-name d)))
		    (type-guts-re-cycle typ))))
	    (t
	     (type-guts-re-cycle typ)))))

(defun type-guts-re-cycle (typ)
   (let ((special-re-cycler (type-feature typ 'type-re-cycler)))
      (cond (special-re-cycler
	     (funcall special-re-cycler typ))
	    (t
	     (anonymous-type-re-cycle typ)))))

(class-feature-datafun type-re-cycler simple-typeform-class
   (defun (_)
      nil))

(defun 1-arg-ignore (_) (values))

(defun anonymous-type-re-cycle (typ)
   (see-circ-or-recurse (Type-freevars typ)
			freevars-re-cycle)
   (cond ((not (eq (Type-super typ) '*as-feature))
	  (see-circ-or-recurse (Type-super typ)
			       type-re-cycle)))
   (repeat :for ((sl :in (Type-slotfns typ)))
      (slot-re-cycle sl))
   (repeat :for ((feat :in (Type-features typ)))
      (see-circ-or-recurse (cadr feat)
			   thing-re-cycle)))

(class-feature-datafun type-re-cycler typeform-class
   (defun (type)
      (cond ((type-feature type 'suspended)
	     (feature-circ-or-recurse type ur-env env-re-cycle))
	    (t
	     (feature-circ-or-recurse type fun tyfun-re-cycle)
	     (list-elts-circ-or-recurse (type-feature type 'args)
					vartype-re-cycle)))))
(defun freevars-re-cycle (vars)
   (repeat :for ((vl vars (cdr vl)))
    :until (null vl)
      (or (see-circularity vl car)
	  (feature-circ-or-recurse (car vl) kind type-re-cycle))))

(defun slot-re-cycle (sl)
   (or (see-circularity sl Slot-type)
       (let ((slty (Slot-type sl)))
          (cond (slty
		 (thing-re-cycle slty)))))
   (see-circ-or-recurse (Slot-atypes sl)
			thing-re-cycle))

(defun arglistspec-re-cycle (spec)
   (list-elts-circ-or-recurse (Arglistspec-argspecs spec)
			      argspec-re-cycle))

(defun argspec-re-cycle (spec)
   (see-circ-or-recurse (Argspec-type spec)
			type-re-cycle)
   (see-circ-or-recurse (Argspec-default spec)
			thing-re-cycle))

(defun tyfun-re-cycle (tyf)
   (cond ((not (loaded-by-name (Tyfun-name tyf)))
	  (see-circ-or-recurse (Tyfun-params tyf)
			       arglistspec-re-cycle)
	  (see-circ-or-recurse (Tyfun-defn tyf)
			       thing-re-cycle)
	  (see-circ-or-recurse (Tyfun-env tyf)
			       env-re-cycle))))

(defun vartype-re-cycle (vt)
   (see-circ-or-recurse (Vartype-type vt)
			type-re-cycle)
   (see-circ-or-recurse (Vartype-val vt)
			thing-re-cycle)
   (repeat :for ((p :in (Vartype-props vt)))
      (see-circ-or-recurse (cadr p)
			   thing-re-cycle)))

(defun type-closure-re-cycle (tc)
   (see-circ-or-recurse (Type-closure-type tc)
			thing-re-cycle)
   (see-circ-or-recurse (Type-closure-env tc)
			env-re-cycle))

(defun env-re-cycle (env)
   (list-elts-circ-or-recurse
      (Env-vartypes env)
      (lambda (vt)
	 (cond ((is-Vartype vt)
		(vartype-re-cycle vt))
	       (t (re-cycle vt))))))

(def-op re-cycle (x)
	  (error-break re-cycle :fatal
	      "Don't know how to re-cycle " x))

(defun thing-re-cycle (x)
   (cond ((listp x)
	  (list-elts-circ-or-recurse x thing-re-cycle))
	 ((not (or (is-Symbol x) (is-String x) (is-Number x) (is-Char x)))
	  (re-cycle x))))

(def-meth re-cycle ((cspec Constructor-spec))
   (constructor-spec-re-cycle cspec))

(defun constructor-spec-re-cycle (cspec)
   (see-circ-or-recurse (Constructor-spec-tyfun cspec)
			tyfun-re-cycle))

(defun constructor-tyfun-circularize (tyf)
   (!= (Constructor-spec-tyfun (Tyfun-defn tyf)) tyf)
   tyf)

(def-meth re-cycle ((cs Canonized-symbol))
   cs)

(def-meth re-cycle ((ty Type))
   (type-re-cycle ty))
	 
(def-meth re-cycle ((x Closable))
   (see-circ-or-recurse (Closable-thing x)
			thing-re-cycle))

(def-meth re-cycle ((x Type-closure))
   (type-closure-re-cycle x))

(def-meth re-cycle ((x Arglistspec))
   (arglistspec-re-cycle x))

(def-meth re-cycle ((x Tyfun))
   (tyfun-re-cycle x))

(def-meth re-cycle ((s Type-system))
   nil)

(def-meth re-cycle ((vt Vartype))
   (vartype-re-cycle vt))

(def-meth re-cycle ((lfn Loadable-function))
   `(make-Loadable-function ',(Loadable-function-name lfn) false))

(defun maybe-change-name (name ty)
   (cond ((not (is-Symbol (Type-desig ty)))
	  (!= (Type-desig ty) name)))
   ty)

(defun loaded-by-name (sym)
   (find-in-obj-frag-table sym))

