;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: tuptypes-valtup.lisp,v 1.3 2006/03/10 15:46:58 dvm Exp $

(depends-on %nity/ foundation vartypes feature alttype)

(end-header :continue-slurping)

;;; Lins are just like Dots except they are used to represent Args and Vals.

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(Empty-tup Empty-row empty-tup-type* empty-row-type* feat-class
	     Lst Rol  Dot Lin  Tup Row
             Arg Val make-lstype)))

(in-type-system universal)

;;; The type of Lisp dot-pairs.
(def-type-class dot-class :universal
  (:var dot-class*))

;;; The type of Arg/Val non-keyword tuples
(def-type-class lin-class :universal
  (:var lin-class*))

;;; The type of lists with fields identified by keywords, such as
;;; (:a "how" :aardvark "cool" :abalone "is" :z "this?").
(def-type-class ktp-class :universal
  (:var ktp-class*))

;;; The type of keyword Arg tuples (as passed to functions).
(def-type-class key-class :universal
  (:var key-class*))

;; (Img (u) (Lst u) ((Row t1 ... tk)))
;; = (Row (Lst t1) ... (Lst tk))
(def-type-class image-class (:system :universal)
   (:var image-class*))

(def-type-class tup-class :universal
   (:var tup-class*))

(def-type-class val-class :universal
   (:var val-class*))

;;; Used only to convert Dots to Lins. 
(def-type-class to-row-class :universal
   (:var to-row-class*))

(defvar empty-tup-type*
    (make-type-of-class boring-class
	'Empty-tup '() univ-type* '()
	(list (tuple 'home universal-type-sys*)
	      (tuple 'is-tester 'null)
	      (tuple 'consts '(empty-list))
	      (tuple 'kind type-tup-type*))))

(define-global-constant-type Empty-tup empty-tup-type* :universal)

(defvar empty-row-type*
    (make-type-of-class boring-class
	'Empty-row '() univ-type* '()
	(list (tuple 'home universal-type-sys*)
	      (tuple 'kind row-type*))))

(define-global-constant-type Empty-row empty-row-type* :universal)

(defun empty-type (med)
   (cond ((eq med 'Tup) empty-tup-type*)
	 (t empty-row-type*)))

;;; Type behaves like a labeled table of its features.
;;; Thrown in here to avoid creating separate files to describe it.
(def-type-class feat-class (:system :universal)
   (:var feat-class*))

;;; Rol is to Lst as Row is to Tup

(defparameter rol-type-fun*
        (let ((rol-tyfun 
		 (new-Tyfun 'Rol
			    (new-Arglistspec
		               (list (new-Argspec
                                        'elty ':required type-type*)))
			    void-type*
			    (empty-vartypes universal-type-sys*)))
	      (elt-param (make-param-type 'elty)))
	   (let ((body-defn
		   (make-nontrivial-alt-type
		      (list empty-row-type*
			    (make-type-of-class lin-class
			       `(Lin (param elty)
				     (Rol (param elty)))
			       (list elt-param)
			       univ-type*
			       '()
			       (list (tuple 'cartype
					    (make-type-closable
					       elt-param false))
				     (tuple 'cdrtype
					    (new-Closable
					       (make-typeform
						  rol-tyfun
						  (list
						     (new-Vartype
						      'elty
						      type-type*
						      elt-param)))
					       (list elt-param)
					       false))
				     (tuple 'medium 'Row)
				     (tuple 'kind type-tup-type*))))
		      univ-type* no-env*)))
	      ;; Cribbed from def-type-fun:
              (!= (Type-features body-defn)
		  (nconc (list (tuple 'defined-fun-args
				      (make-type-closable
					 (make-param-type
					    'defined-fun-args-param)
					 'param-close-elements))
			       (tuple 'defined-fun
				      (make-type-closable
					 (make-param-type 'defined-fun-param)
					 false)))
		       *-*))
	      (!= (Tyfun-defn rol-tyfun)
		  body-defn)
	      rol-tyfun)))

(typedecl 'Rol
	  tyfun-type*
	  rol-type-fun*
	  universal-type-sys*)

(defun make-roltype (elt-param)
	  (make-typeform
	     rol-type-fun*
	     (list (new-Vartype 'elty type-type* elt-param))))

(defun Lst-is-builder (|IS U|)
   (\\ (x) (is-list-of x |IS U|)))

;;;;                #'(lambda (x)
;;;;                    (labels ((is-tester-14 (x)
;;;;                               ((lambda (x)
;;;;                                  (or (null x)
;;;;                                      ((lambda
;;;;                                        (x)
;;;;                                        (and
;;;;                                         (is-pair x)
;;;;                                         ((lambda (x) (funcall |is u| x))
;;;;                                          (car x))
;;;;                                         (is-tester-14 (cdr x))))
;;;;                                       x)))
;;;;                                x)))
;;;;                      (is-tester-14 x))))


(defparameter lst-type-fun*
        (let ((lst-tyfun 
		 (new-Tyfun 'Lst
			    (new-Arglistspec
		               (list (new-Argspec 'elty ':required type-type*)))
			    void-type*
			    (empty-vartypes universal-type-sys*)))
	      (elt-param (make-param-type 'elty)))
	   (let ((body-defn
		   (make-nontrivial-alt-type
		      (list empty-tup-type*
			    (make-type-of-class dot-class
			       `(Dot (param elty)
				     (Lst (param elty)))
			       (list elt-param)
			       univ-type*
			       '()
			       (list (tuple 'cartype
					    (make-type-closable
					       elt-param false))
				     (tuple 'cdrtype
					    (new-Closable
					       (make-typeform
						  lst-tyfun
						  (list
						     (new-Vartype
						      'elty
						      type-type*
						      elt-param)))
					       (list elt-param)
					       false))
				     (tuple 'medium 'Tup)
				     (tuple 'kind type-type*))))
		      univ-type* no-env*)))
	      (!= (type-feature body-defn 'is-builder-name)
		  'Lst-is-builder)
	      ;; Cribbed from def-type-fun:
              (!= (Type-features body-defn)
		  (nconc (list (tuple 'defined-fun-args
				      (make-type-closable
					 (make-param-type
					    'defined-fun-args-param)
					 'param-close-elements))
			       (tuple 'defined-fun
				      (make-type-closable
					 (make-param-type 'defined-fun-param)
					 false)))
		       *-*))
	      (!= (Tyfun-defn lst-tyfun)
		  body-defn)
	      lst-tyfun)))
				
(typedecl 'Lst
	  tyfun-type*
	  lst-type-fun*
	  universal-type-sys*)
		     
(defun make-lstype (elt-param)
	  (make-typeform
	     lst-type-fun*
	     (list (new-Vartype 'elty type-type* elt-param))))

;;; Replace each argspec A with one in which argument is _type_ of argspec
;;; of A.
(defun arglistspec-type-ify (alspec env)
   (let ((old-argspecs 
	    (Arglistspec-argspecs alspec)))
      (let ((new-argspecs (<# argspec-copy old-argspecs)))
	 (repeat :for ((new-a :in new-argspecs))
	    (let ((old-type (Argspec-type new-a)))
	       (!= (Argspec-type new-a)
		   (or (type-find-kind old-type env)
		       type-type*))
	       (!= (argspec-feat new-a 'constraints)
		   (list (tuple '< old-type)))))
	 (new-Arglistspec new-argspecs))))
