;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: tvarelim.lisp,v 1.10 2005/12/26 20:34:25 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ typeintsect feature dotformers 
			        subtype tupformers funformers altformer)

;;; Set type variables (with scope-time > after)
;;; so as to maximize all contravariant parts,
;;; or minimize all orthovariant parts, of ty.
;;; in-ortho is true iff we're in an orthovariant part of a type.
;;; Instructions is a list (ortho-instructions contra-instructions),
;;; where each element is either ':maximize', ':max-if-constrained',
;;; ':minimize', ':min-if-constrained', or false.
;;; Returns < resulting type, unconstrained-vars, undo-stack >
;;; Unconstrained vars are those tvars surviving because they were 
;;; unconstrained, in the case where the relevant instruction was
;;; ':max-if-constrained' or ':min-if-constrained'.
(defun tvar-elim (ty in-ortho instructions after undo-stack env)
   (multi-let (((ty env)
		(follow-var-ref ty env)))
      (cond ((is-tvar-type ty)
	     (cond ((=< (type-must-find-feature ty 'scope-time env)
			after)
		    (values ty '() undo-stack))
		   (t
		    (let ((act (cond (in-ortho (car instructions))
				     (t (cadr instructions)))))
		       (cond (act
			      (this-tvar-elim
			         ty act in-ortho instructions
				 after undo-stack env))
			     (t
			      (values ty '() undo-stack)))))))
	    (t
	     (let ((h (type-find-feature ty 'tvar-flusher env)))
	        (cond (h
		       (funcall h ty in-ortho instructions after
				  undo-stack env))
		      ((and (is-CSymbol (Type-desig ty))
			    (not (eq (Type-desig ty) '*as-feature)))
;;;;		       (cond ((null undo-stack)
;;;;			      (out "Null undo-stack for " ty)))
		       (values ty '() undo-stack))
		      (t
		       (signal-problem tvar-elim
			  "No tvar-flusher handler for " ty))))))))

(defun this-tvar-elim (tvar act in-ortho instructions after
		       undo-stack env)
   (let ((tvbdg (tvar-type-bdg tvar)))
      (cond ((and (null (Typebdg-constraints tvbdg))
		  (or (and (eq act ':max-if-constrained)
			   (null (Typebdg-supers tvbdg)))
		      (and (eq act ':min-if-constrained)
			   (null (Typebdg-subs tvbdg)))))
	     (values tvar (list tvar) undo-stack))
	    (t
	     (let ((wantmax (memq act '(:maximize :max-if-constrained))))
		(multi-let (((newval rem-tvars undo-stack-1)
			     (multi-let (((okay bounds rem-tvars undo-stack-1)
					  (typelist-tvar-flusher
					     (cond (wantmax
						    (Typebdg-supers tvbdg))
						   (t
						    (Typebdg-subs tvbdg)))
					     in-ortho instructions after
					     undo-stack env)))
				(cond (okay
				       (values
					  (cond (wantmax
						 (try-types-intersect
						    bounds))
						(t
						 (make-alt-type bounds empty-env*)))
					  rem-tvars undo-stack-1))
				      (t
				       (values false nil
					       (undo undo-stack-1 undo-stack)))))))
		   (multi-let (((okay undo-stack-2)
				(set-bdg-val tvbdg
					     newval
					     (tvar-type-varname tvar)
					     undo-stack-1)))
		      (cond ((not okay)
			     (signal-problem this-tvar-elim
				"Can't set value of " tvar
				" to " :% newval
				:% "  without violating constraints"
				(:novalue "I will set it anyway"))))
		      (values newval rem-tvars undo-stack-2))))))))

(defun try-types-intersect (tcl)
   (repeat :for ((tc :in tcl)
		expr (intsofar univ-type*))
    :result intsofar
      (!= < expr intsofar _ >
	  (types-intersect false true intsofar true tc !(Undo)
			   global-env* empty-env* !()))
    :while expr
    :else :result (cond ((null tcl) univ-type*)
			;; More accurate, but arbitrary, would be
			;; any element of tcl.
			(t (types-common-supertype
			      false tcl (empty-undo-stack) empty-env*)))))

(class-feature-datafun tvar-flusher const-class
   (defun (ty _ _ _ undo-stack _)
      (values ty !() undo-stack)))

(class-feature-datafun tvar-finder const-class
   (defun (ty _ _ _ undo-stack _)
      (values ty !() undo-stack)))

(class-feature-datafun tvar-flusher boring-class
   (defun (ty _ _ _ undo-stack _)
      (values ty !() undo-stack)))

(class-feature-datafun tvar-flusher fun-class
   (defun (ty in-ortho instructions after undo-stack env)
;;;;      (multi-let (((rty rem us)
      (cond ((> (type-must-find-feature ty 'level env)
		0)
	     ;; For now we assume level-1 functions have no variables.
	     (values ty !() undo-stack))
	    (t
	     (multi-let (((aty _ undo-stack-1)
			  (tvar-elim (type-must-find-feature ty 'argtype env)
					(not in-ortho) instructions
					after undo-stack env))
			 (arglist (funtype-extract-arglist ty env))
			 (reslist (funtype-extract-reslist ty env))
			 (rem-tvar-a nil))
		(cond (aty
		       (cond (arglist
			      (!= < arglist rem-tvar-a undo-stack-1 >
				  (arglistspec-tvar-elim
				     arglist (not in-ortho) instructions
				     after undo-stack-1 env))
			      (cond ((not arglist)
				     (signal-problem fun-class-tvar-flusher
					"Failed to tvar-flusher arglist of "
					ty)))))
		       (multi-let (((rty rem-tvar-r undo-stack-2)
				    (tvar-elim
				       (type-find-feature ty 'resulttype env)
				       in-ortho instructions after
				       undo-stack-1 env)))
			  (cond (rty
				 (cond (reslist
					(!= < reslist _ undo-stack-2 >
					    (arglistspec-tvar-elim
					       reslist in-ortho instructions
					       after undo-stack-2 env))
					(cond ((not reslist)
					       (signal-problem fun-class-tvar-flusher
						  "Failed to tvar-flusher reslist of "
						  ty)))))
;;;;				 (out "undo-stack-2 = " undo-stack-2 :%)
				 (values (make-funtype
					    0 rty aty reslist arglist
					    (type-feature ty 'side-effects env)
					    env)
					 (union  rem-tvar-a rem-tvar-r :test #'eq)
					 undo-stack-2))
				(t
				 (values false
					 '()
					 (undo undo-stack-2
					       undo-stack))))))
		      (t
		       (values false !() (undo undo-stack-1 undo-stack)))))))
;;;;      ))
;;;;      (cond ((and (null us)
;;;;		  (not (null undo-stack)))
;;;;	     (dbg-save rem rty undo-stack)
;;;;	     (breakpoint fun-class-tvar-elim
;;;;		"ustack screwed up; rty = " rty)))
;;;;      (values rty rem us))
))

(class-feature-datafun tvar-flusher unk-class
   (defun (ty _ _ _ undo-stack _)
       (values ty !() undo-stack)))

(class-feature-datafun tvar-flusher typeform-class
   (defun (ty in-ortho instructions after undo-stack env)
      (cond ((type-find-feature ty 'nonmonotonic env)
	     (tvar-elim (type-find-feature ty 'expansion env)
			in-ortho instructions after undo-stack env))
            (t
	     (multi-let (((okay evtl rem-tvars ustack)
			  (vartypes-tvar-elim (type-find-feature ty 'args env)
					      in-ortho instructions after
					      undo-stack env)))
		(cond (okay
		       (values (make-typeform
				  (type-find-feature ty 'fun env)
				  evtl)
			       rem-tvars
			       ustack))
		      (t
		       (values false '() (undo ustack undo-stack)))))))))

(defun vartypes-tvar-elim (vtl in-ortho instructions after undo-stack env)
   (repeat :for ((vt :in vtl)
		 val-maxmin
		 (rem-tvars '())
		 (ustack undo-stack)
		 rtl okay
		 :collector maxmin-args)
    :result (values true maxmin-args rem-tvars ustack)
      (cond ((type-find-feature (Vartype-type vt) 'meta env)
	     (!= < val-maxmin rtl ustack >
		 (tvar-elim (Vartype-val vt)
			    in-ortho instructions after ustack env))
	     (!= rem-tvars
		 (union  rtl *-* :test #'eq))
	     (!= okay val-maxmin))
	    (t
	     (!= okay true)
	     (!= val-maxmin (Vartype-val vt))))
    :while okay 
    :collect (new-Vartype (Vartype-var vt)
			  (Vartype-type vt)
			  val-maxmin)
    :else :result (values false '() rem-tvars (undo ustack undo-stack))))

(class-feature-datafun tvar-flusher alt-class
   (defun (ty in-ortho instructions after undo-stack env)
      (multi-let (((okay alts rem-tvars undo1)
		   (typelist-tvar-flusher
		      (type-find-feature ty 'alt-types env)
		      in-ortho instructions after undo-stack env)))
	 (cond (okay
		(values
		   (make-alt-type alts env)
		   rem-tvars
		   undo1))
	       (t
		(values false '() undo1))))))

(defun dot-lin-tvar-flusher (ty in-ortho instructions after undo-stack env)
   (multi-let (((okay adl rem-tvars undo1)
		(typelist-tvar-flusher
		   (<# (\\ (n) (type-feature ty n env))
		       '(cartype cdrtype))
		   in-ortho instructions after undo-stack env)))
      (cond (okay
	     (values (make-dot-type (type-find-feature ty 'medium env)
				    (car adl) (cadr adl) env)
		     rem-tvars
		     undo1))
	    (t
	     (values false '() (undo undo1 undo-stack))))))

(declare-class-feature tvar-flusher dot-class #'dot-lin-tvar-flusher)
(declare-class-feature tvar-flusher lin-class #'dot-lin-tvar-flusher)

(class-feature-datafun tvar-flusher to-row-class
   (defun (ty in-ortho instructions after undo-stack env)
      (let ((uty (type-find-feature ty 'source-type env)))
	 (multi-let (((uty rem-tvars undo1)
		      (tvar-elim uty in-ortho instructions after
				    undo-stack env)))
	    (cond (uty
		   (values (make-to-row-type uty env)
			   rem-tvars
			   undo1))
		  (t
		   (values false '() (undo undo1 undo-stack))))))))

(defun tup-val-tvar-flusher (ty in-ortho instructions after undo-stack env)
   (multi-let (((alspec rem-tvars undo-stack-1)
		(arglistspec-tvar-elim
		   (type-must-find-feature ty 'elt-types env)
		   in-ortho instructions after undo-stack env)))
      (cond (alspec
	     (values (make-tup-type
		        (type-find-feature ty 'medium env)
			alspec false env)
		     rem-tvars
		     undo-stack-1))
	    (t
	     (values false '() (undo undo-stack-1 undo-stack))))))

(declare-class-feature tvar-flusher tup-class #'tup-val-tvar-flusher)
(declare-class-feature tvar-flusher val-class #'tup-val-tvar-flusher)

(class-feature-datafun tvar-flusher constructed-class
   (defun :^ (ty in-ortho instructions after undo-stack env)
      (let ((cspec (type-find-feature ty 'constructor env))
	    (argtypes (type-find-feature ty 'arg-vartypes env)))
	 (multi-let (((okay arg-vartypes rem-tvars undo-stack-1)
		      (vartypes-tvar-elim argtypes in-ortho instructions after
					  undo-stack env)))
	    (cond (okay
		   (values (finish-constructed-type
			      cspec arg-vartypes env)
			   rem-tvars undo-stack-1))
		  (t
		   (values false '() (undo undo-stack-1 undo-stack))))))))

(class-feature-datafun tvar-flusher modified-class
   (defun (ty in-ortho instructions after undo-stack env)
      (multi-let (((suptype undo-stack-1)
		   (tvar-elim (type-find-supertype ty env)
				 in-ortho instructions after undo-stack env)))
	 (cond (suptype
		(multi-let (((okay slots rem-tvars undo-stack-2)
			     (slots-tvar-flusher
			        (Type-slotfns ty)
				in-ortho instructions after undo-stack env)))
		   (cond (okay
			  (values 
			     (make-modified-type
				(Type-desig ty)
				suptype slots
				(Type-features ty))
			     rem-tvars undo-stack-2))
			 (t
			  (values false '() (undo undo-stack-2 undo-stack))))))
	       (t
		(values false '() (undo undo-stack-1 undo-stack)))))))

(defun slots-tvar-flusher (slots in-ortho instructions after undo-stack env)
   (repeat :for ((sl :in slots)
		(new-slots '())
		(undo1 undo-stack)
		(rem-tvars '())
		slty slalspec rtl)
    :result (values true (reverse new-slots) rem-tvars undo1)
      (cond ((Slot-type sl)
	     (!= < slty rtl undo1 >
		 (tvar-elim (Slot-type sl)
			       in-ortho instructions after undo1 env)))
	    (t
	     (!= slty false)))
    :while (or (not (Slot-type sl))
	      slty)
      (!= rem-tvars (union  rtl *-* :test #'eq))
      (!= < slalspec rtl undo1 >
	  (arglistspec-tvar-elim
	     (Slot-atypes sl) in-ortho instructions after undo1 env))
    :while slalspec
      (!= rem-tvars (union  rtl *-* :test #'eq))
      (!= new-slots
	  (cons (make-Slot (Slot-name sl)
			   (Slot-acc sl)
			   (Slot-set sl)
			   slty slalspec
			   (Slot-invert sl)
			   (Slot-descriptor sl))
		*-*))
    else :result (values false nil '() (undo undo1 undo-stack))))

(class-feature-datafun tvar-flusher image-class
   (defun (ty in-ortho instructions after undo-stack env)
      (let ((itervars (type-find-feature ty 'template-params env))
	    (tmtype (type-find-feature ty 'template-type env))
	    (args (type-find-feature ty 'template-args env)))
	 (multi-let (((okay args rem-tvars undo-stack-1)
		      (typelist-tvar-flusher
		          args in-ortho instructions after undo-stack env)))
	    (cond (okay
		   (values (make-image-type itervars tmtype args
					    false env)
			   rem-tvars undo-stack-1))
		  (t
		   (values false '() (undo undo-stack-1 undo-stack))))))))

(defun arglistspec-tvar-elim (alspec in-ortho instructions after undo-stack env)
   (repeat :for ((aspec :in (Arglistspec-argspecs alspec))
		aty rtl
		(new-argspecs '())
		(rem-tvars '())
		(ustack undo-stack))
    :result (values (new-Arglistspec (reverse new-argspecs))
		    rem-tvars
		    ustack)
      (!= < aty rtl ustack >
	  (tvar-elim (Argspec-type aspec)
			in-ortho instructions after ustack env))
    :while aty
      (!= rem-tvars (union  rtl *-* :test #'eq))
      (!= new-argspecs (cons (new-Argspec
			        (Argspec-name aspec)
				(cond ((memq (Argspec-mode aspec)
					     '(:required :optional :rest))
				       (Argspec-mode aspec))
				      (t (Argspec-position aspec)))
				aty
				(Argspec-default aspec)
				(<# list-copy (Argspec-props aspec)))
			     *-*))
    else :result (values false nil '() (undo ustack undo-stack))))

(class-feature-datafun tvar-flusher subtype-class
   (defun (ty _ _ _ undo-stack _)
      (values ty '() undo-stack)))

(class-feature-datafun tvar-flusher simple-typeform-class
   (defun (ty _ _ _ undo-stack _)
      (values ty '() undo-stack)))

(defun typelist-tvar-flusher (types in-ortho instructions after undo-stack env)
	     (repeat :for ((arg :in types)
			  arg-maxmin rtl
			  (maxmin-args '())
			  (rem-tvars '())
			  (ustack undo-stack))
	      :result (values
		        true
		        (reverse maxmin-args)
			rem-tvars
			ustack)
	        (!= < arg-maxmin rtl ustack >
		    (tvar-elim arg in-ortho instructions after ustack env))
	      :while arg-maxmin
	        (!= maxmin-args (cons arg-maxmin *-*))
		(!= rem-tvars (union  rtl *-* :test #'eq))
	      else :result (values false nil '() (undo ustack undo-stack))))

