;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: tyfun.lisp,v 1.4 2004/03/13 04:20:46 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

;; Types of functions and types on functions

(depends-on :at-run-time %nity/  argspair slot feature desig)

(end-header)

;; Tyfuns are no longer defined here, but in vartypes.

;;; slot-filler features are currently not used
(class-feature-datafun slot-filler param-type-class
   (defun (param slot vartypes)
      (multiple-value-let (ty vartypes)
                          (var-val-spread (param-name param) vartypes)
	  (cond ((same-param ty param)
		 (values false false))
		(t
		 (type-slot-find false ty slot !() vartypes))))))

(class-feature-datafun feature-filler param-type-class
   (defun (oty feat vartypes)
      (multiple-value-let (ty vartypes)
                          (var-val-spread (param-name oty) vartypes)
	  (cond ((same-param ty oty)
		 (values false false))
		(t
		 (type-find-feature ty feat vartypes))))))

(class-feature-datafun feature-filler typeform-class
  (defun (ty feat vartypes)
     (cond ((eq feat 'expansion)
	    (values (typeform-expand ty vartypes)
		    true))
           ((memq feat '(fun args kind))
	    ;; These are put on *this* type by the expander, so we
	    ;; don't do the usual thing.
	    ;; This is executed for the possible caching side effect 
	    (type-find-feature ty 'expansion vartypes)
	    (multi-let (((val found)
			 (type-feature ty feat)))
	       (cond ((not found)
		      (error-break typeform-class-feature-filler-fcn
			 "Expanding typeform failed to cache feature "
			 feat
			 :% ty
			 (:novalue
			  "I will proceed with NIL value; expect trouble"))))
	       (values val found)))
           ((eq feat 'is-tester)
	    ;; Kludge: You don't need the expansion to find the is-tester,
	    ;; so you have to return false here and run the is-builder.
	    (values false false))
           (t
	    (multiple-value-let (e got-ex)
				(type-find-feature ty 'expansion vartypes)
               (cond (got-ex
		      (cond ((eq feat 'supertype)
			     (let ((sup (type-find-supertype e vartypes)))
				(values sup (not (not sup)))))
;;;;			    ((eq feat 'designator)
;;;;			     `(,(Tyfun-name
;;;;				   (type-feature ty 'fun))
;;;;			       ,@(<# (\\ (a)
;;;;					(designator-if-type a vartypes))
;;;;				     (type-feature ty 'args))))
			    (t
			     (multi-let (((a got)
					  (type-find-feature e feat vartypes)))
				(cond (got
				       (values a true))
				      (t
				       (values false false)))))))
		     (t
		      (error-break typeform-feature-filler-fcn :fatal
			 "Typeform with no expansion: " ty))))))))

;;; There are three environments we have to keep straight:
;;; ur-env: The env in effect when the typeform was built.  Has bindings
;;;   of symbols to Unks giving params
;;; Tyfun-env: The env in effect when the Tyfun was defined.  Has bindings
;;;   of free params in Tyfun.
;;; expansion-envs: The various envs in effect when the tyform is expanded.
;;;   It can be expanded zero or more times.

(defun typeform-expand (ty expansion-env)
   (let ((r (multi-let (((tyf args)
			 (typeform-fun-and-args ty)))
	       (tyfun-apply tyf args expansion-env))))
      r))   

;;; This sets fun, args, and kind feature of typeform type.
(defun typeform-fun-and-args (tyform)
   (let ((tyf (type-feature tyform 'fun))
	 (ur-env false))
      (cond ((not tyf)
	     (!= ur-env (type-feature tyform 'ur-env))
	     (let ((fun-name (type-feature tyform 'fun-name)))
		(cond (fun-name
		       (let ((vt (var-lookup fun-name ur-env)))
			  (let ((def (and vt (Vartype-val vt))))
			     (cond ((and def (eq (Vartype-type vt)
						 tyfun-type*))
				    (!= tyf def)
				    (!= (type-feature tyform 'kind)
					(Type-kind (Tyfun-defn tyf)))
				    (!= (type-feature tyform 'fun)
					tyf))
				   (t
				    (error-break typeform-fun-and-args
						 :fatal
				       "Bad tyfun: " fun-name " in "
				       tyform))))))
		      (t
		       (error-break typeform-fun-and-args :fatal
			  "Typeform with no fun-name: " tyform))))))
      (values tyf
	      (cond ((type-feature tyform 'suspended)
		     ;; args still in desig form
		     (multi-let (((arg-desigs found)
				  (type-feature tyform 'arg-desigs)))
			 (cond ((not found)
				(error-break typeform-fun-and-args :fatal
				   "Suspended typeform with no arg-desigs: "
				   tyform)))
			 (cond ((not ur-env)
				(!= ur-env (type-feature tyform 'ur-env))))
			 (multi-let (((tyargs freevars)
				      (tyfun-args-resolve tyf arg-desigs ur-env
							  tyform)))
			    (!= (Type-freevars tyform) freevars)
			    (!= (type-feature tyform 'args)
				tyargs)
			    (!= (type-feature tyform 'suspended)
				false)
			    tyargs)))
		    (t
		     (type-feature tyform 'args))))))

;; If non-false, the function to be applied to rest args in tyfuns.
(defvar tyfun-rest-handler* false)

		   
;;; Note that expansion-env affects only the args, not the defn,
;;; which obeys lexical scoping rules.
(defun tyfun-apply (tyf args expansion-env)
   (let ((new-vartypes (tyfun-args-close args expansion-env)))
;;;;      (dbg-save (nvl new-vartypes) tyf args (eenv expansion-env))
;;;;      (breakpoint tyfun-apply
;;;;	 "defn = " (Tyfun-defn tyf))
      (new-Type-closure
	 (Tyfun-defn tyf)
	 (env-bindings-nconc
	    true
	    (append new-vartypes
		    (list (new-Vartype 'defined-fun-param tyfun-type* tyf)
			  (new-Vartype 'defined-fun-args-param
				       univ-type*
				       (<# Vartype-val new-vartypes))))
	    (Tyfun-env tyf)))))

(defun tyfun-args-close (arg-vtl env)
	    (repeat :for ((a :in arg-vtl))
	     :collect (new-Vartype (Vartype-var a)
				   (Vartype-type a)
				   (new-Type-closure (Vartype-val a)
						     env))))

(class-feature-datafun is-builder typeform-class
   (defun (ft must-be-defined vartypes is-tester-stack warn-if-undefined)
      (multi-let (((tyf _) (type-find-feature ft 'fun vartypes))
		  ((args _) (type-find-feature ft 'args vartypes)))
	 (multi-let (((isbn _)
		      (type-feature (Tyfun-defn tyf)
				    'is-builder-name)))
;;;;	    (dbg-save tyf isbn)
;;;;	    (breakpoint typeform-class-is-builder-fcn
;;;;	       "isbn = " isbn)
	    (cond ((or (not isbn)
		       (eq isbn '*untestable))
;;		   (out (:to *error-output*) "*untestable is-builder-name for "
;;			tyf :%)
		   '*untestable)
		  (t
		   (repeat :for ((vt :in args)
				(tester-args '())
				(all-testable true)
				vtype vval)
		    :result `(lambda (x)
				(funcall (,isbn
					  ,@(reverse tester-args))
					 x))
		       (!= vtype (Vartype-type vt))
		       (!= vval (Vartype-val vt))
		       (cond ((type-feature vtype 'meta)
			      (cond ((and (is-Closable vval)
					  (is-Type (Closable-thing vval)))
				     (!= vval (Closable-thing *-*))))
			      (cond ((is-Type vval)
				     (let ((at (get-is-tester
						  vval must-be-defined
						  vartypes is-tester-stack
						  warn-if-undefined)))
					(cond ((eq at '*untestable)
					       (out (:to *error-output*)
						  "No is-tester: "
						  vval :%)
					       (!= all-testable
						   false))
					      (t
					       (!= tester-args
						   (cons `#',at
							 *-*))))))
				    (t
				     (out (:to *error-output*)
					  "Unexpected typeform arg "
					  vt :%)
				     (!= all-testable false))))
			     (t
			      (!= tester-args
				  (cons vval *-*))))
		      :while all-testable
		      :result '*untestable)))))))

(defvar thing-kind-finder* false)
;; This should get set in builtin.

(defun thing-of-right-kind (x type vartypes)
   (cond ((or (is-Type x)
	      (is-Type-closure x))
	  (cond ((eq (type-find-feature type 'meta vartypes)
		     'type)
		 (type-of-kind x type vartypes))
		(t
		 (eq (error-break thing-of-right-kind
			"Attempt to check whether " x
			" is of non-kind " type
			(:prompt-for
			 "y if it should be considered to be of that kind, n if not"
				     'n))
		     'y))))
	 ((or (is-CSymbol x)
	      (consp x))
	  (eq (error-break thing-of-right-kind
		 "Impossible to determine if " x " is of kind " type
		 (:prompt-for "y if I should assume it is, n if not"
			      'n))
	      'y))
	 (t
	  (const-of-type x type vartypes))))

(defun type-of-kind (ty kind vartypes)
   (cond ((is-Type-closure ty)
	  (type-of-kind (Type-closure-type ty)
			kind
			(Type-closure-env ty)))
	 ((is-param-type ty)
	  (let ((vt (var-lookup (param-type-varname ty)
				vartypes)))
	     (try-accept (Vartype-type vt) kind
			 '() vartypes vartypes '())))
	 ((is-tvar-type ty)
	  (let ((k (or (type-feature ty 'type)
		       type-type*)))
	     (try-accept k kind '() vartypes '() '())))
	 (t
	  (const-of-type ty kind vartypes))))	  

; Returns < whether-okay, revised-undo-stack >
(defun got-typeform-want-other (ctv me other undo-stack my-env other-env
				accept-stack)
   (multi-let (((my-type my-env _)
		(type-find-feature-spread me 'expansion my-env)))
      (type-acceptable ctv my-type other undo-stack
		       my-env other-env accept-stack)))

(declare-class-feature got-me-want-other typeform-class
                       (make-Accept-handler 100 #'got-typeform-want-other))

(defun got-other-want-typeform (ctv other me undo-stack other-env my-env
				accept-stack)
   (multi-let (((my-type my-env _)
		(type-find-feature-spread me 'expansion my-env)))
      (type-acceptable ctv other my-type undo-stack
		       other-env my-env accept-stack)))
      
(declare-class-feature got-other-want-me typeform-class
   (make-Accept-handler 100 #'got-other-want-typeform))

(class-feature-datafun equal-tester subtype-class
   (defun (sub other env1 env2)
      (cond ((subs-have-same-fun sub other env1 env2)
	     (multiple-value-let (sub-args enva1 oa-1)
				 (type-find-feature-spread
				    sub 'defined-fun-args env1)
		(multiple-value-let (other-args enva2 oa-2)
				    (type-find-feature-spread
				       other 'defined-fun-args env2)
		   (and oa-1
			oa-2
			(= (len sub-args)
			   (len other-args))
			(every (\\ (a1 a2)
				  (types-etc-equal a1 a2 enva1 enva2))
			       sub-args
			       other-args)))))
	     (t false))))

(defun sub-is-sub (ctv got wanted undo0 envg envw accept-stack)
   (cond ((subs-have-same-fun got wanted envg envw)
	  (multiple-value-let (got-args enva1 found-a1)
			      (type-find-feature-spread
				 got 'defined-fun-args envg)
	     (multiple-value-let (wanted-args enva2 found-a2)
				 (type-find-feature-spread
				    wanted 'defined-fun-args envw)
		(cond ((and found-a1 found-a2
			    (= (len got-args)
			       (len wanted-args)))
		       (do ((argl1 got-args (cdr argl1))
			    (argl2 wanted-args (cdr argl2))
			    (a1) (a2) (a1-is-type) (a2-is-type)
			    (undo-stack undo0)
			    (okay true))
			   ((or (not okay) (null argl1))
			    (okay-or-undo okay undo-stack undo0))
			 (!= a1 (car argl1))    (!= a2 (car argl2))
			 (!= a1-is-type (is-Typoid a1))
			 (!= a2-is-type (is-Typoid a2))
			 (cond ((or a1-is-type a2-is-type)
				(cond ((and a1-is-type a2-is-type)
				       (!= < okay undo-stack >
					   (type-acceptable
					      ctv
					      (car argl1) (car argl2)
					      undo-stack
					      enva1 enva2
					      accept-stack)))
				      (t (!= okay false))))
			       (t (!= okay (equal a1 a2))))))
		      (t
		       (values false undo0))))))
	 (t
	  (values false undo0))))

(declare-class-feature got-me-want-other subtype-class
   (make-Accept-handler 7
      (\\ (ctv got wanted undo-stack envg envw accept-stack)
	 (multiple-value-let (is-sub undo-stack-2)
			     (sub-is-sub ctv got wanted
					 undo-stack envg envw accept-stack)
	    (cond (is-sub
		   (values true undo-stack-2))
		  (t
		   (type-acceptable ctv (Type-super got) wanted undo-stack-2
				    envg envw accept-stack)))))))

(declare-class-feature got-other-want-me subtype-class
   (make-Accept-handler 7 #'sub-is-sub))

(defun subs-have-same-fun (sub other env1 env2)
   (letrec ()
      (multiple-value-let (sub-fun  envf1 of-1)
			  (type-find-feature-spread sub 'defined-fun env1)
			  (ignore envf1 of-1)
	 (check-legality sub-fun sub env1)
         (multiple-value-let (other-fun envf2 of-2)
			     (type-find-feature-spread other
						       'defined-fun  env2)
			     (ignore envf2 of-2)
	    (check-legality other-fun other env2)
	    (cond ((and (is-Type sub-fun) (is-Type other-fun))
		   (eq (Type-desig sub-fun)
		       (Type-desig other-fun)))
		  ((and (is-Tyfun sub-fun) (is-Tyfun other-fun))
		   (eq (Tyfun-name sub-fun)
		       (Tyfun-name other-fun)))
		  (t false))))
		   
;;;;		   (cond ((and (not (eq sub-fun other-fun))
;;;;			       (eq (Tyfun-name sub-fun)
;;;;				   (Tyfun-name other-fun)))
;;;;			  (dbg-save :run-loud sub other env1 env2
;;;;				    sub-fun other-fun)
;;;;			  (out "defined-fun of " sub " = " sub-fun
;;;;			       :% "defined-fun of " other " = " other-fun
;;;;			       :% " eq? " (eq sub-fun other-fun) :%)
;;;;			  (error-break subs-have-same-fun :novalue
;;;;				       "break")))
;;;;		   (eq sub-fun other-fun)

    :where
      (check-legality (thing owner env)
	 ;;(out (tr check-legality ("Checking " thing)
         (cond ((not (or (not thing)
			 (is-Tyfun thing)
			 (is-Type thing)
			 ))
		(dbg-save :run-loud thing owner env)
		(error-break subs-have-same-fun
		   "This thing: " thing
		   :% "  is Illegal as value of defined-fun feature of " owner
		   (:novalue "I'll go on, but expect nonsense to ensue"))))
	 ;;("Checked " (car out-vals*))))
	 )))
		   
