;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: typebounds.lisp,v 1.6 2005/12/26 20:34:25 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ subtype altformer funformers vartypes
	                       parmutils)

(defun type-bounds (type strict vartypes)
      (values (type-bound type false strict ':all vartypes)
	      (type-bound type true strict ':all vartypes)))

;; The setup in this file has the following defect: We don't collect
;; all the bounds in a table, so we compute the bound for a given tvar
;; repeatedly.

;; Somewhat hackish.
(defun term-bound (term upper elim vartypes)
   (cond ((is-Qvar term)
	  (let ((tv (symboid-tvar (Qvar-sym term))))
	     (cond (tv
		    (type-find-designator
		       (type-bound tv upper false elim vartypes)
		       vartypes))
		   (t term))))
	 ((atom term) term)
	 (t
	  (<# (\\ (a) (term-bound a upper elim vartypes))
	      term))))

;;; Eliminate only tvars in the list elim (a list of var names)
;;; If strict is true, take tightest deducible bound, even if 
;;; you have to leave some tvars in.
;;; If strict is false, replace all elim tvars with their tvar-free
;;; bounds.
;;; If elim is ':all, eliminate all tvars.
(defun type-bound (type upper strict elim vartypes)
   (multi-let (((type vartypes)
		(follow-var-ref type vartypes)))
 ; (out (tr type-bound ("type-bound> "
; 		       (q (upper "upper") (t "lower"))
; 		       1 
; 		       (q (strict "strict "))
; 		       type
; 		       t "elim = " elim
; 		       t vartypes t)
      (cond ((is-tvar-type type)
	     (cond ((or (eq elim ':all)
			(memq (tvar-type-varname type) elim))
		    (let ((tvbdg (tvar-type-bdg type)))
		      (cond (upper
			     (cond ((null (Typebdg-supers tvbdg))
				    univ-type*)
				   (t
				    (types-common-supertype
				       false
				       (<# (\\ (sup)
					      (type-bound sup true strict
							  elim empty-env*))
					   (Typebdg-supers tvbdg))
				       (empty-undo-stack) empty-env*))))
			    (t
			     (make-alt-type
			        (<# (\\ (sub)
				       (type-bound sub false strict
						   elim empty-env*))
				    (Typebdg-subs tvbdg))
				empty-env*)))))
		   (t type)))
	    ((and (is-CSymbol (Type-desig type))
		  (not (eq (Type-desig type) '*as-feature)))
	     type)
	    (t
	     (let ((h (type-find-feature type 'bound-finder vartypes)))
		(cond (h
		       (funcall h type upper strict elim vartypes))
		      (t
		       (cond ((or (not (is-CSymbol (Type-desig type)))
				  (eq (Type-desig type) '*as-feature))
			      (signal-problem type-bound
				 "No bound finder for " type
				 (:novalue
				  "I will assume the type is its own bound"))))
		       type)))))
  ;("type-bound< " (car out-vals*))))
  ))

(defun types-bound (types upper strict elim vartypes)
    (let ((extyl
	    (<# (\\ (exty)
		   (type-bound exty upper strict elim vartypes))
		(types-extreme-elements types upper vartypes))))
      (cond ((= (len extyl) 0)
	     (cond (upper void-type*)
		   (t univ-type*)))
	    ((= (len extyl) 1)
	     (car extyl))
	    (upper
	     (cond (strict
		    (make-alt-type extyl vartypes))
		   (t (</ (\\ (sup ty)
			     (types-common-supertype
				  false (list sup ty) (empty-undo-stack)
				  vartypes))
			  (car extyl) (cdr extyl)))))
	    (t
	     void-type*))))
;;; We can't do better than Void for a lower bound on two or more 
;;; types whose relationship is unknown, because we don't have an 
;;; intersection type.

(class-feature-datafun bound-finder unk-class
   (defun (unkty _ _ _ _)
      unkty))

(class-feature-datafun bound-finder const-class
   (defun (ty _ _ _ _)
      ty))

(class-feature-datafun bound-finder boring-class
   (defun (ty _ _ _ _)
      ty))

(class-feature-datafun bound-finder typeform-class
   (defun (ty upper strict elim vartypes)
      (cond ((type-find-feature ty 'nonmonotonic vartypes)
	     (type-bound (type-find-feature ty 'expansion vartypes) 
 			 upper strict elim vartypes))
            (t
	     (let ((argbounds
		      (<# (\\ (vt)
			     (let ((arg (Vartype-val vt)))
			        (new-Vartype
				   (Vartype-var vt)
				   (Vartype-type vt)
				   (cond ((or (is-Type arg)
					      (is-Type-closure arg))
					  (type-bound arg upper strict
						      elim vartypes))
					 (t arg)))))
			  (type-find-feature ty 'args vartypes))))
	        ;;(out "argbounds = " argbounds t)
		(type-clos/able
		   (make-typeform
		      (type-find-feature ty 'fun vartypes)
		      argbounds)
		   vartypes))))))

(class-feature-datafun bound-finder alt-class
   (defun (ty upper strict elim vartypes)
      (multiple-value-let (atl env)
                          (follow-var-ref (type-find-feature
					     ty 'alt-types vartypes)
					  vartypes)
	 (make-alt-type
	    (<# (\\ (aty)
		   (type-bound aty upper strict elim env))
		atl)
	    vartypes))))

(class-feature-datafun bound-finder fun-class
   (defun (ty upper strict elim vartypes)
      (let ((abound 
               (type-bound (type-find-feature ty 'argtype vartypes)
			   (not upper)
			   strict elim vartypes))
	    (alspec
	       (funtype-extract-arglist ty vartypes))
	    (alspec-bound false))
	 (cond (alspec
		(!= alspec-bound
		    (arglistspec-bound
		       alspec (not upper) strict elim vartypes))
		(multiple-value-let (abty _)
				    (follow-var-ref abound vartypes)
		   (!= (type-feature abty 'elt-types)
		       alspec-bound))))
	 (let ((renv
		  (high-argspecs-declare-placeholder-params
		     ty alspec vartypes)))
	    (let ((rtype (type-find-feature ty 'resulttype renv)))
	       (let ((rbound (type-bound rtype upper strict elim renv))
		     (rlspec (funtype-extract-reslist ty renv))
		     (rlspec-bound false))
		  (cond (rlspec
			 (!= rlspec-bound 
			     (arglistspec-bound rlspec upper strict elim
						renv))
			 (multiple-value-let (rbty _)
					     (follow-var-ref rbound vartypes)
			    (!= (type-feature rbty 'elt-types)
				rlspec-bound))))
;;;;		     (dbg-save :run-loud
;;;;			rbound abound vartypes rlspec-bound alspec-bound)
;;;;		     (breakpoint fun-class-bound-finder
;;;;			"About to make funtype")
		  (make-funtype
		     (type-feature ty 'level)
		     rbound abound rlspec-bound alspec-bound
		     (type-feature ty 'side-effects)
		     vartypes)))))))

(defun arglistspec-bound (alspec0 upper strict elim vartypes)
   (multiple-value-let (alspec _)
		       (follow-var-ref alspec0 vartypes)
      (letrec ((argspec-bound (a)
		  ;;(out (tr argspec-bound ("bound> " a)
		  (make-Argspec
		     (Argspec-name a)
		     (Argspec-mode a)
		     (argspec-pos-zero a)
		     (type-bound (Argspec-type a)
				 upper strict elim vartypes)
		     (Argspec-default a)
		     (Argspec-props a))
		  ;;("bound< " (car out-vals*))))
		  ))
	       ;;(arglistspec-simplify ...)
	 (new-Arglistspec
	    (<# #'argspec-bound (Arglistspec-argspecs alspec))
	    :allow-other-keys (Arglistspec-allow-other-keys alspec)
	    :names-matter (Arglistspec-names-matter alspec)))))

;;;;(defun arglistspec-freevars (alspec vartypes)
;;;;   (let ((fvl '()))
;;;;      (letrec ((note-vars (as)
;;;;		  (!= fvl
;;;;		      (freevars-union
;;;;		         (type-find-freevars (Argspec-type as)
;;;;					     vartypes)
;;;;			 *-*))))
;;;;	 (repeat for ((as in (Arglistspec-argspecs alspec)))
;;;;	    (note-vars as))
;;;;	 fvl)))

(class-feature-datafun bound-finder dot-class
   (defun (ty upper strict elim vartypes)
       (dot-lin-bound-finder 'Tup ty upper strict elim vartypes)))

(class-feature-datafun bound-finder lin-class
   (defun (ty upper strict elim vartypes)
       (dot-lin-bound-finder 'Row ty upper strict elim vartypes)))

(defun dot-lin-bound-finder (med ty upper strict elim vartypes)
   (cond ((is-Closable (type-find-feature ty 'cdrtype vartypes))
	  (breakpoint dot-lin-bound-finder
	     "Sanity near snapping point: " ty)))
   (make-dot-type
      med
      (type-bound (type-find-feature ty 'cartype vartypes)
		  upper strict elim vartypes)
      (type-bound (type-find-feature ty 'cdrtype vartypes)
		  upper strict elim vartypes)
      vartypes))

(class-feature-datafun bound-finder to-row-class
   (defun (ty upper strict elim vartypes)
      (make-to-row-type
         (type-bound (type-find-feature ty 'source-type vartypes)
		     upper strict elim vartypes)
         vartypes)))

(defun tup-val-bound-finder (ty upper strict elim vartypes)
   (let ((alspec (arglistspec-bound
		    (type-find-feature ty 'elt-types vartypes)
		    upper strict elim vartypes)))
      (make-tup-type
         (type-find-feature ty 'medium vartypes)
	 alspec false vartypes)))	 

(declare-class-feature bound-finder tup-class #'tup-val-bound-finder)
(declare-class-feature bound-finder val-class #'tup-val-bound-finder)

(class-feature-datafun bound-finder constructed-class
   (defun :^ (ty upper strict elim vartypes)
      (finish-constructed-type
         (type-must-find-feature ty 'constructor vartypes)
	 (<# (\\ (vt)
		(new-Vartype (Vartype-var vt)
			     (Vartype-type vt)
			     (type-bound (Vartype-val vt) upper strict elim vartypes)))
	     (type-find-feature ty 'arg-vartypes vartypes))
	 vartypes)))

(class-feature-datafun bound-finder modified-class
   (defun (ty upper strict elim vartypes)
      (let ((suptype (type-bound (type-find-supertype ty vartypes)
				 upper strict elim vartypes)))
	 (let ((resty
		  (type-clos/able
		     (make-modified-type
			(Type-desig ty)
			suptype 
			(<# (\\ (s) (slot-bound s upper strict elim vartypes))
			    (Type-slotfns ty))
			(Type-features ty))
		     vartypes)))
	    resty))))

(defun maybe-add-supertype-to-feats (suptype supenv feats)
   (cond (supenv
	  (cons (tuple 'supertype
		       (type-clos/able
			  suptype supenv))
		feats))
	 (t feats)))

(defun slot-bound (s upper strict elim vartypes)
   (make-Slot (Slot-name s)
	      (Slot-acc s)
	      (Slot-set s)
	      (cond ((Slot-type s)
		     (type-bound (new-Type-closure (Slot-type s) vartypes)
				 upper strict elim vartypes))
		    (t false))
	      (arglistspec-bound (close-if-closable (Slot-atypes s) vartypes)
				 (not upper)  ;; feel the vibes
				 strict elim vartypes)
	      (Slot-invert s)
	      (Slot-descriptor s)))

(class-feature-datafun bound-finder image-class
   (defun (ty upper strict elim vartypes)
      ;; We have to assume that the template is not contravariant
      ;; (it's okay if it has contravariant pieces).
      (let ((itervars (type-find-feature ty 'template-params vartypes))
	    (tmtype (type-find-feature ty 'template-type vartypes))
	    (args (type-find-feature ty 'template-args vartypes)))
	 (make-image-type itervars tmtype
			  (<# (\\ (a) (type-bound a upper strict elim
						  vartypes))
			      args)
			  false vartypes))))


(defun cheap-types-equal (ty1 ty2)
   (multi-let (((ty1 env1)
		(follow-var-ref ty1 global-env*))
	       ((ty2 env2)
		(follow-var-ref ty2 global-env*)))
      (cond ((eq env1 env2)
	     ;;(out (to ttyout*) "envs equal" t)
	     (cond ((equal (Type-desig ty1)
			   (Type-desig ty2))
		    ;;(out (to ttyout*) "desigs equal" t)
		    true)
		   (t false)))
	    (t false))))

(class-feature-datafun bound-finder subtype-class
   (defun (ty _ _ _ _)
      ty))

(class-feature-datafun bound-finder simple-typeform-class
   (defun (ty _ _ _ _)
      ty))


