;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: typeintsect.lisp,v 1.4 2005/12/26 20:34:25 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ tloader istester types altformer
	                       unkelim funacceptors tupformers
			       conformer typebounds types)

(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel)
   (export '(types-intersect
	     check-expressibility-and-voidness)))

(def-class Intersect-handler
      (:handler
          (print-object (ah srm)
	     (out (:to srm) "#{Intersect-handler "
		  (Intersect-handler-priority ah) "}")))
  priority fcn)

(defun types-complement (ctv ty1 ty2 undo-stack env1 env2 intersect-stack)
   (types-intersect ctv true ty1 false ty2 undo-stack
		    env1 env2 intersect-stack))

(def-class Recursive-pair-frame
    (:handler (print (rpf srm)
		 (out (:to srm) "#{Rec-pair "
		      (:q ((Recursive-pair-frame-recurser rpf)
			  (Recursive-pair-frame-recurser rpf) ": "))
		      (Recursive-pair-frame-clo1 rpf)
		      "/" (Recursive-pair-frame-clo2 rpf))))
   mode1 clo1 mode2 clo2 recurser env)
;;; If recursion detected (clo1, clo2) is being processed above us
;;; in the tree, set the recurser to a new symbol that will be used
;;; to build a 'Label-types' type.

(def-class Recursive-pair-stack
   (:options (:include List-with-length))
   (:handler (inside-print (ll srm)
		(let ((cl (Recursive-pair-stack-contents ll)))
		   (out (:to srm) 1 (car cl)
			(:q ((not (null (cdr cl)))
			    "... ")))))
	     (ll-add (rp-stack pair)
		(make-Recursive-pair-stack
		   (+ (Recursive-pair-stack-size rp-stack)
		      1)
		   (cons pair
			 (Recursive-pair-stack-contents rp-stack))))))

(defvar rectypeno* 0)

;;; Returns < expressible, intersection, undo-stack >.
;;; If not expressible, the intersection can't be expressed
;;; precisely in this type system.
;;; If sign1 = true and sign2 = true, finds (intsect ty1 ty2)
;;; If sign1 = true and sign2 = false, finds (setdif ty1 ty2)
;;; If sign1 = false and sign2 = false, find (setdif ty2 ty1)
;;; If both signs are false, it's inexpressible
(defun types-intersect (constrain-tvars sign1 ty1 sign2 ty2
			undo-stack env1 env2 intersect-stack)
   (cond ((null intersect-stack)
	  (!= intersect-stack
	      (make-Recursive-pair-stack 0 '()))))
   (multi-let (((ty1 env1)
		(follow-var-ref ty1 env1))
	       ((ty2 env2)
		(follow-var-ref ty2 env2)))
      (cond (sign1 
;;;;	     (dbg-save ty1 ty2)
;;;;	     (breakpoint types-intersect
;;;;		"About to call")
	     (types-intersect-subtype-check
	        constrain-tvars ty1 sign2 ty2 undo-stack
		env1 env2 intersect-stack sign2))
	    (sign2
	     (types-intersect-subtype-check
	        constrain-tvars ty2 sign1 ty1 undo-stack
		env2 env1 intersect-stack false))
	    (t
	     (values false nil undo-stack)))))

(defun types-intersect-subtype-check (ctv subty supersign superty undo-stack
				      subenv superenv intersect-stack
				      check-other-way)
   (multi-let (((okay undo-stack-1)
		(type-acceptable
		   (and ctv supersign)
		   subty superty undo-stack subenv superenv '())))
      (cond (okay
	     (values true
		     (cond (supersign
			    (new-Type-closure subty subenv))
			   (t
			    void-type*))
		     undo-stack-1))
	    (check-other-way
	     (types-intersect-subtype-check
	        ctv superty true subty undo-stack
		superenv subenv intersect-stack false))
	    (t
	     (let ((r (make-Recursive-pair-frame
			 true
			 (identifying-type-closure subty subenv)
			 supersign
			 (identifying-type-closure superty superenv)
			 false nil)))
		(let ((prev (find-in-recursive-pair-stack
			       r intersect-stack)))
		   (cond (prev
			  ;; Recursion detected
			  (values true
				  (create-recursive-typeform prev)
				  undo-stack))
			 (t
			  (try-intersect-handlers
			     ctv true subty supersign superty undo-stack
			     subenv superenv
			     (ll-add intersect-stack r))))))))))

(defun find-in-recursive-pair-stack (newframe intersect-stack)
   (repeat :for ((iframe :in (Recursive-pair-stack-contents intersect-stack))
		(new-mode1 (Recursive-pair-frame-mode1 newframe))
		(new-mode2 (Recursive-pair-frame-mode2 newframe))
		(new-clo1 (Recursive-pair-frame-clo1 newframe))
		(new-clo2 (Recursive-pair-frame-clo2 newframe))
		prev-clo1 prev-clo2 prev-mode1 prev-mode2)
    :result false
      (!= prev-mode1 (Recursive-pair-frame-mode1 iframe))
      (!= prev-mode2 (Recursive-pair-frame-mode2 iframe))
      (!= prev-clo1 (Recursive-pair-frame-clo1 iframe))
      (!= prev-clo2 (Recursive-pair-frame-clo2 iframe))
    :until (or (and (eq prev-mode1 new-mode1)
		   (eq prev-mode2 new-mode2)
		   (type-closures-equal prev-clo1 new-clo1)
		   (type-closures-equal prev-clo2 new-clo2))
	      (and (eq prev-mode2 new-mode1)
		   (eq prev-mode1 new-mode2)
		   (type-closures-equal prev-clo2 new-clo1)
		   (type-closures-equal prev-clo1 new-clo2)))
    :result iframe))

;;; Presumably it matters not what the signs were.
(defun create-recursive-typeform (prev-frame)
   (let ((typefun-name
	    (build-symbol rectyp (++ rectypeno*))))
      (let ((ur-env
	       (make-Env true
			 (list (new-Vartype
				  typefun-name
				  tyfun-type* nil)))))
	 (!= (Recursive-pair-frame-recurser prev-frame)
	     typefun-name)
	 (!= (Recursive-pair-frame-env prev-frame)
	     ur-env)
	 (build-typeform
	    typefun-name '() ur-env))))

(defun try-intersect-handlers (ctv sign1 ty1 sign2 ty2 undo-stack
			       env1 env2 intersect-stack)
   (let ((h1 (type-find-feature ty1 'intersector env1))
	 (h2 (type-find-feature ty2 'intersector env2)))
;;;;      (out "h1 = " h1 (:q (h1 "/" (Intersect-handler-fcn h1)))
;;;;	   " --- h2 = " h2 (:q (h2 "/" (intersect-handler-fcn h2))) :%)
      (multi-let (((expressible restype undo-stack-1)
		   (cond ((and
			     h1
			     (not
				(and h2
				     (> (Intersect-handler-priority h2)
					(Intersect-handler-priority h1)))))
			  (funcall (coerce-to-fun (Intersect-handler-fcn h1))
				   ctv sign1 ty1 sign2 ty2 undo-stack
				   env1 env2 intersect-stack))
			 (h2
			  (funcall (coerce-to-fun (Intersect-handler-fcn h2))
				   ctv sign2 ty2 sign1 ty1 undo-stack
				   env2 env1 intersect-stack))
			 (t
			  (signal-problem types-intersect
			     "Intersecting two types without"
			     " intersect-handlers: "
			     :% ty1 
			     :% ty2)))))
	 (cond (expressible
		(values true
			(let ((pair-frame (ll-car intersect-stack)))
			   (cond ((Recursive-pair-frame-recurser pair-frame)
				  (make-type-recursive
				     (Recursive-pair-frame-recurser pair-frame)
				     (Recursive-pair-frame-env pair-frame)
				     restype))
				 (t restype)))
			undo-stack-1))
	       (t
		(values false nil (undo undo-stack-1 undo-stack)))))))

(defun make-type-recursive (name env type)
   (multi-let ((tyf (new-Tyfun name no-args-spec* nil empty-env*)))
      (let ((tyfun-binding (car (Env-vartypes env))))
	 ;; Should be binding of name
	 (cond ((and (is-Vartype tyfun-binding)
		     (eq (Vartype-var tyfun-binding)
			 name))
		(!= (Vartype-val tyfun-binding)
		    tyf)
		(!= (Tyfun-defn tyf) type)
		(cond ((atom (Type-desig type))
		       (err-out "Unexpected designator " (Type-desig type)
				" for recursive type"))
		      (t
		       (!= (Type-desig type)
			   `(Label-types ((,name () ,(Type-desig type)))
			       (,name)))))
		type)
	       (t
		(signal-problem make-type-recursive
		   "Recursive Tyfun env screwed up"))))))

;;; Useful macro
(defmacro check-expressibility-and-voidness (e^ i^ unew^ uold^ &rest body)
   `(cond (,e^
	   (cond ((eq ,i^ void-type*)
		  (values true void-type* ,unew^))
		 (t
		  ,@body)))
	  (t
	   (values false nil (undo ,unew^ ,uold^)))))

;;; This is for intersecting types (such as those of class 'subtype-class'),
;;; that are assumed disjoint unless provable otherwise).
(defun indep-types-intersect (_ sign1 ty1 sign2 ty2 undo-stack env1 env2 _)
	 ;; We already checked subtype relationships before
	 ;; trying the handler, so we know for both ty1 and ty2
	 ;; that if its sign is positive it isn't a subtype of the
	 ;; other.
	 ;; But you can't express the subtraction of
	 ;; a subtype from its supertype,
	 ;; so we have to do the corresponding tests for the
	 ;; negative cases.
   (cond ((not sign2)
	  (cond ((type-acceptable false ty2 ty1 '()
				  env2 env1 '())
		 (values false nil undo-stack))
		(t
		 (values true (new-Type-closure ty1 env1) undo-stack))))
	 ((not sign1)
	  (cond ((type-acceptable false ty1 ty2 '()
				  env1 env2 '())
		 (values false nil undo-stack))
		(t
		 (values true (new-Type-closure ty2 env2) undo-stack))))
	 (t
	  ;; Positive intersection of two random subtypes
	  ;; is Void
	  (values true void-type* undo-stack))))


(declare-class-feature intersector boring-class
   (make-Intersect-handler 4 #'indep-types-intersect))

(declare-class-feature intersector unk-class
   (make-Intersect-handler 4 (\\ (_ _ _ _ _ undo-stack _ _ _)
				(values true void-type* undo-stack))))

(declare-class-feature intersector simple-typeform-class
   (make-Intersect-handler 4 (\\ (_ _ _ _ _ undo-stack _ _ _)
				(values true void-type* undo-stack))))

(defun tvar-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			 env1 env2 intersect-stack)
                        (ignore intersect-stack)
   (cond (sign1
	  (cond (sign2
		 (multi-let (((consistent undo-stack-1)
			      (tvar-try-accept ty1 'got
					       ty2 undo-stack env1 env2 '())))
		    (cond (consistent
			   (let ((tybdg (tvar-type-bdg ty1)))
			      (let ((val (Typebdg-val tybdg)))
				 ;; Discard changes unless 'ctv'
				 (cond ((not ctv)
					(!= undo-stack-1
					    (undo undo-stack-1 undo-stack))))
				 ;; Expressible only if tvar got set,
				 ;; in which case the value is what
				 ;; it got set to!
				 (values val val undo-stack-1))))
			  (t
			   (values true void-type*
				   (undo undo-stack-1 undo-stack))))))
		(t
		 (tvar-complement ctv ty1 ty2 undo-stack))))
	 ((and sign2
	       (is-tvar-type ty2))
	  (tvar-complement ctv ty2 ty1 undo-stack))
	 (t
	  (values false nil undo-stack))))

(declare-class-feature intersector tvar-type-class
   (make-Intersect-handler 101 #'tvar-intersector))      

(defvar tvar-complement-dbg* true)

(defun tvar-complement (ctv typ tym _)
   (dbg-out tvar-complement-dbg*
      "Attempt to complement [" ctv "] " typ " \ " tym)
   (values false nil))

(declare-class-feature intersector const-class
   (make-Intersect-handler 7
      (\\ (_ sign1 ty1 sign2 ty2 undo-stack env1 env2 _)
	 (multi-let (((cl1 found1) (type-feature ty1 'consts))
		     ((cl2 found2) (type-feature ty2 'consts)))
	    (letrec ()
	       (cond ((and found1 found2)
		      (values true
			      (make-const-type
				 (cond ((and sign1 sign2)
					(intersection  cl1 cl2 :test #'equal))
				       (sign1
					(set-difference cl1 cl2 :test #'equal))
				       (t
					(set-difference cl2 cl1 :test #'equal))))
			      undo-stack))
		     ((and sign1 found1)
		      (try-consts-test
			 cl1 sign2 ty2 env2))
		     ((and sign2 found2)
		      (try-consts-test
			 cl2 sign1 ty1 env1))
		     (t
		      (values false nil)))
	     :where
	       (try-consts-test (consts sign ty env)
		  (let ((is-tester (get-callable-is-tester ty env)))
		     (cond (is-tester
			    (values
			       true
			       (make-const-type
				  (<? (\\ (c)
					 (boole-eq (funcall
						      is-tester
						      (official-const->val
						         c))
						   sign))
				      consts))
			       undo-stack))
			   (t
			    (values false nil undo-stack))))))))))

(defun alt-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			env1 env2 intersect-stack)
   (let ((alt-types (type-find-feature ty1 'alt-types env1)))
      (cond (sign1
	     (repeat :for ((aty :in alt-types)
			  (res '())
			  expressible it (ustack undo-stack))
	      :result (values true
			      (make-alt-type (reverse res) empty-env*)
			      ustack)
		(!= < expressible it ustack >
		    (types-intersect ctv true aty sign2 ty2 ustack
				     env1 env2 intersect-stack))
	      :while expressible
		(!= res (cons it *-*))
	      :else :result (values false nil (undo ustack undo-stack))))
	    (t
	     (repeat :for ((aty :in alt-types)
			  okay (ustack undo-stack))
	      :result (values true ty2 ustack)
		(!= < okay ty2 ustack >
		    (types-intersect ctv true ty2 false aty ustack
				     env2 env1 intersect-stack))
	      :while okay
	      :else :result (values false nil (undo ustack undo-stack)))))))
         
(declare-class-feature intersector alt-class
   (make-Intersect-handler 35 #'alt-intersector))

(defun dot-lin-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			    env1 env2 intersect-stack)
   (let ((sty1 (row-type-simplify ty1 env1))
	 (sty2 (row-type-simplify ty2 env2)))
      (cond ((or (not (eq sty1 ty1))
		 (not (eq sty2 ty2)))
	     (types-intersect ctv sign1 sty1 sign2 sty2 undo-stack
			      env1 env2 intersect-stack))
	    (t
	     (let ((med1 (type-find-feature ty1 'medium env1))
		   (med2 (type-find-feature ty2 'medium env2)))
		(cond ((and med1 med2)
		       (cond ((eq med1 med2)
			      (cond ((and sign1 sign2)
				     (dot-lin-intersection
					ctv ty1 ty2 undo-stack
					env1 env2 intersect-stack))
				    (sign1
				     ;; Subtracting ty2
				     (dot-lin-complement
					ctv med1 ty1 ty2 undo-stack
					env1 env2 intersect-stack))
				    (t
				     (dot-lin-complement
					ctv med1 ty2 ty1 undo-stack
					env2 env1 intersect-stack))))
			     (t
			      (values true void-type* undo-stack))))
		      ((and sign1 sign2)
		       (values true void-type* undo-stack))
		      (sign1
		       (values true (new-Type-closure ty1 env1)
			       undo-stack))
		      (t
		       (values true (new-Type-closure ty2 env2)
			       undo-stack))))))))

(declare-class-feature intersector dot-class
		       (make-Intersect-handler 9 #'dot-lin-intersector))
(declare-class-feature intersector lin-class
		       (make-Intersect-handler 10 #'dot-lin-intersector))

;;; This can only happen if ty was (Row x ?tvar), and the tvar got 
;;; bound to something else.
(defun row-type-simplify (ty env)
   (let ((med (type-find-feature ty 'medium env)))
      (cond ((eq med 'Row)
	     (let ((cdrtype (row-type-simplify
			       (type-find-feature ty 'cdrtype env)
			       env)))
	        (cond ((types-equal cdrtype empty-row-type* env global-env*)
		       (type-find-feature ty 'cartype env))
		      (t ty))))
	    (t ty))))		       

(defun dot-lin-intersection (ctv ty1 ty2 undo-stack
			     env1 env2 intersect-stack)   
   (let ((cartype1 (type-find-feature ty1 'cartype env1))
	 (cdrtype1 (type-find-feature ty1 'cdrtype env1))
	 (cartype2 (type-find-feature ty2 'cartype env2))
	 (cdrtype2 (type-find-feature ty2 'cdrtype env2)))
      (multi-let (((aconsis ai undo-stack-1)
		   (types-intersect
		      ctv true cartype1 true cartype2 undo-stack
		      env1 env2 intersect-stack)))
	 (check-expressibility-and-voidness aconsis ai undo-stack-1 undo-stack
	    (multi-let (((dconsis di undo-stack-2)
			 (types-intersect
			    ctv true cdrtype1 true cdrtype2 undo-stack-1
			    env1 env2 intersect-stack)))
	       (check-expressibility-and-voidness
		  dconsis di undo-stack-2 undo-stack
		  (values
		     true
		     (cond ((eq di void-type*)
			    void-type*)
			   (t
			    (make-dot-type
			       (type-find-feature ty1 'medium env1)
			       ai di empty-env*)))
		     undo-stack-2)))))))

(defun dot-lin-complement (ctv med plus-ty minus-ty undo-stack
			   envp envm intersect-stack)
   (let ((p-cartype (type-find-feature plus-ty 'cartype envp))
	 (p-cdrtype (type-find-feature plus-ty 'cdrtype envp))
	 (m-cartype (type-find-feature minus-ty 'cartype envm))
	 (m-cdrtype (type-find-feature minus-ty 'cdrtype envm)))
      (multi-let (((aie car-intersect undo-stack-ai)
		   (types-intersect ctv true p-cartype true m-cartype undo-stack
				    envp envm intersect-stack)))
	 (multi-let (((dce cdr-complement undo-stack-dc)
			(types-intersect ctv true p-cdrtype false m-cdrtype
					 undo-stack-ai
					 envp envm intersect-stack)))
	    (multi-let (((ace car-complement undo-stack-ac)
			 (types-intersect ctv true p-cartype false m-cartype
					  undo-stack-dc
					  envp envm intersect-stack)))
	       (multi-let (((die cdr-intersect undo-stack-di)
			    (types-intersect ctv true p-cdrtype true m-cdrtype
					     undo-stack-ac
					     envp envm intersect-stack)))
		  (cond ((and aie dce ace die)
			 (values
			    true
			    (make-alt-type
			       (list (make-dot-type
					med car-intersect cdr-complement
					empty-env*)
				     (make-dot-type
					med car-complement cdr-intersect
					empty-env*))
			       global-env*)
			    undo-stack-di))
			(t
			 (values false nil
				 (undo undo-stack-di undo-stack))))))))))

(defun typeform-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			     env1 env2 intersect-stack)
   (types-intersect
      ctv sign1 (type-find-feature ty1 'expansion env1)
      sign2 ty2 undo-stack env1 env2 intersect-stack))

(declare-class-feature intersector typeform-class
		       (make-Intersect-handler 100 #'typeform-intersector))

(defun fun-intersector (ctv sign1 fty1 sign2 fty2 undo-stack
			env1 env2 intersect-stack)
   (cond ((and sign1 sign2)
	  (let ((l1 (type-find-feature fty1 'level env1))
		(l2 (type-find-feature fty2 'level env2))
		(a1 (type-find-feature fty1 'argtype env1))
		(a2 (type-find-feature fty2 'argtype env2))
		(s1 (type-find-feature fty1 'side-effects env1))
		(s2 (type-find-feature fty2 'side-effects env2)))
	     (cond ((and l1 l2 
			 (= l1 l2)
			 (eq (not s1) (not s2)))
		    (let ((arg-u
			   (make-alt-type
			      (list (new-Type-closure a1 env1)
				    (new-Type-closure a2 env2))
			      empty-env*)))
		       (letrec ()
			  (let ((pl1 (funtype-extract-arglist fty1 env1))
				(pl2 (funtype-extract-arglist fty2 env2)))
			     (cond ((and pl1 pl2)
				    (cond ((arglistspecs-congruent pl1 pl2)
					   (cond ((> l1 0)
						  (multi-let
						       (((augenv1 augenv2)
							 (paramlists-declare-identically
							    pl1 pl2 env1 env2)))
						     (resulttypes-intersect
							augenv1 augenv2)))
						 (t
						  (resulttypes-intersect
						     env1 env2))))
					  (t
					   (values false nil))))
				   (t
				    (values false nil))))
		     :where
		       (resulttypes-intersect (e1 e2)
			  (multi-let (((res-e res-i undo-stack-1)
				       (types-intersect
					  ctv true
					  (type-find-feature
					     fty1 'resulttype e1)
					  true
					  (type-find-feature
					     fty2 'resulttype e2)
					  undo-stack e1 e2 intersect-stack)))
			     (cond (res-e
				    (values
				       true
				       (make-funtype
					  1
					  res-i arg-u
					  (type-arglistspec res-i empty-env*)
					  (type-arglistspec arg-u empty-env*)
					  s1 empty-env*)
				       undo-stack-1))
				   (t
				    (values false nil
					    (undo undo-stack-1
						  undo-stack)))))))))
		   (t
		    (values true void-type* undo-stack)))))
	 (t
	  (values false nil undo-stack))))

(declare-class-feature intersector fun-class
		       (make-Intersect-handler 5 #'fun-intersector))

(declare-class-feature intersector to-row-class
   (make-Intersect-handler 12
      (\\ (ctv sign1 ty1 sign2 ty2 undo-stack env1 env2 intersect-stack)
	 (multi-let (((source-e source-i undo-stack-1)
			(types-intersect
			   ctv sign1 (type-find-feature ty1 'source-type env1)
			   sign2 (type-find-feature ty2 'source-type env2)
			   undo-stack env1 env2 intersect-stack)))
	    (check-expressibility-and-voidness
	       source-e source-i undo-stack-1 undo-stack
	       (values true
		       (make-to-row-type
			  source-i empty-env*)
		       undo-stack-1))))))

(declare-class-feature intersector image-class
   (make-Intersect-handler 12
      (\\ (ctv sign1 ty1 sign2 ty2 undo-stack env1 env2 intersect-stack)
         (cond ((and sign1 sign2)
		(let ((itervars1 (type-find-feature ty1 'template-params env1))
		      (tmtype1 (type-find-feature ty1 'template-type env1))
		      (args1 (type-find-feature ty1 'template-args env1))
		      (itervars2 (type-find-feature ty2 'template-params env2))
		      (tmtype2 (type-find-feature ty2 'template-type env2))
		      (args2 (type-find-feature ty2 'template-args env2)))
		   (cond ((and tmtype1 tmtype2)
			  (cond ((= (len itervars1)
				    (len itervars2))
				 (repeat :for ((a1 :in args1)
					       (a2 :in args2)
					       (args-il '())
					       (ustack undo-stack)
					       expressible args-i)
				  :result (finish-image-intersect
					    ctv itervars1 tmtype1 
					    itervars2 tmtype2 
					    (reverse args-il)
					    ustack env1 env2 intersect-stack)
				    (!= < expressible args-i ustack >
					(types-intersect
					   ctv true a1 true a2 ustack
					   env1 env2 intersect-stack))
				  :while expressible
				  :else :result (values false nil
						      (undo ustack undo-stack))
				  :until (eq args-i void-type*)
				  :result (values true void-type* ustack)
				    (!= args-il (cons args-i *-*))))
				(t
				 (values false nil undo-stack))))
			 (t
			  (values true void-type* undo-stack)))))
	       (t
		(values false nil undo-stack))))))

(defun finish-image-intersect (ctv itervars1 tmtype1 
			       itervars2 tmtype2 
			       arg-intersections
			       undo-stack env1 env2 intersect-stack)
   (let ((prev-scope-time scope-time*))
      (!= scope-time* (+ *-* 1))
      (multi-let ((numargs (len itervars1))
		  ((augenv1 augenv2)
		   (letrec ()
		      (paramlists-declare-identically
			 (itervars-arglist
			     itervars1)
			 (itervars-arglist
			     itervars2)
			 env1 env2)
		    :where
		      (itervars-arglist (paramlist)
			 (new-Arglistspec
			    (<# (\\ (ip)
				   (new-Argspec
				      (param-type-varname
					 ip)
				      ':required
				      type-type*))
				paramlist))))))
	 (multi-let (((templ-e templ-i undo-stack-1)
		      (types-intersect
			 ctv true tmtype1 true tmtype2 undo-stack
			 augenv1 augenv2 intersect-stack)))
	    (check-expressibility-and-voidness
	       templ-e templ-i undo-stack-1 undo-stack
	       (let ((unks (<# Vartype-val 
			       (take numargs augenv1)))
		     ;; Somewhat crude
		     )
		 (let ((unk-ids (<# unknown-type-id unks)))
		    (values
		       true
		       (make-image-type
			  (<# (\\ (u) (type-feature u 'param))
			      unks)
			  (unk-elim templ-i unk-ids prev-scope-time
				    empty-env*)
			  arg-intersections
			  false
			  empty-env*)
		       undo-stack-1))))))))

(defun tup-val-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			    env1 env2 intersect-stack)
   (let ((med1 (type-find-feature ty1 'medium env1))
	 (ur1 (tup-val-to-dot-lin ty1 env1))
	 (ur2 (tup-val-to-dot-lin ty2 env2)))
      (cond ((and ur1 ur2)
	     (tup-val-ur-intersect ctv sign1 sign2 ur1 ur2 ty1 ty2 med1
				   undo-stack env1 env2 intersect-stack))
	    ((and ur1 (eq med1 'Row))
	     (tup-val-ur-intersect ctv sign1 sign2 ur1 ty2 ty1 ty2 med1
				   undo-stack env1 env2 intersect-stack))
	    ((and sign1 sign2)
	     (values true void-type* undo-stack))
	    (sign1
	     (values true ty1 undo-stack))
	    (t
	     (values true ty2 undo-stack)))))

(declare-class-feature intersector tup-class
   (make-Intersect-handler 14 #'tup-val-intersector))

(declare-class-feature intersector val-class
   (make-Intersect-handler 15 #'tup-val-intersector))

;;;;(defun tup-val-to-dot-lin (ty env)
;;;;   (let ((tcl (Type-class ty)))
;;;;      (cond ((or (eq tcl dot-class*)
;;;;		 (eq tcl val-class*))
;;;;	     ty)
;;;;	    ((or (eq tcl tup-class*)
;;;;		 (eq tcl val-class*))
;;;;	     (type-find-supertype ty env))
;;;;	    (t false))))

(defun tup-val-ur-intersect (ctv sign1 sign2 ur1 ur2 ty1 ty2 med undo-stack
			     env1 env2 intersect-stack)
   (multi-let (((ur-e ur-i undo-stack-1)
		(types-intersect ctv sign1 ur1 sign2 ur2 undo-stack
				 env1 env2 intersect-stack)))
      (check-expressibility-and-voidness
         ur-e ur-i undo-stack-1 undo-stack
	 (let ((alspec1 (type-find-feature ty1 'elt-types env1))
	       (alspec2 (type-find-feature ty2 'elt-types env2))
	       (ur-argspecs (dot-type-argspecs ur-i empty-env*))
	       (freevars (Type-freevars ur-i)))
	    (let ((combined-argspecs
		     (append
			(consistent-slotnames alspec1 alspec2 true)
			(consistent-slotnames alspec2 alspec1 false)))
		  (ur-alspec (new-Arglistspec ur-argspecs))
		  (ur-argspec-names '()))
	       ;; Calling new-Arglistspec sets the positions of the
	       ;; ur-argspecs correctly
	       (repeat :for ((ur-as :in ur-argspecs))
;;;;		  (err-out "Looking for match to " ur-as 
;;;;			   :% " among " combined-argspecs)
		  (repeat :for ((c-as :in combined-argspecs)
			       (this-argspecs-names '()))
;;;;		     (err-out "Checking c-as")
		     (cond ((equal (Argspec-position c-as)
				   (Argspec-position ur-as))
			    (!= this-argspecs-names
				(cons (Argspec-name c-as)
				      *-*))))
;;;;		     (err-out "Names so far: " this-argspecs-names :%)
		   :result (progn
			     (!= ur-argspec-names
				 (cons this-argspecs-names *-*))
			     (cond ((not (null this-argspecs-names))
				    (!= (Argspec-name ur-as)
					(car this-argspecs-names)))))))
	       (!= ur-argspec-names (reverse *-*))
;;;;	       (dbg-save ur-argspec-names ur-alspec)
;;;;	       (breakpoint tup-val-ur-intersect
;;;;		  "ur-argspec-names = " ur-argspec-names)
	       (values
		  true
		  (new-Type
		     (cond ((eq med 'Tup) tup-class*)
			   (t val-class*))
		     `(,(cond ((eq med 'Tup) 'Tup)
			      (t 'Val))
		       ,@(arglistspec-typed-arglist ur-alspec))
		     freevars
		     ur-i
		     (<! (\\ (as names)
			    (tup-type-slots as names ur-alspec))
			 ur-argspecs
			 ur-argspec-names)
		     (list (tuple 'medium med)
			   (tuple 'kind (cond ((eq med 'Tup) type-type*)
					      (t type-tup-type*)))
			   (tuple 'elt-types
				  (closure-or-closable
				     ur-alspec empty-env*
				     'arglistspec-close freevars))))
		  undo-stack-1))))))

(defun dot-type-argspecs (dtype env)
   (multi-let (((dtype env)
		(follow-var-ref dtype env)))
      (cond ((or (eq dtype empty-row-type*)
		 (eq dtype empty-tup-type*))
	     '())
	    ((or (eq (Type-class dtype) dot-class*)
		 (eq (Type-class dtype) lin-class*))
	     (let ((aty (type-find-feature dtype 'cartype env))
		   (dty (type-find-feature dtype 'cdrtype env)))
		(cons (new-Argspec '_ ':required (new-Type-closure aty env))
		      (dot-type-argspecs dty env))))
	    (t
	     (list (new-Argspec '_ ':rest (new-Type-closure dtype env)))))))

(defun consistent-slotnames (ap1 ap2 keep-intersects)
   (cond (ap1
	  (<! (\\ (as1)
		 (cond ((eq (Argspec-name as1) '_)
			'())
		       (t
			(cond (ap2
			       (repeat :for ((as2 :in (Arglistspec-argspecs ap2))
					    (name1 (Argspec-name as1))
					    (found false))
				  (!= found (eq (Argspec-name as2)
						name1))
				:until found
				:result (cond ((or (not found)
						  (and keep-intersects
						       (equal
							  (Argspec-position
							     as2)
							  (Argspec-position
							     as1))))
					      (list as1))
					     (t '()))))
			      (t (list as1))))))
	      (Arglistspec-argspecs ap1)))
	 (t '())))

(defun tup-type-slots (as names alspec)
   (let ((pos (Argspec-position as)))
      (repeat :for ((name :in (cons (position-slot-name as pos)
				    names)))
       :collect
	    (make-Slot name
		       (position-fun-name 'Tup pos)
		       (position-setter-name false pos)
		       (argspec-slot-type-closable as)
		       no-args-spec*
		       (make-Loadable-function-with-extras
			  :name 'tup-val-pos-type-invert
			  :fcn false
			  :towit (list alspec 'Tup pos))
		       `(,name ,pos)))))

(defun modified-intersector (ctv sign1 ty1 sign2 ty2 undo-stack
			     env1 env2 intersect-stack)
   (let ((cl1 (Type-class ty1))
	 (cl2 (Type-class ty2)))
      (cond ((and sign1 sign2
		  (eq cl1 modified-class*)
		  (eq cl2 modified-class*))
	     (multi-let (((base-e base-i undo-stack-1)
			  (types-intersect ctv
					   true (type-find-supertype ty1 env1)
					   true (type-find-supertype ty2 env2)
					   undo-stack
					   env1 env2 intersect-stack)))
		(check-expressibility-and-voidness
		   base-e base-i undo-stack-1 undo-stack
		   (multi-let (((slots-expressible slots-consis
				 slots-i undo-stack-2)
				(slots-intersect
				    ctv
				    (Type-slotfns ty1)
				    (Type-slotfns ty2)
				    ty1 ty2 undo-stack-1
				    env1 env2 intersect-stack)))
		      (check-expressibility-and-voidness
			 slots-expressible slots-consis undo-stack-2 undo-stack
			 (values
			    true
			    (make-type-of-class modified-class
			       '(Clump)
			       (Type-freevars base-i)
			       base-i
			       slots-i
			       '())
			    undo-stack-2))))))
	    (t
	     (letrec ()
		(cond ((and sign1 (eq cl1 modified-class*))
		       (super-intsect ty1 sign2 ty2 env1 env2))
		      ((and sign2 (eq cl2 modified-class*))
		       (super-intsect ty2 sign1 ty1 env2 env1))
		      (t
		       (values true '())))
	      :where
		(super-intsect (pos sign other env-p env-o)
		   (multi-let (((base-e base-i undo-stack-1)
				(types-intersect
				   ctv
				   true (type-find-supertype pos env-p)
				   sign other
				   undo-stack
				   env-p env-o intersect-stack)))
		      (check-expressibility-and-voidness
		         base-e base-i undo-stack-1 undo-stack
			 (values
			    true
			    (make-type-of-class modified-class
			       '(Clump)
			       (Type-freevars base-i)
			       base-i
			       (Type-slotfns ty1)
			       '())
			    undo-stack-1)))))))))
		   
(declare-class-feature intersector modified-class
   (make-Intersect-handler 99 #'modified-intersector))

(declare-class-feature intersector subtype-class
   (make-Intersect-handler 8 #'indep-types-intersect))

(defun hide-subtype (subty env)
   (let ((vl (Type-freevars subty)))
      (cond ((forall (n in '(defined-fun-param defined-fun-args-param))
		(exists (v in vl)
		   (eq (param-type-varname v)
		       n)))
	     (let ((dfun (var-lookup 'defined-fun-param env))
		   (dargs (var-lookup 'defined-fun-args-param env)))
	        (cond ((and dfun dargs)
		       (make-typeform
			  (Vartype-val dfun) (Vartype-val dargs)))
		      (t
		       (dbg-save :run-loud subty env)
		       (signal-problem hide-subtype :fatal
			  "Subtype with unbound fun and arg params:"
			  :% 3 subty
			  :% 3 env)))))
	    (t
	     (new-Type-closure subty env)))))

(defun slots-intersect (ctv sl1 sl2 ty1 ty2 undo-stack
			env1 env2 intersect-stack)
   (let ((combined-slots '())
	 (expressible true)
	 (consistent true))
      (repeat :for ((slot1 :in sl1) (ustack undo-stack))
       :result (values true
		       (nconc (<? (\\ (slot2)
				     (not (exists (slot1 :in sl1)
					     (eq (Slot-name slot1)
						 (Slot-name slot2)))))
				  sl2)
			      combined-slots)
		       ustack)
	 (repeat :for ((slot2 :in sl2))
	  :result (!= combined-slots
		      (cons slot1 *-*))
	  :until (eq (Slot-name slot2) (Slot-name slot1))
	  :result (cond ((and (equal (Slot-acc slot1) (Slot-acc slot2))
			      (equal (Slot-set slot1) (Slot-set slot2)))
			 (multi-let (((fun-e fun-i ustack-1)
				      (fun-intersector
					 ctv
					 true (slot-funtype slot1 ty1 env1)
					 true (slot-funtype slot2 ty2 env2)
					 ustack env1 env2 intersect-stack)))
			    (!= ustack ustack-1)
			    (cond (fun-e
				   (cond (fun-i
					  (!= combined-slots
					      (cons (intersection-slot
						       slot1 fun-i
						       (or (Slot-invert
							      slot1)
							   (Slot-invert
							      slot2))
						       (or (Slot-descriptor
							      slot1)
							   (Slot-descriptor
							      slot2)))
						    *-*)))
					 (t
					  (!= consistent false))))
				  (t
				   (!= expressible false)))))
		       (t
			(!= expressible false))))
       :while expressible
       :else :result (values false nil (undo ustack undo-stack))
       :while consistent
       :else :result (values true void-type* ustack))))

(defun slot-funtype (slot ty env)
   (let ((alspec 
            (new-Arglistspec
	       (cons (new-Argspec 'self ':required (new-Type-closure ty env))
		     (<# argspec-copy
			 (Arglistspec-argspecs (Slot-atypes slot)))))))
      (make-funtype
	 0
	 (Slot-type slot)
	 (arglistspec-argtype alspec env)
	 (type-arglistspec (Slot-type slot) env)
	 alspec false env)))

(defun intersection-slot (slot intersection-funtype invert descrip)
   (let ((arglist (type-find-feature intersection-funtype
				     'arglist empty-env*)))
      (let ((res-alspec
	       (new-Arglistspec (cdr (Arglistspec-argspecs arglist)))))
	 (make-Slot
	    (Slot-name slot)
	    (Slot-acc slot)
	    (Slot-set slot)
	    (type-find-feature intersection-funtype 'resulttype empty-env*)
	    res-alspec invert
	    descrip))))

(needed-by-macros

(def-meth make-loader ((ih Intersect-handler) tl-stack)
   `(make-Intersect-handler
       ',(Intersect-handler-priority ih)
       ,(make-loader (Intersect-handler-fcn ih)
		     tl-stack)))

(def-meth re-cycle ((ih Intersect-handler))
   nil)

)