;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: types.lisp,v 1.17 2005/12/26 20:34:25 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat %langutils/ namespace)

;;; Slurp to get macro definitions
(end-header :continue-slurping)

;;; This is a bit lame.  Other files depend on this one at compile time,
;;; so multilet will generally be loaded. 
(depends-on :at-run-time %ytools/ debug)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(Argspec make-Argspec Argspec-name Argspec-mode
	     Argspec-kind Argspec-position
	     Argspec-type Argspec-default argspec-feat Argspec-seg-type
	     argspec-copy 
	     is-Type Type type-displace
	     arglistspec new-Arglistspec new-Argspec
	     is-Arglistspec Arglistspec-argspecs
	     Arglistspec-required Arglistspec-optional Arglistspec-rest
	     Arglistspec-key Arglistspec-all-nonrest-argspecs
	     Arglistspec-allow-other-keys args->spec
	     type-type* 
	     Vartype new-Vartype
	     Vartype-var Vartype-type Vartype-val Vartype-initial
	     vartype-feat is-Vartype 
	     def-Type-system new-Type-system is-Type-system Type-system-name
	     find-type-system Type-system-namespace Type-system
	     Type-class Type-class-name new-Type-class find-type-class
	     Type-desig Type-freevars Type-super Type-slotfns Type-features
	     universal-type-sys* universal
	     global-type-define global-ration global-declare global-declaration
	     define-global-constant-type globally-defined-type typedecl
	     def-type-class new-Type make-type-of-class
	     declare-class-feature class-feature-datafun
	     name-lookup name-lookup-or-bind
	    )))

(end-header)

(defvar type-system-tab* '())

;;; For use in printers
(def-op short-label (x)
   x)

;;; Returns < value, found >  'value' is meaningless if 'found' is false.
(def-op name-lookup (some-sort-of-globalish-var-binder name))

;;; Returns value associated with name; if none, calls 'creator' to make one
;;; and returns the new value.  'creator' takes name as arg.
(def-op name-lookup-or-bind
        (some-sort-of-globalish-var-binder name creator))

(def-op get-namespace (x) false)

(def-class Type-system (:options (:kind :object))
    name
    var
    global-decl-tab
    parents    ;; List of other type systems whose symbols are visible here
    env        ;; Convenience; see file 'vartypes'
    namespace
    (:handler
     (print (ts srm)
	(out (:to srm) "#{Type-system " (Type-system-name ts) "}"))
     (name-lookup (ts varname)
        (global-declaration varname ts false))
     (name-lookup-or-bind (ts varname creator)
	(or (global-declaration varname ts false)
	    (and creator
		 (global-declare varname (funcall creator varname)
				 ts true))))
     (get-namespace (ts) (Type-system-namespace ts))
     (short-label (ts) (Type-system-name ts))))

(defun new-Type-system (name var parents &optional (namespace ':create))
         ;;;   (let ((var (type-system-var name))) ...)
   (make-Type-system
      name var (make-eq-hash-table) parents false
      (cond ((eq namespace ':create)
	     (place-Namespace-with-exports name false +all-symbols+))
	    (t namespace))))

(defvar top-level-type-system* false)

;;; (keywords-clauses-let ((v1 [f1] s1) (v2 [f2] s2) ...) v l ---body---)
;;; 
(defmacro keywords-clauses-let (specs remnant l &body b)
   (multiple-value-let (decls b)
		       (ytools::declarations-separate (ytools::ignore-convert b))
      (letrec ()
	 `(let ((,remnant ,l))
	     ,(letrec ()
		(repeat :for ((vspec :in specs)
			      (e b))
		   (multiple-value-let (var fcn s)
				       (unpack-spec vspec)
		      (!= e (wrap-one var fcn s decls *-*)))
		   (!= decls '())
		:result `(progn ,@e))))
       :where
	 (unpack-spec (vs)
	    (cond ((= (len vs) 3)
		   (values (car vs) (cadr vs) (caddr vs)))
		  (t
		   (values (car vs) false (cadr vs)))))

	 (wrap-one (var fcn spec decls e)
;;;;	    (trace-around wrap-one
;;;;	       (:> "(wrap-one: " var 1 fcn ")")
	    `((multiple-value-let (,var ,remnant)
				 (extract-keyword-clauses
				     ',(car spec) ,remnant
				     ,(cadr spec)
				     ,(cond ((cddr spec)
					     (caddr spec))
					    (t 'false)))
				 ,@decls
		 ,@(cond (fcn
			 `((let ((,var (,fcn ,var)))
			     ,@e)))
			 (t e))))
;;;;	       (:< (val &rest _) "wrap-one: " val))
	    ))))

(needed-by-macros

(defun one-key-list (l) (cond ((null l) '()) (t (car l))))
   
(defun one-key (l)
   (cond ((null l) false)
	 (t
	  (!= l (car l))
	  (cond ((not (null (cdr l)))
		 (error-break one-key
		    "More than one thing found where one expected: " l
		    (:novalue "I will ignore all but the first"))))
	  (car l))))

(datafun :slurp-types def-Type-system
   (defun :^ (dts _)
      (keywords-clauses-let ((par one-key-list (:parents 0 1))
			     (var one-key (:var 0 1)))
			    leftover (cddr dts)
			    (ignore leftover)
	 (cond ((not (eq (cadr dts) 'universal))
		(!= par (adjoin 'universal par))))
	 ;;(out "Slurping... par = " par " var = " var :%)
	 (renew-Type-system (cadr dts)
			    var
			    (<# (\\ (p) (find-type-system p true))
				par)))
     false))
)

(defmacro def-Type-system (name &rest clauses)
   (keywords-clauses-let ((par one-key-list (:parents 0 1))
			  (var one-key (:var 0 1)))
			 leftover clauses
      (cond ((not (eq name 'universal))
	     (!= par (adjoin 'universal par))))
      (cond ((not (null leftover))
	     (out (:to *error-output*) "Warning: unparseable junk " leftover
		  " in def-Type-system " name :%)))
      `(progn ,@(include-if (and var (not (eq name 'universal)))
			    `(defvar ,var))
	      (renew-Type-system
		  ',name
		  ',var
		  (list ,@(<# (\\ (p) (type-system-loader (find-type-system p true)))
			      par))))))

(needed-by-macros

(defun renew-Type-system (name var parents &optional (namespace ':create))
   (let ((ts (alist-entry name type-system-tab* false)))
      (cond (ts
	     (!= (Type-system-parents ts) parents))
	    (t
	     (!= ts (new-Type-system name false parents namespace))
	     (!= type-system-tab* (cons (tuple name ts) *-*))))
      (cond (var
	     (!= (Type-system-var ts) var)
	     ;;(out "setting " var :%)
	     (!= (symbol-value var) ts)))
      ts))

;;; Produce a form that at some later date will evaluate to sys.

(defun type-system-loader (sys)
   (cond ((not (is-Symbol (Type-system-name sys)))
	  (error-break type-system-loader :fatal
	     "Bogus sys name: " (Type-system-name sys) " in " sys)))
   (or (Type-system-var sys)
       `(find-type-system ',(Type-system-name sys) true)))

(defun find-type-system (name must-exist)
   (or (alist-entry (cond ((or (not name)
			       (eq name ':universal))
			   'universal)
			  (t name))
		    type-system-tab*
		    false)
       (cond (must-exist
	      (error-break find-type-system :fatal
		   "Undefined system name: " name))
	     (t false))))
)

(defun global-declaration (name ts must-exist)
   (let ((chk (list ts)))
      (letrec ((check-ancestor (a)
		 (multiple-value-let (e found)
				     (table-entry (Type-system-global-decl-tab a)
						  name)
		     (cond (found
			    (values e true))
			   (t
			    (!= chk (cons a *-*))
			    (repeat :for ((p :in (Type-system-parents a)))
			     :result (values false false)
			       (cond ((not (memq p chk))
				      (!= < e found >
					  (check-ancestor p))))
			     :until found
			     :result (values e true)))))))
	 (multiple-value-let (d found)
			     (check-ancestor ts)
	    (cond ((or d (not must-exist))
		   (values d found))
		  (t
		   (cond (must-exist
			  (error-break global-declaration :fatal
			     "Undefined name: " name " in " ts))
			 (t false))))))))

(defun (setf global-declaration) (new name ts)
   (!= (table-entry (Type-system-global-decl-tab ts)
		    name)
       new))

(defun global-declare (name defn ts break-on-change)
   (cond (break-on-change
	  (let ((old (table-entry (Type-system-global-decl-tab ts)
				  name)))
	     (cond (old
		    (error-break global-declare :novalue
		       "Changing definition of " name))))))
   (!= (table-entry (Type-system-global-decl-tab ts)
		    name)
       defn)
   defn)

;;; This is the type system that contains all (and only) types that are visible 
;;; everywhere.
(defvar universal-type-sys*)

(def-class Type-class
       (:handler
	(print (tc srm)
	   (out (:to srm)
		(:pp-block 
		   "#{Type-class " (:a (symbol->string (Type-class-name tc)))
		   (:q ((Type-class-system tc)
			"/" (:a (symbol->string
				   (Type-system-name (Type-class-system tc)))))
		      (t "[?]"))
		   1
		   (:pp-ind :block 10)
		   (:pp-nl :linear)
		   (<# car (Type-class-featurelist tc)) "}"))))
    name
    var
    system     ;; Actually, anything that responds to name-lookup
    featurelist)

(defun new-Type-class (name var system)
   (make-Type-class name
		    var
		    system
		    '()))

(defun renew-Type-class (name var system)
   (let ((tc (find-type-class name system false)))
      (cond ((not tc)
	     (!= tc (new-Type-class name var system))
	     (global-declare name tc system false)))
      (cond (var
	     (!= (Type-class-var tc) var)
	     (!= (symbol-value var) tc)))
      tc))

(defmacro def-type-class (name &rest props)
   (keywords-clauses-let ((var one-key (:var 0 1))
			  (sysname one-key (:system 0 1))
			  (simp-feats (:feature 0))
			  (fun-feats (:feature-datafun 0)))
			 props props
      (let ((class-feats '()))
	 (cond ((not sysname)
		(!= sysname (car props))
		(!= props (cdr *-*))))
	 (let ((sys (find-type-system sysname true)))
	    (let ((unambigname (cond (sys `(,name ,sysname))
				     (t name))))
	       (repeat :for ((p :in fun-feats))
		  (!= class-feats
		      (cons `(class-feature-datafun ,(car p)
						    ,unambigname ,@(cdr p))
			    *-*)))
	       (repeat :for ((p :in simp-feats))
		  (!= class-feats
			     (cons `(declare-class-feature ,(car p)
							   ,unambigname ,(cadr p))
				   *-*)))
	       (cond ((not (null props))
		      (error-break define-type-class
			 "Illegal in define-type-class " name ": " props
			 (:novalue "I will ignore it"))))
	       `(progn ,@(include-if var `(defvar ,var))
		       (renew-Type-class ',name ',var
					 ,(type-system-loader
						    (find-type-system sysname true)))
		       ,@(reverse class-feats)))))))

(defmacro class-feature-datafun (featname classname def)
   (multiple-value-let (cn sys)
		       (unpack-sys classname)
      (let ((funname (cond ((eq sys universal-type-sys*)
			    (build-symbol (< cn) - (< featname) -fcn))
			   (t
			    (build-symbol (< (Type-system-name sys))
					  - (< cn) - (< featname)
					  -fcn)))))
	 (match-let ?(:\| (?defcom :^ ?@defn)
			   (?defcom ?@defn))
		    def
	    `(progn
		(,defcom ,funname ,@defn)
		(declare-class-feature ,featname ,classname #',funname))))))

(defmacro declare-class-feature (featname classname val^)
    `(!= (alist-entry ',featname
		      (Type-class-featurelist 
		          ,(type-classname-loader classname))
		      nil)
	 ,val^))

(defun find-type-class (name sys must-exist)
   (cond ((not sys)
	  (!= sys universal-type-sys*)))
   (cond ((is-Symbol sys)
	  (error-break find-type-class
	     "Use of symbol for system: " sys
	     (:novalue "I'll try to figure out what it denotes"))
	  (!= sys (find-type-system sys true))))
   (let ((x (global-declaration name sys must-exist)))
      (cond ((and x (not (is-Type-class x)))
	     (error-break find-type-class :fatal
		"Symbol " name " is supposed to denote a type class in system " sys ","
		:% " but it denotes: " x)))
      x))		

(defun type-class-loader (class)
   (or (Type-class-var class)
       `(find-type-class ',(Type-class-name class)
			 ,(type-system-loader (Type-class-system class))
			 true)))

;;; Types, at last

(with-packages-unlocked           ; :common-lisp

;;; We don't know what freevars are yet, so we can't print
;;; them until someone tells us how.    
(defvar type-freevar-printer* false)

(def-class Type
     (:handler
      (print (ty stream)
	     (out (:to stream) "#{Type " 
			      (Type-desig ty)
			      (:e (cond (type-freevar-printer*
					(funcall type-freevar-printer*
						 (Type-freevars ty)
						 stream))))
			      "}")))
   desig
   freevars
   super     ; supertype
   slotfns
   features
   cached-features   ; kept separate.  See tloader.nsp for explanation
   class  ; Type-class (may not be omitted)
))



;;;(defun Type-home (ty)
;;;   (let ((cl (Type-class ty)))
;;;      (cond (cl (type-class-system cl))
;;;	    ((is-Symbol (Type-desig ty))
;;;	     (error-break Type-home
;;;		"Type with no class: " ty
;;;		(:novalue "I'll assume it's in the universal type system"))
;;;	     universal-type-sys*)
;;;	    (t false))))

(defun type-displace (old new)
   (!= (Type-desig old) (Type-desig new))
   (!= (Type-freevars old) (Type-freevars new))
   (!= (Type-super old) (Type-super new))
   (!= (Type-slotfns old) (Type-slotfns new))
   (!= (Type-features old) (Type-features new))
   (!= (Type-cached-features old) '())
   (!= (Type-class old) (Type-class new))
   old)

(defun new-Type (class desig freevars super slotfns features)
   (cond ((and (not class) (is-CSymbol desig)
	       (not (assq 'forward features)))
	  (error-break new-Type :fatal
	     "Type with false class: " desig))
	 ((exists (x :in freevars)
	     (is-Vartype x))
	  (error-break new-Type :fatal
	     "Bogus freevars: " freevars))
	 ((desig-contains-types desig)
	  (error-break new-type :fatal
	     "Bogus type desig: " desig)))
   (cond ((matchq (>Row (<Row ?_)) desig)
	  (error-break new-Type :novalue
	     ">Row (<Row ...)")))
   (cond ((and (is-CSymbol desig)
	       (not (assq 'home features)))
	  (!= features
	      (cons (tuple 'home (ensure-top-level-type-sys))
		    *-*))))
   (let ((r (make-Type desig freevars super slotfns features '() class)))
      ;;(check-for-bogosity r)
      r))

(defun desig-contains-types (d)
   (cond ((atom d)
	  (is-Type d))
	 (t
	  (or (desig-contains-types (car d))
	      (desig-contains-types (cdr d))))))

;; Useful at one point, but the bogosity is sometimes okay.
; (defun check-for-bogosity (ty)
;    (letrec ((find-&rest-error (d)
; 	       (cond ((atom d) false)
; 		     ((and (eq (car d) '&rest)
; 			   (symbolp (cadr d))
; 			   (symbolp (caddr d))
; 			   (not (eq (caddr d) '-)))
; 		      true)
; 		     (t
; 		      (or (find-&rest-error (car d))
; 			  (find-&rest-error (cdr d))))))
; 	    (find-tup-error (d)
; 	       (equal (butlast d)
; 		      '(tup x &rest (u) l -))))
;       (cond ((find-&rest-error (Type-desig ty))
; 		;; (find-tup-error (Type-desig ty))
; 	     (error-break check-for-bogosity :novalue
; 		"Creating type with malformed desig: " ty)))))
      
(defmacro make-type-of-class (classname
			      desig^ freevars^ super^ slotfns^ features^)
  `(new-Type  ,(type-classname-loader classname)
	      ,desig^ ,freevars^ ,super^ ,slotfns^ ,features^))

;;; This is the only type defined before foundation.nsp.  We need it to be
;;; able to define any others.  
(defvar type-type* (make-Type '(T) '() false '() '() '() false))
;;; It's not complete until we give it a supertype and type class, in
;;; foundation.

;; A Vartype is a record of a bound variable.  val is its
;; value;  props is an association list.  If the variable is not really
;; bound until run time, then val is its initial value, or false if not
;; known.
(def-class Vartype 
	  (:handler
	   (print (vt s)
	     (let ((x (Vartype-typ vt)))
	        (out (:to s) 
		     "#{Vartype " (Vartype-var vt)
				 " - " (cond ((is-Type x)
					      (Type-desig x))
					     (t x))   ; can be a type-closure
				 " - " (Vartype-initial vt) "}"))))
   var type val props)

(defun Vartype-typ (vt) (Vartype-type vt))

(defun Vartype-initial (vt) (Vartype-val vt))

(defun (setf Vartype-initial) (x vt) (setf (Vartype-val vt) x))

(defun new-Vartype (var type i)
;;;;   (cond ((and (not (is-Symbol var))
;;;;	       (not (is-Typed-exp var)))
;;;;	  (signal-problem new-Vartype
;;;;	     "Binding non-symbol: " var)))
;;;;   (cond ((and (not (is-Type type))
;;;;	       (not (matchq (not ?+(?_ is-Type))
;;;;			    type)))
;;;;	  (signal-problem new-Vartype
;;;;	     type " is not a type")))
;;;;   (cond ((not type)
;;;;	  (signal-problem new-Vartype :fatal
;;;;	     "Null type for " var)))
   (make-inst Vartype :var var :type type :val i :props '()))

(defun vartype-feat (vt n)
   (let ((p (assq n (Vartype-props vt))))
      (cond (p
	     (values (cadr p) true))
	    (t
	     (values false false)))))

(defun (setf vartype-feat) (v vt n)
   (let ((p (assq n (Vartype-props vt))))
      (cond ((not p)
	     (!= p (lrecord n nil))
	     (!= (Vartype-props vt)
		 (cons p *-*))))
      (!= (cadr p) v)
      v))

(defmacro define-global-constant-type (name def^ ts)
   `(global-type-define ',name ,def^ (find-type-system ',ts true)))

(defun global-type-define (name typ ts)
   (typedecl name type-type* typ ts))

(defun globally-defined-type (sym ts must-exist)
   (let ((d (global-declaration sym ts must-exist)))
      (cond (d
	     (cond ((is-Vartype d)
		    (let ((def (Vartype-val d)))
		       (cond ((and (eq (Vartype-type d) type-type*)
				   (is-Type def))
			      def)
			     (must-exist
			      (error-break globally-defined-type :fatal
				 "Attempt to find type " sym ", bound thus: " d))
			     (t false))))
		   (must-exist
		    (error-break globally-defined-type :fatal
		       "Attempt to find type " sym ", declared as: " d))
		   (t
		    false)))
	    (t false))))

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
   (defun type-classname-loader (classname)
      (multiple-value-let (classname sys)
			  (unpack-sys classname)
	 `(find-type-class ',classname
			   ,(type-system-loader sys)
			   true)))
)

(defun extract-sys (thing dotdotdot &optional (use-top-level-default false))
   (multiple-value-let (sys dotdotdot)
		       (find-system-clause dotdotdot)
      (multiple-value-let (real-thing sys1)
			  (unpack-sys thing
				      (and use-top-level-default
					   (not sys) ))
	 (values real-thing
		 dotdotdot
		 (or sys sys1)))))

(defun find-system-clause (def)
	 (repeat :for ((dl def (cdr dl))
		       d)
	  :until (null dl)
	  :result (values nil def)
	    (!= d (car dl))
	  :until (car-eq d ':system)
	  :result (values (find-type-system (cadr d) true)
			  (remove d def :test #'eq :count 1))))

(defun ensure-top-level-type-sys ()
   (or top-level-type-system*
       (prompt-for-top-level-type-system)))

(defun prompt-for-top-level-type-system ()
   (let ((ts (find-type-system
		(error-break quest-for-top-level-type-system
		   "No top-level type-system set"
		   :% "Use in-type-system to set it properly"		   
		   (:prompt-for
		    "the name of the type system to use from now on at top-level"
		    (\\ (x) (and (is-Symbol x)
				 (find-type-system x false)))))
		true)))
     (!= top-level-type-system* ts)
     ts))

(datafun yt::general-slurper in-type-system
   (defun :^ (form tasks states)
      (switch-into-type-system (cadr form))
      (values tasks states)))

(defmacro in-type-system (name)
   `(eval-when (:compile-toplevel :execute :load-toplevel)
      (switch-into-type-system ',name)))

(defun switch-into-type-system (ts-name)
   (let ((tts top-level-type-system*))
      (cond ((during-file-transduction)
	     (after-file-transduction 
		(!= top-level-type-system* tts))))
      (!= top-level-type-system* (find-type-system ts-name true))))

(defvar dofiles* true)

(defvar typedecl-checker* false)

(defun typedecl (sym typ initial sys)
   (let ((actual-sym
	    (cond ((is-Symbol sym) sym)
		  ((is-Canonized-symbol sym)
		   (Canonized-symbol-sym sym))
		  (t (signal-problem typedecl
			"Attempt to declare non-name " sym
			" as type in " sys)))))
      (cond ((and dofiles* now-loading*)
		(!= (get actual-sym 'files)
		    (cons `(,now-loading* ,(Type-desig typ) ,initial)
			  *-*))))
      (let ((d (name-lookup-or-bind
		  sys sym
		  (\\ (sym) (new-Vartype sym false false)))))
			   ;; (global-declaration sym sys false)
	 (let ((dt (Vartype-type d))
	       (old-val (Vartype-val d)))
	    (cond ((not dt)
		   ;; New
		   (!= (Vartype-type d) typ)
		   (!= (Vartype-initial d) initial))
		  (t
		   (cond (typedecl-checker*
			  (funcall typedecl-checker*
				   sym dt typ sys)))
		   (cond ((and (is-Type initial)
			       (is-Type old-val)
			       (or (alref (Type-features old-val) 'forward)
				   (not (is-symbolish (Type-desig initial)))))
			  (type-displace old-val initial))
			 (t
			  (cond ((and (is-Type old-val)
				      (is-Type initial))
;;;;				 (cond ((and (not (alref (Type-features old-val)
;;;;							 'forward))
;;;;					     (is-symbolish (Type-desig initial)))					(err-out
;;;;					   "Old not declared forward: "
;;;;					   old-val
;;;;					   :% " New? " (alref (Type-features
;;;;							         initial)
;;;;							      'forward)
;;;;					   :% "Name is symbolish: "
;;;;					   initial)))
				 ))
			  (!= (Vartype-type d) typ)
			  (!= (Vartype-val d) initial))))))))
   sym)

;;; Handy in macros.

(defun extract-keywords (l specs)
   (repeat :for ((s :in specs)
		 (acc '()))
    :result (values (reverse acc) l)
      (multiple-value-let (k r)
			  (extract-keyword-clauses (car s) l
						   (cadr s)
						   (and (cddr s) (caddr s)))
	 (!= acc (cons k *-*))
	 (!= l r))))

(needed-by-macros
;;; Given a list l = (... (:kw ...) .... (:kw ...)
;;; return < k, l'>,
;;; where k is list of all clauses starting with :kw, and l' is l with all those
;;; clauses removed.  If number found is < lo or > hi, complain.  If hi = false,
;;; there is no upper bound.
(defun extract-keyword-clauses (kw l lo hi)
   (letrec ((check-bounds (k)
	       (let ((n (len k)))
		  (cond ((< n lo)
			 (error-break extract-keyword-clauses
			    "Too few occurrences of " kw " in " l
			    :fatal))
			((and hi (> n hi))
			 (error-break extract-keyword-clauses
			    "Too many occurrences of " kw " in " l
			    :fatal)))
		  k)))
      (values (check-bounds
		 (<# cdr (<? (\\ (c) (car-eq c kw))
			     l)))
	      (<? (\\ (c) (not (car-eq c kw)))
		  l))))

;;; Analyze exp of form (classname sysname).  
;;; Return the classname and the type-system (NOT its name).
;;; Used occasionally when the classname isn't the name of a class.
;;; We actually don't care what it is, so long as it's not a list.
;;; If classname without sys, return false for sys, unless
;;; use-top-level-default = true, when return top-level-type-system*,
;;; bound elsewhere.
(defun unpack-sys (classname &optional (use-top-level-default false))
   (cond ((atom classname)
	  (values classname
		  (cond (use-top-level-default
			 (ensure-top-level-type-sys))
			(t
			 universal-type-sys*))))
	 (t
	  (values (car classname)
		  (find-type-system (cadr classname) true)))))
)

;;; Description of a single function arg
;;; A given Argspec should appear in exactly one Arglistspec, because
;;; it may have a different position or props in different Arglistspecs. 
(def-class Argspec
      (:handler
       (print (a srm)
          (let ((k (Argspec-mode a))
		(pos (Argspec-position a)))
	        ;(is-key (not (memq (Argspec-mode a)
			;	   '(:required :optional :rest)))))
	     (letrec ()
	         (out (:to srm)
		      "#{Argspec " (Argspec-name a)
		      1
		      (:q ((eq k ':required)
			   pos)
			  ((eq k ':optional)
			   pos "?")
			  ((eq k ':rest)
			   "(" (:q ((atom pos) "?? " pos)
				   (t (car pos)))
			       "..)")
			  (t pos))
		      1 (:q ((argspec-feat a 'seg-type)
			    "= "))
		      (Argspec-type a) "}")))))
   name
   mode ; :required, :optional, :rest, :key
   position  ; number or keyword, or (number) for :rest argspec
   type ; may be a type closure (see vartypes)
   default  
   props)

(defun Argspec-kind (a) (Argspec-mode a))

(define-compiler-macro Argspec-kind (a)
   (signal-problem Argspec-kind "Occurrence of (Argspec-kind " a ")")
   `(Argspec-mode ,a))

(defun new-Argspec (name k type &optional (default false) (props '()))
   (cond ((not (is-CSymbol name))
	  (signal-problem new-Argspec
	     "Creating Argspec with non-symbolic param name " name
	     (:proceed "I'll allow it"))))
   (multiple-value-let (mode position)
		       (cond ((memq k '(:required :optional :rest))
			      (values k 0))
			     ((is-CSymbol k)
			      (values ':key k))
			     (t
			      (error-break new-Argspec :fatal
				 "Illegal argspec mode: " k)))
      (cond ((not (memq mode '(:key :required :optional :rest)))
	     (dbg-save :run-loud name mode type)
	     (error-break new-Argspec
		"Bogus mode: " mode)))
      (make-Argspec name mode position type default props)))

(defun argspec-copy (as)
   (make-Argspec (Argspec-name as)
		 (Argspec-mode as)
		 ;;(Argspec-position as)
		 (argspec-pos-zero as)
		 (Argspec-type as)
		 (Argspec-default as)
		 (<# list-copy (Argspec-props as))))

(defun argspec-pos-zero (as)
   (cond ((eq (Argspec-mode as) ':key)
	  (let ((pos (Argspec-position as)))
	     (cond ((is-Argspec pos)
		    (signal-problem argspec-pos-zero :fatal
		       "Argspec " as
		       :% "  has pos " pos)))
	     pos))
	 ;; will be set when incorporated into an Arglistspec
	 (t 0)))

(defun argspec-pos-clear (as)
   (!= (Argspec-position as)
       (argspec-cleared-pos as))
   as)

(defun argspec-cleared-pos (as)
   (cond ((eq (Argspec-mode as) ':key)
	  (Argspec-position as))
	 (t
	  (!= (Argspec-position as) 0)
	  0)))

(defun argspec-feat (vt n)
   (let ((p (assq n (Argspec-props vt))))
      (cond (p
	     (values (cadr p) true))
	    (t
	     (values false false)))))

(defun (setf argspec-feat) (v vt n)
   (let ((p (assq n (Argspec-props vt))))
      (cond ((not p)
	     (!= p (lrecord n nil))
	     (!= (Argspec-props vt)
		 (cons p *-*))))
      (!= (cadr p) v)
      v))

(defun Argspec-supplied-param (as)
   (argspec-feat as 'supplied-param))

(defun new-seg-Argspec (name ty)
   (new-Argspec name ':rest ty
		false (list (tuple 'seg-type true))))

;;; This is meaningful only for :rest args.
(defun Argspec-seg-type (as)
  (argspec-feat as 'seg-type))

(defun Argspec-seg-type-set (as v)
  (!= (argspec-feat as 'seg-type) v))

(defsetf Argspec-seg-type Argspec-seg-type-set)

;;; Description of the args of a function
(def-class Arglistspec
    (:handler
     (print (als srm)
	(let ((ak (Arglistspec-allow-other-keys als))
	      ;(nm (Arglistspec-names-matter als))
	      )
	   (out (:to srm)
		"#{Arglist " (Arglistspec-required als)
		   (:q ((not (null (Arglistspec-optional als)))
			" ?- "
			(Arglistspec-optional als)))
		   (:q ((not (null (Arglistspec-rest als)))
			" ..- "
			(car (Arglistspec-rest als))))
		   (:q ((not (null (Arglistspec-key als)))
			" :- " (Arglistspec-key als)))
		   (:q (ak
			" +:"))
		   "}"))))
   argspecs
   required
   optional
   rest     ;; a list with at most one element
   key
   allow-other-keys
   names-matter) ;; obsolete.  

;;; 'required' must include the '&required' Argspecs, but may include
;;; others as well.  'new-Arglistspec' will sort it out
(defun new-Arglistspec (required
			&key (optional '())
			     (rest '())
			     (key '())
			     (allow-other-keys false)
			     (names-matter false))
   (cond ((not (forall (a :in required)
		  (eq (Argspec-mode a) ':required)))
	  (let ((req1 '()) (opt1 '()) (key1 '()) (rest1 '()))
	     (repeat :for ((a :in required))
		(case (Argspec-mode a)
		   (:required (!= req1 (cons a *-*)))
		   (:optional (!= opt1 (cons a *-*)))
		   (:rest (!= rest1 (cons a *-*)))
		   (:key (!= key1 (cons a *-*)))
		   (t
		    (dbg-save :run-loud a)
		    (error-break new-Arglistspec
		       "Argspec with weird mode " (Argspec-mode a) ": " a
		       (:proceed "I will ignore the weird argspec")))))
	     (!= required (nreverse req1))
	     (!= optional (nreconc opt1 *-*))
	     (!= rest (nreconc rest1 *-*))
	     (!= key (nreconc key1 *-*)))))
   ;;(out "Sorted args" :%)
   (cond ((> (len rest) 1)
	  (error-break new-Arglistspec :fatal
	     "Too many &rest args: " rest))
	 (t
	  (let ((alspec
		   (make-Arglistspec
		      (append required optional rest key)
		      required optional rest
		      key allow-other-keys names-matter)))
	     (arglistspec-number alspec)))))

(defun arglistspec-number (alspec)
	     (repeat :for ((a :in (Arglistspec-argspecs alspec))
			   (i = 1 :by 1))
	      :until (eq (Argspec-mode a) ':key)
		(cond ((not (eql (Argspec-position a) 0))
		       (dbg-save :run-loud a)
		       (signal-problem new-Arglistspec 
			    "Warning: Changing position of argspec "
			    a
			    :proceed)))
		(!= (Argspec-position a)
		    (cond ((equal (Argspec-mode a) ':rest)
			   `(,i))
			  (t i))))
	     alspec)

;;; The most usual case is a list of required args
(defun args->spec (req)
   (new-Arglistspec (<# (\\ (aty)
			   (new-Argspec '_ ':required aty))
			req)))

(defvar no-args-spec* (args->spec '()))

;;; The following are for use at run time by various type testers

(defun find-key (x k ist)
   (cond ((null x) false)
	 ((and (consp x)
	       (eq (car x) k))
	  (and (consp (cdr x))
	       (funcall ist (cadr x))))
	 (t
	  (find-key (cddr x) k ist))))

(defun all-keys-from (x keys)
   (cond ((null x) true)
	 ((and (consp x)
	       (consp (cdr x)))
	  (and (memq (car x) keys)
	       (all-keys-from (cddr x) keys)))
	 (t false)))

