;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: typesee.lisp,v 1.6 2006/03/14 17:19:32 dvm Exp $

;;; Debugging

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ types desig parmsgen)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(typesee symtype dtp dtsee)))

;;(defvar debug-type-sys* nil)

;;(defvar debug-env* nil)

;;; Testbed
(defun ta (ty1 ty2)
   (let ((e (ensure-debug-env-set)))
      (!= ty1 (coerce-to-type *-* e))
      (!= ty2 (coerce-to-type *-* e))
      (try-accept ty1 ty2 '() e e
		  '())))

(defun ensure-debug-env-set ()
  (or debug-env*
      (progn
	  (!= debug-type-sys*
	      (find-type-system
			  (error-break ensure-debug-env-set 
			     "No debug environment set"
			     (:prompt-for
				"Type system name"
				(\\ (x) (and (is-Symbol x)
					     (find-type-system x false)))))
			  true))
	  (!= debug-env* (place-type-system-env debug-type-sys*))
	  debug-env*)))

(defun gt (sym)
   (var-val sym (ensure-debug-env-set)))
;;   (globally-defined-type sym (ensure-debug-type-sys-set) true))

(defvar typesee-classes-only-names* true)

(defun typesee (typ &optional (depth 0))
   (let ((env (ensure-debug-env-set)))
      (letrec ()
	 (cond ((is-type-desig typ env)
		(!= typ (designated-type typ false env)))
	       ((is-Type-closure typ)
		(out "Env: " :% (Type-closure-env typ) :% :%)
		(!= typ (Type-closure-type typ)))
	       ((is-CSymbol typ)
		(let ((ty (symtype typ)))
		   (cond ((not ty)
			  (out "No type: " typ :%)))
		   (!= typ ty))))
	 (cond ((not typ))
	       ((is-Type typ)
		(out "Designator: " (td-massage (Type-desig typ)))
		(let ((fv (Type-freevars typ)))
		   (out :% 5 "Freevars: " (:q ((atom fv) fv)
					    (t (<# freevar-massage fv)))))
		(cond ((Type-slotfns typ)
		       (out
			:% "Slots: " (:e (slots-print (Type-slotfns typ)))))
		      (t (out :% "NO SLOTS"))   )
		(cond ((not (null (Type-features typ)))
		       (out :% "Features: ")
		       (features-see (Type-features typ))))
		(cond ((not (null (Type-cached-features typ)))
		       (out :% "Cached features: ")
		       (features-see (Type-cached-features typ))))
		(out :% "Class: "
                     (:q (typesee-classes-only-names*
                          (Type-class-name (Type-class typ)))
                         (t (Type-class typ))))
                (cond ((Type-super typ)
                       (cond ((> depth 0)
                              (out :% :% "Supertype: " :%)
                              (typesee (Type-super typ) (- depth 1)))
                             (t
                              (out :% :% "Supertype:" (Type-super typ) :%))))))
	     (t (out (ytools::condense typ) " is not a type")))
	 '*   
       :where
	  (features-see (feats)
	      (cond ((atom feats)
		     (out feats))
		    (t
		     (repeat :for ((f :in feats))
			 (out :% (car f) ": " (cadr f)))))))))

(defun slots-print (sl)
   (repeat :for ((s :in sl) (pc 1))
;;;;		 (dbg-save s)
;;;;		 (breakpoint slots-print
;;;;		    "Got " s)			     
      (selq (Slot-name s)
         (conser
	  (out (:pp-block 
		  "CONSER: " (Slot-acc s)
		  (:q ((not (atom (Slot-acc s)))
		      (:t (+ pc 8)))
		     (:t 1))
		  (:pp-nl :linear)
		  "Arg Types: "
		  (:q ((Slot-atypes s)
;;;;		      (dbg-save s)
;;;;		      (breakpoint slots-print
;;;;			 "Ready to print atypes of " s)
		      ;;(setq foo3* 'hello)
		      (arglistspec-typed-arglist
			 (follow-var-ref (Slot-atypes s) debug-env*)))))))
         (is
          (out "IS-tester: " (Slot-acc s)))
         (t
	  ;(out " name>>> " (slot-name s) :%)
          (out (Slot-name s))
          (cond ((Slot-type s)
                 (out " [" (Type-desig (Slot-type s)) "]: "))
                (t (out ": "))   )
          (cond ((Slot-acc s)
                 (out (Slot-acc s)))
                (t (out "No accessor"))   )
          (cond ((and (not (atom (Slot-acc s)))
                      (or (Slot-set s) (Slot-atypes s)))
                 (out (:t (+ pc 8))))
                (t (out 1))   )
          (cond ((Slot-atypes s)
;;;;		 (dbg-save s)
;;;;		 (breakpoint slots-print
;;;;		    "Ready to print atypes of " s)
                 (out "Arg Types: "
		      (arglistspec-typed-arglist
		         (follow-var-ref (Slot-atypes s) debug-env*)))))
          (cond ((Slot-set s)
                 (out :% 1 "Set by " (Slot-set s)))   ))   )
     (out :% (:t pc))   ))

(defun symtype (sym
		&optional (env (ensure-debug-env-set)))
   (let ((vt (var-lookup sym env)))
      (cond (vt
	     (Vartype-type vt))
	    (t
	     false))))

;;; For debugging
(def-type-former tq (p _ _) (:system universal)
   (eval (car p)))

(defun dtp (td &optional (env (ensure-debug-env-set)))
   (designated-type td true env))

(defun dtsee (td env)
   (typesee (designated-type td true env)))
