;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: typevar.lisp,v 1.9 2006/03/13 19:20:08 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on %nity/ types foundation %langutils/ :symboid)

;;; Slurp to get macro definitions
(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(scope-time* param tvar-type* tvar undo !=? empty-undo-stack
	     Typebdg-subs Typebdg-supers Make-typebdg Typebdg-val
	     Typebdg-id Typebdg-varname Typebdg-unique-sub
	     tvar-type make-tvar-type is-tvar-type tvar-type-varname
	     tvar-type-bdg
	     make-param-type is-param-type param-type-varname param-name
	     Type-closure is-Type-closure Type-closure-type 
	     Type-closure-env Type-closure-freevars
	     is-Env make-Env env-bindings-append env-bindings-nconc
	     env-cons
	     Env-is-closing Env-vartypes Env
	     new-Type-closure  
	     anon-tvar unknown-type is-unknown-type unknown-type-id
	     type-feature type-local-feature unknown-tvar-type
	     create-tvar-for-bvar unknown-type-for-param
	     )))

(defvar binding-dbg* false)

(def-class Undo
    undo
    redo
    (:handler
       (:print-object (u srm)
	  (ignore u)
	  (out (:to srm) "{?}"))))

(defun empty-undo-stack () !())

;; Can't be used to set a form with no current value.
;; Optional forward-dbg and backward-dbg are things to do (usually, 
;; things to print) going each way.
(defmacro !=? (^form ^val ^undo-list &optional forward-dbg backward-dbg)
   (let ((v-old (gensym))
	 (v-u (gensym)))
      (multi-let (((vars vals newlftvar writer-form reader-form)
		   (get-setf-expansion ^form)))
	 (setq newlftvar (car newlftvar))
	 `(let* (,@(mapcar #'list vars vals)
		 ,@(cond ((occurs-in '*-* ^val)
			  `((*-* ,reader-form)
			    (,v-old *-*)))
			 (t `((,v-old ,reader-form))))
		 (,newlftvar ,^val))
	     (let ((,v-u (make-Undo (\\ ()
				       ,(subst v-old newlftvar writer-form)
				       ,@(include-if backward-dbg))
				    (\\ ()
				       ,writer-form
				       ,@(include-if forward-dbg)))))
		(funcall (Undo-redo ,v-u))
		(push ,v-u ,^undo-list))))))

;;; Conventions for functions that take an undo-stack as argument:
;;; The function returns some value indicating whether the function
;;; succeeded or fails.
;;; If the function succeeds, it returns a new undo-stack of which the
;;; input one is a tail.
;;; If the function fails, it undoes all the undoable side effects that
;;; occurred on its watch, and returns the undo-stack it was given.

(defun undo (current-list back-to-list)
;   (cond ((not (eq current-list back-to-list))
	  ;(out (:to *error-output*) "Undoing " (len (ldiff current-list back-to-list))
	  ;     " actions" t)
	  ;(error-break undo :novalue
	  ;   "Undo")))
   (repeat :for ((ul current-list (cdr ul)))
    :until (or (null ul) (eq ul back-to-list))
    :result (cond ((not (eq ul back-to-list))
		  (error-break undo
		     "Unexpectedly ran off end of undo stack"
		     (:novalue "I will proceed"))))
      (funcall (Undo-undo (car ul))))
   back-to-list)

(defun redo (past-list future-list)
;;;;   (cond ((not (eq future-list past-list))
;;;;	  (out (:to *error-output*) "Redoing " (len (ldiff future-list past-list))
;;;;	       " actions" t)))
  (letrec ((redo-backwards (fut)
	      (cond ((not (eq fut past-list))
		     (cond ((null fut)
			    (error-break undo
			       "Unexpectedly ran off end of undo stack"
			       (:novalue "I will proceed")))
			   (t
			    (redo-backwards (cdr fut))
			    (funcall (Undo-redo (car fut)))))))))
     (redo-backwards future-list))
  future-list)

;; Common idioms
(defun maybe-undo (boolean current-list back-to-list)
   (cond (boolean
	  current-list)
	 (t
	  (undo current-list back-to-list))))

(defun okay-or-undo (successful new-undo old-undo)
   (values successful
	   (maybe-undo successful new-undo old-undo)))

(defun fail-and-undo (new-undo old-undo)
   (values false (undo new-undo old-undo)))

(def-type-class param-type-class (:system :universal))

;;; Variables bound to types at expansion time.
(defun make-param-type (sym &optional (kind type-type*))
   (let ((ty (make-type-of-class param-type-class
		       `(param ,sym)
		       '()
		       univ-type*
		       '()
		       (list (tuple 'param sym)
			     (tuple 'kind kind)))))
     (!= (Type-freevars ty)
	 (list ty))
     ty))

(defun is-param-type (ty)
  (and (is-Type ty)
       (car-eq (Type-desig ty) 'param)))

(defun param-type-varname (ty)
  (cadr (Type-desig ty)))

(defun same-param (p1 p2)
   (eq (param-type-varname p1)
       (param-type-varname p2)))

(defun param-name (x)
   (cond ((is-Symbol x) x)
	 ((and (is-Type x)
	       (is-param-type x))
	  (param-type-varname x))
	 (t
	  (error-break param-name :fatal
	     "Can't figure out parameter " x))))	 

(defun param-pprint (srm p)
   (cond ((and (null (cddr p)) (is-Symbol (cadr p)))
	  (out (:to srm) "!/" (cadr p)))
	 (t
	  (out (:to srm) "!/<?>" (cdr p)))))

(defun is-param-exp (x) (car-eq x 'param))

(set-pprint-dispatch '(satisfies is-param-exp)
		     #'param-pprint
		     1)

(yt::def-excl-dispatch #\/ (srm _)
   `(param ,(read srm)))

;;;;(set-dispatch-macro-character
;;;;    #\! #\/
;;;;    (\\ (srm _ _)
;;;;       `(param ,(read srm)))
;;;;    ytools::ytools-readtable*)

;; A constraint on a super, sub, or val of a tvar.
(def-class Tvar-constraint
   status  ;; (Const satisfied violated unknown)
   pred)  ;; (Fun (Val (Const satisfied violated unknown) (Lst Undo))
          ;;      (Arg (Const super sub val) (Lst Type-closure) (Lst Undo)))


(defun new-Tvar-constraint (pred)
  (make-Tvar-constraint 'unknown pred))
     
;;; We have to "bind" tvars so we know what their scope is.
(defvar tvar-type* (new-Type boring-class* 'Tvar () univ-type* '()
			     (list (tuple 'home universal-type-sys*))))

(def-type-class tvar-type-class (:system :universal) (:var tvar-type-class*))

;;; Unknown types and tvar types are complementary.  Unks play the role
;;; of universally quantified variables during type checking.  They're
;;; an unknown type that will be filled in later, which has no
;;; properties, and hence can only be incorporated into larger types,
;;; never analyzed.

;;; tvars play the role of existentially quantified variables.  After
;;; they're created, they become more constrained, until we find out which
;;; type they actually denote.

(defvar scope-time* 0)
;;; Every unk and tvar has a scope-time property that tells when it was
;;; bound.  It's illegal to bind a tvar with scope-time t1 to an unk with
;;; scope time > t1, because such an unk corresponds to a variable that 
;;; wasn't bound when the tvar was created.
;;; Binders are responsible for incrementing this counter.  It doesn't 
;;; happen here.

(def-type-class unk-class (:system :universal)
  (:var unk-class*))

(defvar unkno* 0)

(defun unknown-type (&optional (kind type-type*) (sup-type univ-type*))
   (!= unkno* (+ *-* 1))
;;;;   (breakpoint unknown-type
;;;;      "Making unknown-type: " unkno*)
   (make-type-of-class unk-class
      `(Unk ,unkno*)
      '() sup-type '()
      (list (tuple 'kind kind))))

(defun is-unknown-type (x)
      (and (is-Type x)
	   (eq (Type-class x) unk-class*)))

(defun unknown-type-id (unk)
   (cadr (Type-desig unk)))

(defun unk-pprint (srm p)
   (match-cond p
      ?( (Unk ?id (param ?v ?d))
	(out (:to srm) "!." (:q (d "=")) v "." id))
      ?( (Unk ?id)
	(out (:to srm) "!." id))
      (t
       (out (:to srm) "!." (cdr p)))))

(defun is-unk-exp (x) (car-eq x 'Unk))

(set-pprint-dispatch '(satisfies is-unk-exp)
		     #'unk-pprint
		     1)


;;; Two special roles for unks: 
;;; 1. At designator-resolution time, they 
;;; are the values of new variables, and record which param each new
;;; variable should be translated into.  These have a 'param' feature
;;; and a true 'desig' feature.
;;; 2. At type-check time, they serve as placeholders for level-1
;;; parameters when checking the body of a level-1 function.
;;; These have a 'param' feature but a false 'desig' feature.

(defun unknown-type-for-param (var kind sup-type desig)
   (!= unkno* (+ *-* 1))
;;   (!= var-timestamp* (+ *-* 1))
;;;;   (breakpoint unknown-type
;;;;      "Making unknown-type for param: " var 1 unkno*)
   (multi-let (((var param)
		;; Hopefully we don't have to worry about var being
		;; a CSymbol or Symboid, because it's bound --
		(cond ((is-Symbol var)
		       (values var (make-param-type var)))
		      (t
		       (values (param-type-varname var)
			       var)))))
;;;;      (breakpoint unknown-type-for-param
;;;;	 "About to create unknown " unkno* " for " var)
      (make-type-of-class unk-class
	 `(Unk ,unkno* (param ,var ,desig))
	 '() ;;;(list param)  -- why did this seem right?
	 sup-type '()
	 (list (tuple 'param param)
	       (tuple 'scope-time scope-time*)
	       (tuple 'desig desig)
	       (tuple 'kind kind)))))

(defun is-desig-unk (ty)
   (and (eq (Type-class ty) unk-class*)
	(type-local-feature ty 'desig)))

;;; 'unk-desig' is the designator of an unk-type.  If it's cddr is 
;;; non-nil, then the caddr is the designator of a desig-unk, and it is
;;; returned.  Otherwise, false is returned.
(defun unk-desig-param (unk-desig)
   (let ((c (cddr unk-desig)))
      (and (not (null c))
	   (car c))))

(defun make-tvar-type (sym val-only kind)
  (let ((unknown-type (make-type-of-class tvar-type-class
				 (make-Qvar sym '())
				 nil univ-type* '()
				 (list (tuple 'scope-time scope-time*)
				       (tuple 'val-only val-only)
				       (tuple 'kind kind)))))
    (!= (symboid-get sym 'tvar)
	unknown-type)
    unknown-type))

(defun is-tvar-type (ty)
   (and (is-Type ty)
	(is-Qvar (Type-desig ty))))

(defun tvar-type-varname (ty)
  (Qvar-sym (Type-desig ty)))

(defvar any-tvar* (make-tvar-type '_ false type-type*))

;;; Binding of a tvar.
(def-class Typebdg
    (:handler
       (print (tb srm)
	  (out (:to srm)
	       "#{Typebdg " (Typebdg-varname tb) 
	                  (:q ((Typebdg-val tb)
			       " = " (Typebdg-val tb)))
			  (:q ((Typebdg-unique-sub tb)
			       " !"))
			  (:q ((not (null (Typebdg-subs tb)))
			       " > " (Typebdg-subs tb)))
			  (:q ((not (null (Typebdg-supers tb)))
			       " < " (Typebdg-supers tb)))
			  (:q ((Typebdg-val-only tb)
			       "[=]")
			      ((not (null (Typebdg-constraints tb)))
			       "[" (len (Typebdg-constraints tb)) "]"))
			  "}")))
    varname 
    val-only     ;; if true, don't accumulate subs, supers, or constraints
                 ;; just bind to val at first opportunity
    val                ; Type-closure
    
    subs supers  ; lists of Type-closures
    ;; If val is specified, subs and supers are irrelevant
    unique-sub
    constraints)  ;; (Lst Tvar-constraint)
;;; If t is a subtype of
;;; u1 and u2, then either u1 < u2 or u2 < u1.
;;; (I use "<" on types to mean "subtype," so that u < u.)
;;; The same restriction applies to subs if unique-sub = true.
    

;;; No type is well specified unless accompanied by an environment.
;;; The env disambiguates all the params.  Note that tvar-types have indefinite
;;; global scope, which is orthogonal to the lexical scope of
;;; params.

(defun typebdg-pprint (srm tb)
   (out (:to srm)
       (:pp-block
	  (:pre "#{Typebdg ")
	  (Typebdg-varname tb)
	  (:q ((Typebdg-val tb)
	       (:pp-nl :linear srm)
	       " = " (Typebdg-val tb)))
	  (:q ((null (Typebdg-subs tb))
	       (:q ((Typebdg-unique-sub tb)
		    " (>!)")))
	      (t
	       (:pp-nl :linear srm)
	       (:q ((Typebdg-unique-sub tb)
		    " !"))
	       " > " (Typebdg-subs tb)))
	  (:q ((not (null (Typebdg-supers tb)))
	       (:pp-nl :linear srm)
	       " < " (Typebdg-supers tb)))
	  (:q ((Typebdg-val-only tb)
	       "[=]")
	      ((not (null (Typebdg-constraints tb)))
	       (:pp-nl :linear srm)
	       1 (len (Typebdg-constraints tb)) " constraint(s)"))
	  (:suf "}"))))

(set-pprint-dispatch 'Typebdg #'typebdg-pprint 1)

;; We assume that the constraint is untestable at this point; it's the caller's
;; responsibility to verify that.
(defun typebdg-add-constraint (bdg con undo-stack)
   (!=? (Typebdg-constraints bdg)
	(cons con *-*)
	undo-stack)
   undo-stack)

(defun typebdg-constrained (bdg)
   (or (Typebdg-val bdg)
       (not (null (Typebdg-subs bdg)))
       (not (null (Typebdg-supers bdg)))))

;;; If the bdg has a value or gets one during the checking process,
;;; this removes all the constraints from it.
(defun typebdg-constraints-check (bdg which new-stuff
				  undo0 orig-undo-stack)
   (let ((conl (Typebdg-constraints bdg)))
      (let ((undo1 undo0))
	 (!=? (Typebdg-constraints bdg) '() undo1)
	 (repeat :for ((con :in conl)
		       (okay true))
	  :result (values true undo1)
	    (!= < okay undo1 >
		(tvar-constraint-check con which new-stuff undo1))
	  :while okay
	    (cond ((and (eq (Tvar-constraint-status con) 'unknown)
			;; Once the val's assigned the constraints
			;; should disappear
			(not (Typebdg-val bdg)))
		   (!=? (Typebdg-constraints bdg)
			(cons con (Typebdg-constraints bdg))
			undo1)))
	  :else :result (progn (dbg-out binding-dbg*
				  "Constraint failure" t)
			       (values false
				       (undo undo1 orig-undo-stack)))))))

(defun tvar-constraint-check (tvc which tc undo-stack)
   (cond ((eq (Tvar-constraint-status tvc)
	      'unknown)
; 	  (dbg-save tvc which tc)
; 	  (error-break tvar-constraint-check :novalue
; 	     "Checking " tvc)
	  (multi-let (((stat new-undo)
		       (funcall (Tvar-constraint-pred tvc)
			       which tc undo-stack)))
	     (cond ((eq stat 'violated)
		    ;; We don't set the status because we would
		    ;; undo it immediately.
		    (values false (undo new-undo undo-stack)))
		   (t
		    (!=? (Tvar-constraint-status tvc) stat
			 new-undo)
		    (values true new-undo)))))
	 (t
	  (values (eq (Tvar-constraint-status tvc) 'satisfied)
		  undo-stack))))

(defun symboid-tvar (s)
   (symboid-get s 'tvar))

;; For debugging
(defun tvb (s) 
   (let ((tv (symboid-get s 'tvar)))
      (cond (tv
	     (values (tvar-type-bdg tv) tv))
	    (t '\?))))

(defvar tvar-may-be-val-only* true)

(defun tvar-type-bdg (ty)
   (let ((b (assq 'bdg (Type-features ty))))
      (cond ((not b)
	     (!= b (tuple 'bdg (make-Typebdg (tvar-type-varname ty)
					     (and tvar-may-be-val-only*
						  (type-feature ty 'val-only))
					     false '() '() false '())))
	     (!= (Type-features ty)
		 (cons b *-*))))
      (cadr b)))

(defun tvar-value (ty)
   (Typebdg-val (tvar-type-bdg ty)))

(defun tvar-add-constraint (tvar con undo-stack)
   (typebdg-add-constraint (tvar-type-bdg tvar) con undo-stack))

;(defun (setf tvar-value) (x ty)
;   (setf (alist-entry 'value (Type-features ty))
;         x))

(defun is-tvar-with-name (ty name)
   (and (is-tvar-type ty)
	(eq (tvar-type-varname ty)
	    name)))

(defun unknown-tvar-type ()
   (anon-tvar "" false type-type*))

(defvar tvar-no* 0)

(defun anon-tvar (name val-only kind)
   (make-tvar-type (build-symboid (< name) "-" (++ tvar-no*))
		   val-only kind))

;;; This program eliminates duplicates from vl1, but leaves any in vl2
;;; undisturbed.
(defun tvars-union (vl1 vl2)
   (repeat :for ((v :in vl1))
      (cond ((not (member v vl2 :test #'same-tvar))
	     (!= vl2 (cons v *-*)))))
   vl2)

(defun tvars-complement (vl1 vl2)
   (<? (\\ (v1) (not (member v1 vl2 :test #'same-tvar)))
       vl1))

(defun same-tvar (tv1 tv2)
   (eq (tvar-type-varname tv1)
       (tvar-type-varname tv2)))

(defun create-tvar-for-bvar (name)
   (let ((new-tvar (anon-tvar (string-concat "v/" (symbol-name name))
			      false type-type*)))
      (dbg-out binding-dbg*
	 "Creating new tvar ?" (tvar-type-varname new-tvar)
	 " for type of var " name t)
      new-tvar))

(defun type-local-feature (ty feat)
  (let ((p (assq feat (Type-features ty))))
     (cond (p (values (cadr p) true))
	   (t (values false false)))))

;;; In a type closure (not defined until vartypes.lisp), 
;;; the type may include parts with params bound in
;;; the env.  Those parts *must* be of type Closable, so that 
;;; type-find-feature can wrap them in the env when they're 
;;; extracted.
;;; trans is a Loadable-function to do the wrapping.
;;; false means just use new-Type-closure.
(def-class Closable
      (:handler
       (print (tc srm)
	  (let ((ty (Closable-thing tc)))
	     (out (:to srm)
		  "[/"              
		  (:q ((is-Type ty) (Type-desig ty))
		      (t ty))
		  "/"
		  (:q ((Closable-trans tc)
		       (Closable-trans tc))
		      (t "_"))
		  (:e (freevars-print (Closable-freevars tc)
				     srm))
		  "/]"
		      ))))
   thing freevars trans)

;;; Moved from feature.lisp so feature.lisp won't be needed at run time.
;;; Get type feature without worrying about having to compute it
;;; Second value says whether it actually found anything
(defun type-feature (ty feat &optional (unwrap-closables false))
   (let ((p (assq feat (Type-features ty))))
      (cond (p
	     (values
		(cond ((is-Closable (cadr p))
		       (cond ((not unwrap-closables)
			      (error-break type-feature
				 "Found closable with no env to close with: "
				 p t " on type " ty
				 (:novalue
				  "I will unwrap it and proceed with no env"))))
		       (Closable-thing (cadr p)))
		      (t
		       (cadr p)))
		true))
	    ((Type-class ty)
	     (!= p (assq feat (Type-class-featurelist (Type-class ty))))
	     (cond (p
		    (values (cadr p) true))
		   (t
		    (values false false))))
	    (t
	     (values false false)))))

(defun (setf type-feature) (newval ty feat)
   (let ((p (assq feat (Type-features ty))))
      (cond ((not p)
	     (!= p (tuple feat nil))
	     (!= (Type-features ty) (cons p *-*))))
      (!= (cadr p) newval)))

;;;;(defun type-feature-set (ty feat val)
;;;;   (let ((fp (assq feat (Type-features ty))))
;;;;      (cond ((not fp)
;;;;             (!= fp (list feat nil) )
;;;;             (!= (Type-features ty) (cons fp *-*)))   )
;;;;      (!= (cadr fp) val)   ))
;;;;
;;;;(defsetf type-feature type-feature-set)

(defun freevars-print (fvl srm)
   (cond ((not (null fvl))
	  (out (:to srm)
	       "^" (:e (repeat 
			 (:o (param-type-varname (car fvl)))
		       :until (null (cdr fvl))
		         (setq fvl (cdr fvl))
		         (:o 1)))))))

(!= type-freevar-printer* #'freevars-print)
