;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: typewalk.lisp,v 1.2 2004/02/22 00:47:07 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on :at-run-time %nity/ subtype altformer funformers dotformers)

;; Utilities for walking through types, building new versions or
;; collecting some results.

;; Returns < new value, whether-changed >
(defun type-subst (otype alist vartypes)
      (multiple-value-let (type vartypes)
			  (follow-var-ref otype vartypes)
	 (multiple-value-let
	            (substval bingo changed)
		    (cond ((is-tvar-type type)
			   (let ((p (assq (tvar-type-varname type)
					  alist)))
			      (cond (p
				     (values (cadr p) true true))
				    (t
				     (values type false false)))))
			  (t
			   (let ((h (type-feature type 'subster)))
			      (cond (h
				     (multiple-value-let
				               (n ch)
					       (funcall h type alist vartypes)
					(values n false ch)))
				    (t
				     (values type false false))))))
	    (cond (bingo
		   (values substval true))
		  (changed
		   (values (type-clos/able substval vartypes)
			   true))
		  (t
		   (values otype false))))))

(class-feature-datafun subster typeform-class
   (defun (ty alist vartypes)
      (multiple-value-let (argsubsts changed)
                          (typelist-subst
			     (type-find-feature ty 'argtypes vartypes)
			     alist vartypes)
	 (cond (changed
		(values (make-typeform
			   (type-feature ty 'fun-name)
			   argsubsts)
			true))
	       (t
		(values ty false))))))

(class-feature-datafun subster alt-class
   (defun (ty alist vartypes)
      (multiple-value-let (altsubsts changed)
                          (typelist-subst
			     (type-find-feature
				ty 'alt-types vartypes)
			     alist vartypes)
	 (cond (changed
		(values (make-alt-type altsubsts vartypes) true))
	       (t
		(values ty false))))))

(class-feature-datafun subster fun-class
   (defun (ty alist vartypes)
      (multi-let (((asubst achanged)
		   (type-subst (type-find-feature ty 'argtype vartypes)
			       alist vartypes))
		  (alspec
		   (funtype-extract-arglist ty vartypes)))
	 (multi-let (((alsubst alchanged)
		      (cond (alspec
			     (arglistspec-subst alspec alist vartypes))
			    (t
			     (values false false)))))
	    (let ((renv (high-argspecs-declare-placeholder-params
			   ty alsubst vartypes)))
	       (multi-let (((rsubst rchanged)
			    (type-subst
			       (type-find-feature
				  ty 'resulttype renv)
			       alist renv))
			   ((rlsubst rlchanged)
			    (arglistspec-subst
			       (type-find-feature
				  ty 'reslist renv)
			       alist renv)))
		  (cond ((or achanged rchanged
			     alchanged rlchanged)
			 (values (make-funtype
				    (type-feature ty 'level)
				    rsubst asubst rlsubst alsubst
				    (type-feature ty 'side-effects)
				    vartypes)
				 true))
			(t
			 (values ty false)))))))))

(defun arglistspec-subst (alspec alist vartypes)
   (multiple-value-let (alspec alenv)
		       (follow-var-ref alspec vartypes)
      (letrec ((argspec-subst (a)
		  (multiple-value-let (sat changed)
				      (type-subst (Argspec-type a) alist alenv)
		     (cond (changed
			    (values (make-Argspec
				       (Argspec-name a)
				       (Argspec-mode a)
				       (Argspec-position a)
				       sat
				       (Argspec-default a)
				       (Argspec-props a))
				    true))
			   (t
			    (values a false))))))
	 (repeat :for ((as :in (Arglistspec-argspecs alspec))
		      (res '())
		      (somechanged false)
		      sa ch)
	    (!= < sa ch > (argspec-subst as))
	    (cond (ch (!= somechanged true)))
	    (!= res (cons sa *-*))
	  :result (cond (somechanged
			(values (new-Arglistspec
				   (dreverse res)
				   :allow-other-keys
				   (Arglistspec-allow-other-keys alspec)
				   :names-matter
				   (Arglistspec-names-matter alspec))
				true))
		       (t
			(values alspec false)))))))

(class-feature-datafun subster dot-class
   (defun (ty alist vartypes)
      (dot-lin-subster 'Tup ty alist vartypes)))

(class-feature-datafun subster lin-class
   (defun (ty alist vartypes)
       (dot-lin-subster 'Row ty alist vartypes)))

(defun dot-lin-subster (med ty alist vartypes)
   (multi-let (((asubst ach)
		(type-subst (type-find-feature ty 'cartype vartypes)
			    alist vartypes))
	       ((dsubst dch)
		(type-subst (type-find-feature ty 'cdrtype vartypes)
			    alist vartypes)))
      (cond ((or ach dch)
	     (values (make-dot-type
		        med asubst dsubst vartypes)
		     true))
	    (t
	     (values ty false)))))

(defun tup-val-subster (ty alist vartypes)
   (multi-let (((argl-subst argl-ch)
		(arglistspec-subst
		   (type-find-feature ty 'elt-types vartypes)
		   alist
		   vartypes)))
      (cond (argl-ch
	     (values (make-tup-type
			(type-feature ty 'medium)
			argl-subst
			false
			vartypes)
		     true))
	    (t
	     (values ty false)))))

(declare-class-feature subster tup-class #'tup-val-subster)
(declare-class-feature subster val-class #'tup-val-subster)

(class-feature-datafun subster modified-class
   (defun (ty alist vartypes)
      (let ((suptype (type-subst (type-find-supertype ty vartypes)
				 alist vartypes)))
	 (multi-let (((suptype supenv)
		      (follow-var-ref suptype vartypes)))
	    (new-Type modified-class*
	       (Type-desig ty)
	       (Type-freevars suptype)
	       (cond (supenv '*as-feature)
		     (t suptype))
	       (Type-slotfns ty)
	       (cond (supenv
		      (cons (tuple 'supertype
				   (type-clos/able
				      suptype supenv))
			    (Type-features ty)))
		     (t (Type-features ty))))))))

;(declare-class-feature subster tup-class #'tup-type-subst)
;(declare-class-feature subster arg-class #'tup-type-subst)

; (class-feature-datafun subster segment-class
;    (defun (ty alist vartypes)
;       (let ((tty (type-find-feature ty 'tup-type vartypes)))
; 	 (multiple-value-let (newtty changed)
; 			     (type-subst tty alist vartypes)
; 	    (cond (changed
; 		   (values (make-segment-type newtty)
; 			   true))
; 		  (t
; 		   (values tty false)))))))

(class-feature-datafun subster image-class
   (defun (ty alist vartypes)
      (let ((itervars (type-find-feature ty 'template-params vartypes))
	    (tmtype (type-find-feature ty 'template-type vartypes))
	    (args (type-find-feature ty 'template-args vartypes)))
	 (multiple-value-let (newargs changed)
			     (typelist-subst args alist vartypes)
	    (cond (changed
		   (values (make-image-type itervars tmtype newargs
					    false vartypes)
			   true))
		  (t
		   (values ty false)))))))

(defun typelist-subst (tyl alist env)
;;;;   (multiple-value-let (tyl env)
;;;;		       (follow-var-ref tyl vartypes) ...)
      (repeat :for ((ty :in tyl) (res '()) (somechanged false)
		   sty ch)
	 (!= < sty ch >
	     (type-subst ty alist env))
	 (cond (ch (!= somechanged true)))
	 (!= res (cons sty *-*))
       :result (cond (somechanged
		     (values (dreverse res)
			     true))
		    (t
		     (values tyl false)))))

;; Collect pieces of type.
;; Stop if you get stopnum.
;; collector - (Fun (res - (Lst Obj) keep-going - Boolean) <- (Type Env))
;; res is list of things found here.
;; keep-going is true if and only if the pieces of the type should 
;; be explored.
(defun type-collect (type collector stopnum vartypes)
   (type-accum-collect type '() collector
		       (cond ((>= stopnum 0)
			      (\\ (sofar) (>= (len sofar) stopnum)))
			     (t (\\ (_) false)))
		       vartypes))

(defun type-accum-collect (type sofar collector stopwhen vartypes)
   (cond ((funcall stopwhen sofar)
	  sofar)
	 (t
	  (multi-let (((type env)
		       (follow-var-ref type vartypes)))
	     (multi-let (((new keep-going)
			  (funcall collector type env)))
		(!= sofar (append new *-*))
		(cond (keep-going
		       (let ((h (type-find-feature type 'collector env)))
			  (cond (h
				 (funcall h type sofar
					  collector stopwhen vartypes))
				(t sofar))))
		      (t sofar)))))))

(class-feature-datafun collector typeform-class
   (defun (ty sofar collector stopwhen vartypes)
      (typelist-accum-collect (type-find-feature ty 'argtypes vartypes)
                        sofar collector stopwhen vartypes)))

(class-feature-datafun collector alt-class
   (defun (ty sofar collector stopwhen vartypes)
      (typelist-accum-collect (type-find-feature
				ty 'alt-types vartypes)
                        sofar collector stopwhen vartypes)))

(class-feature-datafun collector fun-class
   (defun (ty sofar collector stopwhen vartypes)
      (let ((argtype (type-find-feature ty 'argtype vartypes)))
	 (let ((ares (type-accum-collect
		        argtype sofar collector stopwhen vartypes)))
	    (cond ((funcall stopwhen ares)
		   ares)
		  (t
		   (let ((renv (funtype-declare-placeholder-params
				  ty vartypes)))
		      (type-accum-collect
			 (type-find-feature ty 'resulttype renv)
			 ares collector stopwhen
			 vartypes))))))))

(defun dot-lin-collector (ty sofar collector stopwhen vartypes)
   (let ((ares (type-accum-collect
		  (type-find-feature ty 'cartype vartypes)
		  sofar collector stopwhen vartypes)))
      (cond ((funcall stopwhen ares)
	     ares)
	    (t
	     (type-accum-collect
	        (type-find-feature ty 'cdrtype vartypes)
		ares collector stopwhen vartypes)))))

(declare-class-feature collector dot-class #'dot-lin-collector)
(declare-class-feature collector lin-class #'dot-lin-collector)


(defun tup-val-collector (ty sofar collector stopwhen vartypes)
   (type-accum-collect
      (type-find-supertype ty vartypes)
      sofar collector stopwhen vartypes))

(declare-class-feature collector tup-class #'tup-val-collector)
(declare-class-feature collector val-class #'tup-val-collector)

(class-feature-datafun collector modified-class
   (defun (ty sofar collector stopwhen vartypes)
      (type-accum-collect
         (type-find-supertype ty vartypes)
         sofar collector stopwhen vartypes)))

(class-feature-datafun collector image-class
   (defun (ty sofar collector stopwhen vartypes)
      (!= sofar (typelist-accum-collect
		   (type-find-feature ty 'template-args vartypes)
		   sofar collector stopwhen vartypes))
      (cond ((funcall stopwhen sofar)
	     sofar)
	    (t
	     (let ((local-env
		      (<# (\\ (p)
			     (new-Vartype (param-type-varname p)
					  type-type*
					  '*unbound))
			  (type-find-feature ty 'template-params vartypes))))
		(type-accum-collect
		   (type-find-feature ty 'template-type vartypes)
		   sofar
		   (\\ (ty env)
		      (cond ((and (is-param-type ty)
				  (assq (param-type-varname ty)
					local-env))
			     (values '() false))
			    (t
			     (funcall collector ty env))))
		   stopwhen
		   (append local-env vartypes)))))))

(defun typelist-accum-collect (tyl sofar collector stopwhen vartypes)
   (multiple-value-let (tyl env)
		       (follow-var-ref tyl vartypes)
      (repeat :for ((ty :in tyl))
		   
	 (!= sofar (type-accum-collect ty sofar collector stopwhen env))
       :until (funcall stopwhen sofar)
       :result sofar)))
