;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: unkelim.lisp,v 1.5 2004/03/13 04:20:47 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

;; Not a regular part of the types module, but required by Opt.

(depends-on :at-run-time %nity/ subtype altformer funformers)

;;; tab is (Lst (Tup Integer param)).  Replace every occurrence of 
;;; (Unk id ...) with the corresponding param.
(defun source-unk-elim (e tab)
   (cond ((atom e) e)
	 ((eq (car e) 'Unk)
	  (let ((p (assoc (cadr e) tab)))
	     (cond (p (cadr p))
		   (t e))))
	 (t
	  (<# (\\ (a)
		 (source-unk-elim a tab))
	      e))))

;;; Replace all unknown types in the list elim with (Closables for)
;;; their parameters.
;;; elim is a list of Integer id flags.  (If it's '()', we'll copy a 
;;; lot of types for nothing; this should be caught before unk-elim is
;;; called.)
;;; Returns < new value, created-free >,
;;; where created-free is true if the resulting value has a free occurrence
;;; of one of the new Closables.
(defun unk-elim (otype elim bind-time vartypes)
   (multi-let (((type env)
		(follow-var-ref otype vartypes)))
      (multi-let (((h _) (type-feature type 'unk-flusher)))
	 (cond (h
		(funcall h type elim bind-time env))
	       ((is-tvar-type type)
		(tvar-unk-elim type elim bind-time vartypes))
	       ((is-param-type type)
		(values type false))
	       ((is-CSymbol (Type-desig type))
		(values type false))
	       (t
		(error-break unk-elim
			  "No unk-flusher for " type
			  (:novalue "I will assume nothing changed")))))))

(defun tvar-unk-elim (tvar elim bind-time vartypes)
                     (ignore elim)
   (cond ((=< (type-find-feature tvar 'scope-time vartypes)
	      bind-time)
	  (values tvar false))
	 (t
	  (signal-problem tvar-unk-elim
	     "Typevar " tvar " has scope-time "
	     (type-find-feature tvar 'scope-time vartypes)
	     " > " bind-time))))

(class-feature-datafun unk-flusher unk-class
   (defun (ty elim bind-time vartypes)
          (ignore bind-time vartypes)
      (cond ((member (unknown-type-id ty) elim :test #'=)
	     (let ((p (type-feature ty 'param)))
	        (cond (p
		       (values p ;;;;(make-type-closable p false)
			       true))
		      (t
		       (error-break unk-class-unk-flusher
			  "Attempt to eliminate unknown type with no param"
			  (:novalue "I'll leave it in"))
		       (values ty false)))))
            (t
	     (values ty false)))))

(class-feature-datafun unk-flusher typeform-class
   (defun (ty elim bind-time vartypes)
      (let ((arg-vts (type-find-feature ty 'args vartypes)))
	 (repeat :for ((arg-vt :in arg-vts)
		      (elim-vtl '())
		      (some-cf false))
	    (cond ((is-Vartype arg-vt)
		   (let ((vt-val (Vartype-val arg-vt)))
		      (cond ((is-Typoid vt-val)
			     (multi-let (((val-elim val-cf)
					  (unk-elim (Vartype-val arg-vt)
						    elim bind-time vartypes)))
				(cond (val-cf
				       (!= some-cf true)))
				(cond ((eq val-elim vt-val)
				       (!= elim-vtl
					   (cons arg-vt *-*)))
				      (t
				       (!= elim-vtl
					   (cons (new-Vartype
						    (Vartype-var arg-vt)
						    (Vartype-type arg-vt)
						    val-elim)
						 *-*))))))
			    (t
			     (!= elim-vtl (cons arg-vt *-*))))))
		  (t
		   (!= elim-vtl (cons arg-vt *-*))))
	  :result
	    (values (make-typeform
		       (type-find-feature ty 'fun vartypes)
		       (dreverse elim-vtl))
		    some-cf)))))

(class-feature-datafun unk-flusher alt-class
   (defun (ty elim bind-time vartypes)
      (multiple-value-let (altsubsts cfree)
                           (typelist-unk-elim
			     (type-find-feature
				ty 'alt-types vartypes)
			     elim bind-time vartypes)
	 (values (make-alt-type
		    altsubsts vartypes)
		 cfree))))

(class-feature-datafun unk-flusher fun-class
   (defun (ty elim bind-time vartypes)
      (multi-let (((asubst acf)
		   (unk-elim (type-find-feature ty 'argtype vartypes)
			     elim bind-time vartypes))
		  (alspec
		   (funtype-extract-arglist ty vartypes))
		  (alsubst false)
		  (alcf false)
		  ((lev _)
		   (type-find-feature ty 'level vartypes)))
         (cond (alspec
		   (!= < alsubst alcf >
		       (arglistspec-unk-elim alspec elim bind-time vartypes))))
         (let ((local-bindings
		  (cond ((> lev 0)
			 (argspecs-placeholder-params
			    alsubst vartypes))
			(t '()))))
	    (let ((new-unk-bindings (find-type-bindings local-bindings)))
	       (let ((renv (env-bindings-nconc
			      false
			      local-bindings vartypes))
		     (relim (nconc (<# (\\ (vt)
					  (unknown-type-id
					     (Vartype-val vt)))
				       new-unk-bindings)
				   elim)))
		  (multi-let (((rsubst rcf)
			       (unk-elim
				  (type-find-feature
				     ty 'resulttype renv)
				  relim bind-time renv))
			      (rlspec
			       (funtype-extract-reslist ty renv))
			      (rlsubst false)
			      (rlcf false))
		     (cond (rlspec
			    (!= < rlsubst rlcf >
				(arglistspec-unk-elim rlspec relim
						      bind-time renv))))
		     (values (make-funtype
				       lev
				       rsubst asubst rlsubst alsubst
				       (type-feature ty 'side-effects)
				       vartypes)
			     (or acf alcf rcf rlcf)))))))))

(defun find-type-bindings (vartypes)
   (<? (\\ (vt)
	  (and (is-Vartype vt)
	       (type-find-feature (Vartype-type vt) 'meta empty-env*)))
       vartypes))

(defun arglistspec-unk-elim (alspec elim bind-time vartypes)
   (multiple-value-let (alspec alenv)
		       (follow-var-ref alspec vartypes)
      (letrec ((argspec-subst (a)
		  (multiple-value-let (sat cfree)
				      (unk-elim (Argspec-type a) elim
						bind-time alenv)
		     (values (make-Argspec
				       (Argspec-name a)
				       (Argspec-mode a)
				       (argspec-cleared-pos a)
				       sat
				       (Argspec-default a)
				       (Argspec-props a))
			     cfree))))
	 (repeat :for ((as :in (Arglistspec-argspecs alspec))
		      (res '())
		      (somecfree false)
		      sa cf)
	    (!= < sa cf > (argspec-subst as))
	    (cond (cf (!= somecfree true)))
	    (!= res (cons sa *-*))
	  :result ;;(cond (somecfree ...)
	         (values (new-Arglistspec
			    (dreverse res)
			    :allow-other-keys
			    (Arglistspec-allow-other-keys alspec))
			 somecfree)))))

(class-feature-datafun unk-flusher dot-class
   (defun (dty elim bind-time vartypes)
       (dot-lin-unk-flusher 'Tup dty elim bind-time vartypes)))

(class-feature-datafun unk-flusher lin-class
   (defun (lty elim bind-time vartypes)
       (dot-lin-unk-flusher 'Row lty elim bind-time vartypes)))

(defun dot-lin-unk-flusher (med ty elim bind-time vartypes)
   (multi-let (((aty acf)
		(unk-elim (type-find-feature ty 'cartype vartypes)
			  elim bind-time vartypes))
	       ((dty dcf)
		(unk-elim (type-find-feature ty 'cdrtype vartypes)
			  elim bind-time vartypes)))
      (let ((rty (make-dot-type med aty dty vartypes)))
;;;;	 (dbg-save aty dty ty rty (env6 vartypes))
;;;;	 (breakpoint dot-lin-unk-flusher
;;;;	    "freevars of ty = " (Type-freevars ty)
;;;;	    :% "freevars of aty = " (Closable-freevars aty)
;;;;	    :% "freevars of dty = " (Closable-freevars dty)
;;;;	    :% "freevars of rty = " (Closable-freevars rty))
	 (values rty (or acf dcf)))))
		  
(class-feature-datafun unk-flusher to-row-class
   (defun (ty elim bind-time vartypes)
      (multi-let (((sty scf)
		   (unk-elim (type-find-feature ty 'source-type vartypes)
			     elim bind-time vartypes)))
         (values (make-to-row-type sty vartypes)
		 scf))))

(class-feature-datafun unk-flusher image-class
   (defun (ty elim bind-time vartypes)
      (let ((itervars (type-find-feature ty 'template-params vartypes))
	    (tmtype (type-find-feature ty 'template-type vartypes))
	    (args (type-find-feature ty 'template-args vartypes)))
	 (multiple-value-let (newargs cfree)
			     (typelist-unk-elim args elim bind-time vartypes)
	    (values (make-image-type itervars tmtype newargs true
				     vartypes)
		    cfree)))))

(defun tup-val-unk-flusher (ty elim bind-time vartypes)
   (multi-let (((elim-argl argl-cf)
		(arglistspec-unk-elim
		   (type-must-find-feature ty 'elt-types vartypes)
		   elim bind-time vartypes)))
      (values
	 (make-tup-type
	    (type-feature ty 'medium)
	    elim-argl false vartypes)
	 argl-cf)))

(declare-class-feature unk-flusher tup-class #'tup-val-unk-flusher)
(declare-class-feature unk-flusher val-class #'tup-val-unk-flusher)

(class-feature-datafun unk-flusher constructed-class
   (defun :^ (ty elim bind-time vartypes)
      (multi-let ((args
		   (type-must-find-feature ty 'arg-vartypes vartypes)))
	 (multi-let (((elim-args arg-cfs)
		      (vartypelist-unk-flusher args elim bind-time vartypes)))
	    (values
	       (finish-constructed-type
		  (type-find-feature ty 'constructor vartypes)
		  elim-args vartypes)
	       arg-cfs)))))
			 
(defun vartypelist-unk-flusher (vtl elim bind-time env)
   (repeat :for ((vt :in vtl)
		 (vt-cf false)
		 :collectors vt-elims)
    :within
      (multi-let (((ty cf)
		   (unk-elim (Vartype-val vt)
			     elim bind-time env)))
	 (:continue
	  :collect (:into vt-elims (new-Vartype (Vartype-var vt)
						(Vartype-type vt)
						ty))
	     (cond (cf (!= vt-cf true)))))
    :result
       (values vt-elims vt-cf)))

(class-feature-datafun unk-flusher modified-class
   (defun (ty elim bind-time vartypes)
      (multi-let (((sup supcf)
		   (unk-elim (type-find-supertype ty vartypes)
			     elim bind-time vartypes)))
	 (multi-let (((slots slcf)
		      (slots-unk-elim
		         (Type-slotfns ty) elim bind-time vartypes)))
	    (values
	       (multi-let (((sup supenv)
			    (follow-var-ref sup vartypes))
			   (feats
			    (mod-class-features-copy 
				      (Type-features ty))))
		  (let ((resty
			   (make-type-of-class modified-class
			      (Type-desig ty)
			      (Type-freevars sup)
			      false
			      slots
			      feats)))
		     (type-set-supertype
			resty sup supenv)
		     resty))
	       (or supcf slcf))))))

(defun slots-unk-elim (slots elim bind-time vartypes)
   (repeat :for ((s :in slots)
		(res '())
		(some-cf false))
      (multi-let (((sty tycf)
		   (cond ((Slot-type s)
			  (unk-elim (Slot-type s) elim bind-time vartypes))
			 (t
			  (values (Slot-type s) false)))))
	 (multi-let (((sal alcf)
		      (arglistspec-unk-elim
			 (Slot-atypes s) elim bind-time vartypes)))
	    (!= res
		(cons (make-Slot (Slot-name s)
				 (Slot-acc s)
				 (Slot-set s)
				 sty sal
				 (Slot-invert s)
				 (Slot-descriptor s))
		      *-*))
	    (cond ((or tycf alcf)
		   (!= some-cf true)))))
    :result (values (dreverse res)
		   some-cf)))

(class-feature-datafun unk-flusher simple-typeform-class
   (defun (ty elim bind-time vartypes)
          (ignore elim bind-time vartypes)
      (values ty false)))

(defun typelist-unk-elim (tyl elim bind-time vartypes)
;;;;   (multiple-value-let (tyl env)
;;;;		       (follow-var-ref tyl vartypes)
      (repeat :for ((ty :in tyl)  (somecfree false)
		   sty cf :collector res)
	 (!= < sty cf >
	     (unk-elim ty elim bind-time vartypes))
	 (cond (cf (!= somecfree true)))
       :collect sty
       :result (values res somecfree)))

