;-*- Mode: Common-lisp; Package: nity; Readtable: ytools; -*-
(in-package :nity)
;;; $Id: vartypes.lisp,v 1.11 2005/12/26 20:34:25 dvm Exp $

(depends-on %module/ ytools %ytools/ nilscompat)

(depends-on %nity/ typevar)

(depends-on :at-run-time %langutils/ symboid)

(end-header :continue-slurping)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(Tyfun new-Tyfun Tyfun-env Tyfun-defn is-Tyfun
	     empty-vartypes vartypes-globals var-lookup var-val vartypes-lookup
	     simple-loadable-function follow-var-ref
	     Type-closure is-Type-closure Type-closure-type 
	     Type-closure-env Type-closure-freevars
	     is-Env make-Env env-bindings-append env-bindings-nconc
	     env-cons envs-append
	     Env-is-closing Env-vartypes Env
	     empty-env* no-env* global-env*
	     new-Type-closure *unbound forward-type
	     )))

(def-class Env
    is-closing				; Boolean
    vartypes  ; (Lst Vartype)
   (:handler
       (print (env srm)
	  (let ((vtl (Env-vartypes env)))
	     (out (:to srm)
		  "#{Env" (:q ((Env-is-closing env) '*)
			      (t "^"))
		   (:q ((null vtl) "()") (t vtl))  "}")))))
;;; An Env is closing if it can be used to supply values for
;;; all the free variables (params) of a type.  
;;; Non-closing Envs are used in contexts where we expect to 
;;; produce a type containing free variables; that is, someone else 
;;; is going to provide a closing Env.
;;; Two key places where this scenario obtains: At designator-resolution
;;; time, when the internalized result may well contain free variables;
;;; and when reasoning about the result type of a level-1 function,
;;; where we have to indicate that the args have known kinds, but have not
;;; been supplied yet.
;;; Non-closing Envs can have Vartypes with an unknown-type as a value, 
;;; indicating
;;; that the Env does not supply a value, but that the (sub)type will be
;;; closed in a "real" Env later.
;;; Any other value is assumed to be the *same* (in all relevant respects)
;;; to what it will be in the "real" Env.

(defvar empty-env* (make-Env true '()))

(defvar no-env* (make-Env false '()))

(defun Env-length (env) (len (Env-vartypes env)))

(defun envs-append (e1 e2)
   (make-Env
	 (cond ((Env-is-closing e2)
		(Env-is-closing e1))
	       (t
		(cond ((Env-is-closing e1)
		       (signal-problem envs-append
			  "Can't append a closing Env to a non-closing one"
			  (:proceed
			   "I will treat the result as non-closing"))))
		false))
	 (append (Env-vartypes e1) (Env-vartypes e2))))

(defun env-cons (vt e)
   (make-Env true (cons vt (Env-vartypes e))))

(defun env-bindings-append (is-closing vtl e)
   (cond ((and (null vtl)
	       (boole-eq is-closing (Env-is-closing e)))
	  e)
	 (t
	  (make-Env
	     is-closing
	     (append vtl (Env-vartypes e))))))

(defun env-bindings-nconc (is-closing vtl e)
   (cond ((and (null vtl)
	       (boole-eq is-closing (Env-is-closing e)))
	  e)
	 (t
	  (make-Env
		is-closing
		(nconc vtl (Env-vartypes e))))))

(def-class Type-closure
    (:handler
       (print (tc srm)
	  (type-closure-condense-print tc srm)))
   type
   freevars
   env)

(defun type-closure-condense-print (tc srm)
	  (let ((ty (Type-closure-type tc)))
	     (out (:to srm)
		  "[/"              ;; "#{Typeclo "
		  (:q ((is-Type ty) (Type-desig ty))
		     (t ty))
		  "/" (env-condense (Type-closure-env tc))
		  (:e (freevars-print (Type-closure-freevars tc)
				     srm))
		  "/]"
		      )))

(defun env-condense (env)
   `(,(cond ((Env-is-closing env)
	     ':*)
	    (' :^))
     ,@(env-var-labels env)))


(defun has-params (desig)
   (and (not (atom desig))
	(or (eq (car desig) 'param)
	    (exists (dd :in desig)
	       (has-params dd)))))


;;; Hack
(defun extract-vars (parms-desig)
   (cond ((null parms-desig) '())
	 ((eq (car parms-desig) '-)
	  (extract-vars (cddr parms-desig)))
	 (t
	  (cons (car parms-desig)
		(extract-vars (cdr parms-desig))))))

(defun env-var-labels (env)
   (<# (\\ (x)
	  (cond ((is-Vartype x)
		 (Vartype-var x))
		(t (short-label x))))
      (Env-vartypes env)))

;;;; At one time I thought ...
;;;; type is not necessarily a Type; occasionally we close something else:
(defun Type-closure-thing (tc)
   (signal-problem Type-closure-thing
      "Occurrence of Type-closure-thing")
   (Type-closure-type tc))

;;(defun is-typish (x) (or (is-Type x) (is-Type-closure x)))

(def-class Loadable-function (:options (:medium :object))
   name
   fcn  ;; Can be false if reconstructable from name.
  (:handler
      (print (lf srm)
	 (out (:to srm) "#'<"
	      (Loadable-function-name lf)
	      (:q ((and (Loadable-function-fcn lf)
		       (not (eq (Loadable-function-fcn lf)
				(symbol->fun (Loadable-function-name lf)))))
		  "?"))
	      ">"))))

(defmacro simple-loadable-function (fcn-name)
   `(make-Loadable-function ',fcn-name false))

(defun loadable-function-place-fcn (lf)
   (or (Loadable-function-fcn lf)
       (let ((fcn (loadable-function-build-fcn lf)))
	  (cond (fcn
		 (!= (Loadable-function-fcn lf)
		     fcn)
		 fcn)
		(t
		 (signal-problem loadable-function-call
		    "Trying to call unknown function "
		    (Loadable-function-name lf)))))))

(def-op loadable-function-build-fcn (lf)
   (symbol->fun (Loadable-function-name lf)))

(defun loadable-function-call (lf &rest args)
   (apply (loadable-function-place-fcn lf)
	  args))

(defun coerce-to-fun (x)
   (cond ((and x (is-Symbol x))
	  (symbol->fun x))
	 ((is-Loadable-function x)
	  (loadable-function-place-fcn x))
	 (t x)))

(def-class Loadable-function-with-extras
    (:options :key (:include Loadable-function))
    towit
    (:handler
        (loadable-function-build-fcn (lf)
	   (apply (symbol->fun (Loadable-function-name lf))
		  (Loadable-function-with-extras-towit lf)))))

(defun new-Closable (x fvl trans)
   (cond ((exists (fv :in fvl)
	     (not (is-param-type fv)))
	  (signal-problem new-Closable
	     "Bogus freevars: " fvl)
	  :novalue)
	 ((not x)
	  (signal-problem new-Closable
	     "Closing nil"))
	 ((and (is-Typoid x)
	       (exists (xv :in (Typoid-freevars x))
		  (not (exists (fv :in fvl)
			  (same-param xv fv)))))
	  (signal-problem new-Closable
	     "Closable freevars " fvl
	     t " don't capture all freevars of " x)))
   (make-Closable x fvl trans))

;;; This function is applied only to c + a closing Env.
(defun closable-closer (c)
   (let ((closer (coerce-to-fun (Closable-trans c))))
      (or closer
	  #'new-Type-closure)))

;;;;   (let ((tr (Closable-trans c)))
;;;;      (cond ((not tr) #'new-Type-closure) ; formerly make-Type-closure
;;;;	    ((is-Symbol tr) (symbol->fun tr))
;;;;	    (t (loadable-function-place-fcn tr)))))

(defun make-type-closable (type trans)
;;;;   (cond ((null (Type-freevars type))
;;;;	  (dbg-save :run-loud type)
;;;;	  (breakpoint type-clos/able
;;;;	     "No freevars for " type)))
   (new-Closable type (Typoid-freevars type) trans))

(defun close-elements (lty env)
   (<# (\\ (ty) (new-Type-closure ty env))
       lty))

;;; pty is not a list, but should evaluate to a list in env
(defun param-close-elements (pty env)
   (cond ((is-param-type pty)
	  (multi-let (((val env)
		       (var-val-spread (param-type-varname pty)
				       env)))
	     (cond ((is-list-of val (\\ (x)
				       (or (is-Type x)
					   (is-Closable x)
					   (is-Type-closure x))))
		    (close-elements val env))
		   (t
		    (signal-problem param-close-elements
		       :fatal
		       "Param value is not list: " val)))))
	 (t
	  (signal-problem param-close-elements :fatal
	     "Not a param-type: " pty))))

(defun arglistspec-freevars (alspec)
   (nodup  (<$ (\\ (aspec)
		  (Typoid-freevars (Argspec-type aspec)))
	       (Arglistspec-argspecs alspec))
	   :test #'same-param))

(defun is-Typoid (x)
   (or (is-Type x) (is-Type-closure x) (is-Closable x)))

(defun Typoid-freevars (x)
   (cond ((is-Type x) (Type-freevars x))
	 ((is-Type-closure x)
	  (Type-closure-freevars x))
	 ((is-Closable x)
	  (Closable-freevars x))
	 (t
	  (signal-problem Typoid-freevars
	     "Not a Typoid: " x))))

;;; If env is non-closing, we just ignore it.
;;; If ty has no free variables, we discard env.
;;; If ty is a Closable, delegate to its closer.
;;; Otherwise, make an actual Type-closure.
(defun new-Type-closure (ty env)
   (cond ((Env-is-closing env)
	  (close-with-env ty env))
	 (t ty)))

(defun close-with-env (ty env)
   (letrec ()
	     (cond ((is-Type ty)
		    (cond ((is-param-type ty)
			   (param-type-lookup ty env))
			  (t
			   (close-if-freevars-bound ty (Type-freevars ty)))))
		   ((is-Type-closure ty)
		    (close-if-freevars-bound ty (Type-closure-freevars ty)))
		   ((is-Closable ty)
		    (funcall (closable-closer ty)
			     (Closable-thing ty)
			     env))
		   (t
		    (signal-problem close-with-env
		       "Attempt to close non-Typoid: " ty)))
    :where
      (param-type-lookup (ty env)
	 (let ((vt (var-lookup (param-type-varname ty)
			       env)))
	   (cond (vt
		  (cond ((eq (Vartype-val vt)
			     '*unbound)
			 ty)
			(t
			 (Vartype-val vt))))
		 (t
		  (signal-problem param-type-lookup
		     "Unbound param: " ty " in " env
		     :fatal)))))

      (close-if-freevars-bound (x fvl)
;;;;	 (trace-around close-if-fv-bound
;;;;	    (:> "(close-if-fv-bound: " x 1 fvl")")
	 (cond ((null fvl)
;;;;		(cond ((is-funtype x)
;;;;		       (dbg-save x)
;;;;		       (breakpoint close-if-freevars-bound
;;;;			  "Funtype with no freevars: " x)))
		x)
	       (t
		(let ((bad-apples
		         (repeat :for (fv :in fvl)
			  :when 
			       (and ;;(not (type-feature fv 'free-param))
				    (let ((vt (var-lookup
					          (param-type-varname fv)
						  env)))
				       (or (not vt)
					   ;;(eq (Vartype-val vt) '*unbound)
					   )))
			  :collect fv)))
		   (cond ((null bad-apples)
			  (make-Type-closure x
					     (env-freevars fvl env false)
					     env))
			 (t
			  (signal-problem close-with-env
			     "Some variables of " x
			     t " are not bound in " env
			     t " to wit: " bad-apples))))))
;;;;	    (:< (val &rest _) "close-if-fv-bound: " val))
	 )))

;;; This version is used for creating tvar bounds, which have to be
;;; Type-closures even if they have no freevars.
(defun force-type-closure (ty env)
   (cond ((is-Type ty)
	  (let ((fvl (Type-freevars ty)))
	     (let ((bad-apples
		      (repeat :for (fv :in fvl)
		       :when
			    (let ((vt (var-lookup
					 (param-type-varname fv)
					 env)))
			       (not vt))
		       :collect fv)))
		(cond ((null bad-apples)
		       (make-Type-closure ty
					  (env-freevars fvl env false)
					  env))
		      (t
		       (signal-problem force-type-closure
			     "Some variables of " ty
			     t " are not bound in " env
			     t " to wit: " bad-apples))))))
	 (t
	  (signal-problem force-type-closure
	     "Trying to force closure of non-type "
	     t 2  ty 
	     t 2 "in " env))))

(defun env-freevars (fvl vartypes unbounds-are-free)
   (nodup  (<! (\\ (fv)
		  (let ((vt (var-lookup (param-type-varname fv)
					vartypes)))
		     (cond (vt
			    (let ((vval (Vartype-val vt)))
			       (cond ((is-Typoid vval)
				      (cond ((and (not
						     (Env-is-closing vartypes))
						  (is-unknown-type vval))
					     (list fv))
					    (t
					     (Typoid-freevars vval))))
				     ((and unbounds-are-free
					   (eq vval '*unbound))
				      (list fv))
				     (t '()))))
			   (unbounds-are-free (list fv))
			   (t
			    (signal-problem env-freevars
			       "Unbound param: " fv)))))
	       fvl)
	   :test #'same-param))

(defun type-clos/able (ty env)
   (closure-or-closable ty env false (Type-freevars ty)))

(defun closure-or-closable (ty env lf freevars)
   (cond ((null freevars) ty)
	 ((not (is-list-of freevars #'is-param-type))
	  (dbg-save :run-loud freevars ty env lf)
	  (signal-problem closure-or-closable
	     "Sigh.  freevars = " freevars))
	 ((Env-is-closing env)
	  (cond ((not lf)
		 (close-with-env ty env))
		(t
		 (let ((closer (cond ((is-Loadable-function lf)
				      (loadable-function-place-fcn lf))
				     ((is-Symbol lf)
				      (symbol->fun lf))
				     (t
				      (signal-problem closure-or-closable
					 !"Illegal function name for env 
                                           closer: "
					 lf)))))
		    (funcall closer ty env)))))
	 (t
	  (new-Closable
	     ty
	     freevars
	     (cond ((or (not lf)
			(is-Loadable-function lf))
		    lf)
		   ((is-Symbol lf)
		    (make-Loadable-function lf false))
		   (t
		    (signal-problem closure-or-closable
		       "Illegal function name for Closable " lf)))))))

(defun close-if-closable (c env)
   (cond ((or (is-Type c) (is-Closable c))
	  (new-Type-closure c env))
	 (t c)))

(defun resolve-if-closable (x env)
   (cond ((is-Closable x)
	  (let ((y (funcall (closable-closer x)
			    (Closable-thing x)
			    env)))
	     y))
	 (t x)))

;;; The following functions, with names ending in '-close', are used as the
;;; 'trans' slot-filler in Closables.  That is, x-close is used when the
;;; 'thing' slot is an x.

(defun arglistspec-close (alspec env)
   (new-Arglistspec
      (argspecs-close
	  (Arglistspec-argspecs alspec)
	  env)
      :allow-other-keys (Arglistspec-allow-other-keys alspec)
      ;;:names-matter (Arglistspec-names-matter alspec
      ))

;;; We *don't* optimize this by checking if env is closing,
;;; because Argspecs cannot be shared among Arglistspecs --- they
;;; must be copied.
(defun argspecs-close (asl env)
   (<# (\\ (as) (argspec-close as env))
       asl))

(defun argspec-close (as env)
   (make-Argspec (Argspec-name as)
		 (Argspec-mode as)
		 (argspec-pos-zero as)
		 (new-Type-closure (Argspec-type as)
				   env)
		 (Argspec-default as)
		 (<# list-copy (Argspec-props as))))

(defun typelist-close (tl env)
   (<# (\\ (ty) (new-Type-closure ty env))
       tl))

(defun vartypelist-close (vtl env)
   (<# (\\ (vt)
	  (let ((vt-ty (Vartype-type vt)))
	     (let ((new-vt (new-Vartype
			      (Vartype-var vt)
			      vt-ty
			      (close-with-env (Vartype-val vt)
					      env))))
	        (!= (Vartype-props new-vt)
		    (<# list-copy (Vartype-props vt)))
		new-vt)))
       vtl))

(defun no-close (x env)
                (ignore env)
   x)

;;; Not clear that this is right; it's used by the typeconseq machinery.
(defun strip-env (x bad-env good-env)
   (cond ((is-Type-closure x)
	  (cond ((eq (Type-closure-env x)
		     bad-env)
		 (cond ((and (not (null bad-env))
			     (not (eq bad-env good-env))
			     (memq (car bad-env) good-env))
			(signal-problem strip-env 
			   "Element from bad env " bad-env
			   t "  remains after strip-env "
			   t "  good env = " good-env
			   (:novalue "I will strip bad env anyway"))))
		 (Type-closure-type x))
		(t
		 x)))
	 (t x)))

;;;;(defun has-freevars (x)
;;;;	 (cond ((is-Type x)
;;;;		(not (null (Type-freevars x))))
;;;;	       ((atom x) false)
;;;;	       (t
;;;;		(exists (y :in x)
;;;;		   (has-freevars y)))))

;;; Vartypes now allow type systems as elements.
;;; Consequently, we can no longer treat () as the empty env.
;;; We always have a type system in the list where globals are kept.
(defun empty-vartypes (ts)
   (make-Env true (list ts)))

(defvar global-env* (empty-vartypes universal-type-sys*))

(!= (Type-system-env universal-type-sys*) global-env*)

(defun place-type-system-env (ts)
   (or (Type-system-env ts)
       (let ((ne (empty-vartypes ts)))
	  (!= (Type-system-env ts) ne)
	  ne)))

(defun vartypes-globals (vartypes)
  (repeat :for ((vtl (Env-vartypes vartypes) (cdr vtl)))
   :until (null vtl)
   :result (progn
   	     (error-break vartypes-globals 
		"Vartypes with no globals " vartypes
		(:novalue "I will go on"))
	     '())
   :until (is-Type-system (car vtl))
   :result (car vtl)))

;;(defun param-vartype-spread (v env) ...)
(defun var-val-spread (v env)
   (follow-var-ref (var-val v env) env))

(defun var-val (v vartypes)
   (let ((vt (var-lookup v vartypes)))
      (let ((vv (and vt (Vartype-val vt))))
	 (cond ((and vt (not (eq vv '*unbound)))
		vv)
	       (t
		(signal-problem var-val
		   "Symbol " v " unbound in " vartypes
		   :fatal))))))

(defun var-lookup (v vartypes)
;;;;   (cond ((eq vartypes '*no-env)
;;;;	  (signal-problem var-lookup
;;;;	     "Looking up " v " with no environment")))
  (vartypes-lookup v (Env-vartypes vartypes)))

(defun vartypes-lookup (v vartypes)
   (repeat :for ((vt :in vartypes)
		(found-vt false))
    :result false
      (cond ((is-Vartype vt)
	     (cond ((eq (Vartype-var vt) v)
		    (!= found-vt vt))))
	    (t
	     (!= found-vt (name-lookup vt v))))
    :until found-vt
    :result found-vt))

;;; El cheapo version of follow-var-ref
(defun unwrap (x)
   (cond ((is-Type-closure x)
	  (unwrap (Type-closure-type x)))
	 ((is-Closable x)
	  (unwrap (Closable-thing x)))
	 ;;((is-Unclosable x)
	 ;; (unwrap (Unclosable-thing x)))
	 (t x)))

(defun type-unwrap (x vartypes)
   (cond ((is-Type-closure x)
	  (multiple-value-let (y env)
			      (type-unwrap (Type-closure-type x)
					   (Type-closure-env x))
	     (values y env)))
	 (t
	  (values x vartypes))))

;;; Returns < ty, env > after resolving variable references.
;;; This is a "spread Type-closure."
;;; Properties: Let input be < input-ty, input-env >, output be
;;; < output-ty, output-env >.
;;; If input-env is non-closing, then < output-ty, output-env > has the 
;;; same freevars as input-ty.
;;; Otherwise, all the freevars of output-ty must be bound in output-env;
;;;   output-env itself could have freevars, though.
;;; output-ty must not be a bound param-type, a tvar-type with a value,
;;; a Type-closure, or a Closable.
(defun follow-var-ref (ty env)
   (cond ((is-Type ty)
	  (cond ((is-tvar-type ty)
		 (let ((ty1 (tvar-value ty)))
		    (cond (ty1 
			   (follow-var-ref ty1 env))
			  (t
			   (values ty env)))))
		((is-param-type ty)
		 (let ((b (var-lookup (param-type-varname ty) env)))
		    (cond (b
			   (let ((val (Vartype-val b)))
			      (cond ((and (not (Env-is-closing env))
					  (is-unknown-type val))
				     (values ty no-env*))
				    ((eq val '*unbound)
				     (values ty env))
				    (t
				     (follow-var-ref val env))))
;;;;			   (let ((nc (Vartype-val b)))
;;;;			      (cond ((eq nc '*unbound)
;;;;				     (values ty no-env*))
;;;;				    ((is-param-type nc)
;;;;				     (values nc env))
;;;;				    (t
;;;;				     (follow-var-ref nc env))))
			   )
			  (t
			   (dbg-save :comp-quiet ty env)
			   (error-break follow-var-ref :fatal
				"Unbound parameter: " ty)))))
		(t (values ty env))))
	 ((is-Type-closure ty)
	  (multi-let (((val venv)
		       (follow-var-ref (Type-closure-type ty)
				       (Type-closure-env ty))))
	     (cond ((Env-is-closing env)
		    (cond ((Env-is-closing venv)
			   (values val (env-bind-freevars val venv env)))
			  (t
			   (values val env))))
		   (t
		    (values val venv)))))
	 ((is-Closable ty)
	  (follow-var-ref 
	     (funcall (closable-closer ty)
		      (Closable-thing ty)
		      env)
	     env))
	 (t
	  (values ty env))))

(defun env-bind-freevars (typoid inner-env outer-env)
   (repeat :for ((fv :in (Typoid-freevars typoid))
		(new-vartypes '()))
      (let ((v (param-type-varname fv)))
	 (let ((fv-vt (cond ((vartypes-lookup v new-vartypes)
			     ;; Ignore duplicates
			     false)
			    (t
			     (var-lookup v inner-env)))))
	    (cond (fv-vt
		   (let ((fv-val (Vartype-val fv-vt)))
		      (cond ((and (is-Typoid fv-val)
				  (not (null (Typoid-freevars fv-val))))
			     (!= new-vartypes
				 (cons (new-Vartype
					  v
					  (Vartype-type fv-vt)
					  (new-Type-closure
					     fv-val
					     outer-env))
				       *-*))))))
		  (t
		   (signal-problem follow-var-ref
		      "Unbound parameter: " fv)))))
    :result (env-bindings-nconc (Env-is-closing inner-env)
			       new-vartypes inner-env)))

;;; For debugging
(defun freevar-massage (v)
   (let ((d (Type-desig v)))
      (cond ((is-Qvar d) d)
	    ((car-eq d 'param) (cadr d))
	    (t d))))
	     
;;; The following two functions should probably be called "params-union"
;;; and "params-complement."  freevars are always lists of params

;;; This program eliminates duplicates from vl1, but leaves any in vl2
;;; undisturbed.
(defun freevars-union (vl1 vl2)
   (repeat :for ((v :in vl1))
      (cond ((not (member v vl2 :test #'same-param))
	     (!= vl2 (cons v *-*)))))
   vl2)

(defun freevars-complement (vl1 vl2)
   (<? (\\ (v1) (not (member v1 vl2 :test #'same-param)))
       vl1))

;;; Types may begin to be defined before their details are known.  
;;; We indicate this with a forward feature.

;;; Return whether it had to define it.
(defun maybe-forward-define (typename type-sys)
  (cond ((not (global-declaration typename type-sys false))
	 (forward-define typename type-sys)
	 true)
	(t false)))

(defun forward-define (typename &optional sys)
   (let ((vt (global-declaration typename sys false)))
      (cond ((and vt
		  (not (and (eq (Vartype-typ vt) type-type*)
			    (type-feature
			       (Vartype-initial vt)
			       'forward))))
;;;	     (out (:to (errout))
;;;		  "Warning -- redefining type " typename
;;;		  " [old binding: " vt "]" t " with features "
;;;		  (Type-features (Vartype-initial vt)) :%)
	     ;(typesee (Vartype-initial vt))
	     ))
      (typedecl typename
		type-type*
		(forward-type typename sys)
		sys)))

(defun forward-type (typename home)
		(new-Type false typename '() univ-type* '()
			  (list (tuple 'forward true)
				(tuple 'home home))))

;;; The "Ty" means "at type-expansion time," connoting the fact that
;;; these functions can be "evaluated" at compile-time.
;;; The defn is something (usually a type)
;;; that contains occurrences of the params, as 
;;; types of class param-type-class.
(def-class Tyfun
    (:handler (print-object (tf stream)
		 (out (:to stream) "#{Tyfun " (Tyfun-name tf) "}")))
   name 
   params  ; An arglistspec
   defn   
   env  ;; Lexical bindings, as God intended
)
;; Typically, we'll get new bindings for the params, and add them
;; to the env.
;; defn is often a Closable, and the env will be modified, so it's
;; not appropriate to roll defn and env together into a Type-closure.

(defun Tyfun-result-type (tyf)
   (repeat :for ((defn (Tyfun-defn tyf)))
    :until (is-Type defn)
    :result defn
      (cond ((is-Type-closure defn)
	     (!= defn (Type-closure-type *-*)))
	    ((is-Closable defn)
	     (!= defn (Closable-thing *-*)))
	    (t
	     (error-break Tyfun-result-type :fatal
		"Can't find type in definition of " tyf)))))
    

(defun new-Tyfun (name params defn env)
   (make-Tyfun name params defn env))

(defun tyfun-forward-define (typefun-name params sys)
         (typedecl typefun-name
		   tyfun-type*
		   (new-Tyfun typefun-name
			      params
			      'forward
			      (empty-vartypes sys))
			 sys))

;;; Most typeform stuff is in tyfun, but the basics are here.
;;; Typeforms are either suspended or unsuspended.
;;; If suspended, we don't know the args yet, just their
;;; designators.

(defun build-typeform (fun-name arg-desigs vartypes)
   (make-type-of-class typeform-class
		       `(,fun-name ,@arg-desigs)
		       ;'*compute
		       '()
		       univ-type* '()
		       (list (tuple 'suspended true)
			     (tuple 'fun-name fun-name)
			     (tuple 'arg-desigs arg-desigs)
			     (tuple 'ur-env vartypes)
			     (tuple 'kind void-type*))))

;; Args is a list of Vartypes.
(defun make-typeform (fun args)
   (make-type-of-class typeform-class
		       `(,(Tyfun-name fun)
			 ,@(<# (\\ (vt)
				  (extract-if-type
				     (Vartype-val vt)
				     #'Type-desig))
			       args))
		       (freevars-union
			  (<$ (\\ (vt)
				 (extract-if-type
				    (Vartype-val vt) #'Type-freevars))
			      args)
			  '())
		       '*as-feature
		       '()
		       (nconc (list (tuple 'suspended false)
				    (tuple 'fun fun)
				    (tuple 'args args))
			      (cond ((eq (Tyfun-defn fun) 'forward)
				     '())
				    (t
				     (list (tuple 'kind
						  (Type-kind
						     (Tyfun-defn fun)))))))))

(defun make-modified-type (desig suptype slots features)
   (multi-let (((suptype supenv)
		(follow-var-ref suptype empty-env*))
	       (feats
		(mod-class-features-copy features)))
      (make-type-of-class modified-class
	 desig
	 (type-find-freevars suptype supenv)
	 (cond (supenv '*as-feature)
	       (t suptype))
	 slots
	 (cond (supenv
		(cons (tuple 'supertype
		       (type-clos/able
			  suptype supenv))
		      feats))
	       (t feats)))))

(defun mod-class-features-copy (feats)
   (<! (\\ (feat)
	  (cond ((eq (car feat) 'supertype)
		 '())
		(t (list (list-copy feat)))))
       feats))

(defun extract-if-type (x extractor)
   (cond ((is-Type x)
	  (funcall extractor x))
	 ((is-Type-closure x)
	  (extract-if-type (Type-closure-type x)
			   extractor))
	 ((is-Closable x)
	  (extract-if-type (Closable-thing x)
			   extractor))
	 (t x)))

(defun is-typeform (x)
   (and (is-Type x)
	(eq (Type-class x) typeform-class*)))

;;; Returns same, proper-part, or false.
;;; tvar is not a tvar, but the name of one
(defun tvar-occurs-in (tvar type top envt)
   (cond ((not (is-Symboid tvar))
	  (error-break tvar-occurs-in :fatal
	     "tvar-occurs-in wants name of tvar, not: " tvar)))
   (cond ((is-tvar-type type)
	  (cond ((eq tvar (tvar-type-varname type))
		 (cond (top 'same)
		       (t 'proper-part)))
		(t
		 (let ((val (tvar-value type)))
		    (cond (val
			   (tvar-occurs-in tvar val true envt))
			  (t
			   false))))))
	 ((is-param-type type)
	  (multiple-value-let (ty env)
			      (var-val-spread (param-type-varname type)
					      envt)
	      (tvar-occurs-in tvar ty top env)))
	 ((is-Type-closure type)
	  (tvar-occurs-in tvar (Type-closure-type type)
		     top
		     (Type-closure-env type)))
	 ((is-unknown-type type) false)
	 (t
	  (repeat :for ((fv :in (Type-freevars type))
		       (occ false))
	   :result false
	     (!= occ 
		 (tvar-occurs-in tvar fv false envt))
	   :until occ
	   :result occ))))

(defun type-find-kind (ty env)
   (multi-let (((ty _)
		(follow-var-ref ty env)))
      (Type-kind ty)))

(defun type-set-supertype (ty sup env)
   (cond ((is-Type sup)
	  (!= (Type-super ty) sup))
	 (t
	  (!= (Type-super ty) '*as-feature)
	  (!= (Type-features ty)
	      (cons (tuple 'supertype
			   (closure-or-closable ty env false
						(Type-freevars ty)))
		    *-*)))))

(defun type-find-class (ty env)
   (multiple-value-let (ty _)
		       (follow-var-ref ty env)
      (cond ((is-Type ty)
	     (Type-class ty))
	    (t
	     (error-break find-class
		"Can't find class of: " ty
		:fatal)))))

(defun type-find-freevars (ty env)
   (env-freevars (Typoid-freevars ty)
		 env
		 (not (Env-is-closing env))))

;;;;   (let ((fvl (Typoid-freevars ty)))
;;;;      (cond ((Env-is-closing env)
;;;;	     (env-freevars fvl env false))
;;;;	    (t
;;;;	     fvl))))

(defun type-find-unbound-vars (ty env)
   (env-freevars (Typoid-freevars ty)
		 env
		 true))

(defun same-freevar-lists (fvl1 fvl2)
   (letrec ()
      (and (freevar-sublist fvl1 fvl2)
	   (freevar-sublist fvl2 fvl1))))

(defun freevar-sublist (z1 z2)
	       (forall (v1 :in z1)
		  (exists (v2 :in z2)
		     (same-param v1 v2))))

;;; Called by code generated by funformers
(defun bound-last (params env)
   (let ((vartypes (Env-vartypes env))
	 (n (len params)))
      (letrec ((find-in-first-n (p)
		  (repeat :for ((b :in vartypes)
			       (i = 1 :to n))
		   :result false
		   :until (and (is-Vartype b)
			      (eq (Vartype-var b) p))
		   :result true)))
	 (forall (p :in params)
	    (find-in-first-n p)))))

(defun Type-kind (ty)
   (or (type-feature ty 'kind)
       type-type*))

(defvar debug-type-sys* nil)

(defvar debug-env* nil)

(defun load-typesee (debug-sys)
   (fload %nity/ typesee)
   (!= debug-type-sys* debug-sys)
   (!= debug-env* (place-type-system-env debug-type-sys*)))


